/* Generated by Cython 0.25.2 */

/* BEGIN: Cython Metadata
{
    "distutils": {
        "depends": [
            "/usr/local/include/cpl_conv.h",
            "/usr/local/include/cpl_error.h",
            "/usr/local/include/cpl_string.h",
            "/usr/local/include/cpl_vsi.h",
            "/usr/local/include/gdal.h",
            "/usr/local/include/gdal_alg.h",
            "/usr/local/include/gdalwarper.h",
            "/usr/local/include/ogr_api.h",
            "/usr/local/include/ogr_core.h",
            "/usr/local/include/ogr_srs_api.h",
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/numpy/core/include/numpy/arrayobject.h",
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/numpy/core/include/numpy/ufuncobject.h"
        ],
        "extra_compile_args": [
            "-Wno-unused-parameter",
            "-Wno-unused-function"
        ],
        "include_dirs": [
            "/var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/numpy/core/include",
            "/usr/local/include"
        ],
        "libraries": [
            "gdal"
        ],
        "library_dirs": [
            "/usr/local/lib"
        ]
    },
    "module_name": "rasterio._features"
}
END: Cython Metadata */

#define PY_SSIZE_T_CLEAN
#include "Python.h"
#ifndef Py_PYTHON_H
    #error Python headers needed to compile C extensions, please install development version of Python.
#elif PY_VERSION_HEX < 0x02060000 || (0x03000000 <= PY_VERSION_HEX && PY_VERSION_HEX < 0x03020000)
    #error Cython requires Python 2.6+ or Python 3.2+.
#else
#define CYTHON_ABI "0_25_2"
#include <stddef.h>
#ifndef offsetof
  #define offsetof(type, member) ( (size_t) & ((type*)0) -> member )
#endif
#if !defined(WIN32) && !defined(MS_WINDOWS)
  #ifndef __stdcall
    #define __stdcall
  #endif
  #ifndef __cdecl
    #define __cdecl
  #endif
  #ifndef __fastcall
    #define __fastcall
  #endif
#endif
#ifndef DL_IMPORT
  #define DL_IMPORT(t) t
#endif
#ifndef DL_EXPORT
  #define DL_EXPORT(t) t
#endif
#ifndef HAVE_LONG_LONG
  #if PY_VERSION_HEX >= 0x03030000 || (PY_MAJOR_VERSION == 2 && PY_VERSION_HEX >= 0x02070000)
    #define HAVE_LONG_LONG
  #endif
#endif
#ifndef PY_LONG_LONG
  #define PY_LONG_LONG LONG_LONG
#endif
#ifndef Py_HUGE_VAL
  #define Py_HUGE_VAL HUGE_VAL
#endif
#ifdef PYPY_VERSION
  #define CYTHON_COMPILING_IN_PYPY 1
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #undef CYTHON_USE_TYPE_SLOTS
  #define CYTHON_USE_TYPE_SLOTS 0
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #undef CYTHON_USE_UNICODE_INTERNALS
  #define CYTHON_USE_UNICODE_INTERNALS 0
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #undef CYTHON_AVOID_BORROWED_REFS
  #define CYTHON_AVOID_BORROWED_REFS 1
  #undef CYTHON_ASSUME_SAFE_MACROS
  #define CYTHON_ASSUME_SAFE_MACROS 0
  #undef CYTHON_UNPACK_METHODS
  #define CYTHON_UNPACK_METHODS 0
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#elif defined(PYSTON_VERSION)
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 1
  #define CYTHON_COMPILING_IN_CPYTHON 0
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #undef CYTHON_USE_ASYNC_SLOTS
  #define CYTHON_USE_ASYNC_SLOTS 0
  #undef CYTHON_USE_PYLIST_INTERNALS
  #define CYTHON_USE_PYLIST_INTERNALS 0
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #undef CYTHON_USE_UNICODE_WRITER
  #define CYTHON_USE_UNICODE_WRITER 0
  #undef CYTHON_USE_PYLONG_INTERNALS
  #define CYTHON_USE_PYLONG_INTERNALS 0
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #undef CYTHON_FAST_THREAD_STATE
  #define CYTHON_FAST_THREAD_STATE 0
  #undef CYTHON_FAST_PYCALL
  #define CYTHON_FAST_PYCALL 0
#else
  #define CYTHON_COMPILING_IN_PYPY 0
  #define CYTHON_COMPILING_IN_PYSTON 0
  #define CYTHON_COMPILING_IN_CPYTHON 1
  #ifndef CYTHON_USE_TYPE_SLOTS
    #define CYTHON_USE_TYPE_SLOTS 1
  #endif
  #if PY_MAJOR_VERSION < 3
    #undef CYTHON_USE_ASYNC_SLOTS
    #define CYTHON_USE_ASYNC_SLOTS 0
  #elif !defined(CYTHON_USE_ASYNC_SLOTS)
    #define CYTHON_USE_ASYNC_SLOTS 1
  #endif
  #if PY_VERSION_HEX < 0x02070000
    #undef CYTHON_USE_PYLONG_INTERNALS
    #define CYTHON_USE_PYLONG_INTERNALS 0
  #elif !defined(CYTHON_USE_PYLONG_INTERNALS)
    #define CYTHON_USE_PYLONG_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_PYLIST_INTERNALS
    #define CYTHON_USE_PYLIST_INTERNALS 1
  #endif
  #ifndef CYTHON_USE_UNICODE_INTERNALS
    #define CYTHON_USE_UNICODE_INTERNALS 1
  #endif
  #if PY_VERSION_HEX < 0x030300F0
    #undef CYTHON_USE_UNICODE_WRITER
    #define CYTHON_USE_UNICODE_WRITER 0
  #elif !defined(CYTHON_USE_UNICODE_WRITER)
    #define CYTHON_USE_UNICODE_WRITER 1
  #endif
  #ifndef CYTHON_AVOID_BORROWED_REFS
    #define CYTHON_AVOID_BORROWED_REFS 0
  #endif
  #ifndef CYTHON_ASSUME_SAFE_MACROS
    #define CYTHON_ASSUME_SAFE_MACROS 1
  #endif
  #ifndef CYTHON_UNPACK_METHODS
    #define CYTHON_UNPACK_METHODS 1
  #endif
  #ifndef CYTHON_FAST_THREAD_STATE
    #define CYTHON_FAST_THREAD_STATE 1
  #endif
  #ifndef CYTHON_FAST_PYCALL
    #define CYTHON_FAST_PYCALL 1
  #endif
#endif
#if !defined(CYTHON_FAST_PYCCALL)
#define CYTHON_FAST_PYCCALL  (CYTHON_FAST_PYCALL && PY_VERSION_HEX >= 0x030600B1)
#endif
#if CYTHON_USE_PYLONG_INTERNALS
  #include "longintrepr.h"
  #undef SHIFT
  #undef BASE
  #undef MASK
#endif
#if CYTHON_COMPILING_IN_PYPY && PY_VERSION_HEX < 0x02070600 && !defined(Py_OptimizeFlag)
  #define Py_OptimizeFlag 0
#endif
#define __PYX_BUILD_PY_SSIZE_T "n"
#define CYTHON_FORMAT_SSIZE_T "z"
#if PY_MAJOR_VERSION < 3
  #define __Pyx_BUILTIN_MODULE_NAME "__builtin__"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a+k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyClass_Type
#else
  #define __Pyx_BUILTIN_MODULE_NAME "builtins"
  #define __Pyx_PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)\
          PyCode_New(a, k, l, s, f, code, c, n, v, fv, cell, fn, name, fline, lnos)
  #define __Pyx_DefaultClassType PyType_Type
#endif
#ifndef Py_TPFLAGS_CHECKTYPES
  #define Py_TPFLAGS_CHECKTYPES 0
#endif
#ifndef Py_TPFLAGS_HAVE_INDEX
  #define Py_TPFLAGS_HAVE_INDEX 0
#endif
#ifndef Py_TPFLAGS_HAVE_NEWBUFFER
  #define Py_TPFLAGS_HAVE_NEWBUFFER 0
#endif
#ifndef Py_TPFLAGS_HAVE_FINALIZE
  #define Py_TPFLAGS_HAVE_FINALIZE 0
#endif
#ifndef METH_FASTCALL
  #define METH_FASTCALL 0x80
  typedef PyObject *(*__Pyx_PyCFunctionFast) (PyObject *self, PyObject **args,
                                              Py_ssize_t nargs, PyObject *kwnames);
#else
  #define __Pyx_PyCFunctionFast _PyCFunctionFast
#endif
#if CYTHON_FAST_PYCCALL
#define __Pyx_PyFastCFunction_Check(func)\
    ((PyCFunction_Check(func) && (METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)))))
#else
#define __Pyx_PyFastCFunction_Check(func) 0
#endif
#if PY_VERSION_HEX > 0x03030000 && defined(PyUnicode_KIND)
  #define CYTHON_PEP393_ENABLED 1
  #define __Pyx_PyUnicode_READY(op)       (likely(PyUnicode_IS_READY(op)) ?\
                                              0 : _PyUnicode_Ready((PyObject *)(op)))
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_LENGTH(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) PyUnicode_READ_CHAR(u, i)
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   PyUnicode_MAX_CHAR_VALUE(u)
  #define __Pyx_PyUnicode_KIND(u)         PyUnicode_KIND(u)
  #define __Pyx_PyUnicode_DATA(u)         PyUnicode_DATA(u)
  #define __Pyx_PyUnicode_READ(k, d, i)   PyUnicode_READ(k, d, i)
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  PyUnicode_WRITE(k, d, i, ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != (likely(PyUnicode_IS_READY(u)) ? PyUnicode_GET_LENGTH(u) : PyUnicode_GET_SIZE(u)))
#else
  #define CYTHON_PEP393_ENABLED 0
  #define PyUnicode_1BYTE_KIND  1
  #define PyUnicode_2BYTE_KIND  2
  #define PyUnicode_4BYTE_KIND  4
  #define __Pyx_PyUnicode_READY(op)       (0)
  #define __Pyx_PyUnicode_GET_LENGTH(u)   PyUnicode_GET_SIZE(u)
  #define __Pyx_PyUnicode_READ_CHAR(u, i) ((Py_UCS4)(PyUnicode_AS_UNICODE(u)[i]))
  #define __Pyx_PyUnicode_MAX_CHAR_VALUE(u)   ((sizeof(Py_UNICODE) == 2) ? 65535 : 1114111)
  #define __Pyx_PyUnicode_KIND(u)         (sizeof(Py_UNICODE))
  #define __Pyx_PyUnicode_DATA(u)         ((void*)PyUnicode_AS_UNICODE(u))
  #define __Pyx_PyUnicode_READ(k, d, i)   ((void)(k), (Py_UCS4)(((Py_UNICODE*)d)[i]))
  #define __Pyx_PyUnicode_WRITE(k, d, i, ch)  (((void)(k)), ((Py_UNICODE*)d)[i] = ch)
  #define __Pyx_PyUnicode_IS_TRUE(u)      (0 != PyUnicode_GET_SIZE(u))
#endif
#if CYTHON_COMPILING_IN_PYPY
  #define __Pyx_PyUnicode_Concat(a, b)      PyNumber_Add(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  PyNumber_Add(a, b)
#else
  #define __Pyx_PyUnicode_Concat(a, b)      PyUnicode_Concat(a, b)
  #define __Pyx_PyUnicode_ConcatSafe(a, b)  ((unlikely((a) == Py_None) || unlikely((b) == Py_None)) ?\
      PyNumber_Add(a, b) : __Pyx_PyUnicode_Concat(a, b))
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyUnicode_Contains)
  #define PyUnicode_Contains(u, s)  PySequence_Contains(u, s)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyByteArray_Check)
  #define PyByteArray_Check(obj)  PyObject_TypeCheck(obj, &PyByteArray_Type)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Format)
  #define PyObject_Format(obj, fmt)  PyObject_CallMethod(obj, "__format__", "O", fmt)
#endif
#if CYTHON_COMPILING_IN_PYPY && !defined(PyObject_Malloc)
  #define PyObject_Malloc(s)   PyMem_Malloc(s)
  #define PyObject_Free(p)     PyMem_Free(p)
  #define PyObject_Realloc(p)  PyMem_Realloc(p)
#endif
#if CYTHON_COMPILING_IN_PYSTON
  #define __Pyx_PyCode_HasFreeVars(co)  PyCode_HasFreeVars(co)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno) PyFrame_SetLineNumber(frame, lineno)
#else
  #define __Pyx_PyCode_HasFreeVars(co)  (PyCode_GetNumFree(co) > 0)
  #define __Pyx_PyFrame_SetLineNumber(frame, lineno)  (frame)->f_lineno = (lineno)
#endif
#define __Pyx_PyString_FormatSafe(a, b)   ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : __Pyx_PyString_Format(a, b))
#define __Pyx_PyUnicode_FormatSafe(a, b)  ((unlikely((a) == Py_None)) ? PyNumber_Remainder(a, b) : PyUnicode_Format(a, b))
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyString_Format(a, b)  PyUnicode_Format(a, b)
#else
  #define __Pyx_PyString_Format(a, b)  PyString_Format(a, b)
#endif
#if PY_MAJOR_VERSION < 3 && !defined(PyObject_ASCII)
  #define PyObject_ASCII(o)            PyObject_Repr(o)
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBaseString_Type            PyUnicode_Type
  #define PyStringObject               PyUnicodeObject
  #define PyString_Type                PyUnicode_Type
  #define PyString_Check               PyUnicode_Check
  #define PyString_CheckExact          PyUnicode_CheckExact
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyBaseString_Check(obj) PyUnicode_Check(obj)
  #define __Pyx_PyBaseString_CheckExact(obj) PyUnicode_CheckExact(obj)
#else
  #define __Pyx_PyBaseString_Check(obj) (PyString_Check(obj) || PyUnicode_Check(obj))
  #define __Pyx_PyBaseString_CheckExact(obj) (PyString_CheckExact(obj) || PyUnicode_CheckExact(obj))
#endif
#ifndef PySet_CheckExact
  #define PySet_CheckExact(obj)        (Py_TYPE(obj) == &PySet_Type)
#endif
#define __Pyx_TypeCheck(obj, type) PyObject_TypeCheck(obj, (PyTypeObject *)type)
#define __Pyx_PyException_Check(obj) __Pyx_TypeCheck(obj, PyExc_Exception)
#if PY_MAJOR_VERSION >= 3
  #define PyIntObject                  PyLongObject
  #define PyInt_Type                   PyLong_Type
  #define PyInt_Check(op)              PyLong_Check(op)
  #define PyInt_CheckExact(op)         PyLong_CheckExact(op)
  #define PyInt_FromString             PyLong_FromString
  #define PyInt_FromUnicode            PyLong_FromUnicode
  #define PyInt_FromLong               PyLong_FromLong
  #define PyInt_FromSize_t             PyLong_FromSize_t
  #define PyInt_FromSsize_t            PyLong_FromSsize_t
  #define PyInt_AsLong                 PyLong_AsLong
  #define PyInt_AS_LONG                PyLong_AS_LONG
  #define PyInt_AsSsize_t              PyLong_AsSsize_t
  #define PyInt_AsUnsignedLongMask     PyLong_AsUnsignedLongMask
  #define PyInt_AsUnsignedLongLongMask PyLong_AsUnsignedLongLongMask
  #define PyNumber_Int                 PyNumber_Long
#endif
#if PY_MAJOR_VERSION >= 3
  #define PyBoolObject                 PyLongObject
#endif
#if PY_MAJOR_VERSION >= 3 && CYTHON_COMPILING_IN_PYPY
  #ifndef PyUnicode_InternFromString
    #define PyUnicode_InternFromString(s) PyUnicode_FromString(s)
  #endif
#endif
#if PY_VERSION_HEX < 0x030200A4
  typedef long Py_hash_t;
  #define __Pyx_PyInt_FromHash_t PyInt_FromLong
  #define __Pyx_PyInt_AsHash_t   PyInt_AsLong
#else
  #define __Pyx_PyInt_FromHash_t PyInt_FromSsize_t
  #define __Pyx_PyInt_AsHash_t   PyInt_AsSsize_t
#endif
#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyMethod_New(func, self, klass) ((self) ? PyMethod_New(func, self) : PyInstanceMethod_New(func))
#else
  #define __Pyx_PyMethod_New(func, self, klass) PyMethod_New(func, self, klass)
#endif
#if CYTHON_USE_ASYNC_SLOTS
  #if PY_VERSION_HEX >= 0x030500B1
    #define __Pyx_PyAsyncMethodsStruct PyAsyncMethods
    #define __Pyx_PyType_AsAsync(obj) (Py_TYPE(obj)->tp_as_async)
  #else
    typedef struct {
        unaryfunc am_await;
        unaryfunc am_aiter;
        unaryfunc am_anext;
    } __Pyx_PyAsyncMethodsStruct;
    #define __Pyx_PyType_AsAsync(obj) ((__Pyx_PyAsyncMethodsStruct*) (Py_TYPE(obj)->tp_reserved))
  #endif
#else
  #define __Pyx_PyType_AsAsync(obj) NULL
#endif
#ifndef CYTHON_RESTRICT
  #if defined(__GNUC__)
    #define CYTHON_RESTRICT __restrict__
  #elif defined(_MSC_VER) && _MSC_VER >= 1400
    #define CYTHON_RESTRICT __restrict
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_RESTRICT restrict
  #else
    #define CYTHON_RESTRICT
  #endif
#endif
#ifndef CYTHON_UNUSED
# if defined(__GNUC__)
#   if !(defined(__cplusplus)) || (__GNUC__ > 3 || (__GNUC__ == 3 && __GNUC_MINOR__ >= 4))
#     define CYTHON_UNUSED __attribute__ ((__unused__))
#   else
#     define CYTHON_UNUSED
#   endif
# elif defined(__ICC) || (defined(__INTEL_COMPILER) && !defined(_MSC_VER))
#   define CYTHON_UNUSED __attribute__ ((__unused__))
# else
#   define CYTHON_UNUSED
# endif
#endif
#ifndef CYTHON_MAYBE_UNUSED_VAR
#  if defined(__cplusplus)
     template<class T> void CYTHON_MAYBE_UNUSED_VAR( const T& ) { }
#  else
#    define CYTHON_MAYBE_UNUSED_VAR(x) (void)(x)
#  endif
#endif
#ifndef CYTHON_NCP_UNUSED
# if CYTHON_COMPILING_IN_CPYTHON
#  define CYTHON_NCP_UNUSED
# else
#  define CYTHON_NCP_UNUSED CYTHON_UNUSED
# endif
#endif
#define __Pyx_void_to_None(void_result) ((void)(void_result), Py_INCREF(Py_None), Py_None)

#ifndef CYTHON_INLINE
  #if defined(__clang__)
    #define CYTHON_INLINE __inline__ __attribute__ ((__unused__))
  #elif defined(__GNUC__)
    #define CYTHON_INLINE __inline__
  #elif defined(_MSC_VER)
    #define CYTHON_INLINE __inline
  #elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define CYTHON_INLINE inline
  #else
    #define CYTHON_INLINE
  #endif
#endif

#if defined(WIN32) || defined(MS_WINDOWS)
  #define _USE_MATH_DEFINES
#endif
#include <math.h>
#ifdef NAN
#define __PYX_NAN() ((float) NAN)
#else
static CYTHON_INLINE float __PYX_NAN() {
  float value;
  memset(&value, 0xFF, sizeof(value));
  return value;
}
#endif
#if defined(__CYGWIN__) && defined(_LDBL_EQ_DBL)
#define __Pyx_truncl trunc
#else
#define __Pyx_truncl truncl
#endif


#define __PYX_ERR(f_index, lineno, Ln_error) \
{ \
  __pyx_filename = __pyx_f[f_index]; __pyx_lineno = lineno; __pyx_clineno = __LINE__; goto Ln_error; \
}

#if PY_MAJOR_VERSION >= 3
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_TrueDivide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceTrueDivide(x,y)
#else
  #define __Pyx_PyNumber_Divide(x,y)         PyNumber_Divide(x,y)
  #define __Pyx_PyNumber_InPlaceDivide(x,y)  PyNumber_InPlaceDivide(x,y)
#endif

#ifndef __PYX_EXTERN_C
  #ifdef __cplusplus
    #define __PYX_EXTERN_C extern "C"
  #else
    #define __PYX_EXTERN_C extern
  #endif
#endif

#define __PYX_HAVE__rasterio___features
#define __PYX_HAVE_API__rasterio___features
#include <string.h>
#include <stdio.h>
#include "cpl_conv.h"
#include "cpl_error.h"
#include "cpl_string.h"
#include "cpl_vsi.h"
#include "ogr_srs_api.h"
#include "gdal.h"
#include "ogr_api.h"
#include "gdalwarper.h"
#include "gdal_alg.h"
#include "ogr_core.h"
#include <stdlib.h>
#include "numpy/arrayobject.h"
#include "numpy/ufuncobject.h"
#ifdef _OPENMP
#include <omp.h>
#endif /* _OPENMP */

#ifdef PYREX_WITHOUT_ASSERTIONS
#define CYTHON_WITHOUT_ASSERTIONS
#endif

typedef struct {PyObject **p; const char *s; const Py_ssize_t n; const char* encoding;
                const char is_unicode; const char is_str; const char intern; } __Pyx_StringTabEntry;

#define __PYX_DEFAULT_STRING_ENCODING_IS_ASCII 0
#define __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT 0
#define __PYX_DEFAULT_STRING_ENCODING ""
#define __Pyx_PyObject_FromString __Pyx_PyBytes_FromString
#define __Pyx_PyObject_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#define __Pyx_uchar_cast(c) ((unsigned char)c)
#define __Pyx_long_cast(x) ((long)x)
#define __Pyx_fits_Py_ssize_t(v, type, is_signed)  (\
    (sizeof(type) < sizeof(Py_ssize_t))  ||\
    (sizeof(type) > sizeof(Py_ssize_t) &&\
          likely(v < (type)PY_SSIZE_T_MAX ||\
                 v == (type)PY_SSIZE_T_MAX)  &&\
          (!is_signed || likely(v > (type)PY_SSIZE_T_MIN ||\
                                v == (type)PY_SSIZE_T_MIN)))  ||\
    (sizeof(type) == sizeof(Py_ssize_t) &&\
          (is_signed || likely(v < (type)PY_SSIZE_T_MAX ||\
                               v == (type)PY_SSIZE_T_MAX)))  )
#if defined (__cplusplus) && __cplusplus >= 201103L
    #include <cstdlib>
    #define __Pyx_sst_abs(value) std::abs(value)
#elif SIZEOF_INT >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) abs(value)
#elif SIZEOF_LONG >= SIZEOF_SIZE_T
    #define __Pyx_sst_abs(value) labs(value)
#elif defined (_MSC_VER) && defined (_M_X64)
    #define __Pyx_sst_abs(value) _abs64(value)
#elif defined (__STDC_VERSION__) && __STDC_VERSION__ >= 199901L
    #define __Pyx_sst_abs(value) llabs(value)
#elif defined (__GNUC__)
    #define __Pyx_sst_abs(value) __builtin_llabs(value)
#else
    #define __Pyx_sst_abs(value) ((value<0) ? -value : value)
#endif
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject*);
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject*, Py_ssize_t* length);
#define __Pyx_PyByteArray_FromString(s) PyByteArray_FromStringAndSize((const char*)s, strlen((const char*)s))
#define __Pyx_PyByteArray_FromStringAndSize(s, l) PyByteArray_FromStringAndSize((const char*)s, l)
#define __Pyx_PyBytes_FromString        PyBytes_FromString
#define __Pyx_PyBytes_FromStringAndSize PyBytes_FromStringAndSize
static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char*);
#if PY_MAJOR_VERSION < 3
    #define __Pyx_PyStr_FromString        __Pyx_PyBytes_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyBytes_FromStringAndSize
#else
    #define __Pyx_PyStr_FromString        __Pyx_PyUnicode_FromString
    #define __Pyx_PyStr_FromStringAndSize __Pyx_PyUnicode_FromStringAndSize
#endif
#define __Pyx_PyObject_AsSString(s)    ((signed char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_AsUString(s)    ((unsigned char*) __Pyx_PyObject_AsString(s))
#define __Pyx_PyObject_FromCString(s)  __Pyx_PyObject_FromString((const char*)s)
#define __Pyx_PyBytes_FromCString(s)   __Pyx_PyBytes_FromString((const char*)s)
#define __Pyx_PyByteArray_FromCString(s)   __Pyx_PyByteArray_FromString((const char*)s)
#define __Pyx_PyStr_FromCString(s)     __Pyx_PyStr_FromString((const char*)s)
#define __Pyx_PyUnicode_FromCString(s) __Pyx_PyUnicode_FromString((const char*)s)
#if PY_MAJOR_VERSION < 3
static CYTHON_INLINE size_t __Pyx_Py_UNICODE_strlen(const Py_UNICODE *u)
{
    const Py_UNICODE *u_end = u;
    while (*u_end++) ;
    return (size_t)(u_end - u - 1);
}
#else
#define __Pyx_Py_UNICODE_strlen Py_UNICODE_strlen
#endif
#define __Pyx_PyUnicode_FromUnicode(u)       PyUnicode_FromUnicode(u, __Pyx_Py_UNICODE_strlen(u))
#define __Pyx_PyUnicode_FromUnicodeAndLength PyUnicode_FromUnicode
#define __Pyx_PyUnicode_AsUnicode            PyUnicode_AsUnicode
#define __Pyx_NewRef(obj) (Py_INCREF(obj), obj)
#define __Pyx_Owned_Py_None(b) __Pyx_NewRef(Py_None)
#define __Pyx_PyBool_FromLong(b) ((b) ? __Pyx_NewRef(Py_True) : __Pyx_NewRef(Py_False))
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject*);
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x);
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject*);
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t);
#if CYTHON_ASSUME_SAFE_MACROS
#define __pyx_PyFloat_AsDouble(x) (PyFloat_CheckExact(x) ? PyFloat_AS_DOUBLE(x) : PyFloat_AsDouble(x))
#else
#define __pyx_PyFloat_AsDouble(x) PyFloat_AsDouble(x)
#endif
#define __pyx_PyFloat_AsFloat(x) ((float) __pyx_PyFloat_AsDouble(x))
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyNumber_Int(x) (PyLong_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Long(x))
#else
#define __Pyx_PyNumber_Int(x) (PyInt_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Int(x))
#endif
#define __Pyx_PyNumber_Float(x) (PyFloat_CheckExact(x) ? __Pyx_NewRef(x) : PyNumber_Float(x))
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
static int __Pyx_sys_getdefaultencoding_not_ascii;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    PyObject* ascii_chars_u = NULL;
    PyObject* ascii_chars_b = NULL;
    const char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    if (strcmp(default_encoding_c, "ascii") == 0) {
        __Pyx_sys_getdefaultencoding_not_ascii = 0;
    } else {
        char ascii_chars[128];
        int c;
        for (c = 0; c < 128; c++) {
            ascii_chars[c] = c;
        }
        __Pyx_sys_getdefaultencoding_not_ascii = 1;
        ascii_chars_u = PyUnicode_DecodeASCII(ascii_chars, 128, NULL);
        if (!ascii_chars_u) goto bad;
        ascii_chars_b = PyUnicode_AsEncodedString(ascii_chars_u, default_encoding_c, NULL);
        if (!ascii_chars_b || !PyBytes_Check(ascii_chars_b) || memcmp(ascii_chars, PyBytes_AS_STRING(ascii_chars_b), 128) != 0) {
            PyErr_Format(
                PyExc_ValueError,
                "This module compiled with c_string_encoding=ascii, but default encoding '%.200s' is not a superset of ascii.",
                default_encoding_c);
            goto bad;
        }
        Py_DECREF(ascii_chars_u);
        Py_DECREF(ascii_chars_b);
    }
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    Py_XDECREF(ascii_chars_u);
    Py_XDECREF(ascii_chars_b);
    return -1;
}
#endif
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT && PY_MAJOR_VERSION >= 3
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_DecodeUTF8(c_str, size, NULL)
#else
#define __Pyx_PyUnicode_FromStringAndSize(c_str, size) PyUnicode_Decode(c_str, size, __PYX_DEFAULT_STRING_ENCODING, NULL)
#if __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT
static char* __PYX_DEFAULT_STRING_ENCODING;
static int __Pyx_init_sys_getdefaultencoding_params(void) {
    PyObject* sys;
    PyObject* default_encoding = NULL;
    char* default_encoding_c;
    sys = PyImport_ImportModule("sys");
    if (!sys) goto bad;
    default_encoding = PyObject_CallMethod(sys, (char*) (const char*) "getdefaultencoding", NULL);
    Py_DECREF(sys);
    if (!default_encoding) goto bad;
    default_encoding_c = PyBytes_AsString(default_encoding);
    if (!default_encoding_c) goto bad;
    __PYX_DEFAULT_STRING_ENCODING = (char*) malloc(strlen(default_encoding_c));
    if (!__PYX_DEFAULT_STRING_ENCODING) goto bad;
    strcpy(__PYX_DEFAULT_STRING_ENCODING, default_encoding_c);
    Py_DECREF(default_encoding);
    return 0;
bad:
    Py_XDECREF(default_encoding);
    return -1;
}
#endif
#endif


/* Test for GCC > 2.95 */
#if defined(__GNUC__)     && (__GNUC__ > 2 || (__GNUC__ == 2 && (__GNUC_MINOR__ > 95)))
  #define likely(x)   __builtin_expect(!!(x), 1)
  #define unlikely(x) __builtin_expect(!!(x), 0)
#else /* !__GNUC__ or GCC < 2.95 */
  #define likely(x)   (x)
  #define unlikely(x) (x)
#endif /* __GNUC__ */

static PyObject *__pyx_m;
static PyObject *__pyx_d;
static PyObject *__pyx_b;
static PyObject *__pyx_empty_tuple;
static PyObject *__pyx_empty_bytes;
static PyObject *__pyx_empty_unicode;
static int __pyx_lineno;
static int __pyx_clineno = 0;
static const char * __pyx_cfilenm= __FILE__;
static const char *__pyx_filename;

/* Header.proto */
#if !defined(CYTHON_CCOMPLEX)
  #if defined(__cplusplus)
    #define CYTHON_CCOMPLEX 1
  #elif defined(_Complex_I)
    #define CYTHON_CCOMPLEX 1
  #else
    #define CYTHON_CCOMPLEX 0
  #endif
#endif
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #include <complex>
  #else
    #include <complex.h>
  #endif
#endif
#if CYTHON_CCOMPLEX && !defined(__cplusplus) && defined(__sun__) && defined(__GNUC__)
  #undef _Complex_I
  #define _Complex_I 1.0fj
#endif


static const char *__pyx_f[] = {
  "rasterio/_features.pyx",
  "__init__.pxd",
  "type.pxd",
  "rasterio/_base.pxd",
  "rasterio/_io.pxd",
};

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":725
 * # in Cython to enable them only on the right systems.
 * 
 * ctypedef npy_int8       int8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 */
typedef npy_int8 __pyx_t_5numpy_int8_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":726
 * 
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t
 */
typedef npy_int16 __pyx_t_5numpy_int16_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":727
 * ctypedef npy_int8       int8_t
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_int64      int64_t
 * #ctypedef npy_int96      int96_t
 */
typedef npy_int32 __pyx_t_5numpy_int32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":728
 * ctypedef npy_int16      int16_t
 * ctypedef npy_int32      int32_t
 * ctypedef npy_int64      int64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_int96      int96_t
 * #ctypedef npy_int128     int128_t
 */
typedef npy_int64 __pyx_t_5numpy_int64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":732
 * #ctypedef npy_int128     int128_t
 * 
 * ctypedef npy_uint8      uint8_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 */
typedef npy_uint8 __pyx_t_5numpy_uint8_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":733
 * 
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t
 */
typedef npy_uint16 __pyx_t_5numpy_uint16_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":734
 * ctypedef npy_uint8      uint8_t
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uint64     uint64_t
 * #ctypedef npy_uint96     uint96_t
 */
typedef npy_uint32 __pyx_t_5numpy_uint32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":735
 * ctypedef npy_uint16     uint16_t
 * ctypedef npy_uint32     uint32_t
 * ctypedef npy_uint64     uint64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_uint96     uint96_t
 * #ctypedef npy_uint128    uint128_t
 */
typedef npy_uint64 __pyx_t_5numpy_uint64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":739
 * #ctypedef npy_uint128    uint128_t
 * 
 * ctypedef npy_float32    float32_t             # <<<<<<<<<<<<<<
 * ctypedef npy_float64    float64_t
 * #ctypedef npy_float80    float80_t
 */
typedef npy_float32 __pyx_t_5numpy_float32_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":740
 * 
 * ctypedef npy_float32    float32_t
 * ctypedef npy_float64    float64_t             # <<<<<<<<<<<<<<
 * #ctypedef npy_float80    float80_t
 * #ctypedef npy_float128   float128_t
 */
typedef npy_float64 __pyx_t_5numpy_float64_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":749
 * # The int types are mapped a bit surprising --
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t
 */
typedef npy_long __pyx_t_5numpy_int_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":750
 * # numpy.int corresponds to 'l' and numpy.long to 'q'
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longlong   longlong_t
 * 
 */
typedef npy_longlong __pyx_t_5numpy_long_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":751
 * ctypedef npy_long       int_t
 * ctypedef npy_longlong   long_t
 * ctypedef npy_longlong   longlong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_ulong      uint_t
 */
typedef npy_longlong __pyx_t_5numpy_longlong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":753
 * ctypedef npy_longlong   longlong_t
 * 
 * ctypedef npy_ulong      uint_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t
 */
typedef npy_ulong __pyx_t_5numpy_uint_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":754
 * 
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t             # <<<<<<<<<<<<<<
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 */
typedef npy_ulonglong __pyx_t_5numpy_ulong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":755
 * ctypedef npy_ulong      uint_t
 * ctypedef npy_ulonglong  ulong_t
 * ctypedef npy_ulonglong  ulonglong_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_intp       intp_t
 */
typedef npy_ulonglong __pyx_t_5numpy_ulonglong_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":757
 * ctypedef npy_ulonglong  ulonglong_t
 * 
 * ctypedef npy_intp       intp_t             # <<<<<<<<<<<<<<
 * ctypedef npy_uintp      uintp_t
 * 
 */
typedef npy_intp __pyx_t_5numpy_intp_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":758
 * 
 * ctypedef npy_intp       intp_t
 * ctypedef npy_uintp      uintp_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_double     float_t
 */
typedef npy_uintp __pyx_t_5numpy_uintp_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":760
 * ctypedef npy_uintp      uintp_t
 * 
 * ctypedef npy_double     float_t             # <<<<<<<<<<<<<<
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t
 */
typedef npy_double __pyx_t_5numpy_float_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":761
 * 
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t             # <<<<<<<<<<<<<<
 * ctypedef npy_longdouble longdouble_t
 * 
 */
typedef npy_double __pyx_t_5numpy_double_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":762
 * ctypedef npy_double     float_t
 * ctypedef npy_double     double_t
 * ctypedef npy_longdouble longdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cfloat      cfloat_t
 */
typedef npy_longdouble __pyx_t_5numpy_longdouble_t;

/* "rasterio/_io.pxd":33
 * 
 * 
 * ctypedef np.uint8_t DTYPE_UBYTE_t             # <<<<<<<<<<<<<<
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t
 */
typedef __pyx_t_5numpy_uint8_t __pyx_t_8rasterio_3_io_DTYPE_UBYTE_t;

/* "rasterio/_io.pxd":34
 * 
 * ctypedef np.uint8_t DTYPE_UBYTE_t
 * ctypedef np.uint16_t DTYPE_UINT16_t             # <<<<<<<<<<<<<<
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t
 */
typedef __pyx_t_5numpy_uint16_t __pyx_t_8rasterio_3_io_DTYPE_UINT16_t;

/* "rasterio/_io.pxd":35
 * ctypedef np.uint8_t DTYPE_UBYTE_t
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t             # <<<<<<<<<<<<<<
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t
 */
typedef __pyx_t_5numpy_int16_t __pyx_t_8rasterio_3_io_DTYPE_INT16_t;

/* "rasterio/_io.pxd":36
 * ctypedef np.uint16_t DTYPE_UINT16_t
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 */
typedef __pyx_t_5numpy_uint32_t __pyx_t_8rasterio_3_io_DTYPE_UINT32_t;

/* "rasterio/_io.pxd":37
 * ctypedef np.int16_t DTYPE_INT16_t
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 * ctypedef np.float64_t DTYPE_FLOAT64_t
 */
typedef __pyx_t_5numpy_int32_t __pyx_t_8rasterio_3_io_DTYPE_INT32_t;

/* "rasterio/_io.pxd":38
 * ctypedef np.uint32_t DTYPE_UINT32_t
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t             # <<<<<<<<<<<<<<
 * ctypedef np.float64_t DTYPE_FLOAT64_t
 * 
 */
typedef __pyx_t_5numpy_float32_t __pyx_t_8rasterio_3_io_DTYPE_FLOAT32_t;

/* "rasterio/_io.pxd":39
 * ctypedef np.int32_t DTYPE_INT32_t
 * ctypedef np.float32_t DTYPE_FLOAT32_t
 * ctypedef np.float64_t DTYPE_FLOAT64_t             # <<<<<<<<<<<<<<
 * 
 * 
 */
typedef __pyx_t_5numpy_float64_t __pyx_t_8rasterio_3_io_DTYPE_FLOAT64_t;
/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< float > __pyx_t_float_complex;
  #else
    typedef float _Complex __pyx_t_float_complex;
  #endif
#else
    typedef struct { float real, imag; } __pyx_t_float_complex;
#endif
static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float, float);

/* Declarations.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    typedef ::std::complex< double > __pyx_t_double_complex;
  #else
    typedef double _Complex __pyx_t_double_complex;
  #endif
#else
    typedef struct { double real, imag; } __pyx_t_double_complex;
#endif
static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double, double);


/*--- Type declarations ---*/
struct __pyx_obj_8rasterio_5_base_DatasetBase;
struct __pyx_obj_8rasterio_3_io_DatasetReaderBase;
struct __pyx_obj_8rasterio_3_io_DatasetWriterBase;
struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase;
struct __pyx_obj_8rasterio_3_io_InMemoryRaster;
struct __pyx_obj_8rasterio_9_features_GeomBuilder;
struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder;
struct __pyx_obj_8rasterio_9_features_ShapeIterator;
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes;
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve;
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr;
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode;
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":764
 * ctypedef npy_longdouble longdouble_t
 * 
 * ctypedef npy_cfloat      cfloat_t             # <<<<<<<<<<<<<<
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t
 */
typedef npy_cfloat __pyx_t_5numpy_cfloat_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":765
 * 
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t             # <<<<<<<<<<<<<<
 * ctypedef npy_clongdouble clongdouble_t
 * 
 */
typedef npy_cdouble __pyx_t_5numpy_cdouble_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":766
 * ctypedef npy_cfloat      cfloat_t
 * ctypedef npy_cdouble     cdouble_t
 * ctypedef npy_clongdouble clongdouble_t             # <<<<<<<<<<<<<<
 * 
 * ctypedef npy_cdouble     complex_t
 */
typedef npy_clongdouble __pyx_t_5numpy_clongdouble_t;

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":768
 * ctypedef npy_clongdouble clongdouble_t
 * 
 * ctypedef npy_cdouble     complex_t             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 */
typedef npy_cdouble __pyx_t_5numpy_complex_t;
struct __pyx_opt_args_8rasterio_3_io_io_auto;

/* "rasterio/_io.pxd":44
 * cdef bint in_dtype_range(value, dtype)
 * 
 * cdef int io_auto(image, GDALRasterBandH band, bint write, int resampling=*)             # <<<<<<<<<<<<<<
 */
struct __pyx_opt_args_8rasterio_3_io_io_auto {
  int __pyx_n;
  int resampling;
};

/* "rasterio/_base.pxd":6
 * 
 * 
 * cdef class DatasetBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef GDALDatasetH _hds
 */
struct __pyx_obj_8rasterio_5_base_DatasetBase {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase *__pyx_vtab;
  GDALDatasetH _hds;
  PyObject *name;
  PyObject *mode;
  PyObject *options;
  PyObject *width;
  PyObject *height;
  PyObject *shape;
  PyObject *driver;
  PyObject *_count;
  PyObject *_dtypes;
  PyObject *_closed;
  PyObject *_crs;
  PyObject *_crs_wkt;
  PyObject *_transform;
  PyObject *_block_shapes;
  PyObject *_nodatavals;
  PyObject *_units;
  PyObject *_descriptions;
  PyObject *_read;
  PyObject *_gcps;
};


/* "rasterio/_io.pxd":8
 * 
 * 
 * cdef class DatasetReaderBase(DatasetBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8rasterio_3_io_DatasetReaderBase {
  struct __pyx_obj_8rasterio_5_base_DatasetBase __pyx_base;
};


/* "rasterio/_io.pxd":12
 * 
 * 
 * cdef class DatasetWriterBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef readonly object _init_dtype
 *     cdef readonly object _init_nodata
 */
struct __pyx_obj_8rasterio_3_io_DatasetWriterBase {
  struct __pyx_obj_8rasterio_3_io_DatasetReaderBase __pyx_base;
  PyObject *_init_dtype;
  PyObject *_init_nodata;
  PyObject *_options;
};


/* "rasterio/_io.pxd":18
 * 
 * 
 * cdef class BufferedDatasetWriterBase(DatasetWriterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */
struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase {
  struct __pyx_obj_8rasterio_3_io_DatasetWriterBase __pyx_base;
};


/* "rasterio/_io.pxd":22
 * 
 * 
 * cdef class InMemoryRaster:             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds
 *     cdef double transform[6]
 */
struct __pyx_obj_8rasterio_3_io_InMemoryRaster {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *__pyx_vtab;
  GDALDatasetH _hds;
  double transform[6];
  int band_ids[1];
  PyArrayObject *_image;
  PyObject *crs;
};


/* "rasterio/_features.pxd":6
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH geom
 */
struct __pyx_obj_8rasterio_9_features_GeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *__pyx_vtab;
  OGRGeometryH geom;
  PyObject *code;
  PyObject *geomtypename;
  PyObject *ndims;
};


/* "rasterio/_features.pxd":22
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL
 */
struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder {
  PyObject_HEAD
  struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *__pyx_vtab;
};


/* "rasterio/_features.pxd":37
 * 
 * 
 * cdef class ShapeIterator:             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRLayerH layer
 */
struct __pyx_obj_8rasterio_9_features_ShapeIterator {
  PyObject_HEAD
  OGRLayerH layer;
  int fieldtype;
};


/* "rasterio/_features.pyx":20
 * 
 * 
 * def _shapes(image, mask, connectivity, transform):             # <<<<<<<<<<<<<<
 *     """
 *     Return a generator of (polygon, value) for each each set of adjacent pixels
 */
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes {
  PyObject_HEAD
  GDALRasterBandH __pyx_v_band;
  int __pyx_v_cols;
  PyObject *__pyx_v_connectivity;
  GDALDriverH __pyx_v_driver;
  OGRFieldDefnH __pyx_v_fielddefn;
  int __pyx_v_fieldtp;
  OGRDataSourceH __pyx_v_fs;
  PyObject *__pyx_v_image;
  int __pyx_v_is_float;
  OGRLayerH __pyx_v_layer;
  PyObject *__pyx_v_mask;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_mask_ds;
  GDALRasterBandH __pyx_v_maskband;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_mem_ds;
  PyObject *__pyx_v_mrdr;
  char **__pyx_v_options;
  PyObject *__pyx_v_rdr;
  int __pyx_v_retval;
  int __pyx_v_rows;
  PyObject *__pyx_v_s;
  struct __pyx_obj_8rasterio_9_features_ShapeIterator *__pyx_v_shape_iter;
  PyObject *__pyx_v_transform;
  PyObject *__pyx_v_v;
  PyObject *__pyx_v_valid_dtypes;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
};


/* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve {
  PyObject_HEAD
  PyObject *__pyx_v_valid_dtypes;
};


/* "rasterio/_features.pyx":182
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t             # <<<<<<<<<<<<<<
 *                                      in valid_dtypes))
 *         raise ValueError(
 */
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr {
  PyObject_HEAD
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *__pyx_outer_scope;
  PyObject *__pyx_v_t;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
};


/* "rasterio/_features.pyx":321
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode {
  PyObject_HEAD
  PyObject *__pyx_v_coords;
  PyObject *__pyx_v_e;
  PyObject *__pyx_v_f;
  PyObject *__pyx_t_0;
  Py_ssize_t __pyx_t_1;
  PyObject *(*__pyx_t_2)(PyObject *);
  PyObject *__pyx_t_3;
  Py_ssize_t __pyx_t_4;
  PyObject *(*__pyx_t_5)(PyObject *);
};


/* "rasterio/_features.pyx":379
 * # Mapping of GeoJSON type names to OGR integer geometry types
 * GEOJSON2OGR_GEOMETRY_TYPES = dict(
 *     (v, k) for k, v in GEOMETRY_TYPES.iteritems()             # <<<<<<<<<<<<<<
 * )
 * 
 */
struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr {
  PyObject_HEAD
  PyObject *__pyx_v_k;
  PyObject *__pyx_v_v;
};



/* "rasterio/_base.pxd":6
 * 
 * 
 * cdef class DatasetBase:             # <<<<<<<<<<<<<<
 * 
 *     cdef GDALDatasetH _hds
 */

struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase {
  GDALDatasetH (*handle)(struct __pyx_obj_8rasterio_5_base_DatasetBase *);
  GDALRasterBandH (*band)(struct __pyx_obj_8rasterio_5_base_DatasetBase *, int);
};
static struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase *__pyx_vtabptr_8rasterio_5_base_DatasetBase;


/* "rasterio/_io.pxd":8
 * 
 * 
 * cdef class DatasetReaderBase(DatasetBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase {
  struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *__pyx_vtabptr_8rasterio_3_io_DatasetReaderBase;


/* "rasterio/_io.pxd":12
 * 
 * 
 * cdef class DatasetWriterBase(DatasetReaderBase):             # <<<<<<<<<<<<<<
 *     cdef readonly object _init_dtype
 *     cdef readonly object _init_nodata
 */

struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase *__pyx_vtabptr_8rasterio_3_io_DatasetWriterBase;


/* "rasterio/_io.pxd":18
 * 
 * 
 * cdef class BufferedDatasetWriterBase(DatasetWriterBase):             # <<<<<<<<<<<<<<
 *     pass
 * 
 */

struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase {
  struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase __pyx_base;
};
static struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase *__pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase;


/* "rasterio/_io.pxd":22
 * 
 * 
 * cdef class InMemoryRaster:             # <<<<<<<<<<<<<<
 *     cdef GDALDatasetH _hds
 *     cdef double transform[6]
 */

struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster {
  GDALDatasetH (*handle)(struct __pyx_obj_8rasterio_3_io_InMemoryRaster *);
  GDALRasterBandH (*band)(struct __pyx_obj_8rasterio_3_io_InMemoryRaster *, int);
};
static struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *__pyx_vtabptr_8rasterio_3_io_InMemoryRaster;


/* "rasterio/_features.pyx":393
 * 
 * 
 * cdef class GeomBuilder:             # <<<<<<<<<<<<<<
 *     """Builds a GeoJSON (Fiona-style) geometry from an OGR geometry."""
 * 
 */

struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder {
  PyObject *(*_buildCoords)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
  PyObject *(*_buildPoint)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLineString)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildLinearRing)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildParts)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
  PyObject *(*_buildPolygon)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*_buildMultiPolygon)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch);
  PyObject *(*build)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH);
};
static struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *__pyx_vtabptr_8rasterio_9_features_GeomBuilder;


/* "rasterio/_features.pyx":457
 * 
 * 
 * cdef class OGRGeomBuilder:             # <<<<<<<<<<<<<<
 *     """
 *     Builds an OGR geometry from GeoJSON geometry.
 */

struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder {
  OGRGeometryH (*_createOgrGeometry)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, int);
  PyObject *(*_addPointToGeometry)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, OGRGeometryH, PyObject *);
  OGRGeometryH (*_buildPoint)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildLineString)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildLinearRing)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildPolygon)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiPoint)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiLineString)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildMultiPolygon)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*_buildGeometryCollection)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
  OGRGeometryH (*build)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *);
};
static struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *__pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder;

/* --- Runtime support code (head) --- */
/* Refnanny.proto */
#ifndef CYTHON_REFNANNY
  #define CYTHON_REFNANNY 0
#endif
#if CYTHON_REFNANNY
  typedef struct {
    void (*INCREF)(void*, PyObject*, int);
    void (*DECREF)(void*, PyObject*, int);
    void (*GOTREF)(void*, PyObject*, int);
    void (*GIVEREF)(void*, PyObject*, int);
    void* (*SetupContext)(const char*, int, const char*);
    void (*FinishContext)(void**);
  } __Pyx_RefNannyAPIStruct;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNanny = NULL;
  static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname);
  #define __Pyx_RefNannyDeclarations void *__pyx_refnanny = NULL;
#ifdef WITH_THREAD
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          if (acquire_gil) {\
              PyGILState_STATE __pyx_gilstate_save = PyGILState_Ensure();\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
              PyGILState_Release(__pyx_gilstate_save);\
          } else {\
              __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__);\
          }
#else
  #define __Pyx_RefNannySetupContext(name, acquire_gil)\
          __pyx_refnanny = __Pyx_RefNanny->SetupContext((name), __LINE__, __FILE__)
#endif
  #define __Pyx_RefNannyFinishContext()\
          __Pyx_RefNanny->FinishContext(&__pyx_refnanny)
  #define __Pyx_INCREF(r)  __Pyx_RefNanny->INCREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_DECREF(r)  __Pyx_RefNanny->DECREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GOTREF(r)  __Pyx_RefNanny->GOTREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_GIVEREF(r) __Pyx_RefNanny->GIVEREF(__pyx_refnanny, (PyObject *)(r), __LINE__)
  #define __Pyx_XINCREF(r)  do { if((r) != NULL) {__Pyx_INCREF(r); }} while(0)
  #define __Pyx_XDECREF(r)  do { if((r) != NULL) {__Pyx_DECREF(r); }} while(0)
  #define __Pyx_XGOTREF(r)  do { if((r) != NULL) {__Pyx_GOTREF(r); }} while(0)
  #define __Pyx_XGIVEREF(r) do { if((r) != NULL) {__Pyx_GIVEREF(r);}} while(0)
#else
  #define __Pyx_RefNannyDeclarations
  #define __Pyx_RefNannySetupContext(name, acquire_gil)
  #define __Pyx_RefNannyFinishContext()
  #define __Pyx_INCREF(r) Py_INCREF(r)
  #define __Pyx_DECREF(r) Py_DECREF(r)
  #define __Pyx_GOTREF(r)
  #define __Pyx_GIVEREF(r)
  #define __Pyx_XINCREF(r) Py_XINCREF(r)
  #define __Pyx_XDECREF(r) Py_XDECREF(r)
  #define __Pyx_XGOTREF(r)
  #define __Pyx_XGIVEREF(r)
#endif
#define __Pyx_XDECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_XDECREF(tmp);\
    } while (0)
#define __Pyx_DECREF_SET(r, v) do {\
        PyObject *tmp = (PyObject *) r;\
        r = v; __Pyx_DECREF(tmp);\
    } while (0)
#define __Pyx_CLEAR(r)    do { PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);} while(0)
#define __Pyx_XCLEAR(r)   do { if((r) != NULL) {PyObject* tmp = ((PyObject*)(r)); r = NULL; __Pyx_DECREF(tmp);}} while(0)

/* PyObjectGetAttrStr.proto */
#if CYTHON_USE_TYPE_SLOTS
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetAttrStr(PyObject* obj, PyObject* attr_name) {
    PyTypeObject* tp = Py_TYPE(obj);
    if (likely(tp->tp_getattro))
        return tp->tp_getattro(obj, attr_name);
#if PY_MAJOR_VERSION < 3
    if (likely(tp->tp_getattr))
        return tp->tp_getattr(obj, PyString_AS_STRING(attr_name));
#endif
    return PyObject_GetAttr(obj, attr_name);
}
#else
#define __Pyx_PyObject_GetAttrStr(o,n) PyObject_GetAttr(o,n)
#endif

/* GetBuiltinName.proto */
static PyObject *__Pyx_GetBuiltinName(PyObject *name);

/* IterFinish.proto */
static CYTHON_INLINE int __Pyx_IterFinish(void);

/* PyFunctionFastCall.proto */
#if CYTHON_FAST_PYCALL
#define __Pyx_PyFunction_FastCall(func, args, nargs)\
    __Pyx_PyFunction_FastCallDict((func), (args), (nargs), NULL)
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs);
#else
#define __Pyx_PyFunction_FastCallDict(func, args, nargs, kwargs) _PyFunction_FastCallDict(func, args, nargs, kwargs)
#endif
#endif

/* PyObjectCall.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw);
#else
#define __Pyx_PyObject_Call(func, arg, kw) PyObject_Call(func, arg, kw)
#endif

/* PyObjectCallMethO.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg);
#endif

/* PyObjectCallNoArg.proto */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func);
#else
#define __Pyx_PyObject_CallNoArg(func) __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL)
#endif

/* PyCFunctionFastCall.proto */
#if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject *__Pyx_PyCFunction_FastCall(PyObject *func, PyObject **args, Py_ssize_t nargs);
#else
#define __Pyx_PyCFunction_FastCall(func, args, nargs)  (assert(0), NULL)
#endif

/* PyObjectCallOneArg.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg);

/* PyObjectCallMethod0.proto */
static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name);

/* RaiseNeedMoreValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index);

/* RaiseTooManyValuesToUnpack.proto */
static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected);

/* UnpackItemEndCheck.proto */
static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected);

/* RaiseNoneIterError.proto */
static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void);

/* UnpackTupleError.proto */
static void __Pyx_UnpackTupleError(PyObject *, Py_ssize_t index);

/* UnpackTuple2.proto */
static CYTHON_INLINE int __Pyx_unpack_tuple2(PyObject* tuple, PyObject** value1, PyObject** value2,
                                             int is_tuple, int has_known_size, int decref_tuple);

/* dict_iter.proto */
static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* dict, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_is_dict);
static CYTHON_INLINE int __Pyx_dict_iter_next(PyObject* dict_or_iter, Py_ssize_t orig_length, Py_ssize_t* ppos,
                                              PyObject** pkey, PyObject** pvalue, PyObject** pitem, int is_dict);

/* GetModuleGlobalName.proto */
static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name);

/* RaiseArgTupleInvalid.proto */
static void __Pyx_RaiseArgtupleInvalid(const char* func_name, int exact,
    Py_ssize_t num_min, Py_ssize_t num_max, Py_ssize_t num_found);

/* RaiseDoubleKeywords.proto */
static void __Pyx_RaiseDoubleKeywordsError(const char* func_name, PyObject* kw_name);

/* ParseKeywords.proto */
static int __Pyx_ParseOptionalKeywords(PyObject *kwds, PyObject **argnames[],\
    PyObject *kwds2, PyObject *values[], Py_ssize_t num_pos_args,\
    const char* function_name);

/* PySequenceContains.proto */
static CYTHON_INLINE int __Pyx_PySequence_ContainsTF(PyObject* item, PyObject* seq, int eq) {
    int result = PySequence_Contains(seq, item);
    return unlikely(result < 0) ? result : (result == (eq == Py_EQ));
}

/* StringJoin.proto */
#if PY_MAJOR_VERSION < 3
#define __Pyx_PyString_Join __Pyx_PyBytes_Join
#define __Pyx_PyBaseString_Join(s, v) (PyUnicode_CheckExact(s) ? PyUnicode_Join(s, v) : __Pyx_PyBytes_Join(s, v))
#else
#define __Pyx_PyString_Join PyUnicode_Join
#define __Pyx_PyBaseString_Join PyUnicode_Join
#endif
#if CYTHON_COMPILING_IN_CPYTHON
    #if PY_MAJOR_VERSION < 3
    #define __Pyx_PyBytes_Join _PyString_Join
    #else
    #define __Pyx_PyBytes_Join _PyBytes_Join
    #endif
#else
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values);
#endif

/* PyThreadStateGet.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyThreadState_declare  PyThreadState *__pyx_tstate;
#define __Pyx_PyThreadState_assign  __pyx_tstate = PyThreadState_GET();
#else
#define __Pyx_PyThreadState_declare
#define __Pyx_PyThreadState_assign
#endif

/* PyErrFetchRestore.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ErrRestoreWithState(type, value, tb)  __Pyx_ErrRestoreInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)    __Pyx_ErrFetchInState(PyThreadState_GET(), type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  __Pyx_ErrRestoreInState(__pyx_tstate, type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)    __Pyx_ErrFetchInState(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
#define __Pyx_ErrRestoreWithState(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetchWithState(type, value, tb)  PyErr_Fetch(type, value, tb)
#define __Pyx_ErrRestore(type, value, tb)  PyErr_Restore(type, value, tb)
#define __Pyx_ErrFetch(type, value, tb)  PyErr_Fetch(type, value, tb)
#endif

/* RaiseException.proto */
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause);

/* ExtTypeTest.proto */
static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type);

/* IncludeStringH.proto */
#include <string.h>

/* BytesEquals.proto */
static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals);

/* UnicodeEquals.proto */
static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals);

/* StrEquals.proto */
#if PY_MAJOR_VERSION >= 3
#define __Pyx_PyString_Equals __Pyx_PyUnicode_Equals
#else
#define __Pyx_PyString_Equals __Pyx_PyBytes_Equals
#endif

/* PyIntBinop.proto */
#if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, long intval, int inplace);
#else
#define __Pyx_PyInt_EqObjC(op1, op2, intval, inplace)\
    PyObject_RichCompare(op1, op2, Py_EQ)
    #endif

/* None.proto */
static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname);

/* GetItemInt.proto */
#define __Pyx_GetItemInt(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Fast(o, (Py_ssize_t)i, is_list, wraparound, boundscheck) :\
    (is_list ? (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL) :\
               __Pyx_GetItemInt_Generic(o, to_py_func(i))))
#define __Pyx_GetItemInt_List(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_List_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "list index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
#define __Pyx_GetItemInt_Tuple(o, i, type, is_signed, to_py_func, is_list, wraparound, boundscheck)\
    (__Pyx_fits_Py_ssize_t(i, type, is_signed) ?\
    __Pyx_GetItemInt_Tuple_Fast(o, (Py_ssize_t)i, wraparound, boundscheck) :\
    (PyErr_SetString(PyExc_IndexError, "tuple index out of range"), (PyObject*)NULL))
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              int wraparound, int boundscheck);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j);
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i,
                                                     int is_list, int wraparound, int boundscheck);

/* SaveResetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSave(type, value, tb)  __Pyx__ExceptionSave(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#define __Pyx_ExceptionReset(type, value, tb)  __Pyx__ExceptionReset(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb);
#else
#define __Pyx_ExceptionSave(type, value, tb)   PyErr_GetExcInfo(type, value, tb)
#define __Pyx_ExceptionReset(type, value, tb)  PyErr_SetExcInfo(type, value, tb)
#endif

/* GetException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_GetException(type, value, tb)  __Pyx__GetException(__pyx_tstate, type, value, tb)
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* PyObjectLookupSpecial.proto */
#if CYTHON_COMPILING_IN_CPYTHON && PY_VERSION_HEX >= 0x02070000
static CYTHON_INLINE PyObject* __Pyx_PyObject_LookupSpecial(PyObject* obj, PyObject* attr_name) {
    PyObject *res;
    PyTypeObject *tp = Py_TYPE(obj);
#if PY_MAJOR_VERSION < 3
    if (unlikely(PyInstance_Check(obj)))
        return __Pyx_PyObject_GetAttrStr(obj, attr_name);
#endif
    res = _PyType_Lookup(tp, attr_name);
    if (likely(res)) {
        descrgetfunc f = Py_TYPE(res)->tp_descr_get;
        if (!f) {
            Py_INCREF(res);
        } else {
            res = f(res, obj, (PyObject *)tp);
        }
    } else {
        PyErr_SetObject(PyExc_AttributeError, attr_name);
    }
    return res;
}
#else
#define __Pyx_PyObject_LookupSpecial(o,n) __Pyx_PyObject_GetAttrStr(o,n)
#endif

/* SwapException.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_ExceptionSwap(type, value, tb)  __Pyx__ExceptionSwap(__pyx_tstate, type, value, tb)
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb);
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb);
#endif

/* ListAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_PyList_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len) & likely(len > (L->allocated >> 1))) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_PyList_Append(L,x) PyList_Append(L,x)
#endif

/* ListCompAppend.proto */
#if CYTHON_USE_PYLIST_INTERNALS && CYTHON_ASSUME_SAFE_MACROS
static CYTHON_INLINE int __Pyx_ListComp_Append(PyObject* list, PyObject* x) {
    PyListObject* L = (PyListObject*) list;
    Py_ssize_t len = Py_SIZE(list);
    if (likely(L->allocated > len)) {
        Py_INCREF(x);
        PyList_SET_ITEM(list, len, x);
        Py_SIZE(list) = len+1;
        return 0;
    }
    return PyList_Append(list, x);
}
#else
#define __Pyx_ListComp_Append(L,x) PyList_Append(L,x)
#endif

/* GetAttr.proto */
static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *, PyObject *);

/* SliceObject.proto */
static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(
        PyObject* obj, Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** py_start, PyObject** py_stop, PyObject** py_slice,
        int has_cstart, int has_cstop, int wraparound);

/* DictGetItem.proto */
#if PY_MAJOR_VERSION >= 3 && !CYTHON_COMPILING_IN_PYPY
static PyObject *__Pyx_PyDict_GetItem(PyObject *d, PyObject* key) {
    PyObject *value;
    value = PyDict_GetItemWithError(d, key);
    if (unlikely(!value)) {
        if (!PyErr_Occurred()) {
            PyObject* args = PyTuple_Pack(1, key);
            if (likely(args))
                PyErr_SetObject(PyExc_KeyError, args);
            Py_XDECREF(args);
        }
        return NULL;
    }
    Py_INCREF(value);
    return value;
}
#else
    #define __Pyx_PyDict_GetItem(d, key) PyObject_GetItem(d, key)
#endif

/* PyErrExceptionMatches.proto */
#if CYTHON_FAST_THREAD_STATE
#define __Pyx_PyErr_ExceptionMatches(err) __Pyx_PyErr_ExceptionMatchesInState(__pyx_tstate, err)
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err);
#else
#define __Pyx_PyErr_ExceptionMatches(err)  PyErr_ExceptionMatches(err)
#endif

/* SetVTable.proto */
static int __Pyx_SetVtable(PyObject *dict, void *vtable);

/* GetVTable.proto */
static void* __Pyx_GetVtable(PyObject *dict);

/* Import.proto */
static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level);

/* ImportFrom.proto */
static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name);

/* CodeObjectCache.proto */
typedef struct {
    PyCodeObject* code_object;
    int code_line;
} __Pyx_CodeObjectCacheEntry;
struct __Pyx_CodeObjectCache {
    int count;
    int max_count;
    __Pyx_CodeObjectCacheEntry* entries;
};
static struct __Pyx_CodeObjectCache __pyx_code_cache = {0,0,NULL};
static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line);
static PyCodeObject *__pyx_find_code_object(int code_line);
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object);

/* AddTraceback.proto */
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value);

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value);

/* RealImag.proto */
#if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    #define __Pyx_CREAL(z) ((z).real())
    #define __Pyx_CIMAG(z) ((z).imag())
  #else
    #define __Pyx_CREAL(z) (__real__(z))
    #define __Pyx_CIMAG(z) (__imag__(z))
  #endif
#else
    #define __Pyx_CREAL(z) ((z).real)
    #define __Pyx_CIMAG(z) ((z).imag)
#endif
#if defined(__cplusplus) && CYTHON_CCOMPLEX\
        && (defined(_WIN32) || defined(__clang__) || (defined(__GNUC__) && (__GNUC__ >= 5 || __GNUC__ == 4 && __GNUC_MINOR__ >= 4 )) || __cplusplus >= 201103)
    #define __Pyx_SET_CREAL(z,x) ((z).real(x))
    #define __Pyx_SET_CIMAG(z,y) ((z).imag(y))
#else
    #define __Pyx_SET_CREAL(z,x) __Pyx_CREAL(z) = (x)
    #define __Pyx_SET_CIMAG(z,y) __Pyx_CIMAG(z) = (y)
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_float(a, b)   ((a)==(b))
    #define __Pyx_c_sum_float(a, b)  ((a)+(b))
    #define __Pyx_c_diff_float(a, b) ((a)-(b))
    #define __Pyx_c_prod_float(a, b) ((a)*(b))
    #define __Pyx_c_quot_float(a, b) ((a)/(b))
    #define __Pyx_c_neg_float(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_float(z) ((z)==(float)0)
    #define __Pyx_c_conj_float(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (::std::abs(z))
        #define __Pyx_c_pow_float(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_float(z) ((z)==0)
    #define __Pyx_c_conj_float(z)    (conjf(z))
    #if 1
        #define __Pyx_c_abs_float(z)     (cabsf(z))
        #define __Pyx_c_pow_float(a, b)  (cpowf(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex, __pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex);
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex);
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex);
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex, __pyx_t_float_complex);
    #endif
#endif

/* Arithmetic.proto */
#if CYTHON_CCOMPLEX
    #define __Pyx_c_eq_double(a, b)   ((a)==(b))
    #define __Pyx_c_sum_double(a, b)  ((a)+(b))
    #define __Pyx_c_diff_double(a, b) ((a)-(b))
    #define __Pyx_c_prod_double(a, b) ((a)*(b))
    #define __Pyx_c_quot_double(a, b) ((a)/(b))
    #define __Pyx_c_neg_double(a)     (-(a))
  #ifdef __cplusplus
    #define __Pyx_c_is_zero_double(z) ((z)==(double)0)
    #define __Pyx_c_conj_double(z)    (::std::conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (::std::abs(z))
        #define __Pyx_c_pow_double(a, b)  (::std::pow(a, b))
    #endif
  #else
    #define __Pyx_c_is_zero_double(z) ((z)==0)
    #define __Pyx_c_conj_double(z)    (conj(z))
    #if 1
        #define __Pyx_c_abs_double(z)     (cabs(z))
        #define __Pyx_c_pow_double(a, b)  (cpow(a, b))
    #endif
 #endif
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex, __pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex);
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex);
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex);
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex);
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex, __pyx_t_double_complex);
    #endif
#endif

/* CIntToPy.proto */
static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value);

/* CIntFromPy.proto */
static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *);

/* CIntFromPy.proto */
static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *);

/* FetchCommonType.proto */
static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type);

/* PyObjectCallMethod1.proto */
static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg);

/* CoroutineBase.proto */
typedef PyObject *(*__pyx_coroutine_body_t)(PyObject *, PyObject *);
typedef struct {
    PyObject_HEAD
    __pyx_coroutine_body_t body;
    PyObject *closure;
    PyObject *exc_type;
    PyObject *exc_value;
    PyObject *exc_traceback;
    PyObject *gi_weakreflist;
    PyObject *classobj;
    PyObject *yieldfrom;
    PyObject *gi_name;
    PyObject *gi_qualname;
    PyObject *gi_modulename;
    int resume_label;
    char is_running;
} __pyx_CoroutineObject;
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
    PyTypeObject *type, __pyx_coroutine_body_t body, PyObject *closure,
    PyObject *name, PyObject *qualname, PyObject *module_name);
static int __Pyx_Coroutine_clear(PyObject *self);
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue);
#else
#define __Pyx_PyGen_FetchStopIterationValue(pvalue) PyGen_FetchStopIterationValue(pvalue)
#endif

/* PatchModuleWithCoroutine.proto */
static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code);

/* PatchGeneratorABC.proto */
static int __Pyx_patch_abc(void);

/* Generator.proto */
#define __Pyx_Generator_USED
static PyTypeObject *__pyx_GeneratorType = 0;
#define __Pyx_Generator_CheckExact(obj) (Py_TYPE(obj) == __pyx_GeneratorType)
#define __Pyx_Generator_New(body, closure, name, qualname, module_name)\
    __Pyx__Coroutine_New(__pyx_GeneratorType, body, closure, name, qualname, module_name)
static PyObject *__Pyx_Generator_Next(PyObject *self);
static int __pyx_Generator_init(void);

/* CheckBinaryVersion.proto */
static int __Pyx_check_binary_version(void);

/* PyIdentifierFromString.proto */
#if !defined(__Pyx_PyIdentifier_FromString)
#if PY_MAJOR_VERSION < 3
  #define __Pyx_PyIdentifier_FromString(s) PyString_FromString(s)
#else
  #define __Pyx_PyIdentifier_FromString(s) PyUnicode_FromString(s)
#endif
#endif

/* ModuleImport.proto */
static PyObject *__Pyx_ImportModule(const char *name);

/* TypeImport.proto */
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name, size_t size, int strict);

/* FunctionImport.proto */
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig);

/* InitStrings.proto */
static int __Pyx_InitStrings(__Pyx_StringTabEntry *t);

static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildCoords(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildLineString(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildLinearRing(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildParts(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder_build(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__createOgrGeometry(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, int __pyx_v_geom_type); /* proto*/
static PyObject *__pyx_f_8rasterio_9_features_14OGRGeomBuilder__addPointToGeometry(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom, PyObject *__pyx_v_coordinate); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLineString(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLinearRing(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPolygon(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPoint(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiLineString(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildGeometryCollection(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates); /* proto*/
static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder_build(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_geometry); /* proto*/

/* Module declarations from 'libc.string' */

/* Module declarations from 'libc.stdio' */

/* Module declarations from 'cpython.buffer' */

/* Module declarations from '__builtin__' */

/* Module declarations from 'cpython.type' */
static PyTypeObject *__pyx_ptype_7cpython_4type_type = 0;

/* Module declarations from 'cpython' */

/* Module declarations from 'cpython.object' */

/* Module declarations from 'cpython.ref' */

/* Module declarations from 'libc.stdlib' */

/* Module declarations from 'numpy' */

/* Module declarations from 'numpy' */
static PyTypeObject *__pyx_ptype_5numpy_dtype = 0;
static PyTypeObject *__pyx_ptype_5numpy_flatiter = 0;
static PyTypeObject *__pyx_ptype_5numpy_broadcast = 0;
static PyTypeObject *__pyx_ptype_5numpy_ndarray = 0;
static PyTypeObject *__pyx_ptype_5numpy_ufunc = 0;
static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *, char *, char *, int *); /*proto*/

/* Module declarations from 'rasterio._err' */
static int (*__pyx_f_8rasterio_4_err_exc_wrap_int)(int); /*proto*/
static void *(*__pyx_f_8rasterio_4_err_exc_wrap_pointer)(void *); /*proto*/

/* Module declarations from 'rasterio._base' */
static PyTypeObject *__pyx_ptype_8rasterio_5_base_DatasetBase = 0;

/* Module declarations from 'rasterio._io' */
static PyTypeObject *__pyx_ptype_8rasterio_3_io_DatasetReaderBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_DatasetWriterBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase = 0;
static PyTypeObject *__pyx_ptype_8rasterio_3_io_InMemoryRaster = 0;
static int (*__pyx_f_8rasterio_3_io_io_auto)(PyObject *, GDALRasterBandH, int, struct __pyx_opt_args_8rasterio_3_io_io_auto *__pyx_optional_args); /*proto*/

/* Module declarations from 'rasterio._features' */
static PyTypeObject *__pyx_ptype_8rasterio_9_features_GeomBuilder = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features_OGRGeomBuilder = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features_ShapeIterator = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features___pyx_scope_struct___shapes = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features___pyx_scope_struct_1__sieve = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features___pyx_scope_struct_2_genexpr = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features___pyx_scope_struct_3__explode = 0;
static PyTypeObject *__pyx_ptype_8rasterio_9_features___pyx_scope_struct_4_genexpr = 0;
static PyObject *__pyx_f_8rasterio_9_features__deleteOgrGeom(OGRGeometryH); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features__deleteOgrFeature(OGRFeatureH); /*proto*/
#define __Pyx_MODULE_NAME "rasterio._features"
int __pyx_module_is_main_rasterio___features = 0;

/* Implementation of 'rasterio._features' */
static PyObject *__pyx_builtin_ValueError;
static PyObject *__pyx_builtin_enumerate;
static PyObject *__pyx_builtin_range;
static PyObject *__pyx_builtin_min;
static PyObject *__pyx_builtin_max;
static PyObject *__pyx_builtin_zip;
static PyObject *__pyx_builtin_StopIteration;
static PyObject *__pyx_builtin_RuntimeError;
static PyObject *__pyx_builtin_ImportError;
static const char __pyx_k_e[] = "e";
static const char __pyx_k_f[] = "f";
static const char __pyx_k_i[] = "i";
static const char __pyx_k_s[] = "s";
static const char __pyx_k_v[] = "v";
static const char __pyx_k__2[] = ", ";
static const char __pyx_k_ds[] = "ds";
static const char __pyx_k_fs[] = "fs";
static const char __pyx_k_np[] = "np";
static const char __pyx_k_get[] = "get";
static const char __pyx_k_log[] = "log";
static const char __pyx_k_max[] = "max";
static const char __pyx_k_mem[] = "mem";
static const char __pyx_k_min[] = "min";
static const char __pyx_k_out[] = "out";
static const char __pyx_k_rdr[] = "rdr";
static const char __pyx_k_udr[] = "udr";
static const char __pyx_k_xyz[] = "xyz";
static const char __pyx_k_zip[] = "zip";
static const char __pyx_k_None[] = "None";
static const char __pyx_k_args[] = "args";
static const char __pyx_k_band[] = "band";
static const char __pyx_k_bidx[] = "bidx";
static const char __pyx_k_bool[] = "bool";
static const char __pyx_k_cols[] = "cols";
static const char __pyx_k_exit[] = "__exit__";
static const char __pyx_k_join[] = "join";
static const char __pyx_k_main[] = "__main__";
static const char __pyx_k_mask[] = "mask";
static const char __pyx_k_mrdr[] = "mrdr";
static const char __pyx_k_name[] = "name";
static const char __pyx_k_read[] = "read";
static const char __pyx_k_rows[] = "rows";
static const char __pyx_k_send[] = "send";
static const char __pyx_k_size[] = "size";
static const char __pyx_k_test[] = "__test__";
static const char __pyx_k_type[] = "type";
static const char __pyx_k_xmax[] = "xmax";
static const char __pyx_k_xmin[] = "xmin";
static const char __pyx_k_ymax[] = "ymax";
static const char __pyx_k_ymin[] = "ymin";
static const char __pyx_k_Point[] = "Point";
static const char __pyx_k_build[] = "_build";
static const char __pyx_k_close[] = "close";
static const char __pyx_k_debug[] = "debug";
static const char __pyx_k_dtype[] = "dtype";
static const char __pyx_k_enter[] = "__enter__";
static const char __pyx_k_error[] = "error";
static const char __pyx_k_geoms[] = "geoms";
static const char __pyx_k_image[] = "image";
static const char __pyx_k_int16[] = "int16";
static const char __pyx_k_int32[] = "int32";
static const char __pyx_k_layer[] = "layer";
static const char __pyx_k_numpy[] = "numpy";
static const char __pyx_k_range[] = "range";
static const char __pyx_k_shape[] = "shape";
static const char __pyx_k_sieve[] = "_sieve";
static const char __pyx_k_throw[] = "throw";
static const char __pyx_k_uint8[] = "uint8";
static const char __pyx_k_value[] = "value";
static const char __pyx_k_xmaxs[] = "xmaxs";
static const char __pyx_k_xmins[] = "xmins";
static const char __pyx_k_ymaxs[] = "ymaxs";
static const char __pyx_k_ymins[] = "ymins";
static const char __pyx_k_astype[] = "astype";
static const char __pyx_k_bounds[] = "_bounds";
static const char __pyx_k_coords[] = "coords";
static const char __pyx_k_driver[] = "driver";
static const char __pyx_k_dtypes[] = "dtypes";
static const char __pyx_k_format[] = "format";
static const char __pyx_k_import[] = "__import__";
static const char __pyx_k_mem_ds[] = "mem_ds";
static const char __pyx_k_name_2[] = "__name__";
static const char __pyx_k_retval[] = "retval";
static const char __pyx_k_shapes[] = "_shapes";
static const char __pyx_k_uint16[] = "uint16";
static const char __pyx_k_Polygon[] = "Polygon";
static const char __pyx_k_Unknown[] = "Unknown";
static const char __pyx_k_explode[] = "_explode";
static const char __pyx_k_feature[] = "feature";
static const char __pyx_k_fieldtp[] = "fieldtp";
static const char __pyx_k_float32[] = "float32";
static const char __pyx_k_genexpr[] = "genexpr";
static const char __pyx_k_in_band[] = "in_band";
static const char __pyx_k_logging[] = "logging";
static const char __pyx_k_mask_ds[] = "mask_ds";
static const char __pyx_k_options[] = "options";
static const char __pyx_k_3D_Point[] = "3D Point";
static const char __pyx_k_features[] = "features";
static const char __pyx_k_geometry[] = "geometry";
static const char __pyx_k_is_float[] = "is_float";
static const char __pyx_k_maskband[] = "maskband";
static const char __pyx_k_out_band[] = "out_band";
static const char __pyx_k_rasterio[] = "rasterio";
static const char __pyx_k_shapes_2[] = "shapes";
static const char __pyx_k_Null_geom[] = "Null geom";
static const char __pyx_k_enumerate[] = "enumerate";
static const char __pyx_k_fielddefn[] = "fielddefn";
static const char __pyx_k_getLogger[] = "getLogger";
static const char __pyx_k_in_mem_ds[] = "in_mem_ds";
static const char __pyx_k_iteritems[] = "iteritems";
static const char __pyx_k_mask_band[] = "mask_band";
static const char __pyx_k_num_geoms[] = "num_geoms";
static const char __pyx_k_rasterize[] = "_rasterize";
static const char __pyx_k_transform[] = "transform";
static const char __pyx_k_3D_Polygon[] = "3D Polygon";
static const char __pyx_k_LineString[] = "LineString";
static const char __pyx_k_LinearRing[] = "LinearRing";
static const char __pyx_k_MultiPoint[] = "MultiPoint";
static const char __pyx_k_NULL_layer[] = "NULL layer";
static const char __pyx_k_ValueError[] = "ValueError";
static const char __pyx_k_all_shapes[] = "all_shapes";
static const char __pyx_k_buildPoint[] = "_buildPoint";
static const char __pyx_k_geometries[] = "geometries";
static const char __pyx_k_is_ndarray[] = "is_ndarray";
static const char __pyx_k_out_mem_ds[] = "out_mem_ds";
static const char __pyx_k_pyx_vtable[] = "__pyx_vtable__";
static const char __pyx_k_rasterio_0[] = "rasterio.{0}";
static const char __pyx_k_shape_iter[] = "shape_iter";
static const char __pyx_k_ImportError[] = "ImportError";
static const char __pyx_k_NULL_driver[] = "NULL driver";
static const char __pyx_k_all_touched[] = "all_touched";
static const char __pyx_k_coordinates[] = "coordinates";
static const char __pyx_k_mask_mem_ds[] = "mask_mem_ds";
static const char __pyx_k_mask_reader[] = "mask_reader";
static const char __pyx_k_out_array_r[] = "out array: %r";
static const char __pyx_k_MultiPolygon[] = "MultiPolygon";
static const char __pyx_k_RuntimeError[] = "RuntimeError";
static const char __pyx_k_buildPolygon[] = "_buildPolygon";
static const char __pyx_k_connectivity[] = "connectivity";
static const char __pyx_k_pixel_values[] = "pixel_values";
static const char __pyx_k_valid_dtypes[] = "valid_dtypes";
static const char __pyx_k_3D_LineString[] = "3D LineString";
static const char __pyx_k_3D_MultiPoint[] = "3D MultiPoint";
static const char __pyx_k_StopIteration[] = "StopIteration";
static const char __pyx_k_GEOMETRY_TYPES[] = "GEOMETRY_TYPES";
static const char __pyx_k_3D_MultiPolygon[] = "3D MultiPolygon";
static const char __pyx_k_MultiLineString[] = "MultiLineString";
static const char __pyx_k_buildLineString[] = "_buildLineString";
static const char __pyx_k_buildLinearRing[] = "_buildLinearRing";
static const char __pyx_k_valid_types_str[] = "valid_types_str";
static const char __pyx_k_Invalid_out_image[] = "Invalid out image";
static const char __pyx_k_buildMultiPolygon[] = "_buildMultiPolygon";
static const char __pyx_k_3D_MultiLineString[] = "3D MultiLineString";
static const char __pyx_k_Feature_extraction[] = "Feature extraction";
static const char __pyx_k_GeometryCollection[] = "GeometryCollection";
static const char __pyx_k_rasterio__features[] = "rasterio._features";
static const char __pyx_k_Invalid_source_image[] = "Invalid source image";
static const char __pyx_k_NULL_feature_dataset[] = "NULL feature dataset";
static const char __pyx_k_sieve_locals_genexpr[] = "_sieve.<locals>.genexpr";
static const char __pyx_k_3D_GeometryCollection[] = "3D GeometryCollection";
static const char __pyx_k_NULL_field_definition[] = "NULL field definition";
static const char __pyx_k_GEOJSON2OGR_GEOMETRY_TYPES[] = "GEOJSON2OGR_GEOMETRY_TYPES";
static const char __pyx_k_Unsupported_geometry_type_s[] = "Unsupported geometry type %s";
static const char __pyx_k_connectivity_must_be_4_or_8[] = "connectivity must be 4 or 8";
static const char __pyx_k_ndarray_is_not_C_contiguous[] = "ndarray is not C contiguous";
static const char __pyx_k_size_must_be_greater_than_0[] = "size must be greater than 0";
static const char __pyx_k_image_dtype_must_be_one_of_0[] = "image dtype must be one of: {0}";
static const char __pyx_k_Users_sean_code_frs_wheel_build[] = "/Users/sean/code/frs-wheel-builds/src/rasterio/rasterio/_features.pyx";
static const char __pyx_k_numpy_core_multiarray_failed_to[] = "numpy.core.multiarray failed to import";
static const char __pyx_k_unknown_dtype_code_in_numpy_pxd[] = "unknown dtype code in numpy.pxd (%d)";
static const char __pyx_k_Connectivity_Option_must_be_4_or[] = "Connectivity Option must be 4 or 8";
static const char __pyx_k_Could_not_create_OGR_Geometry_of[] = "Could not create OGR Geometry of type: %i";
static const char __pyx_k_Format_string_allocated_too_shor[] = "Format string allocated too short, see comment in numpy.pxd";
static const char __pyx_k_Geometry_r_at_index_d_with_value[] = "Geometry %r at index %d with value %d skipped";
static const char __pyx_k_Input_is_not_a_valid_geometry_ob[] = "Input is not a valid geometry object";
static const char __pyx_k_Mask_must_be_dtype_rasterio_bool[] = "Mask must be dtype rasterio.bool_ or rasterio.uint8";
static const char __pyx_k_Mask_must_have_same_shape_as_ima[] = "Mask must have same shape as image";
static const char __pyx_k_Non_native_byte_order_not_suppor[] = "Non-native byte order not supported";
static const char __pyx_k_ndarray_is_not_Fortran_contiguou[] = "ndarray is not Fortran contiguous";
static const char __pyx_k_numpy_core_umath_failed_to_impor[] = "numpy.core.umath failed to import";
static const char __pyx_k_out_raster_must_match_dtype_of_i[] = "out raster must match dtype of image";
static const char __pyx_k_out_raster_shape_must_be_same_as[] = "out raster shape must be same as image shape";
static const char __pyx_k_size_must_be_an_integer_number_o[] = "size must be an integer number of pixels";
static const char __pyx_k_size_must_be_smaller_than_size_o[] = "size must be smaller than size of image";
static const char __pyx_k_Format_string_allocated_too_shor_2[] = "Format string allocated too short.";
static PyObject *__pyx_kp_s_3D_GeometryCollection;
static PyObject *__pyx_kp_s_3D_LineString;
static PyObject *__pyx_kp_s_3D_MultiLineString;
static PyObject *__pyx_kp_s_3D_MultiPoint;
static PyObject *__pyx_kp_s_3D_MultiPolygon;
static PyObject *__pyx_kp_s_3D_Point;
static PyObject *__pyx_kp_s_3D_Polygon;
static PyObject *__pyx_kp_s_Connectivity_Option_must_be_4_or;
static PyObject *__pyx_kp_s_Could_not_create_OGR_Geometry_of;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor;
static PyObject *__pyx_kp_u_Format_string_allocated_too_shor_2;
static PyObject *__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES;
static PyObject *__pyx_n_s_GEOMETRY_TYPES;
static PyObject *__pyx_n_s_GeometryCollection;
static PyObject *__pyx_kp_s_Geometry_r_at_index_d_with_value;
static PyObject *__pyx_n_s_ImportError;
static PyObject *__pyx_kp_s_Input_is_not_a_valid_geometry_ob;
static PyObject *__pyx_kp_s_Invalid_out_image;
static PyObject *__pyx_kp_s_Invalid_source_image;
static PyObject *__pyx_n_s_LineString;
static PyObject *__pyx_n_s_LinearRing;
static PyObject *__pyx_kp_s_Mask_must_be_dtype_rasterio_bool;
static PyObject *__pyx_kp_s_Mask_must_have_same_shape_as_ima;
static PyObject *__pyx_n_s_MultiLineString;
static PyObject *__pyx_n_s_MultiPoint;
static PyObject *__pyx_n_s_MultiPolygon;
static PyObject *__pyx_kp_s_NULL_driver;
static PyObject *__pyx_kp_s_NULL_feature_dataset;
static PyObject *__pyx_kp_s_NULL_field_definition;
static PyObject *__pyx_kp_s_NULL_layer;
static PyObject *__pyx_kp_u_Non_native_byte_order_not_suppor;
static PyObject *__pyx_n_s_None;
static PyObject *__pyx_kp_s_Null_geom;
static PyObject *__pyx_n_s_Point;
static PyObject *__pyx_n_s_Polygon;
static PyObject *__pyx_n_s_RuntimeError;
static PyObject *__pyx_n_s_StopIteration;
static PyObject *__pyx_n_s_Unknown;
static PyObject *__pyx_kp_s_Unsupported_geometry_type_s;
static PyObject *__pyx_kp_s_Users_sean_code_frs_wheel_build;
static PyObject *__pyx_n_s_ValueError;
static PyObject *__pyx_kp_s__2;
static PyObject *__pyx_n_s_all_shapes;
static PyObject *__pyx_n_s_all_touched;
static PyObject *__pyx_n_s_args;
static PyObject *__pyx_n_s_astype;
static PyObject *__pyx_n_s_band;
static PyObject *__pyx_n_s_bidx;
static PyObject *__pyx_n_s_bool;
static PyObject *__pyx_n_s_bounds;
static PyObject *__pyx_n_s_build;
static PyObject *__pyx_n_s_buildLineString;
static PyObject *__pyx_n_s_buildLinearRing;
static PyObject *__pyx_n_s_buildMultiPolygon;
static PyObject *__pyx_n_s_buildPoint;
static PyObject *__pyx_n_s_buildPolygon;
static PyObject *__pyx_n_s_close;
static PyObject *__pyx_n_s_cols;
static PyObject *__pyx_n_s_connectivity;
static PyObject *__pyx_kp_s_connectivity_must_be_4_or_8;
static PyObject *__pyx_n_s_coordinates;
static PyObject *__pyx_n_s_coords;
static PyObject *__pyx_n_s_debug;
static PyObject *__pyx_n_s_driver;
static PyObject *__pyx_n_s_ds;
static PyObject *__pyx_n_s_dtype;
static PyObject *__pyx_n_s_dtypes;
static PyObject *__pyx_n_s_e;
static PyObject *__pyx_n_s_enter;
static PyObject *__pyx_n_s_enumerate;
static PyObject *__pyx_n_s_error;
static PyObject *__pyx_n_s_exit;
static PyObject *__pyx_n_s_explode;
static PyObject *__pyx_n_s_f;
static PyObject *__pyx_n_s_feature;
static PyObject *__pyx_n_s_features;
static PyObject *__pyx_n_s_fielddefn;
static PyObject *__pyx_n_s_fieldtp;
static PyObject *__pyx_n_s_float32;
static PyObject *__pyx_n_s_format;
static PyObject *__pyx_n_s_fs;
static PyObject *__pyx_n_s_genexpr;
static PyObject *__pyx_n_s_geometries;
static PyObject *__pyx_n_s_geometry;
static PyObject *__pyx_n_s_geoms;
static PyObject *__pyx_n_s_get;
static PyObject *__pyx_n_s_getLogger;
static PyObject *__pyx_n_s_i;
static PyObject *__pyx_n_s_image;
static PyObject *__pyx_kp_s_image_dtype_must_be_one_of_0;
static PyObject *__pyx_n_s_import;
static PyObject *__pyx_n_s_in_band;
static PyObject *__pyx_n_s_in_mem_ds;
static PyObject *__pyx_n_s_int16;
static PyObject *__pyx_n_s_int32;
static PyObject *__pyx_n_s_is_float;
static PyObject *__pyx_n_s_is_ndarray;
static PyObject *__pyx_n_s_iteritems;
static PyObject *__pyx_n_s_join;
static PyObject *__pyx_n_s_layer;
static PyObject *__pyx_n_s_log;
static PyObject *__pyx_n_s_logging;
static PyObject *__pyx_n_s_main;
static PyObject *__pyx_n_s_mask;
static PyObject *__pyx_n_s_mask_band;
static PyObject *__pyx_n_s_mask_ds;
static PyObject *__pyx_n_s_mask_mem_ds;
static PyObject *__pyx_n_s_mask_reader;
static PyObject *__pyx_n_s_maskband;
static PyObject *__pyx_n_s_max;
static PyObject *__pyx_n_s_mem;
static PyObject *__pyx_n_s_mem_ds;
static PyObject *__pyx_n_s_min;
static PyObject *__pyx_n_s_mrdr;
static PyObject *__pyx_n_s_name;
static PyObject *__pyx_n_s_name_2;
static PyObject *__pyx_kp_u_ndarray_is_not_C_contiguous;
static PyObject *__pyx_kp_u_ndarray_is_not_Fortran_contiguou;
static PyObject *__pyx_n_s_np;
static PyObject *__pyx_n_s_num_geoms;
static PyObject *__pyx_n_s_numpy;
static PyObject *__pyx_kp_s_numpy_core_multiarray_failed_to;
static PyObject *__pyx_kp_s_numpy_core_umath_failed_to_impor;
static PyObject *__pyx_n_s_options;
static PyObject *__pyx_n_s_out;
static PyObject *__pyx_kp_s_out_array_r;
static PyObject *__pyx_n_s_out_band;
static PyObject *__pyx_n_s_out_mem_ds;
static PyObject *__pyx_kp_s_out_raster_must_match_dtype_of_i;
static PyObject *__pyx_kp_s_out_raster_shape_must_be_same_as;
static PyObject *__pyx_n_s_pixel_values;
static PyObject *__pyx_n_s_pyx_vtable;
static PyObject *__pyx_n_s_range;
static PyObject *__pyx_n_s_rasterio;
static PyObject *__pyx_kp_s_rasterio_0;
static PyObject *__pyx_n_s_rasterio__features;
static PyObject *__pyx_n_s_rasterize;
static PyObject *__pyx_n_s_rdr;
static PyObject *__pyx_n_s_read;
static PyObject *__pyx_n_s_retval;
static PyObject *__pyx_n_s_rows;
static PyObject *__pyx_n_s_s;
static PyObject *__pyx_n_s_send;
static PyObject *__pyx_n_s_shape;
static PyObject *__pyx_n_s_shape_iter;
static PyObject *__pyx_n_s_shapes;
static PyObject *__pyx_n_s_shapes_2;
static PyObject *__pyx_n_s_sieve;
static PyObject *__pyx_n_s_sieve_locals_genexpr;
static PyObject *__pyx_n_s_size;
static PyObject *__pyx_kp_s_size_must_be_an_integer_number_o;
static PyObject *__pyx_kp_s_size_must_be_greater_than_0;
static PyObject *__pyx_kp_s_size_must_be_smaller_than_size_o;
static PyObject *__pyx_n_s_test;
static PyObject *__pyx_n_s_throw;
static PyObject *__pyx_n_s_transform;
static PyObject *__pyx_n_s_type;
static PyObject *__pyx_n_s_udr;
static PyObject *__pyx_n_s_uint16;
static PyObject *__pyx_n_s_uint8;
static PyObject *__pyx_kp_u_unknown_dtype_code_in_numpy_pxd;
static PyObject *__pyx_n_s_v;
static PyObject *__pyx_n_s_valid_dtypes;
static PyObject *__pyx_n_s_valid_types_str;
static PyObject *__pyx_n_s_value;
static PyObject *__pyx_n_s_xmax;
static PyObject *__pyx_n_s_xmaxs;
static PyObject *__pyx_n_s_xmin;
static PyObject *__pyx_n_s_xmins;
static PyObject *__pyx_n_s_xyz;
static PyObject *__pyx_n_s_ymax;
static PyObject *__pyx_n_s_ymaxs;
static PyObject *__pyx_n_s_ymin;
static PyObject *__pyx_n_s_ymins;
static PyObject *__pyx_n_s_zip;
static PyObject *__pyx_pf_8rasterio_9_features_12genexpr(CYTHON_UNUSED PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features__shapes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_mask, PyObject *__pyx_v_connectivity, PyObject *__pyx_v_transform); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_6_sieve_genexpr(PyObject *__pyx_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_3_sieve(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_size, PyObject *__pyx_v_out, PyObject *__pyx_v_mask, PyObject *__pyx_v_connectivity); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_5_rasterize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_shapes, PyObject *__pyx_v_image, PyObject *__pyx_v_transform, PyObject *__pyx_v_all_touched); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_7_explode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_10_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_geometry); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_2_buildLineString(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_4_buildLinearRing(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_6_buildPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_8_buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_13ShapeIterator___iter__(struct __pyx_obj_8rasterio_9_features_ShapeIterator *__pyx_v_self); /* proto */
static PyObject *__pyx_pf_8rasterio_9_features_13ShapeIterator_2__next__(struct __pyx_obj_8rasterio_9_features_ShapeIterator *__pyx_v_self); /* proto */
static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /* proto */
static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info); /* proto */
static PyObject *__pyx_tp_new_8rasterio_9_features_GeomBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features_OGRGeomBuilder(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features_ShapeIterator(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct___shapes(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_1__sieve(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_2_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_3__explode(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_4_genexpr(PyTypeObject *t, PyObject *a, PyObject *k); /*proto*/
static PyObject *__pyx_int_0;
static PyObject *__pyx_int_1;
static PyObject *__pyx_int_2;
static PyObject *__pyx_int_3;
static PyObject *__pyx_int_4;
static PyObject *__pyx_int_5;
static PyObject *__pyx_int_6;
static PyObject *__pyx_int_7;
static PyObject *__pyx_int_8;
static PyObject *__pyx_int_100;
static PyObject *__pyx_int_101;
static PyObject *__pyx_int_2147483648;
static PyObject *__pyx_int_2147483649;
static PyObject *__pyx_int_2147483650;
static PyObject *__pyx_int_2147483651;
static PyObject *__pyx_int_2147483652;
static PyObject *__pyx_int_2147483653;
static PyObject *__pyx_int_2147483654;
static PyObject *__pyx_int_2147483655;
static PyObject *__pyx_tuple_;
static PyObject *__pyx_tuple__3;
static PyObject *__pyx_tuple__4;
static PyObject *__pyx_tuple__5;
static PyObject *__pyx_tuple__6;
static PyObject *__pyx_tuple__7;
static PyObject *__pyx_tuple__8;
static PyObject *__pyx_tuple__9;
static PyObject *__pyx_slice__28;
static PyObject *__pyx_tuple__10;
static PyObject *__pyx_tuple__11;
static PyObject *__pyx_tuple__12;
static PyObject *__pyx_tuple__13;
static PyObject *__pyx_tuple__14;
static PyObject *__pyx_tuple__15;
static PyObject *__pyx_tuple__16;
static PyObject *__pyx_tuple__17;
static PyObject *__pyx_tuple__18;
static PyObject *__pyx_tuple__19;
static PyObject *__pyx_tuple__20;
static PyObject *__pyx_tuple__21;
static PyObject *__pyx_tuple__22;
static PyObject *__pyx_tuple__23;
static PyObject *__pyx_tuple__24;
static PyObject *__pyx_tuple__25;
static PyObject *__pyx_tuple__26;
static PyObject *__pyx_tuple__27;
static PyObject *__pyx_tuple__29;
static PyObject *__pyx_tuple__30;
static PyObject *__pyx_tuple__31;
static PyObject *__pyx_tuple__32;
static PyObject *__pyx_tuple__33;
static PyObject *__pyx_tuple__34;
static PyObject *__pyx_tuple__35;
static PyObject *__pyx_tuple__36;
static PyObject *__pyx_tuple__37;
static PyObject *__pyx_tuple__38;
static PyObject *__pyx_tuple__39;
static PyObject *__pyx_tuple__40;
static PyObject *__pyx_tuple__41;
static PyObject *__pyx_tuple__43;
static PyObject *__pyx_tuple__45;
static PyObject *__pyx_tuple__47;
static PyObject *__pyx_tuple__49;
static PyObject *__pyx_codeobj__42;
static PyObject *__pyx_codeobj__44;
static PyObject *__pyx_codeobj__46;
static PyObject *__pyx_codeobj__48;
static PyObject *__pyx_codeobj__50;
static PyObject *__pyx_gb_8rasterio_9_features_14generator3(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_features.pyx":379
 * # Mapping of GeoJSON type names to OGR integer geometry types
 * GEOJSON2OGR_GEOMETRY_TYPES = dict(
 *     (v, k) for k, v in GEOMETRY_TYPES.iteritems()             # <<<<<<<<<<<<<<
 * )
 * 
 */

static PyObject *__pyx_pf_8rasterio_9_features_12genexpr(CYTHON_UNUSED PyObject *__pyx_self) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_4_genexpr(__pyx_ptype_8rasterio_9_features___pyx_scope_struct_4_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 379, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_9_features_14generator3, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_genexpr, __pyx_n_s_rasterio__features); if (unlikely(!gen)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._features.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_9_features_14generator3(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  Py_ssize_t __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_r = PyDict_New(); if (unlikely(!__pyx_r)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_r);
  __pyx_t_2 = 0;
  __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOMETRY_TYPES); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (unlikely(__pyx_t_5 == Py_None)) {
    PyErr_Format(PyExc_AttributeError, "'NoneType' object has no attribute '%s'", "iteritems");
    __PYX_ERR(0, 379, __pyx_L1_error)
  }
  __pyx_t_6 = __Pyx_dict_iterator(__pyx_t_5, 0, __pyx_n_s_iteritems, (&__pyx_t_3), (&__pyx_t_4)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __pyx_t_1 = __pyx_t_6;
  __pyx_t_6 = 0;
  while (1) {
    __pyx_t_7 = __Pyx_dict_iter_next(__pyx_t_1, __pyx_t_3, &__pyx_t_2, &__pyx_t_6, &__pyx_t_5, NULL, __pyx_t_4);
    if (unlikely(__pyx_t_7 == 0)) break;
    if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 379, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_k);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_k, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_5);
    __Pyx_GIVEREF(__pyx_t_5);
    __pyx_t_5 = 0;
    if (unlikely(PyDict_SetItem(__pyx_r, (PyObject*)__pyx_cur_scope->__pyx_v_v, (PyObject*)__pyx_cur_scope->__pyx_v_k))) __PYX_ERR(0, 379, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8rasterio_9_features_2generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_features.pyx":20
 * 
 * 
 * def _shapes(image, mask, connectivity, transform):             # <<<<<<<<<<<<<<
 *     """
 *     Return a generator of (polygon, value) for each each set of adjacent pixels
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_1_shapes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_9_features__shapes[] = "\n    Return a generator of (polygon, value) for each each set of adjacent pixels\n    of the same value.\n\n    Parameters\n    ----------\n    image : numpy ndarray or rasterio Band object\n        (RasterReader, bidx namedtuple).\n        Data type must be one of rasterio.int16, rasterio.int32,\n        rasterio.uint8, rasterio.uint16, or rasterio.float32.\n    mask : numpy ndarray or rasterio Band object\n        Values of False or 0 will be excluded from feature generation\n        Must evaluate to bool (rasterio.bool_ or rasterio.uint8)\n    connectivity : int\n        Use 4 or 8 pixel connectivity for grouping pixels into features\n    transform : Affine transformation\n        If not provided, feature coordinates will be generated based on pixel\n        coordinates\n\n    Returns\n    -------\n    Generator of (polygon, value)\n        Yields a pair of (polygon, value) for each feature found in the image.\n        Polygons are GeoJSON-like dicts and the values are the associated value\n        from the image, in the data type of the image.\n        Note: due to floating point precision issues, values returned from a\n        floating point image may not exactly match the original values.\n\n    ";
static PyMethodDef __pyx_mdef_8rasterio_9_features_1_shapes = {"_shapes", (PyCFunction)__pyx_pw_8rasterio_9_features_1_shapes, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_9_features__shapes};
static PyObject *__pyx_pw_8rasterio_9_features_1_shapes(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_mask = 0;
  PyObject *__pyx_v_connectivity = 0;
  PyObject *__pyx_v_transform = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_shapes (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_mask,&__pyx_n_s_connectivity,&__pyx_n_s_transform,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_shapes", 1, 4, 4, 1); __PYX_ERR(0, 20, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connectivity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_shapes", 1, 4, 4, 2); __PYX_ERR(0, 20, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transform)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_shapes", 1, 4, 4, 3); __PYX_ERR(0, 20, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_shapes") < 0)) __PYX_ERR(0, 20, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_image = values[0];
    __pyx_v_mask = values[1];
    __pyx_v_connectivity = values[2];
    __pyx_v_transform = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_shapes", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 20, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("rasterio._features._shapes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_9_features__shapes(__pyx_self, __pyx_v_image, __pyx_v_mask, __pyx_v_connectivity, __pyx_v_transform);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features__shapes(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_mask, PyObject *__pyx_v_connectivity, PyObject *__pyx_v_transform) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_shapes", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)__pyx_tp_new_8rasterio_9_features___pyx_scope_struct___shapes(__pyx_ptype_8rasterio_9_features___pyx_scope_struct___shapes, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 20, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_image = __pyx_v_image;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_image);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_image);
  __pyx_cur_scope->__pyx_v_mask = __pyx_v_mask;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_mask);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_mask);
  __pyx_cur_scope->__pyx_v_connectivity = __pyx_v_connectivity;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_connectivity);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_connectivity);
  __pyx_cur_scope->__pyx_v_transform = __pyx_v_transform;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_transform);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_transform);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_9_features_2generator, (PyObject *) __pyx_cur_scope, __pyx_n_s_shapes, __pyx_n_s_shapes, __pyx_n_s_rasterio__features); if (unlikely(!gen)) __PYX_ERR(0, 20, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._features._shapes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_9_features_2generator(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  GDALRasterBandH __pyx_t_9;
  GDALRasterBandH __pyx_t_10;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  PyObject *(*__pyx_t_13)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L25_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 20, __pyx_L1_error)

  /* "rasterio/_features.pyx":53
 *     cdef int rows
 *     cdef int cols
 *     cdef GDALRasterBandH band = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALRasterBandH maskband = NULL
 *     cdef GDALDriverH driver = NULL
 */
  __pyx_cur_scope->__pyx_v_band = NULL;

  /* "rasterio/_features.pyx":54
 *     cdef int cols
 *     cdef GDALRasterBandH band = NULL
 *     cdef GDALRasterBandH maskband = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALDriverH driver = NULL
 *     cdef OGRDataSourceH fs = NULL
 */
  __pyx_cur_scope->__pyx_v_maskband = NULL;

  /* "rasterio/_features.pyx":55
 *     cdef GDALRasterBandH band = NULL
 *     cdef GDALRasterBandH maskband = NULL
 *     cdef GDALDriverH driver = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRDataSourceH fs = NULL
 *     cdef OGRLayerH layer = NULL
 */
  __pyx_cur_scope->__pyx_v_driver = NULL;

  /* "rasterio/_features.pyx":56
 *     cdef GDALRasterBandH maskband = NULL
 *     cdef GDALDriverH driver = NULL
 *     cdef OGRDataSourceH fs = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRLayerH layer = NULL
 *     cdef OGRFieldDefnH fielddefn = NULL
 */
  __pyx_cur_scope->__pyx_v_fs = NULL;

  /* "rasterio/_features.pyx":57
 *     cdef GDALDriverH driver = NULL
 *     cdef OGRDataSourceH fs = NULL
 *     cdef OGRLayerH layer = NULL             # <<<<<<<<<<<<<<
 *     cdef OGRFieldDefnH fielddefn = NULL
 *     cdef char **options = NULL
 */
  __pyx_cur_scope->__pyx_v_layer = NULL;

  /* "rasterio/_features.pyx":58
 *     cdef OGRDataSourceH fs = NULL
 *     cdef OGRLayerH layer = NULL
 *     cdef OGRFieldDefnH fielddefn = NULL             # <<<<<<<<<<<<<<
 *     cdef char **options = NULL
 *     cdef InMemoryRaster mem_ds = None
 */
  __pyx_cur_scope->__pyx_v_fielddefn = NULL;

  /* "rasterio/_features.pyx":59
 *     cdef OGRLayerH layer = NULL
 *     cdef OGRFieldDefnH fielddefn = NULL
 *     cdef char **options = NULL             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster mem_ds = None
 *     cdef InMemoryRaster mask_ds = None
 */
  __pyx_cur_scope->__pyx_v_options = NULL;

  /* "rasterio/_features.pyx":60
 *     cdef OGRFieldDefnH fielddefn = NULL
 *     cdef char **options = NULL
 *     cdef InMemoryRaster mem_ds = None             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster mask_ds = None
 *     cdef ShapeIterator shape_iter = None
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_mem_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":61
 *     cdef char **options = NULL
 *     cdef InMemoryRaster mem_ds = None
 *     cdef InMemoryRaster mask_ds = None             # <<<<<<<<<<<<<<
 *     cdef ShapeIterator shape_iter = None
 * 
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_mask_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":62
 *     cdef InMemoryRaster mem_ds = None
 *     cdef InMemoryRaster mask_ds = None
 *     cdef ShapeIterator shape_iter = None             # <<<<<<<<<<<<<<
 * 
 *     cdef bint is_float = np.dtype(image.dtype).kind == 'f'
 */
  __Pyx_INCREF(Py_None);
  __Pyx_GIVEREF(Py_None);
  __pyx_cur_scope->__pyx_v_shape_iter = ((struct __pyx_obj_8rasterio_9_features_ShapeIterator *)Py_None);

  /* "rasterio/_features.pyx":64
 *     cdef ShapeIterator shape_iter = None
 * 
 *     cdef bint is_float = np.dtype(image.dtype).kind == 'f'             # <<<<<<<<<<<<<<
 *     cdef int fieldtp = 2 if is_float else 0
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_image, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 64, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_cur_scope->__pyx_v_is_float = (((PyArray_Descr *)__pyx_t_1)->kind == 'f');
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":65
 * 
 *     cdef bint is_float = np.dtype(image.dtype).kind == 'f'
 *     cdef int fieldtp = 2 if is_float else 0             # <<<<<<<<<<<<<<
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16', 'float32')
 */
  if ((__pyx_cur_scope->__pyx_v_is_float != 0)) {
    __pyx_t_3 = 2;
  } else {
    __pyx_t_3 = 0;
  }
  __pyx_cur_scope->__pyx_v_fieldtp = __pyx_t_3;

  /* "rasterio/_features.pyx":67
 *     cdef int fieldtp = 2 if is_float else 0
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16', 'float32')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 */
  __Pyx_INCREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);
  __pyx_cur_scope->__pyx_v_valid_dtypes = __pyx_tuple_;

  /* "rasterio/_features.pyx":69
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16', 'float32')
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:             # <<<<<<<<<<<<<<
 *         raise ValueError("image dtype must be one of: {0}".format(
 *             ', '.join(valid_dtypes)))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_image, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_cur_scope->__pyx_v_valid_dtypes, Py_NE)); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 69, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "rasterio/_features.pyx":70
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         raise ValueError("image dtype must be one of: {0}".format(             # <<<<<<<<<<<<<<
 *             ', '.join(valid_dtypes)))
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_image_dtype_must_be_one_of_0, __pyx_n_s_format); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);

    /* "rasterio/_features.pyx":71
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         raise ValueError("image dtype must be one of: {0}".format(
 *             ', '.join(valid_dtypes)))             # <<<<<<<<<<<<<<
 * 
 *     if connectivity not in (4, 8):
 */
    __pyx_t_6 = __Pyx_PyString_Join(__pyx_kp_s__2, __pyx_cur_scope->__pyx_v_valid_dtypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 71, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_1))) {
      __pyx_t_7 = PyMethod_GET_SELF(__pyx_t_1);
      if (likely(__pyx_t_7)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_1, function);
      }
    }
    if (!__pyx_t_7) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_t_6); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
        PyObject *__pyx_temp[2] = {__pyx_t_7, __pyx_t_6};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      } else
      #endif
      {
        __pyx_t_8 = PyTuple_New(1+1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 70, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_GIVEREF(__pyx_t_7); PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_7); __pyx_t_7 = NULL;
        __Pyx_GIVEREF(__pyx_t_6);
        PyTuple_SET_ITEM(__pyx_t_8, 0+1, __pyx_t_6);
        __pyx_t_6 = 0;
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_8, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":70
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         raise ValueError("image dtype must be one of: {0}".format(             # <<<<<<<<<<<<<<
 *             ', '.join(valid_dtypes)))
 * 
 */
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 70, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 70, __pyx_L1_error)

    /* "rasterio/_features.pyx":69
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16', 'float32')
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:             # <<<<<<<<<<<<<<
 *         raise ValueError("image dtype must be one of: {0}".format(
 *             ', '.join(valid_dtypes)))
 */
  }

  /* "rasterio/_features.pyx":73
 *             ', '.join(valid_dtypes)))
 * 
 *     if connectivity not in (4, 8):             # <<<<<<<<<<<<<<
 *         raise ValueError("Connectivity Option must be 4 or 8")
 * 
 */
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_connectivity);
  __pyx_t_2 = __pyx_cur_scope->__pyx_v_connectivity;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_int_4, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {
  } else {
    __pyx_t_5 = __pyx_t_4;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_2, __pyx_int_8, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 73, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 73, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_5 = __pyx_t_4;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":74
 * 
 *     if connectivity not in (4, 8):
 *         raise ValueError("Connectivity Option must be 4 or 8")             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(image):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__3, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 74, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 74, __pyx_L1_error)

    /* "rasterio/_features.pyx":73
 *             ', '.join(valid_dtypes)))
 * 
 *     if connectivity not in (4, 8):             # <<<<<<<<<<<<<<
 *         raise ValueError("Connectivity Option must be 4 or 8")
 * 
 */
  }

  /* "rasterio/_features.pyx":76
 *         raise ValueError("Connectivity Option must be 4 or 8")
 * 
 *     if dtypes.is_ndarray(image):             # <<<<<<<<<<<<<<
 *         mem_ds = InMemoryRaster(image=image, transform=transform)
 *         band = mem_ds.band(1)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_8);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
    __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_8);
    if (likely(__pyx_t_1)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_1);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_8, function);
    }
  }
  if (!__pyx_t_1) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_image); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_cur_scope->__pyx_v_image};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
      PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_cur_scope->__pyx_v_image};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_1); __pyx_t_1 = NULL;
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_image);
      __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_cur_scope->__pyx_v_image);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 76, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 76, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":77
 * 
 *     if dtypes.is_ndarray(image):
 *         mem_ds = InMemoryRaster(image=image, transform=transform)             # <<<<<<<<<<<<<<
 *         band = mem_ds.band(1)
 *     elif isinstance(image, tuple):
 */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_image, __pyx_cur_scope->__pyx_v_image) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_transform, __pyx_cur_scope->__pyx_v_transform) < 0) __PYX_ERR(0, 77, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 77, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_GOTREF(((PyObject *)__pyx_cur_scope->__pyx_v_mem_ds));
    __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_mem_ds, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_8));
    __Pyx_GIVEREF(__pyx_t_8);
    __pyx_t_8 = 0;

    /* "rasterio/_features.pyx":78
 *     if dtypes.is_ndarray(image):
 *         mem_ds = InMemoryRaster(image=image, transform=transform)
 *         band = mem_ds.band(1)             # <<<<<<<<<<<<<<
 *     elif isinstance(image, tuple):
 *         rdr = image.ds
 */
    __pyx_t_9 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_cur_scope->__pyx_v_mem_ds->__pyx_vtab)->band(__pyx_cur_scope->__pyx_v_mem_ds, 1); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(0, 78, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_band = __pyx_t_9;

    /* "rasterio/_features.pyx":76
 *         raise ValueError("Connectivity Option must be 4 or 8")
 * 
 *     if dtypes.is_ndarray(image):             # <<<<<<<<<<<<<<
 *         mem_ds = InMemoryRaster(image=image, transform=transform)
 *         band = mem_ds.band(1)
 */
    goto __pyx_L8;
  }

  /* "rasterio/_features.pyx":79
 *         mem_ds = InMemoryRaster(image=image, transform=transform)
 *         band = mem_ds.band(1)
 *     elif isinstance(image, tuple):             # <<<<<<<<<<<<<<
 *         rdr = image.ds
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)
 */
  __pyx_t_4 = PyTuple_Check(__pyx_cur_scope->__pyx_v_image); 
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "rasterio/_features.pyx":80
 *         band = mem_ds.band(1)
 *     elif isinstance(image, tuple):
 *         rdr = image.ds             # <<<<<<<<<<<<<<
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_image, __pyx_n_s_ds); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 80, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    __pyx_cur_scope->__pyx_v_rdr = __pyx_t_8;
    __pyx_t_8 = 0;

    /* "rasterio/_features.pyx":81
 *     elif isinstance(image, tuple):
 *         rdr = image.ds
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Invalid source image")
 */
    if (!(likely(__Pyx_TypeTest(__pyx_cur_scope->__pyx_v_rdr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 81, __pyx_L1_error)
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_image, __pyx_n_s_bidx); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_8); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 81, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_10 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_cur_scope->__pyx_v_rdr)->__pyx_base.__pyx_vtab)->__pyx_base.band(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_cur_scope->__pyx_v_rdr)), __pyx_t_3); if (unlikely(__pyx_t_10 == NULL)) __PYX_ERR(0, 81, __pyx_L1_error)
    __pyx_cur_scope->__pyx_v_band = __pyx_t_10;

    /* "rasterio/_features.pyx":79
 *         mem_ds = InMemoryRaster(image=image, transform=transform)
 *         band = mem_ds.band(1)
 *     elif isinstance(image, tuple):             # <<<<<<<<<<<<<<
 *         rdr = image.ds
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)
 */
    goto __pyx_L8;
  }

  /* "rasterio/_features.pyx":83
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 *         raise ValueError("Invalid source image")             # <<<<<<<<<<<<<<
 * 
 *     if mask is not None:
 */
  /*else*/ {
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 83, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_Raise(__pyx_t_8, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __PYX_ERR(0, 83, __pyx_L1_error)
  }
  __pyx_L8:;

  /* "rasterio/_features.pyx":85
 *         raise ValueError("Invalid source image")
 * 
 *     if mask is not None:             # <<<<<<<<<<<<<<
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")
 */
  __pyx_t_5 = (__pyx_cur_scope->__pyx_v_mask != Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":86
 * 
 *     if mask is not None:
 *         if mask.shape != image.shape:             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must have same shape as image")
 * 
 */
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_mask, __pyx_n_s_shape); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = PyObject_RichCompare(__pyx_t_8, __pyx_t_2, Py_NE); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 86, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_4) {

      /* "rasterio/_features.pyx":87
 *     if mask is not None:
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")             # <<<<<<<<<<<<<<
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):
 */
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__5, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 87, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(0, 87, __pyx_L1_error)

      /* "rasterio/_features.pyx":86
 * 
 *     if mask is not None:
 *         if mask.shape != image.shape:             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must have same shape as image")
 * 
 */
    }

    /* "rasterio/_features.pyx":89
 *             raise ValueError("Mask must have same shape as image")
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "
 *                              "rasterio.uint8")
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_mask, __pyx_n_s_dtype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_6);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_6);
    __pyx_t_6 = 0;
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_2, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 89, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_bool, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 89, __pyx_L1_error)
    if (__pyx_t_5) {
    } else {
      __pyx_t_4 = __pyx_t_5;
      goto __pyx_L12_bool_binop_done;
    }
    __pyx_t_5 = (__Pyx_PyString_Equals(__pyx_t_2, __pyx_n_s_uint8, Py_NE)); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 89, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_5;
    __pyx_L12_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_5 = (__pyx_t_4 != 0);
    if (__pyx_t_5) {

      /* "rasterio/_features.pyx":90
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "             # <<<<<<<<<<<<<<
 *                              "rasterio.uint8")
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 90, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 90, __pyx_L1_error)

      /* "rasterio/_features.pyx":89
 *             raise ValueError("Mask must have same shape as image")
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "
 *                              "rasterio.uint8")
 */
    }

    /* "rasterio/_features.pyx":93
 *                              "rasterio.uint8")
 * 
 *         if dtypes.is_ndarray(mask):             # <<<<<<<<<<<<<<
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),
 */
    __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_8);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_8, function);
      }
    }
    if (!__pyx_t_6) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_mask};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
        PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_cur_scope->__pyx_v_mask};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_1 = PyTuple_New(1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_6); __pyx_t_6 = NULL;
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_mask);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_mask);
        PyTuple_SET_ITEM(__pyx_t_1, 0+1, __pyx_cur_scope->__pyx_v_mask);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 93, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 93, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_5) {

      /* "rasterio/_features.pyx":95
 *         if dtypes.is_ndarray(mask):
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),             # <<<<<<<<<<<<<<
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)
 */
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_8 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_mask, __pyx_n_s_astype); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_tuple__7, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_image, __pyx_t_1) < 0) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

      /* "rasterio/_features.pyx":96
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),
 *                                      transform=transform)             # <<<<<<<<<<<<<<
 *             maskband = mask_ds.band(1)
 *         elif isinstance(mask, tuple):
 */
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_transform, __pyx_cur_scope->__pyx_v_transform) < 0) __PYX_ERR(0, 95, __pyx_L1_error)

      /* "rasterio/_features.pyx":95
 *         if dtypes.is_ndarray(mask):
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),             # <<<<<<<<<<<<<<
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)
 */
      __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 95, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(((PyObject *)__pyx_cur_scope->__pyx_v_mask_ds));
      __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_mask_ds, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_1));
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_t_1 = 0;

      /* "rasterio/_features.pyx":97
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)             # <<<<<<<<<<<<<<
 *         elif isinstance(mask, tuple):
 *             mrdr = mask.ds
 */
      __pyx_t_9 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_cur_scope->__pyx_v_mask_ds->__pyx_vtab)->band(__pyx_cur_scope->__pyx_v_mask_ds, 1); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(0, 97, __pyx_L1_error)
      __pyx_cur_scope->__pyx_v_maskband = __pyx_t_9;

      /* "rasterio/_features.pyx":93
 *                              "rasterio.uint8")
 * 
 *         if dtypes.is_ndarray(mask):             # <<<<<<<<<<<<<<
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),
 */
      goto __pyx_L14;
    }

    /* "rasterio/_features.pyx":98
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)
 *         elif isinstance(mask, tuple):             # <<<<<<<<<<<<<<
 *             mrdr = mask.ds
 *             maskband = (<DatasetReaderBase?>mrdr).band(mask.bidx)
 */
    __pyx_t_5 = PyTuple_Check(__pyx_cur_scope->__pyx_v_mask); 
    __pyx_t_4 = (__pyx_t_5 != 0);
    if (__pyx_t_4) {

      /* "rasterio/_features.pyx":99
 *             maskband = mask_ds.band(1)
 *         elif isinstance(mask, tuple):
 *             mrdr = mask.ds             # <<<<<<<<<<<<<<
 *             maskband = (<DatasetReaderBase?>mrdr).band(mask.bidx)
 * 
 */
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_mask, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 99, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_GIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_v_mrdr = __pyx_t_1;
      __pyx_t_1 = 0;

      /* "rasterio/_features.pyx":100
 *         elif isinstance(mask, tuple):
 *             mrdr = mask.ds
 *             maskband = (<DatasetReaderBase?>mrdr).band(mask.bidx)             # <<<<<<<<<<<<<<
 * 
 *     # Create an in-memory feature store.
 */
      if (!(likely(__Pyx_TypeTest(__pyx_cur_scope->__pyx_v_mrdr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 100, __pyx_L1_error)
      __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_cur_scope->__pyx_v_mask, __pyx_n_s_bidx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 100, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      __pyx_t_10 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_cur_scope->__pyx_v_mrdr)->__pyx_base.__pyx_vtab)->__pyx_base.band(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_cur_scope->__pyx_v_mrdr)), __pyx_t_3); if (unlikely(__pyx_t_10 == NULL)) __PYX_ERR(0, 100, __pyx_L1_error)
      __pyx_cur_scope->__pyx_v_maskband = __pyx_t_10;

      /* "rasterio/_features.pyx":98
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)
 *         elif isinstance(mask, tuple):             # <<<<<<<<<<<<<<
 *             mrdr = mask.ds
 *             maskband = (<DatasetReaderBase?>mrdr).band(mask.bidx)
 */
    }
    __pyx_L14:;

    /* "rasterio/_features.pyx":85
 *         raise ValueError("Invalid source image")
 * 
 *     if mask is not None:             # <<<<<<<<<<<<<<
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")
 */
  }

  /* "rasterio/_features.pyx":103
 * 
 *     # Create an in-memory feature store.
 *     driver = OGRGetDriverByName("Memory")             # <<<<<<<<<<<<<<
 *     if driver == NULL:
 *         raise ValueError("NULL driver")
 */
  __pyx_cur_scope->__pyx_v_driver = OGRGetDriverByName(((char const *)"Memory"));

  /* "rasterio/_features.pyx":104
 *     # Create an in-memory feature store.
 *     driver = OGRGetDriverByName("Memory")
 *     if driver == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL driver")
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 */
  __pyx_t_4 = ((__pyx_cur_scope->__pyx_v_driver == NULL) != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":105
 *     driver = OGRGetDriverByName("Memory")
 *     if driver == NULL:
 *         raise ValueError("NULL driver")             # <<<<<<<<<<<<<<
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__8, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 105, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 105, __pyx_L1_error)

    /* "rasterio/_features.pyx":104
 *     # Create an in-memory feature store.
 *     driver = OGRGetDriverByName("Memory")
 *     if driver == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL driver")
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 */
  }

  /* "rasterio/_features.pyx":106
 *     if driver == NULL:
 *         raise ValueError("NULL driver")
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)             # <<<<<<<<<<<<<<
 *     if fs == NULL:
 *         raise ValueError("NULL feature dataset")
 */
  __pyx_cur_scope->__pyx_v_fs = OGR_Dr_CreateDataSource(__pyx_cur_scope->__pyx_v_driver, ((char const *)"temp"), NULL);

  /* "rasterio/_features.pyx":107
 *         raise ValueError("NULL driver")
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL feature dataset")
 * 
 */
  __pyx_t_4 = ((__pyx_cur_scope->__pyx_v_fs == NULL) != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":108
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:
 *         raise ValueError("NULL feature dataset")             # <<<<<<<<<<<<<<
 * 
 *     # And a layer.
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__9, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 108, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 108, __pyx_L1_error)

    /* "rasterio/_features.pyx":107
 *         raise ValueError("NULL driver")
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL feature dataset")
 * 
 */
  }

  /* "rasterio/_features.pyx":111
 * 
 *     # And a layer.
 *     layer = OGR_DS_CreateLayer(fs, "polygons", NULL, 3, NULL)             # <<<<<<<<<<<<<<
 *     if layer == NULL:
 *         raise ValueError("NULL layer")
 */
  __pyx_cur_scope->__pyx_v_layer = OGR_DS_CreateLayer(__pyx_cur_scope->__pyx_v_fs, ((char const *)"polygons"), NULL, 3, NULL);

  /* "rasterio/_features.pyx":112
 *     # And a layer.
 *     layer = OGR_DS_CreateLayer(fs, "polygons", NULL, 3, NULL)
 *     if layer == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL layer")
 * 
 */
  __pyx_t_4 = ((__pyx_cur_scope->__pyx_v_layer == NULL) != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":113
 *     layer = OGR_DS_CreateLayer(fs, "polygons", NULL, 3, NULL)
 *     if layer == NULL:
 *         raise ValueError("NULL layer")             # <<<<<<<<<<<<<<
 * 
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__10, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 113, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 113, __pyx_L1_error)

    /* "rasterio/_features.pyx":112
 *     # And a layer.
 *     layer = OGR_DS_CreateLayer(fs, "polygons", NULL, 3, NULL)
 *     if layer == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL layer")
 * 
 */
  }

  /* "rasterio/_features.pyx":115
 *         raise ValueError("NULL layer")
 * 
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)             # <<<<<<<<<<<<<<
 *     if fielddefn == NULL:
 *         raise ValueError("NULL field definition")
 */
  __pyx_cur_scope->__pyx_v_fielddefn = OGR_Fld_Create(((char const *)"image_value"), __pyx_cur_scope->__pyx_v_fieldtp);

  /* "rasterio/_features.pyx":116
 * 
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 *     if fielddefn == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL field definition")
 *     OGR_L_CreateField(layer, fielddefn, 1)
 */
  __pyx_t_4 = ((__pyx_cur_scope->__pyx_v_fielddefn == NULL) != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":117
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 *     if fielddefn == NULL:
 *         raise ValueError("NULL field definition")             # <<<<<<<<<<<<<<
 *     OGR_L_CreateField(layer, fielddefn, 1)
 *     OGR_Fld_Destroy(fielddefn)
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__11, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 117, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 117, __pyx_L1_error)

    /* "rasterio/_features.pyx":116
 * 
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 *     if fielddefn == NULL:             # <<<<<<<<<<<<<<
 *         raise ValueError("NULL field definition")
 *     OGR_L_CreateField(layer, fielddefn, 1)
 */
  }

  /* "rasterio/_features.pyx":118
 *     if fielddefn == NULL:
 *         raise ValueError("NULL field definition")
 *     OGR_L_CreateField(layer, fielddefn, 1)             # <<<<<<<<<<<<<<
 *     OGR_Fld_Destroy(fielddefn)
 * 
 */
  OGR_L_CreateField(__pyx_cur_scope->__pyx_v_layer, __pyx_cur_scope->__pyx_v_fielddefn, 1);

  /* "rasterio/_features.pyx":119
 *         raise ValueError("NULL field definition")
 *     OGR_L_CreateField(layer, fielddefn, 1)
 *     OGR_Fld_Destroy(fielddefn)             # <<<<<<<<<<<<<<
 * 
 *     if connectivity == 8:
 */
  OGR_Fld_Destroy(__pyx_cur_scope->__pyx_v_fielddefn);

  /* "rasterio/_features.pyx":121
 *     OGR_Fld_Destroy(fielddefn)
 * 
 *     if connectivity == 8:             # <<<<<<<<<<<<<<
 *         options = CSLSetNameValue(options, "8CONNECTED", "8")
 * 
 */
  __pyx_t_1 = __Pyx_PyInt_EqObjC(__pyx_cur_scope->__pyx_v_connectivity, __pyx_int_8, 8, 0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 121, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":122
 * 
 *     if connectivity == 8:
 *         options = CSLSetNameValue(options, "8CONNECTED", "8")             # <<<<<<<<<<<<<<
 * 
 *     if is_float:
 */
    __pyx_cur_scope->__pyx_v_options = CSLSetNameValue(__pyx_cur_scope->__pyx_v_options, ((char *)"8CONNECTED"), ((char *)"8"));

    /* "rasterio/_features.pyx":121
 *     OGR_Fld_Destroy(fielddefn)
 * 
 *     if connectivity == 8:             # <<<<<<<<<<<<<<
 *         options = CSLSetNameValue(options, "8CONNECTED", "8")
 * 
 */
  }

  /* "rasterio/_features.pyx":124
 *         options = CSLSetNameValue(options, "8CONNECTED", "8")
 * 
 *     if is_float:             # <<<<<<<<<<<<<<
 *         GDALFPolygonize(band, maskband, layer, 0, options, NULL, NULL)
 *     else:
 */
  __pyx_t_4 = (__pyx_cur_scope->__pyx_v_is_float != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":125
 * 
 *     if is_float:
 *         GDALFPolygonize(band, maskband, layer, 0, options, NULL, NULL)             # <<<<<<<<<<<<<<
 *     else:
 *         GDALPolygonize(band, maskband, layer, 0, options, NULL, NULL)
 */
    GDALFPolygonize(__pyx_cur_scope->__pyx_v_band, __pyx_cur_scope->__pyx_v_maskband, __pyx_cur_scope->__pyx_v_layer, 0, __pyx_cur_scope->__pyx_v_options, NULL, NULL);

    /* "rasterio/_features.pyx":124
 *         options = CSLSetNameValue(options, "8CONNECTED", "8")
 * 
 *     if is_float:             # <<<<<<<<<<<<<<
 *         GDALFPolygonize(band, maskband, layer, 0, options, NULL, NULL)
 *     else:
 */
    goto __pyx_L20;
  }

  /* "rasterio/_features.pyx":127
 *         GDALFPolygonize(band, maskband, layer, 0, options, NULL, NULL)
 *     else:
 *         GDALPolygonize(band, maskband, layer, 0, options, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Yield Fiona-style features
 */
  /*else*/ {
    GDALPolygonize(__pyx_cur_scope->__pyx_v_band, __pyx_cur_scope->__pyx_v_maskband, __pyx_cur_scope->__pyx_v_layer, 0, __pyx_cur_scope->__pyx_v_options, NULL, NULL);
  }
  __pyx_L20:;

  /* "rasterio/_features.pyx":130
 * 
 *     # Yield Fiona-style features
 *     shape_iter = ShapeIterator()             # <<<<<<<<<<<<<<
 *     shape_iter.layer = layer
 *     shape_iter.fieldtype = fieldtp
 */
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_ShapeIterator), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 130, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GOTREF(((PyObject *)__pyx_cur_scope->__pyx_v_shape_iter));
  __Pyx_DECREF_SET(__pyx_cur_scope->__pyx_v_shape_iter, ((struct __pyx_obj_8rasterio_9_features_ShapeIterator *)__pyx_t_1));
  __Pyx_GIVEREF(__pyx_t_1);
  __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":131
 *     # Yield Fiona-style features
 *     shape_iter = ShapeIterator()
 *     shape_iter.layer = layer             # <<<<<<<<<<<<<<
 *     shape_iter.fieldtype = fieldtp
 *     for s, v in shape_iter:
 */
  __pyx_cur_scope->__pyx_v_shape_iter->layer = __pyx_cur_scope->__pyx_v_layer;

  /* "rasterio/_features.pyx":132
 *     shape_iter = ShapeIterator()
 *     shape_iter.layer = layer
 *     shape_iter.fieldtype = fieldtp             # <<<<<<<<<<<<<<
 *     for s, v in shape_iter:
 *         yield s, v
 */
  __pyx_cur_scope->__pyx_v_shape_iter->fieldtype = __pyx_cur_scope->__pyx_v_fieldtp;

  /* "rasterio/_features.pyx":133
 *     shape_iter.layer = layer
 *     shape_iter.fieldtype = fieldtp
 *     for s, v in shape_iter:             # <<<<<<<<<<<<<<
 *         yield s, v
 * 
 */
  if (likely(PyList_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_v_shape_iter))) || PyTuple_CheckExact(((PyObject *)__pyx_cur_scope->__pyx_v_shape_iter))) {
    __pyx_t_1 = ((PyObject *)__pyx_cur_scope->__pyx_v_shape_iter); __Pyx_INCREF(__pyx_t_1); __pyx_t_11 = 0;
    __pyx_t_12 = NULL;
  } else {
    __pyx_t_11 = -1; __pyx_t_1 = PyObject_GetIter(((PyObject *)__pyx_cur_scope->__pyx_v_shape_iter)); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 133, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_12 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 133, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_12)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_11); __Pyx_INCREF(__pyx_t_2); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 133, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_1, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 133, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_12(__pyx_t_1);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 133, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    if ((likely(PyTuple_CheckExact(__pyx_t_2))) || (PyList_CheckExact(__pyx_t_2))) {
      PyObject* sequence = __pyx_t_2;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 133, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_8 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_8 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_6 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_8);
      __Pyx_INCREF(__pyx_t_6);
      #else
      __pyx_t_8 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_6 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      #endif
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_7 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 133, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_13 = Py_TYPE(__pyx_t_7)->tp_iternext;
      index = 0; __pyx_t_8 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_8)) goto __pyx_L23_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_8);
      index = 1; __pyx_t_6 = __pyx_t_13(__pyx_t_7); if (unlikely(!__pyx_t_6)) goto __pyx_L23_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_6);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_13(__pyx_t_7), 2) < 0) __PYX_ERR(0, 133, __pyx_L1_error)
      __pyx_t_13 = NULL;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      goto __pyx_L24_unpacking_done;
      __pyx_L23_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_13 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 133, __pyx_L1_error)
      __pyx_L24_unpacking_done:;
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_s, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_8);
    __pyx_t_8 = 0;
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_v, __pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_6);
    __pyx_t_6 = 0;

    /* "rasterio/_features.pyx":134
 *     shape_iter.fieldtype = fieldtp
 *     for s, v in shape_iter:
 *         yield s, v             # <<<<<<<<<<<<<<
 * 
 *     if mem_ds is not None:
 */
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 134, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_s);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_s);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_cur_scope->__pyx_v_s);
    __Pyx_INCREF(__pyx_cur_scope->__pyx_v_v);
    __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_v);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_cur_scope->__pyx_v_v);
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_11;
    __pyx_cur_scope->__pyx_t_2 = __pyx_t_12;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L25_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_11 = __pyx_cur_scope->__pyx_t_1;
    __pyx_t_12 = __pyx_cur_scope->__pyx_t_2;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 134, __pyx_L1_error)

    /* "rasterio/_features.pyx":133
 *     shape_iter.layer = layer
 *     shape_iter.fieldtype = fieldtp
 *     for s, v in shape_iter:             # <<<<<<<<<<<<<<
 *         yield s, v
 * 
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":136
 *         yield s, v
 * 
 *     if mem_ds is not None:             # <<<<<<<<<<<<<<
 *         mem_ds.close()
 *     if mask_ds is not None:
 */
  __pyx_t_4 = (((PyObject *)__pyx_cur_scope->__pyx_v_mem_ds) != Py_None);
  __pyx_t_5 = (__pyx_t_4 != 0);
  if (__pyx_t_5) {

    /* "rasterio/_features.pyx":137
 * 
 *     if mem_ds is not None:
 *         mem_ds.close()             # <<<<<<<<<<<<<<
 *     if mask_ds is not None:
 *         mask_ds.close()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_mem_ds), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 137, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 137, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":136
 *         yield s, v
 * 
 *     if mem_ds is not None:             # <<<<<<<<<<<<<<
 *         mem_ds.close()
 *     if mask_ds is not None:
 */
  }

  /* "rasterio/_features.pyx":138
 *     if mem_ds is not None:
 *         mem_ds.close()
 *     if mask_ds is not None:             # <<<<<<<<<<<<<<
 *         mask_ds.close()
 *     if fs != NULL:
 */
  __pyx_t_5 = (((PyObject *)__pyx_cur_scope->__pyx_v_mask_ds) != Py_None);
  __pyx_t_4 = (__pyx_t_5 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":139
 *         mem_ds.close()
 *     if mask_ds is not None:
 *         mask_ds.close()             # <<<<<<<<<<<<<<
 *     if fs != NULL:
 *         OGR_DS_Destroy(fs)
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_cur_scope->__pyx_v_mask_ds), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 139, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_1 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 139, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":138
 *     if mem_ds is not None:
 *         mem_ds.close()
 *     if mask_ds is not None:             # <<<<<<<<<<<<<<
 *         mask_ds.close()
 *     if fs != NULL:
 */
  }

  /* "rasterio/_features.pyx":140
 *     if mask_ds is not None:
 *         mask_ds.close()
 *     if fs != NULL:             # <<<<<<<<<<<<<<
 *         OGR_DS_Destroy(fs)
 *     if options:
 */
  __pyx_t_4 = ((__pyx_cur_scope->__pyx_v_fs != NULL) != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":141
 *         mask_ds.close()
 *     if fs != NULL:
 *         OGR_DS_Destroy(fs)             # <<<<<<<<<<<<<<
 *     if options:
 *         CSLDestroy(options)
 */
    OGR_DS_Destroy(__pyx_cur_scope->__pyx_v_fs);

    /* "rasterio/_features.pyx":140
 *     if mask_ds is not None:
 *         mask_ds.close()
 *     if fs != NULL:             # <<<<<<<<<<<<<<
 *         OGR_DS_Destroy(fs)
 *     if options:
 */
  }

  /* "rasterio/_features.pyx":142
 *     if fs != NULL:
 *         OGR_DS_Destroy(fs)
 *     if options:             # <<<<<<<<<<<<<<
 *         CSLDestroy(options)
 * 
 */
  __pyx_t_4 = (__pyx_cur_scope->__pyx_v_options != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":143
 *         OGR_DS_Destroy(fs)
 *     if options:
 *         CSLDestroy(options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
    CSLDestroy(__pyx_cur_scope->__pyx_v_options);

    /* "rasterio/_features.pyx":142
 *     if fs != NULL:
 *         OGR_DS_Destroy(fs)
 *     if options:             # <<<<<<<<<<<<<<
 *         CSLDestroy(options)
 * 
 */
  }
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "rasterio/_features.pyx":20
 * 
 * 
 * def _shapes(image, mask, connectivity, transform):             # <<<<<<<<<<<<<<
 *     """
 *     Return a generator of (polygon, value) for each each set of adjacent pixels
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("_shapes", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_4_sieve(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_9_features_3_sieve[] = "\n    Replaces small polygons in `image` with the value of their largest\n    neighbor.  Polygons are found for each set of neighboring pixels of the\n    same value.\n\n    Parameters\n    ----------\n    image : numpy ndarray or rasterio Band object\n        (RasterReader, bidx namedtuple)\n        Must be of type rasterio.int16, rasterio.int32, rasterio.uint8,\n        rasterio.uint16, or rasterio.float32.\n    size : int\n        minimum polygon size (number of pixels) to retain.\n    out : numpy ndarray\n        Array of same shape and data type as `image` in which to store results.\n    mask : numpy ndarray or rasterio Band object\n        Values of False or 0 will be excluded from feature generation.\n        Must evaluate to bool (rasterio.bool_ or rasterio.uint8)\n    connectivity : int\n        Use 4 or 8 pixel connectivity for grouping pixels into features.\n\n    ";
static PyMethodDef __pyx_mdef_8rasterio_9_features_4_sieve = {"_sieve", (PyCFunction)__pyx_pw_8rasterio_9_features_4_sieve, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_9_features_3_sieve};
static PyObject *__pyx_pw_8rasterio_9_features_4_sieve(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_size = 0;
  PyObject *__pyx_v_out = 0;
  PyObject *__pyx_v_mask = 0;
  PyObject *__pyx_v_connectivity = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_sieve (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_image,&__pyx_n_s_size,&__pyx_n_s_out,&__pyx_n_s_mask,&__pyx_n_s_connectivity,0};
    PyObject* values[5] = {0,0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  5: values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_size)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sieve", 1, 5, 5, 1); __PYX_ERR(0, 146, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_out)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sieve", 1, 5, 5, 2); __PYX_ERR(0, 146, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_mask)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sieve", 1, 5, 5, 3); __PYX_ERR(0, 146, __pyx_L3_error)
        }
        case  4:
        if (likely((values[4] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_connectivity)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_sieve", 1, 5, 5, 4); __PYX_ERR(0, 146, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_sieve") < 0)) __PYX_ERR(0, 146, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 5) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
      values[4] = PyTuple_GET_ITEM(__pyx_args, 4);
    }
    __pyx_v_image = values[0];
    __pyx_v_size = values[1];
    __pyx_v_out = values[2];
    __pyx_v_mask = values[3];
    __pyx_v_connectivity = values[4];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_sieve", 1, 5, 5, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 146, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("rasterio._features._sieve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_9_features_3_sieve(__pyx_self, __pyx_v_image, __pyx_v_size, __pyx_v_out, __pyx_v_mask, __pyx_v_connectivity);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8rasterio_9_features_6_sieve_2generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_features.pyx":182
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t             # <<<<<<<<<<<<<<
 *                                      in valid_dtypes))
 *         raise ValueError(
 */

static PyObject *__pyx_pf_8rasterio_9_features_6_sieve_genexpr(PyObject *__pyx_self) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("genexpr", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_2_genexpr(__pyx_ptype_8rasterio_9_features___pyx_scope_struct_2_genexpr, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 182, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_outer_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *) __pyx_self;
  __Pyx_INCREF(((PyObject *)__pyx_cur_scope->__pyx_outer_scope));
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_outer_scope);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_9_features_6_sieve_2generator2, (PyObject *) __pyx_cur_scope, __pyx_n_s_genexpr, __pyx_n_s_sieve_locals_genexpr, __pyx_n_s_rasterio__features); if (unlikely(!gen)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._features._sieve.genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_9_features_6_sieve_2generator2(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L6_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 182, __pyx_L1_error)

  /* "rasterio/_features.pyx":183
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t
 *                                      in valid_dtypes))             # <<<<<<<<<<<<<<
 *         raise ValueError(
 *             "image dtype must be one of: {0}".format(valid_types_str))
 */
  if (unlikely(!__pyx_cur_scope->__pyx_outer_scope->__pyx_v_valid_dtypes)) { __Pyx_RaiseClosureNameError("valid_dtypes"); __PYX_ERR(0, 183, __pyx_L1_error) }
  if (unlikely(__pyx_cur_scope->__pyx_outer_scope->__pyx_v_valid_dtypes == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(0, 183, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_cur_scope->__pyx_outer_scope->__pyx_v_valid_dtypes; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 183, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 183, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_t);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_t, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_3);
    __pyx_t_3 = 0;

    /* "rasterio/_features.pyx":182
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t             # <<<<<<<<<<<<<<
 *                                      in valid_dtypes))
 *         raise ValueError(
 */
    __pyx_t_4 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_rasterio_0, __pyx_n_s_format); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_4))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_4);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_4);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_4, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_4, __pyx_cur_scope->__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_t};
        __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_4)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_cur_scope->__pyx_v_t};
        __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_4, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_t);
        __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_t);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_cur_scope->__pyx_v_t);
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_4, __pyx_t_6, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 182, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_r = __pyx_t_3;
    __pyx_t_3 = 0;
    __Pyx_XGIVEREF(__pyx_t_1);
    __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
    __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
    __Pyx_XGIVEREF(__pyx_r);
    __Pyx_RefNannyFinishContext();
    /* return from generator, yielding value */
    __pyx_generator->resume_label = 1;
    return __pyx_r;
    __pyx_L6_resume_from_yield:;
    __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
    __pyx_cur_scope->__pyx_t_0 = 0;
    __Pyx_XGOTREF(__pyx_t_1);
    __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
    if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 182, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("genexpr", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */

static PyObject *__pyx_pf_8rasterio_9_features_3_sieve(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_image, PyObject *__pyx_v_size, PyObject *__pyx_v_out, PyObject *__pyx_v_mask, PyObject *__pyx_v_connectivity) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *__pyx_cur_scope;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_in_mem_ds = 0;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_out_mem_ds = 0;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_mask_mem_ds = 0;
  GDALRasterBandH __pyx_v_in_band;
  GDALRasterBandH __pyx_v_out_band;
  GDALRasterBandH __pyx_v_mask_band;
  PyObject *__pyx_v_valid_types_str = NULL;
  PyObject *__pyx_v_rdr = NULL;
  PyObject *__pyx_v_udr = NULL;
  PyObject *__pyx_v_mask_reader = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  GDALRasterBandH __pyx_t_7;
  int __pyx_t_8;
  GDALRasterBandH __pyx_t_9;
  int __pyx_t_10;
  __Pyx_RefNannySetupContext("_sieve", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_1__sieve(__pyx_ptype_8rasterio_9_features___pyx_scope_struct_1__sieve, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 146, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }

  /* "rasterio/_features.pyx":172
 *     cdef int rows
 *     cdef int cols
 *     cdef InMemoryRaster in_mem_ds = None             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster out_mem_ds = None
 *     cdef InMemoryRaster mask_mem_ds = None
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_in_mem_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":173
 *     cdef int cols
 *     cdef InMemoryRaster in_mem_ds = None
 *     cdef InMemoryRaster out_mem_ds = None             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster mask_mem_ds = None
 *     cdef GDALRasterBandH in_band = NULL
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_out_mem_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":174
 *     cdef InMemoryRaster in_mem_ds = None
 *     cdef InMemoryRaster out_mem_ds = None
 *     cdef InMemoryRaster mask_mem_ds = None             # <<<<<<<<<<<<<<
 *     cdef GDALRasterBandH in_band = NULL
 *     cdef GDALRasterBandH out_band = NULL
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_mask_mem_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":175
 *     cdef InMemoryRaster out_mem_ds = None
 *     cdef InMemoryRaster mask_mem_ds = None
 *     cdef GDALRasterBandH in_band = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALRasterBandH out_band = NULL
 *     cdef GDALRasterBandH mask_band = NULL
 */
  __pyx_v_in_band = NULL;

  /* "rasterio/_features.pyx":176
 *     cdef InMemoryRaster mask_mem_ds = None
 *     cdef GDALRasterBandH in_band = NULL
 *     cdef GDALRasterBandH out_band = NULL             # <<<<<<<<<<<<<<
 *     cdef GDALRasterBandH mask_band = NULL
 * 
 */
  __pyx_v_out_band = NULL;

  /* "rasterio/_features.pyx":177
 *     cdef GDALRasterBandH in_band = NULL
 *     cdef GDALRasterBandH out_band = NULL
 *     cdef GDALRasterBandH mask_band = NULL             # <<<<<<<<<<<<<<
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16')
 */
  __pyx_v_mask_band = NULL;

  /* "rasterio/_features.pyx":179
 *     cdef GDALRasterBandH mask_band = NULL
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 */
  __Pyx_INCREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);
  __pyx_cur_scope->__pyx_v_valid_dtypes = __pyx_tuple__12;

  /* "rasterio/_features.pyx":181
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16')
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:             # <<<<<<<<<<<<<<
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t
 *                                      in valid_dtypes))
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_dtype); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_1);
  PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
  __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_name); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = (__Pyx_PySequence_ContainsTF(__pyx_t_2, __pyx_cur_scope->__pyx_v_valid_dtypes, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 181, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":182
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t             # <<<<<<<<<<<<<<
 *                                      in valid_dtypes))
 *         raise ValueError(
 */
    __pyx_t_2 = __pyx_pf_8rasterio_9_features_6_sieve_genexpr(((PyObject*)__pyx_cur_scope)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyString_Join(__pyx_kp_s__2, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 182, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_v_valid_types_str = ((PyObject*)__pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":185
 *                                      in valid_dtypes))
 *         raise ValueError(
 *             "image dtype must be one of: {0}".format(valid_types_str))             # <<<<<<<<<<<<<<
 * 
 *     if size <= 0:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_kp_s_image_dtype_must_be_one_of_0, __pyx_n_s_format); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 185, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (!__pyx_t_5) {
      __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_v_valid_types_str); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_valid_types_str};
        __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_2)) {
        PyObject *__pyx_temp[2] = {__pyx_t_5, __pyx_v_valid_types_str};
        __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_2, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
        __Pyx_GOTREF(__pyx_t_1);
      } else
      #endif
      {
        __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_6);
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_5); __pyx_t_5 = NULL;
        __Pyx_INCREF(__pyx_v_valid_types_str);
        __Pyx_GIVEREF(__pyx_v_valid_types_str);
        PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_valid_types_str);
        __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 185, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "rasterio/_features.pyx":184
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t
 *                                      in valid_dtypes))
 *         raise ValueError(             # <<<<<<<<<<<<<<
 *             "image dtype must be one of: {0}".format(valid_types_str))
 * 
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 184, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 184, __pyx_L1_error)

    /* "rasterio/_features.pyx":181
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16')
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:             # <<<<<<<<<<<<<<
 *         valid_types_str = ', '.join(('rasterio.{0}'.format(t) for t
 *                                      in valid_dtypes))
 */
  }

  /* "rasterio/_features.pyx":187
 *             "image dtype must be one of: {0}".format(valid_types_str))
 * 
 *     if size <= 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:
 */
  __pyx_t_1 = PyObject_RichCompare(__pyx_v_size, __pyx_int_0, Py_LE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 187, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 187, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":188
 * 
 *     if size <= 0:
 *         raise ValueError('size must be greater than 0')             # <<<<<<<<<<<<<<
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__13, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 188, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 188, __pyx_L1_error)

    /* "rasterio/_features.pyx":187
 *             "image dtype must be one of: {0}".format(valid_types_str))
 * 
 *     if size <= 0:             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:
 */
  }

  /* "rasterio/_features.pyx":189
 *     if size <= 0:
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):
 */
  __pyx_t_1 = PyObject_RichCompare(((PyObject *)Py_TYPE(__pyx_v_size)), ((PyObject *)(&PyFloat_Type)), Py_EQ); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 189, __pyx_L1_error)
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 189, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":190
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')             # <<<<<<<<<<<<<<
 *     elif size > (image.shape[0] * image.shape[1]):
 *         raise ValueError('size must be smaller than size of image')
 */
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__14, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 190, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 190, __pyx_L1_error)

    /* "rasterio/_features.pyx":189
 *     if size <= 0:
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):
 */
  }

  /* "rasterio/_features.pyx":191
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be smaller than size of image')
 * 
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_GetItemInt(__pyx_t_1, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_6 = __Pyx_GetItemInt(__pyx_t_1, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = PyNumber_Multiply(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = PyObject_RichCompare(__pyx_v_size, __pyx_t_1, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 191, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":192
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):
 *         raise ValueError('size must be smaller than size of image')             # <<<<<<<<<<<<<<
 * 
 *     if connectivity not in (4, 8):
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__15, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 192, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 192, __pyx_L1_error)

    /* "rasterio/_features.pyx":191
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):             # <<<<<<<<<<<<<<
 *         raise ValueError('size must be smaller than size of image')
 * 
 */
  }

  /* "rasterio/_features.pyx":194
 *         raise ValueError('size must be smaller than size of image')
 * 
 *     if connectivity not in (4, 8):             # <<<<<<<<<<<<<<
 *         raise ValueError('connectivity must be 4 or 8')
 * 
 */
  __Pyx_INCREF(__pyx_v_connectivity);
  __pyx_t_6 = __pyx_v_connectivity;
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_6, __pyx_int_4, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_3) {
  } else {
    __pyx_t_4 = __pyx_t_3;
    goto __pyx_L6_bool_binop_done;
  }
  __pyx_t_1 = PyObject_RichCompare(__pyx_t_6, __pyx_int_8, Py_NE); __Pyx_XGOTREF(__pyx_t_1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 194, __pyx_L1_error)
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 194, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_4 = __pyx_t_3;
  __pyx_L6_bool_binop_done:;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":195
 * 
 *     if connectivity not in (4, 8):
 *         raise ValueError('connectivity must be 4 or 8')             # <<<<<<<<<<<<<<
 * 
 *     if out.shape != image.shape:
 */
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__16, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 195, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 195, __pyx_L1_error)

    /* "rasterio/_features.pyx":194
 *         raise ValueError('size must be smaller than size of image')
 * 
 *     if connectivity not in (4, 8):             # <<<<<<<<<<<<<<
 *         raise ValueError('connectivity must be 4 or 8')
 * 
 */
  }

  /* "rasterio/_features.pyx":197
 *         raise ValueError('connectivity must be 4 or 8')
 * 
 *     if out.shape != image.shape:             # <<<<<<<<<<<<<<
 *         raise ValueError('out raster shape must be same as image shape')
 * 
 */
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 197, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":198
 * 
 *     if out.shape != image.shape:
 *         raise ValueError('out raster shape must be same as image shape')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__17, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 198, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 198, __pyx_L1_error)

    /* "rasterio/_features.pyx":197
 *         raise ValueError('connectivity must be 4 or 8')
 * 
 *     if out.shape != image.shape:             # <<<<<<<<<<<<<<
 *         raise ValueError('out raster shape must be same as image shape')
 * 
 */
  }

  /* "rasterio/_features.pyx":200
 *         raise ValueError('out raster shape must be same as image shape')
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:             # <<<<<<<<<<<<<<
 *         raise ValueError('out raster must match dtype of image')
 * 
 */
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_6 = PyTuple_New(1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_GIVEREF(__pyx_t_2);
  PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
  __pyx_t_2 = 0;
  __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_6, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_name); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = PyObject_RichCompare(__pyx_t_1, __pyx_t_6, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 200, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":201
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:
 *         raise ValueError('out raster must match dtype of image')             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(image):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__18, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 201, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 201, __pyx_L1_error)

    /* "rasterio/_features.pyx":200
 *         raise ValueError('out raster shape must be same as image shape')
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:             # <<<<<<<<<<<<<<
 *         raise ValueError('out raster must match dtype of image')
 * 
 */
  }

  /* "rasterio/_features.pyx":203
 *         raise ValueError('out raster must match dtype of image')
 * 
 *     if dtypes.is_ndarray(image):             # <<<<<<<<<<<<<<
 *         in_mem_ds = InMemoryRaster(image=image)
 *         in_band = in_mem_ds.band(1)
 */
  __pyx_t_6 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_6);
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_t_6, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_1))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_1);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_1);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_1, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_1, __pyx_v_image); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_image};
      __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_1)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_v_image};
      __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_1, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_2);
    } else
    #endif
    {
      __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_INCREF(__pyx_v_image);
      __Pyx_GIVEREF(__pyx_v_image);
      PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_image);
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 203, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 203, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":204
 * 
 *     if dtypes.is_ndarray(image):
 *         in_mem_ds = InMemoryRaster(image=image)             # <<<<<<<<<<<<<<
 *         in_band = in_mem_ds.band(1)
 *     elif isinstance(image, tuple):
 */
    __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_image, __pyx_v_image) < 0) __PYX_ERR(0, 204, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 204, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_in_mem_ds, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_1));
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":205
 *     if dtypes.is_ndarray(image):
 *         in_mem_ds = InMemoryRaster(image=image)
 *         in_band = in_mem_ds.band(1)             # <<<<<<<<<<<<<<
 *     elif isinstance(image, tuple):
 *         rdr = image.ds
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_v_in_mem_ds->__pyx_vtab)->band(__pyx_v_in_mem_ds, 1); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(0, 205, __pyx_L1_error)
    __pyx_v_in_band = __pyx_t_7;

    /* "rasterio/_features.pyx":203
 *         raise ValueError('out raster must match dtype of image')
 * 
 *     if dtypes.is_ndarray(image):             # <<<<<<<<<<<<<<
 *         in_mem_ds = InMemoryRaster(image=image)
 *         in_band = in_mem_ds.band(1)
 */
    goto __pyx_L10;
  }

  /* "rasterio/_features.pyx":206
 *         in_mem_ds = InMemoryRaster(image=image)
 *         in_band = in_mem_ds.band(1)
 *     elif isinstance(image, tuple):             # <<<<<<<<<<<<<<
 *         rdr = image.ds
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)
 */
  __pyx_t_3 = PyTuple_Check(__pyx_v_image); 
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":207
 *         in_band = in_mem_ds.band(1)
 *     elif isinstance(image, tuple):
 *         rdr = image.ds             # <<<<<<<<<<<<<<
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 */
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_ds); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 207, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_v_rdr = __pyx_t_1;
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":208
 *     elif isinstance(image, tuple):
 *         rdr = image.ds
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Invalid source image")
 */
    if (!(likely(__Pyx_TypeTest(__pyx_v_rdr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 208, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_bidx); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_1); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 208, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_9 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_rdr)->__pyx_base.__pyx_vtab)->__pyx_base.band(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_rdr)), __pyx_t_8); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(0, 208, __pyx_L1_error)
    __pyx_v_in_band = __pyx_t_9;

    /* "rasterio/_features.pyx":206
 *         in_mem_ds = InMemoryRaster(image=image)
 *         in_band = in_mem_ds.band(1)
 *     elif isinstance(image, tuple):             # <<<<<<<<<<<<<<
 *         rdr = image.ds
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)
 */
    goto __pyx_L10;
  }

  /* "rasterio/_features.pyx":210
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 *         raise ValueError("Invalid source image")             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(out):
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__19, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 210, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 210, __pyx_L1_error)
  }
  __pyx_L10:;

  /* "rasterio/_features.pyx":212
 *         raise ValueError("Invalid source image")
 * 
 *     if dtypes.is_ndarray(out):             # <<<<<<<<<<<<<<
 *         log.debug("out array: %r", out)
 *         out_mem_ds = InMemoryRaster(image=out)
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_2, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_t_2 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_2)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_2);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_2) {
    __pyx_t_1 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_v_out); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_out};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_2, __pyx_v_out};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_6 = PyTuple_New(1+1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GIVEREF(__pyx_t_2); PyTuple_SET_ITEM(__pyx_t_6, 0, __pyx_t_2); __pyx_t_2 = NULL;
      __Pyx_INCREF(__pyx_v_out);
      __Pyx_GIVEREF(__pyx_v_out);
      PyTuple_SET_ITEM(__pyx_t_6, 0+1, __pyx_v_out);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_6, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 212, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_1); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 212, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":213
 * 
 *     if dtypes.is_ndarray(out):
 *         log.debug("out array: %r", out)             # <<<<<<<<<<<<<<
 *         out_mem_ds = InMemoryRaster(image=out)
 *         out_band = out_mem_ds.band(1)
 */
    __pyx_t_5 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_5, __pyx_n_s_debug); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 213, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
    __pyx_t_5 = NULL;
    __pyx_t_8 = 0;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_5)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_5);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
        __pyx_t_8 = 1;
      }
    }
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_out_array_r, __pyx_v_out};
      __pyx_t_1 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
      PyObject *__pyx_temp[3] = {__pyx_t_5, __pyx_kp_s_out_array_r, __pyx_v_out};
      __pyx_t_1 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-__pyx_t_8, 2+__pyx_t_8); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_GOTREF(__pyx_t_1);
    } else
    #endif
    {
      __pyx_t_2 = PyTuple_New(2+__pyx_t_8); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (__pyx_t_5) {
        __Pyx_GIVEREF(__pyx_t_5); PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_5); __pyx_t_5 = NULL;
      }
      __Pyx_INCREF(__pyx_kp_s_out_array_r);
      __Pyx_GIVEREF(__pyx_kp_s_out_array_r);
      PyTuple_SET_ITEM(__pyx_t_2, 0+__pyx_t_8, __pyx_kp_s_out_array_r);
      __Pyx_INCREF(__pyx_v_out);
      __Pyx_GIVEREF(__pyx_v_out);
      PyTuple_SET_ITEM(__pyx_t_2, 1+__pyx_t_8, __pyx_v_out);
      __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 213, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_1);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":214
 *     if dtypes.is_ndarray(out):
 *         log.debug("out array: %r", out)
 *         out_mem_ds = InMemoryRaster(image=out)             # <<<<<<<<<<<<<<
 *         out_band = out_mem_ds.band(1)
 *     elif isinstance(out, tuple):
 */
    __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_image, __pyx_v_out) < 0) __PYX_ERR(0, 214, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_1); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 214, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_DECREF_SET(__pyx_v_out_mem_ds, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "rasterio/_features.pyx":215
 *         log.debug("out array: %r", out)
 *         out_mem_ds = InMemoryRaster(image=out)
 *         out_band = out_mem_ds.band(1)             # <<<<<<<<<<<<<<
 *     elif isinstance(out, tuple):
 *         udr = out.ds
 */
    __pyx_t_7 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_v_out_mem_ds->__pyx_vtab)->band(__pyx_v_out_mem_ds, 1); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(0, 215, __pyx_L1_error)
    __pyx_v_out_band = __pyx_t_7;

    /* "rasterio/_features.pyx":212
 *         raise ValueError("Invalid source image")
 * 
 *     if dtypes.is_ndarray(out):             # <<<<<<<<<<<<<<
 *         log.debug("out array: %r", out)
 *         out_mem_ds = InMemoryRaster(image=out)
 */
    goto __pyx_L11;
  }

  /* "rasterio/_features.pyx":216
 *         out_mem_ds = InMemoryRaster(image=out)
 *         out_band = out_mem_ds.band(1)
 *     elif isinstance(out, tuple):             # <<<<<<<<<<<<<<
 *         udr = out.ds
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)
 */
  __pyx_t_4 = PyTuple_Check(__pyx_v_out); 
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":217
 *         out_band = out_mem_ds.band(1)
 *     elif isinstance(out, tuple):
 *         udr = out.ds             # <<<<<<<<<<<<<<
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)
 *     else:
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_ds); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 217, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_v_udr = __pyx_t_6;
    __pyx_t_6 = 0;

    /* "rasterio/_features.pyx":218
 *     elif isinstance(out, tuple):
 *         udr = out.ds
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)             # <<<<<<<<<<<<<<
 *     else:
 *         raise ValueError("Invalid out image")
 */
    if (!(likely(__Pyx_TypeTest(__pyx_v_udr, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 218, __pyx_L1_error)
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_out, __pyx_n_s_bidx); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_6); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 218, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_9 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_udr)->__pyx_base.__pyx_vtab)->__pyx_base.band(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_udr)), __pyx_t_8); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(0, 218, __pyx_L1_error)
    __pyx_v_out_band = __pyx_t_9;

    /* "rasterio/_features.pyx":216
 *         out_mem_ds = InMemoryRaster(image=out)
 *         out_band = out_mem_ds.band(1)
 *     elif isinstance(out, tuple):             # <<<<<<<<<<<<<<
 *         udr = out.ds
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)
 */
    goto __pyx_L11;
  }

  /* "rasterio/_features.pyx":220
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)
 *     else:
 *         raise ValueError("Invalid out image")             # <<<<<<<<<<<<<<
 * 
 *     if mask is not None:
 */
  /*else*/ {
    __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__20, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 220, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_Raise(__pyx_t_6, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __PYX_ERR(0, 220, __pyx_L1_error)
  }
  __pyx_L11:;

  /* "rasterio/_features.pyx":222
 *         raise ValueError("Invalid out image")
 * 
 *     if mask is not None:             # <<<<<<<<<<<<<<
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")
 */
  __pyx_t_3 = (__pyx_v_mask != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":223
 * 
 *     if mask is not None:
 *         if mask.shape != image.shape:             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must have same shape as image")
 * 
 */
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_shape); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_image, __pyx_n_s_shape); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyObject_RichCompare(__pyx_t_6, __pyx_t_1, Py_NE); __Pyx_XGOTREF(__pyx_t_2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 223, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_4) {

      /* "rasterio/_features.pyx":224
 *     if mask is not None:
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")             # <<<<<<<<<<<<<<
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__21, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 224, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 224, __pyx_L1_error)

      /* "rasterio/_features.pyx":223
 * 
 *     if mask is not None:
 *         if mask.shape != image.shape:             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must have same shape as image")
 * 
 */
    }

    /* "rasterio/_features.pyx":226
 *             raise ValueError("Mask must have same shape as image")
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "
 *                              "rasterio.uint8")
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_dtype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = PyTuple_New(1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_GIVEREF(__pyx_t_2);
    PyTuple_SET_ITEM(__pyx_t_1, 0, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_5numpy_dtype), __pyx_t_1, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 226, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_3 = (__Pyx_PyString_Equals(((PyObject *)__pyx_t_2), __pyx_n_s_bool, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 226, __pyx_L1_error)
    if (__pyx_t_3) {
    } else {
      __pyx_t_4 = __pyx_t_3;
      goto __pyx_L15_bool_binop_done;
    }
    __pyx_t_3 = (__Pyx_PyString_Equals(((PyObject *)__pyx_t_2), __pyx_n_s_uint8, Py_NE)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 226, __pyx_L1_error)
    __pyx_t_4 = __pyx_t_3;
    __pyx_L15_bool_binop_done:;
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_3 = (__pyx_t_4 != 0);
    if (__pyx_t_3) {

      /* "rasterio/_features.pyx":227
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "             # <<<<<<<<<<<<<<
 *                              "rasterio.uint8")
 * 
 */
      __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__22, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 227, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_Raise(__pyx_t_2, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __PYX_ERR(0, 227, __pyx_L1_error)

      /* "rasterio/_features.pyx":226
 *             raise ValueError("Mask must have same shape as image")
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):             # <<<<<<<<<<<<<<
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "
 *                              "rasterio.uint8")
 */
    }

    /* "rasterio/_features.pyx":230
 *                              "rasterio.uint8")
 * 
 *         if dtypes.is_ndarray(mask):             # <<<<<<<<<<<<<<
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_mem_ds = InMemoryRaster(image=mask.astype('uint8'))
 */
    __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_dtypes); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_t_1, __pyx_n_s_is_ndarray); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_t_1 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_6))) {
      __pyx_t_1 = PyMethod_GET_SELF(__pyx_t_6);
      if (likely(__pyx_t_1)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_1);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_6, function);
      }
    }
    if (!__pyx_t_1) {
      __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_v_mask); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_mask};
        __pyx_t_2 = __Pyx_PyFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_6)) {
        PyObject *__pyx_temp[2] = {__pyx_t_1, __pyx_v_mask};
        __pyx_t_2 = __Pyx_PyCFunction_FastCall(__pyx_t_6, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
        __Pyx_GOTREF(__pyx_t_2);
      } else
      #endif
      {
        __pyx_t_5 = PyTuple_New(1+1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_5);
        __Pyx_GIVEREF(__pyx_t_1); PyTuple_SET_ITEM(__pyx_t_5, 0, __pyx_t_1); __pyx_t_1 = NULL;
        __Pyx_INCREF(__pyx_v_mask);
        __Pyx_GIVEREF(__pyx_v_mask);
        PyTuple_SET_ITEM(__pyx_t_5, 0+1, __pyx_v_mask);
        __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_t_5, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 230, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    __pyx_t_3 = __Pyx_PyObject_IsTrue(__pyx_t_2); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 230, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    if (__pyx_t_3) {

      /* "rasterio/_features.pyx":232
 *         if dtypes.is_ndarray(mask):
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_mem_ds = InMemoryRaster(image=mask.astype('uint8'))             # <<<<<<<<<<<<<<
 *             mask_band = mask_mem_ds.band(1)
 * 
 */
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_6 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_astype); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_5 = __Pyx_PyObject_Call(__pyx_t_6, __pyx_tuple__23, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_image, __pyx_t_5) < 0) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 232, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __Pyx_DECREF_SET(__pyx_v_mask_mem_ds, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_5));
      __pyx_t_5 = 0;

      /* "rasterio/_features.pyx":233
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_mem_ds = InMemoryRaster(image=mask.astype('uint8'))
 *             mask_band = mask_mem_ds.band(1)             # <<<<<<<<<<<<<<
 * 
 *         elif isinstance(mask, tuple):
 */
      __pyx_t_7 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_v_mask_mem_ds->__pyx_vtab)->band(__pyx_v_mask_mem_ds, 1); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(0, 233, __pyx_L1_error)
      __pyx_v_mask_band = __pyx_t_7;

      /* "rasterio/_features.pyx":230
 *                              "rasterio.uint8")
 * 
 *         if dtypes.is_ndarray(mask):             # <<<<<<<<<<<<<<
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_mem_ds = InMemoryRaster(image=mask.astype('uint8'))
 */
      goto __pyx_L17;
    }

    /* "rasterio/_features.pyx":235
 *             mask_band = mask_mem_ds.band(1)
 * 
 *         elif isinstance(mask, tuple):             # <<<<<<<<<<<<<<
 *             mask_reader = mask.ds
 *             mask_band = (<DatasetReaderBase?>mask_reader).band(mask.bidx)
 */
    __pyx_t_3 = PyTuple_Check(__pyx_v_mask); 
    __pyx_t_4 = (__pyx_t_3 != 0);
    if (__pyx_t_4) {

      /* "rasterio/_features.pyx":236
 * 
 *         elif isinstance(mask, tuple):
 *             mask_reader = mask.ds             # <<<<<<<<<<<<<<
 *             mask_band = (<DatasetReaderBase?>mask_reader).band(mask.bidx)
 * 
 */
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_ds); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 236, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_v_mask_reader = __pyx_t_5;
      __pyx_t_5 = 0;

      /* "rasterio/_features.pyx":237
 *         elif isinstance(mask, tuple):
 *             mask_reader = mask.ds
 *             mask_band = (<DatasetReaderBase?>mask_reader).band(mask.bidx)             # <<<<<<<<<<<<<<
 * 
 *     GDALSieveFilter(in_band, mask_band, out_band, size, connectivity,
 */
      if (!(likely(__Pyx_TypeTest(__pyx_v_mask_reader, __pyx_ptype_8rasterio_3_io_DatasetReaderBase)))) __PYX_ERR(0, 237, __pyx_L1_error)
      __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_v_mask, __pyx_n_s_bidx); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_t_5); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 237, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_9 = ((struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_mask_reader)->__pyx_base.__pyx_vtab)->__pyx_base.band(((struct __pyx_obj_8rasterio_5_base_DatasetBase *)((struct __pyx_obj_8rasterio_3_io_DatasetReaderBase *)__pyx_v_mask_reader)), __pyx_t_8); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(0, 237, __pyx_L1_error)
      __pyx_v_mask_band = __pyx_t_9;

      /* "rasterio/_features.pyx":235
 *             mask_band = mask_mem_ds.band(1)
 * 
 *         elif isinstance(mask, tuple):             # <<<<<<<<<<<<<<
 *             mask_reader = mask.ds
 *             mask_band = (<DatasetReaderBase?>mask_reader).band(mask.bidx)
 */
    }
    __pyx_L17:;

    /* "rasterio/_features.pyx":222
 *         raise ValueError("Invalid out image")
 * 
 *     if mask is not None:             # <<<<<<<<<<<<<<
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")
 */
  }

  /* "rasterio/_features.pyx":239
 *             mask_band = (<DatasetReaderBase?>mask_reader).band(mask.bidx)
 * 
 *     GDALSieveFilter(in_band, mask_band, out_band, size, connectivity,             # <<<<<<<<<<<<<<
 *                           NULL, NULL, NULL)
 * 
 */
  __pyx_t_8 = __Pyx_PyInt_As_int(__pyx_v_size); if (unlikely((__pyx_t_8 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)
  __pyx_t_10 = __Pyx_PyInt_As_int(__pyx_v_connectivity); if (unlikely((__pyx_t_10 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 239, __pyx_L1_error)

  /* "rasterio/_features.pyx":240
 * 
 *     GDALSieveFilter(in_band, mask_band, out_band, size, connectivity,
 *                           NULL, NULL, NULL)             # <<<<<<<<<<<<<<
 * 
 *     # Read from out_band into out
 */
  GDALSieveFilter(__pyx_v_in_band, __pyx_v_mask_band, __pyx_v_out_band, __pyx_t_8, __pyx_t_10, NULL, NULL, NULL);

  /* "rasterio/_features.pyx":243
 * 
 *     # Read from out_band into out
 *     io_auto(out, out_band, False)             # <<<<<<<<<<<<<<
 * 
 *     if in_mem_ds is not None:
 */
  __pyx_f_8rasterio_3_io_io_auto(__pyx_v_out, __pyx_v_out_band, 0, NULL);

  /* "rasterio/_features.pyx":245
 *     io_auto(out, out_band, False)
 * 
 *     if in_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         in_mem_ds.close()
 *     if out_mem_ds is not None:
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_in_mem_ds) != Py_None);
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":246
 * 
 *     if in_mem_ds is not None:
 *         in_mem_ds.close()             # <<<<<<<<<<<<<<
 *     if out_mem_ds is not None:
 *         out_mem_ds.close()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_in_mem_ds), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 246, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 246, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "rasterio/_features.pyx":245
 *     io_auto(out, out_band, False)
 * 
 *     if in_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         in_mem_ds.close()
 *     if out_mem_ds is not None:
 */
  }

  /* "rasterio/_features.pyx":247
 *     if in_mem_ds is not None:
 *         in_mem_ds.close()
 *     if out_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         out_mem_ds.close()
 *     if mask_mem_ds is not None:
 */
  __pyx_t_3 = (((PyObject *)__pyx_v_out_mem_ds) != Py_None);
  __pyx_t_4 = (__pyx_t_3 != 0);
  if (__pyx_t_4) {

    /* "rasterio/_features.pyx":248
 *         in_mem_ds.close()
 *     if out_mem_ds is not None:
 *         out_mem_ds.close()             # <<<<<<<<<<<<<<
 *     if mask_mem_ds is not None:
 *         mask_mem_ds.close()
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_out_mem_ds), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 248, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 248, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 248, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "rasterio/_features.pyx":247
 *     if in_mem_ds is not None:
 *         in_mem_ds.close()
 *     if out_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         out_mem_ds.close()
 *     if mask_mem_ds is not None:
 */
  }

  /* "rasterio/_features.pyx":249
 *     if out_mem_ds is not None:
 *         out_mem_ds.close()
 *     if mask_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         mask_mem_ds.close()
 * 
 */
  __pyx_t_4 = (((PyObject *)__pyx_v_mask_mem_ds) != Py_None);
  __pyx_t_3 = (__pyx_t_4 != 0);
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":250
 *         out_mem_ds.close()
 *     if mask_mem_ds is not None:
 *         mask_mem_ds.close()             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mask_mem_ds), __pyx_n_s_close); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 250, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = NULL;
    if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_2))) {
      __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_2);
      if (likely(__pyx_t_6)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_2, function);
      }
    }
    if (__pyx_t_6) {
      __pyx_t_5 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_6); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 250, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    } else {
      __pyx_t_5 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 250, __pyx_L1_error)
    }
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

    /* "rasterio/_features.pyx":249
 *     if out_mem_ds is not None:
 *         out_mem_ds.close()
 *     if mask_mem_ds is not None:             # <<<<<<<<<<<<<<
 *         mask_mem_ds.close()
 * 
 */
  }

  /* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("rasterio._features._sieve", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_in_mem_ds);
  __Pyx_XDECREF((PyObject *)__pyx_v_out_mem_ds);
  __Pyx_XDECREF((PyObject *)__pyx_v_mask_mem_ds);
  __Pyx_XDECREF(__pyx_v_valid_types_str);
  __Pyx_XDECREF(__pyx_v_rdr);
  __Pyx_XDECREF(__pyx_v_udr);
  __Pyx_XDECREF(__pyx_v_mask_reader);
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":253
 * 
 * 
 * def _rasterize(shapes, image, transform, all_touched):             # <<<<<<<<<<<<<<
 *     """
 *     Burns input geometries into `image`.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_6_rasterize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds); /*proto*/
static char __pyx_doc_8rasterio_9_features_5_rasterize[] = "\n    Burns input geometries into `image`.\n\n    Parameters\n    ----------\n    shapes : iterable of (geometry, value) pairs\n        `geometry` is a GeoJSON-like object.\n    image : numpy ndarray\n        Array in which to store results.\n    transform : Affine transformation object, optional\n        Transformation from pixel coordinates of `image` to the\n        coordinate system of the input `shapes`. See the `transform`\n        property of dataset objects.\n    all_touched : boolean, optional\n        If True, all pixels touched by geometries will be burned in.\n        If false, only pixels whose center is within the polygon or\n        that are selected by Bresenham's line algorithm will be burned\n        in.\n    ";
static PyMethodDef __pyx_mdef_8rasterio_9_features_6_rasterize = {"_rasterize", (PyCFunction)__pyx_pw_8rasterio_9_features_6_rasterize, METH_VARARGS|METH_KEYWORDS, __pyx_doc_8rasterio_9_features_5_rasterize};
static PyObject *__pyx_pw_8rasterio_9_features_6_rasterize(PyObject *__pyx_self, PyObject *__pyx_args, PyObject *__pyx_kwds) {
  PyObject *__pyx_v_shapes = 0;
  PyObject *__pyx_v_image = 0;
  PyObject *__pyx_v_transform = 0;
  PyObject *__pyx_v_all_touched = 0;
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_rasterize (wrapper)", 0);
  {
    static PyObject **__pyx_pyargnames[] = {&__pyx_n_s_shapes_2,&__pyx_n_s_image,&__pyx_n_s_transform,&__pyx_n_s_all_touched,0};
    PyObject* values[4] = {0,0,0,0};
    if (unlikely(__pyx_kwds)) {
      Py_ssize_t kw_args;
      const Py_ssize_t pos_args = PyTuple_GET_SIZE(__pyx_args);
      switch (pos_args) {
        case  4: values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
        case  3: values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
        case  2: values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
        case  1: values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
        case  0: break;
        default: goto __pyx_L5_argtuple_error;
      }
      kw_args = PyDict_Size(__pyx_kwds);
      switch (pos_args) {
        case  0:
        if (likely((values[0] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_shapes_2)) != 0)) kw_args--;
        else goto __pyx_L5_argtuple_error;
        case  1:
        if (likely((values[1] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_image)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rasterize", 1, 4, 4, 1); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        case  2:
        if (likely((values[2] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_transform)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rasterize", 1, 4, 4, 2); __PYX_ERR(0, 253, __pyx_L3_error)
        }
        case  3:
        if (likely((values[3] = PyDict_GetItem(__pyx_kwds, __pyx_n_s_all_touched)) != 0)) kw_args--;
        else {
          __Pyx_RaiseArgtupleInvalid("_rasterize", 1, 4, 4, 3); __PYX_ERR(0, 253, __pyx_L3_error)
        }
      }
      if (unlikely(kw_args > 0)) {
        if (unlikely(__Pyx_ParseOptionalKeywords(__pyx_kwds, __pyx_pyargnames, 0, values, pos_args, "_rasterize") < 0)) __PYX_ERR(0, 253, __pyx_L3_error)
      }
    } else if (PyTuple_GET_SIZE(__pyx_args) != 4) {
      goto __pyx_L5_argtuple_error;
    } else {
      values[0] = PyTuple_GET_ITEM(__pyx_args, 0);
      values[1] = PyTuple_GET_ITEM(__pyx_args, 1);
      values[2] = PyTuple_GET_ITEM(__pyx_args, 2);
      values[3] = PyTuple_GET_ITEM(__pyx_args, 3);
    }
    __pyx_v_shapes = values[0];
    __pyx_v_image = values[1];
    __pyx_v_transform = values[2];
    __pyx_v_all_touched = values[3];
  }
  goto __pyx_L4_argument_unpacking_done;
  __pyx_L5_argtuple_error:;
  __Pyx_RaiseArgtupleInvalid("_rasterize", 1, 4, 4, PyTuple_GET_SIZE(__pyx_args)); __PYX_ERR(0, 253, __pyx_L3_error)
  __pyx_L3_error:;
  __Pyx_AddTraceback("rasterio._features._rasterize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __Pyx_RefNannyFinishContext();
  return NULL;
  __pyx_L4_argument_unpacking_done:;
  __pyx_r = __pyx_pf_8rasterio_9_features_5_rasterize(__pyx_self, __pyx_v_shapes, __pyx_v_image, __pyx_v_transform, __pyx_v_all_touched);

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_5_rasterize(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_shapes, PyObject *__pyx_v_image, PyObject *__pyx_v_transform, PyObject *__pyx_v_all_touched) {
  size_t __pyx_v_i;
  size_t __pyx_v_num_geoms;
  OGRGeometryH *__pyx_v_geoms;
  char **__pyx_v_options;
  double *__pyx_v_pixel_values;
  struct __pyx_obj_8rasterio_3_io_InMemoryRaster *__pyx_v_mem = 0;
  PyObject *__pyx_v_all_shapes = NULL;
  PyObject *__pyx_v_geometry = NULL;
  PyObject *__pyx_v_value = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  Py_ssize_t __pyx_t_3;
  size_t __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *(*__pyx_t_9)(PyObject *);
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  OGRGeometryH __pyx_t_13;
  double __pyx_t_14;
  PyObject *__pyx_t_15 = NULL;
  PyObject *__pyx_t_16 = NULL;
  PyObject *__pyx_t_17 = NULL;
  int __pyx_t_18;
  PyObject *__pyx_t_19 = NULL;
  PyObject *__pyx_t_20 = NULL;
  GDALDatasetH __pyx_t_21;
  PyObject *__pyx_t_22 = NULL;
  int __pyx_t_23;
  size_t __pyx_t_24;
  int __pyx_t_25;
  char const *__pyx_t_26;
  PyObject *__pyx_t_27 = NULL;
  __Pyx_RefNannySetupContext("_rasterize", 0);
  __Pyx_INCREF(__pyx_v_image);

  /* "rasterio/_features.pyx":275
 *     cdef int retval
 *     cdef size_t i
 *     cdef size_t num_geoms = 0             # <<<<<<<<<<<<<<
 *     cdef OGRGeometryH *geoms = NULL
 *     cdef char **options = NULL
 */
  __pyx_v_num_geoms = 0;

  /* "rasterio/_features.pyx":276
 *     cdef size_t i
 *     cdef size_t num_geoms = 0
 *     cdef OGRGeometryH *geoms = NULL             # <<<<<<<<<<<<<<
 *     cdef char **options = NULL
 *     cdef double *pixel_values = NULL
 */
  __pyx_v_geoms = NULL;

  /* "rasterio/_features.pyx":277
 *     cdef size_t num_geoms = 0
 *     cdef OGRGeometryH *geoms = NULL
 *     cdef char **options = NULL             # <<<<<<<<<<<<<<
 *     cdef double *pixel_values = NULL
 *     cdef InMemoryRaster mem = None
 */
  __pyx_v_options = NULL;

  /* "rasterio/_features.pyx":278
 *     cdef OGRGeometryH *geoms = NULL
 *     cdef char **options = NULL
 *     cdef double *pixel_values = NULL             # <<<<<<<<<<<<<<
 *     cdef InMemoryRaster mem = None
 * 
 */
  __pyx_v_pixel_values = NULL;

  /* "rasterio/_features.pyx":279
 *     cdef char **options = NULL
 *     cdef double *pixel_values = NULL
 *     cdef InMemoryRaster mem = None             # <<<<<<<<<<<<<<
 * 
 *     try:
 */
  __Pyx_INCREF(Py_None);
  __pyx_v_mem = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None);

  /* "rasterio/_features.pyx":281
 *     cdef InMemoryRaster mem = None
 * 
 *     try:             # <<<<<<<<<<<<<<
 *         if all_touched:
 *             options = CSLSetNameValue(options, "ALL_TOUCHED", "TRUE")
 */
  /*try:*/ {

    /* "rasterio/_features.pyx":282
 * 
 *     try:
 *         if all_touched:             # <<<<<<<<<<<<<<
 *             options = CSLSetNameValue(options, "ALL_TOUCHED", "TRUE")
 * 
 */
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_v_all_touched); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 282, __pyx_L4_error)
    if (__pyx_t_1) {

      /* "rasterio/_features.pyx":283
 *     try:
 *         if all_touched:
 *             options = CSLSetNameValue(options, "ALL_TOUCHED", "TRUE")             # <<<<<<<<<<<<<<
 * 
 *         # GDAL needs an array of geometries.
 */
      __pyx_v_options = CSLSetNameValue(__pyx_v_options, ((char *)"ALL_TOUCHED"), ((char *)"TRUE"));

      /* "rasterio/_features.pyx":282
 * 
 *     try:
 *         if all_touched:             # <<<<<<<<<<<<<<
 *             options = CSLSetNameValue(options, "ALL_TOUCHED", "TRUE")
 * 
 */
    }

    /* "rasterio/_features.pyx":288
 *         # For now, we'll build a Python list on the way to building that
 *         # C array. TODO: make this more efficient.
 *         all_shapes = list(shapes)             # <<<<<<<<<<<<<<
 *         num_geoms = len(all_shapes)
 * 
 */
    __pyx_t_2 = PySequence_List(__pyx_v_shapes); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 288, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_v_all_shapes = ((PyObject*)__pyx_t_2);
    __pyx_t_2 = 0;

    /* "rasterio/_features.pyx":289
 *         # C array. TODO: make this more efficient.
 *         all_shapes = list(shapes)
 *         num_geoms = len(all_shapes)             # <<<<<<<<<<<<<<
 * 
 *         geoms = <OGRGeometryH *>CPLMalloc(
 */
    __pyx_t_3 = PyList_GET_SIZE(__pyx_v_all_shapes); if (unlikely(__pyx_t_3 == -1)) __PYX_ERR(0, 289, __pyx_L4_error)
    __pyx_v_num_geoms = __pyx_t_3;

    /* "rasterio/_features.pyx":291
 *         num_geoms = len(all_shapes)
 * 
 *         geoms = <OGRGeometryH *>CPLMalloc(             # <<<<<<<<<<<<<<
 *             num_geoms * sizeof(OGRGeometryH))
 *         pixel_values = <double *>CPLMalloc(num_geoms * sizeof(double))
 */
    __pyx_v_geoms = ((OGRGeometryH *)CPLMalloc((__pyx_v_num_geoms * (sizeof(OGRGeometryH)))));

    /* "rasterio/_features.pyx":293
 *         geoms = <OGRGeometryH *>CPLMalloc(
 *             num_geoms * sizeof(OGRGeometryH))
 *         pixel_values = <double *>CPLMalloc(num_geoms * sizeof(double))             # <<<<<<<<<<<<<<
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):
 */
    __pyx_v_pixel_values = ((double *)CPLMalloc((__pyx_v_num_geoms * (sizeof(double)))));

    /* "rasterio/_features.pyx":295
 *         pixel_values = <double *>CPLMalloc(num_geoms * sizeof(double))
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):             # <<<<<<<<<<<<<<
 *             try:
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 */
    __pyx_t_4 = 0;
    __pyx_t_2 = __pyx_v_all_shapes; __Pyx_INCREF(__pyx_t_2); __pyx_t_3 = 0;
    for (;;) {
      if (__pyx_t_3 >= PyList_GET_SIZE(__pyx_t_2)) break;
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_5 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_3); __Pyx_INCREF(__pyx_t_5); __pyx_t_3++; if (unlikely(0 < 0)) __PYX_ERR(0, 295, __pyx_L4_error)
      #else
      __pyx_t_5 = PySequence_ITEM(__pyx_t_2, __pyx_t_3); __pyx_t_3++; if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 295, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_5);
      #endif
      if ((likely(PyTuple_CheckExact(__pyx_t_5))) || (PyList_CheckExact(__pyx_t_5))) {
        PyObject* sequence = __pyx_t_5;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 2)) {
          if (size > 2) __Pyx_RaiseTooManyValuesError(2);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 295, __pyx_L4_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_6 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 1); 
        } else {
          __pyx_t_6 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_7 = PyList_GET_ITEM(sequence, 1); 
        }
        __Pyx_INCREF(__pyx_t_6);
        __Pyx_INCREF(__pyx_t_7);
        #else
        __pyx_t_6 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 295, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_6);
        __pyx_t_7 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 295, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_7);
        #endif
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        Py_ssize_t index = -1;
        __pyx_t_8 = PyObject_GetIter(__pyx_t_5); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 295, __pyx_L4_error)
        __Pyx_GOTREF(__pyx_t_8);
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
        __pyx_t_9 = Py_TYPE(__pyx_t_8)->tp_iternext;
        index = 0; __pyx_t_6 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_6)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_6);
        index = 1; __pyx_t_7 = __pyx_t_9(__pyx_t_8); if (unlikely(!__pyx_t_7)) goto __pyx_L9_unpacking_failed;
        __Pyx_GOTREF(__pyx_t_7);
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_9(__pyx_t_8), 2) < 0) __PYX_ERR(0, 295, __pyx_L4_error)
        __pyx_t_9 = NULL;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        goto __pyx_L10_unpacking_done;
        __pyx_L9_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __pyx_t_9 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 295, __pyx_L4_error)
        __pyx_L10_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_geometry, __pyx_t_6);
      __pyx_t_6 = 0;
      __Pyx_XDECREF_SET(__pyx_v_value, __pyx_t_7);
      __pyx_t_7 = 0;
      __pyx_v_i = __pyx_t_4;
      __pyx_t_4 = (__pyx_t_4 + 1);

      /* "rasterio/_features.pyx":296
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):
 *             try:             # <<<<<<<<<<<<<<
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 *                 pixel_values[i] = <double>value
 */
      {
        __Pyx_PyThreadState_declare
        __Pyx_PyThreadState_assign
        __Pyx_ExceptionSave(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
        __Pyx_XGOTREF(__pyx_t_10);
        __Pyx_XGOTREF(__pyx_t_11);
        __Pyx_XGOTREF(__pyx_t_12);
        /*try:*/ {

          /* "rasterio/_features.pyx":297
 *         for i, (geometry, value) in enumerate(all_shapes):
 *             try:
 *                 geoms[i] = OGRGeomBuilder().build(geometry)             # <<<<<<<<<<<<<<
 *                 pixel_values[i] = <double>value
 *             except:
 */
          __pyx_t_5 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_OGRGeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 297, __pyx_L11_error)
          __Pyx_GOTREF(__pyx_t_5);
          __pyx_t_13 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_5)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_5), __pyx_v_geometry); if (unlikely(__pyx_t_13 == NULL)) __PYX_ERR(0, 297, __pyx_L11_error)
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          (__pyx_v_geoms[__pyx_v_i]) = __pyx_t_13;

          /* "rasterio/_features.pyx":298
 *             try:
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 *                 pixel_values[i] = <double>value             # <<<<<<<<<<<<<<
 *             except:
 *                 log.error("Geometry %r at index %d with value %d skipped",
 */
          __pyx_t_14 = __pyx_PyFloat_AsDouble(__pyx_v_value); if (unlikely((__pyx_t_14 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 298, __pyx_L11_error)
          (__pyx_v_pixel_values[__pyx_v_i]) = ((double)__pyx_t_14);

          /* "rasterio/_features.pyx":296
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):
 *             try:             # <<<<<<<<<<<<<<
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 *                 pixel_values[i] = <double>value
 */
        }
        __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
        __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
        __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
        goto __pyx_L18_try_end;
        __pyx_L11_error:;
        __Pyx_PyThreadState_assign
        __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
        __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
        __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;

        /* "rasterio/_features.pyx":299
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 *                 pixel_values[i] = <double>value
 *             except:             # <<<<<<<<<<<<<<
 *                 log.error("Geometry %r at index %d with value %d skipped",
 *                     geometry, i, value)
 */
        /*except:*/ {
          __Pyx_AddTraceback("rasterio._features._rasterize", __pyx_clineno, __pyx_lineno, __pyx_filename);
          if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_7, &__pyx_t_6) < 0) __PYX_ERR(0, 299, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_5);
          __Pyx_GOTREF(__pyx_t_7);
          __Pyx_GOTREF(__pyx_t_6);

          /* "rasterio/_features.pyx":300
 *                 pixel_values[i] = <double>value
 *             except:
 *                 log.error("Geometry %r at index %d with value %d skipped",             # <<<<<<<<<<<<<<
 *                     geometry, i, value)
 * 
 */
          __pyx_t_15 = __Pyx_GetModuleGlobalName(__pyx_n_s_log); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 300, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_16 = __Pyx_PyObject_GetAttrStr(__pyx_t_15, __pyx_n_s_error); if (unlikely(!__pyx_t_16)) __PYX_ERR(0, 300, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_16);
          __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;

          /* "rasterio/_features.pyx":301
 *             except:
 *                 log.error("Geometry %r at index %d with value %d skipped",
 *                     geometry, i, value)             # <<<<<<<<<<<<<<
 * 
 *         with InMemoryRaster(image=image, transform=transform) as mem:
 */
          __pyx_t_15 = __Pyx_PyInt_FromSize_t(__pyx_v_i); if (unlikely(!__pyx_t_15)) __PYX_ERR(0, 301, __pyx_L13_except_error)
          __Pyx_GOTREF(__pyx_t_15);
          __pyx_t_17 = NULL;
          __pyx_t_18 = 0;
          if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_16))) {
            __pyx_t_17 = PyMethod_GET_SELF(__pyx_t_16);
            if (likely(__pyx_t_17)) {
              PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_16);
              __Pyx_INCREF(__pyx_t_17);
              __Pyx_INCREF(function);
              __Pyx_DECREF_SET(__pyx_t_16, function);
              __pyx_t_18 = 1;
            }
          }
          #if CYTHON_FAST_PYCALL
          if (PyFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[5] = {__pyx_t_17, __pyx_kp_s_Geometry_r_at_index_d_with_value, __pyx_v_geometry, __pyx_t_15, __pyx_v_value};
            __pyx_t_8 = __Pyx_PyFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_18, 4+__pyx_t_18); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L13_except_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          } else
          #endif
          #if CYTHON_FAST_PYCCALL
          if (__Pyx_PyFastCFunction_Check(__pyx_t_16)) {
            PyObject *__pyx_temp[5] = {__pyx_t_17, __pyx_kp_s_Geometry_r_at_index_d_with_value, __pyx_v_geometry, __pyx_t_15, __pyx_v_value};
            __pyx_t_8 = __Pyx_PyCFunction_FastCall(__pyx_t_16, __pyx_temp+1-__pyx_t_18, 4+__pyx_t_18); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L13_except_error)
            __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_15); __pyx_t_15 = 0;
          } else
          #endif
          {
            __pyx_t_19 = PyTuple_New(4+__pyx_t_18); if (unlikely(!__pyx_t_19)) __PYX_ERR(0, 300, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_19);
            if (__pyx_t_17) {
              __Pyx_GIVEREF(__pyx_t_17); PyTuple_SET_ITEM(__pyx_t_19, 0, __pyx_t_17); __pyx_t_17 = NULL;
            }
            __Pyx_INCREF(__pyx_kp_s_Geometry_r_at_index_d_with_value);
            __Pyx_GIVEREF(__pyx_kp_s_Geometry_r_at_index_d_with_value);
            PyTuple_SET_ITEM(__pyx_t_19, 0+__pyx_t_18, __pyx_kp_s_Geometry_r_at_index_d_with_value);
            __Pyx_INCREF(__pyx_v_geometry);
            __Pyx_GIVEREF(__pyx_v_geometry);
            PyTuple_SET_ITEM(__pyx_t_19, 1+__pyx_t_18, __pyx_v_geometry);
            __Pyx_GIVEREF(__pyx_t_15);
            PyTuple_SET_ITEM(__pyx_t_19, 2+__pyx_t_18, __pyx_t_15);
            __Pyx_INCREF(__pyx_v_value);
            __Pyx_GIVEREF(__pyx_v_value);
            PyTuple_SET_ITEM(__pyx_t_19, 3+__pyx_t_18, __pyx_v_value);
            __pyx_t_15 = 0;
            __pyx_t_8 = __Pyx_PyObject_Call(__pyx_t_16, __pyx_t_19, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 300, __pyx_L13_except_error)
            __Pyx_GOTREF(__pyx_t_8);
            __Pyx_DECREF(__pyx_t_19); __pyx_t_19 = 0;
          }
          __Pyx_DECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
          __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
          goto __pyx_L12_exception_handled;
        }
        __pyx_L13_except_error:;

        /* "rasterio/_features.pyx":296
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):
 *             try:             # <<<<<<<<<<<<<<
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 *                 pixel_values[i] = <double>value
 */
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        goto __pyx_L4_error;
        __pyx_L12_exception_handled:;
        __Pyx_PyThreadState_assign
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
        __pyx_L18_try_end:;
      }

      /* "rasterio/_features.pyx":295
 *         pixel_values = <double *>CPLMalloc(num_geoms * sizeof(double))
 * 
 *         for i, (geometry, value) in enumerate(all_shapes):             # <<<<<<<<<<<<<<
 *             try:
 *                 geoms[i] = OGRGeomBuilder().build(geometry)
 */
    }
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "rasterio/_features.pyx":303
 *                     geometry, i, value)
 * 
 *         with InMemoryRaster(image=image, transform=transform) as mem:             # <<<<<<<<<<<<<<
 *             exc_wrap_int(
 *                 GDALRasterizeGeometries(
 */
    /*with:*/ {
      __pyx_t_2 = PyDict_New(); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_image, __pyx_v_image) < 0) __PYX_ERR(0, 303, __pyx_L4_error)
      if (PyDict_SetItem(__pyx_t_2, __pyx_n_s_transform, __pyx_v_transform) < 0) __PYX_ERR(0, 303, __pyx_L4_error)
      __pyx_t_6 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_3_io_InMemoryRaster), __pyx_empty_tuple, __pyx_t_2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 303, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_t_12 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_exit); if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 303, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_12);
      __pyx_t_7 = __Pyx_PyObject_LookupSpecial(__pyx_t_6, __pyx_n_s_enter); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 303, __pyx_L21_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_5 = NULL;
      if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_5 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_5)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_5);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (__pyx_t_5) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_5); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L21_error)
        __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_7); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 303, __pyx_L21_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      __pyx_t_7 = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      /*try:*/ {
        {
          __Pyx_PyThreadState_declare
          __Pyx_PyThreadState_assign
          __Pyx_ExceptionSave(&__pyx_t_11, &__pyx_t_10, &__pyx_t_20);
          __Pyx_XGOTREF(__pyx_t_11);
          __Pyx_XGOTREF(__pyx_t_10);
          __Pyx_XGOTREF(__pyx_t_20);
          /*try:*/ {
            if (!(likely(((__pyx_t_7) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_7, __pyx_ptype_8rasterio_3_io_InMemoryRaster))))) __PYX_ERR(0, 303, __pyx_L25_error)
            __Pyx_DECREF_SET(__pyx_v_mem, ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)__pyx_t_7));
            __pyx_t_7 = 0;

            /* "rasterio/_features.pyx":306
 *             exc_wrap_int(
 *                 GDALRasterizeGeometries(
 *                     mem.handle(), 1, mem.band_ids,num_geoms, geoms, NULL,             # <<<<<<<<<<<<<<
 *                     mem.transform, pixel_values, options, NULL, NULL))
 * 
 */
            __pyx_t_21 = ((struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster *)__pyx_v_mem->__pyx_vtab)->handle(__pyx_v_mem); if (unlikely(__pyx_t_21 == NULL)) __PYX_ERR(0, 306, __pyx_L25_error)

            /* "rasterio/_features.pyx":304
 * 
 *         with InMemoryRaster(image=image, transform=transform) as mem:
 *             exc_wrap_int(             # <<<<<<<<<<<<<<
 *                 GDALRasterizeGeometries(
 *                     mem.handle(), 1, mem.band_ids,num_geoms, geoms, NULL,
 */
            __pyx_t_18 = __pyx_f_8rasterio_4_err_exc_wrap_int(GDALRasterizeGeometries(__pyx_t_21, 1, __pyx_v_mem->band_ids, __pyx_v_num_geoms, __pyx_v_geoms, NULL, __pyx_v_mem->transform, __pyx_v_pixel_values, __pyx_v_options, NULL, NULL)); if (unlikely(__pyx_t_18 == -1)) __PYX_ERR(0, 304, __pyx_L25_error)

            /* "rasterio/_features.pyx":310
 * 
 *             # Read in-memory data back into image
 *             image = mem.read()             # <<<<<<<<<<<<<<
 * 
 *     finally:
 */
            __pyx_t_6 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_mem), __pyx_n_s_read); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 310, __pyx_L25_error)
            __Pyx_GOTREF(__pyx_t_6);
            __pyx_t_2 = NULL;
            if (CYTHON_UNPACK_METHODS && likely(PyMethod_Check(__pyx_t_6))) {
              __pyx_t_2 = PyMethod_GET_SELF(__pyx_t_6);
              if (likely(__pyx_t_2)) {
                PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_6);
                __Pyx_INCREF(__pyx_t_2);
                __Pyx_INCREF(function);
                __Pyx_DECREF_SET(__pyx_t_6, function);
              }
            }
            if (__pyx_t_2) {
              __pyx_t_7 = __Pyx_PyObject_CallOneArg(__pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L25_error)
              __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            } else {
              __pyx_t_7 = __Pyx_PyObject_CallNoArg(__pyx_t_6); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 310, __pyx_L25_error)
            }
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF_SET(__pyx_v_image, __pyx_t_7);
            __pyx_t_7 = 0;

            /* "rasterio/_features.pyx":303
 *                     geometry, i, value)
 * 
 *         with InMemoryRaster(image=image, transform=transform) as mem:             # <<<<<<<<<<<<<<
 *             exc_wrap_int(
 *                 GDALRasterizeGeometries(
 */
          }
          __Pyx_XDECREF(__pyx_t_11); __pyx_t_11 = 0;
          __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
          __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
          goto __pyx_L32_try_end;
          __pyx_L25_error:;
          __Pyx_PyThreadState_assign
          __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
          __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
          __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
          __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
          __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
          __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
          __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
          __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
          __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
          /*except:*/ {
            __Pyx_AddTraceback("rasterio._features._rasterize", __pyx_clineno, __pyx_lineno, __pyx_filename);
            if (__Pyx_GetException(&__pyx_t_7, &__pyx_t_6, &__pyx_t_2) < 0) __PYX_ERR(0, 303, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_7);
            __Pyx_GOTREF(__pyx_t_6);
            __Pyx_GOTREF(__pyx_t_2);
            __pyx_t_5 = PyTuple_Pack(3, __pyx_t_7, __pyx_t_6, __pyx_t_2); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 303, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_5);
            __pyx_t_22 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_t_5, NULL);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
            if (unlikely(!__pyx_t_22)) __PYX_ERR(0, 303, __pyx_L27_except_error)
            __Pyx_GOTREF(__pyx_t_22);
            __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_22);
            __Pyx_DECREF(__pyx_t_22); __pyx_t_22 = 0;
            if (__pyx_t_1 < 0) __PYX_ERR(0, 303, __pyx_L27_except_error)
            __pyx_t_23 = ((!(__pyx_t_1 != 0)) != 0);
            if (__pyx_t_23) {
              __Pyx_GIVEREF(__pyx_t_7);
              __Pyx_GIVEREF(__pyx_t_6);
              __Pyx_XGIVEREF(__pyx_t_2);
              __Pyx_ErrRestoreWithState(__pyx_t_7, __pyx_t_6, __pyx_t_2);
              __pyx_t_7 = 0; __pyx_t_6 = 0; __pyx_t_2 = 0; 
              __PYX_ERR(0, 303, __pyx_L27_except_error)
            }
            __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
            __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
            __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
            goto __pyx_L26_exception_handled;
          }
          __pyx_L27_except_error:;
          __Pyx_PyThreadState_assign
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_20);
          goto __pyx_L4_error;
          __pyx_L26_exception_handled:;
          __Pyx_PyThreadState_assign
          __Pyx_XGIVEREF(__pyx_t_11);
          __Pyx_XGIVEREF(__pyx_t_10);
          __Pyx_XGIVEREF(__pyx_t_20);
          __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_10, __pyx_t_20);
          __pyx_L32_try_end:;
        }
      }
      /*finally:*/ {
        /*normal exit:*/{
          if (__pyx_t_12) {
            __pyx_t_20 = __Pyx_PyObject_Call(__pyx_t_12, __pyx_tuple__24, NULL);
            __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
            if (unlikely(!__pyx_t_20)) __PYX_ERR(0, 303, __pyx_L4_error)
            __Pyx_GOTREF(__pyx_t_20);
            __Pyx_DECREF(__pyx_t_20); __pyx_t_20 = 0;
          }
          goto __pyx_L24;
        }
        __pyx_L24:;
      }
      goto __pyx_L36;
      __pyx_L21_error:;
      __Pyx_DECREF(__pyx_t_12); __pyx_t_12 = 0;
      goto __pyx_L4_error;
      __pyx_L36:;
    }
  }

  /* "rasterio/_features.pyx":313
 * 
 *     finally:
 *         for i in range(num_geoms):             # <<<<<<<<<<<<<<
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)
 */
  /*finally:*/ {
    /*normal exit:*/{
      __pyx_t_4 = __pyx_v_num_geoms;
      for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_4; __pyx_t_24+=1) {
        __pyx_v_i = __pyx_t_24;

        /* "rasterio/_features.pyx":314
 *     finally:
 *         for i in range(num_geoms):
 *             _deleteOgrGeom(geoms[i])             # <<<<<<<<<<<<<<
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 */
        __pyx_t_2 = __pyx_f_8rasterio_9_features__deleteOgrGeom((__pyx_v_geoms[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }

      /* "rasterio/_features.pyx":315
 *         for i in range(num_geoms):
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)             # <<<<<<<<<<<<<<
 *         CPLFree(pixel_values)
 *         if options:
 */
      CPLFree(__pyx_v_geoms);

      /* "rasterio/_features.pyx":316
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)             # <<<<<<<<<<<<<<
 *         if options:
 *             CSLDestroy(options)
 */
      CPLFree(__pyx_v_pixel_values);

      /* "rasterio/_features.pyx":317
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 *         if options:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 * 
 */
      __pyx_t_23 = (__pyx_v_options != 0);
      if (__pyx_t_23) {

        /* "rasterio/_features.pyx":318
 *         CPLFree(pixel_values)
 *         if options:
 *             CSLDestroy(options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
        CSLDestroy(__pyx_v_options);

        /* "rasterio/_features.pyx":317
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 *         if options:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 * 
 */
      }
      goto __pyx_L5;
    }
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_12 = 0; __pyx_t_20 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_22 = 0; __pyx_t_27 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_17); __pyx_t_17 = 0;
      __Pyx_XDECREF(__pyx_t_15); __pyx_t_15 = 0;
      __Pyx_XDECREF(__pyx_t_19); __pyx_t_19 = 0;
      __Pyx_XDECREF(__pyx_t_16); __pyx_t_16 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_5); __pyx_t_5 = 0;
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_11, &__pyx_t_22, &__pyx_t_27);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_12, &__pyx_t_20, &__pyx_t_10) < 0)) __Pyx_ErrFetch(&__pyx_t_12, &__pyx_t_20, &__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_12);
      __Pyx_XGOTREF(__pyx_t_20);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_22);
      __Pyx_XGOTREF(__pyx_t_27);
      __pyx_t_18 = __pyx_lineno; __pyx_t_25 = __pyx_clineno; __pyx_t_26 = __pyx_filename;
      {

        /* "rasterio/_features.pyx":313
 * 
 *     finally:
 *         for i in range(num_geoms):             # <<<<<<<<<<<<<<
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)
 */
        __pyx_t_4 = __pyx_v_num_geoms;
        for (__pyx_t_24 = 0; __pyx_t_24 < __pyx_t_4; __pyx_t_24+=1) {
          __pyx_v_i = __pyx_t_24;

          /* "rasterio/_features.pyx":314
 *     finally:
 *         for i in range(num_geoms):
 *             _deleteOgrGeom(geoms[i])             # <<<<<<<<<<<<<<
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 */
          __pyx_t_2 = __pyx_f_8rasterio_9_features__deleteOgrGeom((__pyx_v_geoms[__pyx_v_i])); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 314, __pyx_L41_error)
          __Pyx_GOTREF(__pyx_t_2);
          __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
        }

        /* "rasterio/_features.pyx":315
 *         for i in range(num_geoms):
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)             # <<<<<<<<<<<<<<
 *         CPLFree(pixel_values)
 *         if options:
 */
        CPLFree(__pyx_v_geoms);

        /* "rasterio/_features.pyx":316
 *             _deleteOgrGeom(geoms[i])
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)             # <<<<<<<<<<<<<<
 *         if options:
 *             CSLDestroy(options)
 */
        CPLFree(__pyx_v_pixel_values);

        /* "rasterio/_features.pyx":317
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 *         if options:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 * 
 */
        __pyx_t_23 = (__pyx_v_options != 0);
        if (__pyx_t_23) {

          /* "rasterio/_features.pyx":318
 *         CPLFree(pixel_values)
 *         if options:
 *             CSLDestroy(options)             # <<<<<<<<<<<<<<
 * 
 * 
 */
          CSLDestroy(__pyx_v_options);

          /* "rasterio/_features.pyx":317
 *         CPLFree(geoms)
 *         CPLFree(pixel_values)
 *         if options:             # <<<<<<<<<<<<<<
 *             CSLDestroy(options)
 * 
 */
        }
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_XGIVEREF(__pyx_t_27);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_22, __pyx_t_27);
      }
      __Pyx_XGIVEREF(__pyx_t_12);
      __Pyx_XGIVEREF(__pyx_t_20);
      __Pyx_XGIVEREF(__pyx_t_10);
      __Pyx_ErrRestore(__pyx_t_12, __pyx_t_20, __pyx_t_10);
      __pyx_t_12 = 0; __pyx_t_20 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_22 = 0; __pyx_t_27 = 0;
      __pyx_lineno = __pyx_t_18; __pyx_clineno = __pyx_t_25; __pyx_filename = __pyx_t_26;
      goto __pyx_L1_error;
      __pyx_L41_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_22);
        __Pyx_XGIVEREF(__pyx_t_27);
        __Pyx_ExceptionReset(__pyx_t_11, __pyx_t_22, __pyx_t_27);
      }
      __Pyx_XDECREF(__pyx_t_12); __pyx_t_12 = 0;
      __Pyx_XDECREF(__pyx_t_20); __pyx_t_20 = 0;
      __Pyx_XDECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_11 = 0; __pyx_t_22 = 0; __pyx_t_27 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L5:;
  }

  /* "rasterio/_features.pyx":253
 * 
 * 
 * def _rasterize(shapes, image, transform, all_touched):             # <<<<<<<<<<<<<<
 *     """
 *     Burns input geometries into `image`.
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_15);
  __Pyx_XDECREF(__pyx_t_16);
  __Pyx_XDECREF(__pyx_t_17);
  __Pyx_XDECREF(__pyx_t_19);
  __Pyx_AddTraceback("rasterio._features._rasterize", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_mem);
  __Pyx_XDECREF(__pyx_v_all_shapes);
  __Pyx_XDECREF(__pyx_v_geometry);
  __Pyx_XDECREF(__pyx_v_value);
  __Pyx_XDECREF(__pyx_v_image);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static PyObject *__pyx_gb_8rasterio_9_features_9generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value); /* proto */

/* "rasterio/_features.pyx":321
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_8_explode(PyObject *__pyx_self, PyObject *__pyx_v_coords); /*proto*/
static char __pyx_doc_8rasterio_9_features_7_explode[] = "Explode a GeoJSON geometry's coordinates object and yield\n    coordinate tuples. As long as the input is conforming, the type of\n    the geometry doesn't matter.  From Fiona 1.4.8";
static PyMethodDef __pyx_mdef_8rasterio_9_features_8_explode = {"_explode", (PyCFunction)__pyx_pw_8rasterio_9_features_8_explode, METH_O, __pyx_doc_8rasterio_9_features_7_explode};
static PyObject *__pyx_pw_8rasterio_9_features_8_explode(PyObject *__pyx_self, PyObject *__pyx_v_coords) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explode (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_7_explode(__pyx_self, ((PyObject *)__pyx_v_coords));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_7_explode(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_coords) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *__pyx_cur_scope;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_explode", 0);
  __pyx_cur_scope = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_3__explode(__pyx_ptype_8rasterio_9_features___pyx_scope_struct_3__explode, __pyx_empty_tuple, NULL);
  if (unlikely(!__pyx_cur_scope)) {
    __pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)Py_None);
    __Pyx_INCREF(Py_None);
    __PYX_ERR(0, 321, __pyx_L1_error)
  } else {
    __Pyx_GOTREF(__pyx_cur_scope);
  }
  __pyx_cur_scope->__pyx_v_coords = __pyx_v_coords;
  __Pyx_INCREF(__pyx_cur_scope->__pyx_v_coords);
  __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_coords);
  {
    __pyx_CoroutineObject *gen = __Pyx_Generator_New((__pyx_coroutine_body_t) __pyx_gb_8rasterio_9_features_9generator1, (PyObject *) __pyx_cur_scope, __pyx_n_s_explode, __pyx_n_s_explode, __pyx_n_s_rasterio__features); if (unlikely(!gen)) __PYX_ERR(0, 321, __pyx_L1_error)
    __Pyx_DECREF(__pyx_cur_scope);
    __Pyx_RefNannyFinishContext();
    return (PyObject *) gen;
  }

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_AddTraceback("rasterio._features._explode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __Pyx_DECREF(((PyObject *)__pyx_cur_scope));
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_gb_8rasterio_9_features_9generator1(__pyx_CoroutineObject *__pyx_generator, PyObject *__pyx_sent_value) /* generator body */
{
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *__pyx_cur_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)__pyx_generator->closure);
  PyObject *__pyx_r = NULL;
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *(*__pyx_t_3)(PyObject *);
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  Py_ssize_t __pyx_t_11;
  PyObject *(*__pyx_t_12)(PyObject *);
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("None", 0);
  switch (__pyx_generator->resume_label) {
    case 0: goto __pyx_L3_first_run;
    case 1: goto __pyx_L9_resume_from_yield;
    case 2: goto __pyx_L12_resume_from_yield;
    default: /* CPython raises the right error here */
    __Pyx_RefNannyFinishContext();
    return NULL;
  }
  __pyx_L3_first_run:;
  if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 321, __pyx_L1_error)

  /* "rasterio/_features.pyx":325
 *     coordinate tuples. As long as the input is conforming, the type of
 *     the geometry doesn't matter.  From Fiona 1.4.8"""
 *     for e in coords:             # <<<<<<<<<<<<<<
 *         if isinstance(e, (float, int)):
 *             yield coords
 */
  if (likely(PyList_CheckExact(__pyx_cur_scope->__pyx_v_coords)) || PyTuple_CheckExact(__pyx_cur_scope->__pyx_v_coords)) {
    __pyx_t_1 = __pyx_cur_scope->__pyx_v_coords; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
    __pyx_t_3 = NULL;
  } else {
    __pyx_t_2 = -1; __pyx_t_1 = PyObject_GetIter(__pyx_cur_scope->__pyx_v_coords); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 325, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_3 = Py_TYPE(__pyx_t_1)->tp_iternext; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 325, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_3)) {
      if (likely(PyList_CheckExact(__pyx_t_1))) {
        if (__pyx_t_2 >= PyList_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyList_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 325, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      } else {
        if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_4); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(0, 325, __pyx_L1_error)
        #else
        __pyx_t_4 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 325, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        #endif
      }
    } else {
      __pyx_t_4 = __pyx_t_3(__pyx_t_1);
      if (unlikely(!__pyx_t_4)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 325, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_4);
    }
    __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_e);
    __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_e, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_4);
    __pyx_t_4 = 0;

    /* "rasterio/_features.pyx":326
 *     the geometry doesn't matter.  From Fiona 1.4.8"""
 *     for e in coords:
 *         if isinstance(e, (float, int)):             # <<<<<<<<<<<<<<
 *             yield coords
 *             break
 */
    __pyx_t_6 = PyFloat_Check(__pyx_cur_scope->__pyx_v_e); 
    __pyx_t_7 = (__pyx_t_6 != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_5 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = PyInt_Check(__pyx_cur_scope->__pyx_v_e); 
    __pyx_t_6 = (__pyx_t_7 != 0);
    __pyx_t_5 = __pyx_t_6;
    __pyx_L7_bool_binop_done:;
    __pyx_t_6 = (__pyx_t_5 != 0);
    if (__pyx_t_6) {

      /* "rasterio/_features.pyx":327
 *     for e in coords:
 *         if isinstance(e, (float, int)):
 *             yield coords             # <<<<<<<<<<<<<<
 *             break
 *         else:
 */
      __Pyx_INCREF(__pyx_cur_scope->__pyx_v_coords);
      __pyx_r = __pyx_cur_scope->__pyx_v_coords;
      __Pyx_XGIVEREF(__pyx_t_1);
      __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
      __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
      __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
      __Pyx_XGIVEREF(__pyx_r);
      __Pyx_RefNannyFinishContext();
      /* return from generator, yielding value */
      __pyx_generator->resume_label = 1;
      return __pyx_r;
      __pyx_L9_resume_from_yield:;
      __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
      __pyx_cur_scope->__pyx_t_0 = 0;
      __Pyx_XGOTREF(__pyx_t_1);
      __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
      __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
      if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 327, __pyx_L1_error)

      /* "rasterio/_features.pyx":328
 *         if isinstance(e, (float, int)):
 *             yield coords
 *             break             # <<<<<<<<<<<<<<
 *         else:
 *             for f in _explode(e):
 */
      goto __pyx_L5_break;

      /* "rasterio/_features.pyx":326
 *     the geometry doesn't matter.  From Fiona 1.4.8"""
 *     for e in coords:
 *         if isinstance(e, (float, int)):             # <<<<<<<<<<<<<<
 *             yield coords
 *             break
 */
    }

    /* "rasterio/_features.pyx":330
 *             break
 *         else:
 *             for f in _explode(e):             # <<<<<<<<<<<<<<
 *                 yield f
 * 
 */
    /*else*/ {
      __pyx_t_8 = __Pyx_GetModuleGlobalName(__pyx_n_s_explode); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 330, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_8))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_8);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_8);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_8, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_8, __pyx_cur_scope->__pyx_v_e); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_cur_scope->__pyx_v_e};
          __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_8)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_cur_scope->__pyx_v_e};
          __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_8, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_4);
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 330, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
          __Pyx_INCREF(__pyx_cur_scope->__pyx_v_e);
          __Pyx_GIVEREF(__pyx_cur_scope->__pyx_v_e);
          PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_cur_scope->__pyx_v_e);
          __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_8, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_4);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      if (likely(PyList_CheckExact(__pyx_t_4)) || PyTuple_CheckExact(__pyx_t_4)) {
        __pyx_t_8 = __pyx_t_4; __Pyx_INCREF(__pyx_t_8); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
      } else {
        __pyx_t_11 = -1; __pyx_t_8 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 330, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_8);
        __pyx_t_12 = Py_TYPE(__pyx_t_8)->tp_iternext; if (unlikely(!__pyx_t_12)) __PYX_ERR(0, 330, __pyx_L1_error)
      }
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      for (;;) {
        if (likely(!__pyx_t_12)) {
          if (likely(PyList_CheckExact(__pyx_t_8))) {
            if (__pyx_t_11 >= PyList_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyList_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_4); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          } else {
            if (__pyx_t_11 >= PyTuple_GET_SIZE(__pyx_t_8)) break;
            #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
            __pyx_t_4 = PyTuple_GET_ITEM(__pyx_t_8, __pyx_t_11); __Pyx_INCREF(__pyx_t_4); __pyx_t_11++; if (unlikely(0 < 0)) __PYX_ERR(0, 330, __pyx_L1_error)
            #else
            __pyx_t_4 = PySequence_ITEM(__pyx_t_8, __pyx_t_11); __pyx_t_11++; if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 330, __pyx_L1_error)
            __Pyx_GOTREF(__pyx_t_4);
            #endif
          }
        } else {
          __pyx_t_4 = __pyx_t_12(__pyx_t_8);
          if (unlikely(!__pyx_t_4)) {
            PyObject* exc_type = PyErr_Occurred();
            if (exc_type) {
              if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
              else __PYX_ERR(0, 330, __pyx_L1_error)
            }
            break;
          }
          __Pyx_GOTREF(__pyx_t_4);
        }
        __Pyx_XGOTREF(__pyx_cur_scope->__pyx_v_f);
        __Pyx_XDECREF_SET(__pyx_cur_scope->__pyx_v_f, __pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_4);
        __pyx_t_4 = 0;

        /* "rasterio/_features.pyx":331
 *         else:
 *             for f in _explode(e):
 *                 yield f             # <<<<<<<<<<<<<<
 * 
 * 
 */
        __Pyx_INCREF(__pyx_cur_scope->__pyx_v_f);
        __pyx_r = __pyx_cur_scope->__pyx_v_f;
        __Pyx_XGIVEREF(__pyx_t_1);
        __pyx_cur_scope->__pyx_t_0 = __pyx_t_1;
        __pyx_cur_scope->__pyx_t_1 = __pyx_t_2;
        __pyx_cur_scope->__pyx_t_2 = __pyx_t_3;
        __Pyx_XGIVEREF(__pyx_t_8);
        __pyx_cur_scope->__pyx_t_3 = __pyx_t_8;
        __pyx_cur_scope->__pyx_t_4 = __pyx_t_11;
        __pyx_cur_scope->__pyx_t_5 = __pyx_t_12;
        __Pyx_XGIVEREF(__pyx_r);
        __Pyx_RefNannyFinishContext();
        /* return from generator, yielding value */
        __pyx_generator->resume_label = 2;
        return __pyx_r;
        __pyx_L12_resume_from_yield:;
        __pyx_t_1 = __pyx_cur_scope->__pyx_t_0;
        __pyx_cur_scope->__pyx_t_0 = 0;
        __Pyx_XGOTREF(__pyx_t_1);
        __pyx_t_2 = __pyx_cur_scope->__pyx_t_1;
        __pyx_t_3 = __pyx_cur_scope->__pyx_t_2;
        __pyx_t_8 = __pyx_cur_scope->__pyx_t_3;
        __pyx_cur_scope->__pyx_t_3 = 0;
        __Pyx_XGOTREF(__pyx_t_8);
        __pyx_t_11 = __pyx_cur_scope->__pyx_t_4;
        __pyx_t_12 = __pyx_cur_scope->__pyx_t_5;
        if (unlikely(!__pyx_sent_value)) __PYX_ERR(0, 331, __pyx_L1_error)

        /* "rasterio/_features.pyx":330
 *             break
 *         else:
 *             for f in _explode(e):             # <<<<<<<<<<<<<<
 *                 yield f
 * 
 */
      }
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    }

    /* "rasterio/_features.pyx":325
 *     coordinate tuples. As long as the input is conforming, the type of
 *     the geometry doesn't matter.  From Fiona 1.4.8"""
 *     for e in coords:             # <<<<<<<<<<<<<<
 *         if isinstance(e, (float, int)):
 *             yield coords
 */
  }
  __pyx_L5_break:;
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  CYTHON_MAYBE_UNUSED_VAR(__pyx_cur_scope);

  /* "rasterio/_features.pyx":321
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */

  /* function exit code */
  PyErr_SetNone(PyExc_StopIteration);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("_explode", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_r); __pyx_r = 0;
  __pyx_generator->resume_label = -1;
  __Pyx_Coroutine_clear((PyObject*)__pyx_generator);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":334
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry.
 *     From Fiona 1.4.8 with updates here to handle feature collections.
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11_bounds(PyObject *__pyx_self, PyObject *__pyx_v_geometry); /*proto*/
static char __pyx_doc_8rasterio_9_features_10_bounds[] = "Bounding box of a GeoJSON geometry.  \n    From Fiona 1.4.8 with updates here to handle feature collections.\n    TODO: add to Fiona.\n    ";
static PyMethodDef __pyx_mdef_8rasterio_9_features_11_bounds = {"_bounds", (PyCFunction)__pyx_pw_8rasterio_9_features_11_bounds, METH_O, __pyx_doc_8rasterio_9_features_10_bounds};
static PyObject *__pyx_pw_8rasterio_9_features_11_bounds(PyObject *__pyx_self, PyObject *__pyx_v_geometry) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_bounds (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_10_bounds(__pyx_self, ((PyObject *)__pyx_v_geometry));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_10_bounds(CYTHON_UNUSED PyObject *__pyx_self, PyObject *__pyx_v_geometry) {
  PyObject *__pyx_v_xmins = NULL;
  PyObject *__pyx_v_ymins = NULL;
  PyObject *__pyx_v_xmaxs = NULL;
  PyObject *__pyx_v_ymaxs = NULL;
  PyObject *__pyx_v_feature = NULL;
  PyObject *__pyx_v_xmin = NULL;
  PyObject *__pyx_v_ymin = NULL;
  PyObject *__pyx_v_xmax = NULL;
  PyObject *__pyx_v_ymax = NULL;
  PyObject *__pyx_v_xyz = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *(*__pyx_t_12)(PyObject *);
  int __pyx_t_13;
  __Pyx_RefNannySetupContext("_bounds", 0);

  /* "rasterio/_features.pyx":339
 *     TODO: add to Fiona.
 *     """
 *     if 'features' in geometry:             # <<<<<<<<<<<<<<
 *         xmins = []
 *         ymins = []
 */
  __pyx_t_1 = (__Pyx_PySequence_ContainsTF(__pyx_n_s_features, __pyx_v_geometry, Py_EQ)); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 339, __pyx_L1_error)
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "rasterio/_features.pyx":340
 *     """
 *     if 'features' in geometry:
 *         xmins = []             # <<<<<<<<<<<<<<
 *         ymins = []
 *         xmaxs = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 340, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_xmins = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "rasterio/_features.pyx":341
 *     if 'features' in geometry:
 *         xmins = []
 *         ymins = []             # <<<<<<<<<<<<<<
 *         xmaxs = []
 *         ymaxs = []
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 341, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ymins = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "rasterio/_features.pyx":342
 *         xmins = []
 *         ymins = []
 *         xmaxs = []             # <<<<<<<<<<<<<<
 *         ymaxs = []
 *         for feature in geometry['features']:
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 342, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_xmaxs = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "rasterio/_features.pyx":343
 *         ymins = []
 *         xmaxs = []
 *         ymaxs = []             # <<<<<<<<<<<<<<
 *         for feature in geometry['features']:
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])
 */
    __pyx_t_3 = PyList_New(0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 343, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_v_ymaxs = ((PyObject*)__pyx_t_3);
    __pyx_t_3 = 0;

    /* "rasterio/_features.pyx":344
 *         xmaxs = []
 *         ymaxs = []
 *         for feature in geometry['features']:             # <<<<<<<<<<<<<<
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])
 *             xmins.append(xmin)
 */
    __pyx_t_3 = PyObject_GetItem(__pyx_v_geometry, __pyx_n_s_features); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (likely(PyList_CheckExact(__pyx_t_3)) || PyTuple_CheckExact(__pyx_t_3)) {
      __pyx_t_4 = __pyx_t_3; __Pyx_INCREF(__pyx_t_4); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
    } else {
      __pyx_t_5 = -1; __pyx_t_4 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 344, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_6 = Py_TYPE(__pyx_t_4)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 344, __pyx_L1_error)
    }
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    for (;;) {
      if (likely(!__pyx_t_6)) {
        if (likely(PyList_CheckExact(__pyx_t_4))) {
          if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyList_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        } else {
          if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_4)) break;
          #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
          __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_4, __pyx_t_5); __Pyx_INCREF(__pyx_t_3); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 344, __pyx_L1_error)
          #else
          __pyx_t_3 = PySequence_ITEM(__pyx_t_4, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 344, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          #endif
        }
      } else {
        __pyx_t_3 = __pyx_t_6(__pyx_t_4);
        if (unlikely(!__pyx_t_3)) {
          PyObject* exc_type = PyErr_Occurred();
          if (exc_type) {
            if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
            else __PYX_ERR(0, 344, __pyx_L1_error)
          }
          break;
        }
        __Pyx_GOTREF(__pyx_t_3);
      }
      __Pyx_XDECREF_SET(__pyx_v_feature, __pyx_t_3);
      __pyx_t_3 = 0;

      /* "rasterio/_features.pyx":345
 *         ymaxs = []
 *         for feature in geometry['features']:
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])             # <<<<<<<<<<<<<<
 *             xmins.append(xmin)
 *             ymins.append(ymin)
 */
      __pyx_t_7 = __Pyx_GetModuleGlobalName(__pyx_n_s_bounds); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __pyx_t_8 = PyObject_GetItem(__pyx_v_feature, __pyx_n_s_geometry); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 345, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_8);
      __pyx_t_9 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_7))) {
        __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_7);
        if (likely(__pyx_t_9)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_7);
          __Pyx_INCREF(__pyx_t_9);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_7, function);
        }
      }
      if (!__pyx_t_9) {
        __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_7, __pyx_t_8); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        __Pyx_GOTREF(__pyx_t_3);
      } else {
        #if CYTHON_FAST_PYCALL
        if (PyFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_t_8};
          __pyx_t_3 = __Pyx_PyFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        #if CYTHON_FAST_PYCCALL
        if (__Pyx_PyFastCFunction_Check(__pyx_t_7)) {
          PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_t_8};
          __pyx_t_3 = __Pyx_PyCFunction_FastCall(__pyx_t_7, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
        } else
        #endif
        {
          __pyx_t_10 = PyTuple_New(1+1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_10);
          __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_9); __pyx_t_9 = NULL;
          __Pyx_GIVEREF(__pyx_t_8);
          PyTuple_SET_ITEM(__pyx_t_10, 0+1, __pyx_t_8);
          __pyx_t_8 = 0;
          __pyx_t_3 = __Pyx_PyObject_Call(__pyx_t_7, __pyx_t_10, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 345, __pyx_L1_error)
          __Pyx_GOTREF(__pyx_t_3);
          __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
        }
      }
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
      if ((likely(PyTuple_CheckExact(__pyx_t_3))) || (PyList_CheckExact(__pyx_t_3))) {
        PyObject* sequence = __pyx_t_3;
        #if !CYTHON_COMPILING_IN_PYPY
        Py_ssize_t size = Py_SIZE(sequence);
        #else
        Py_ssize_t size = PySequence_Size(sequence);
        #endif
        if (unlikely(size != 4)) {
          if (size > 4) __Pyx_RaiseTooManyValuesError(4);
          else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
          __PYX_ERR(0, 345, __pyx_L1_error)
        }
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        if (likely(PyTuple_CheckExact(sequence))) {
          __pyx_t_7 = PyTuple_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyTuple_GET_ITEM(sequence, 1); 
          __pyx_t_8 = PyTuple_GET_ITEM(sequence, 2); 
          __pyx_t_9 = PyTuple_GET_ITEM(sequence, 3); 
        } else {
          __pyx_t_7 = PyList_GET_ITEM(sequence, 0); 
          __pyx_t_10 = PyList_GET_ITEM(sequence, 1); 
          __pyx_t_8 = PyList_GET_ITEM(sequence, 2); 
          __pyx_t_9 = PyList_GET_ITEM(sequence, 3); 
        }
        __Pyx_INCREF(__pyx_t_7);
        __Pyx_INCREF(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_8);
        __Pyx_INCREF(__pyx_t_9);
        #else
        {
          Py_ssize_t i;
          PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_10,&__pyx_t_8,&__pyx_t_9};
          for (i=0; i < 4; i++) {
            PyObject* item = PySequence_ITEM(sequence, i); if (unlikely(!item)) __PYX_ERR(0, 345, __pyx_L1_error)
            __Pyx_GOTREF(item);
            *(temps[i]) = item;
          }
        }
        #endif
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      } else {
        Py_ssize_t index = -1;
        PyObject** temps[4] = {&__pyx_t_7,&__pyx_t_10,&__pyx_t_8,&__pyx_t_9};
        __pyx_t_11 = PyObject_GetIter(__pyx_t_3); if (unlikely(!__pyx_t_11)) __PYX_ERR(0, 345, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_11);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __pyx_t_12 = Py_TYPE(__pyx_t_11)->tp_iternext;
        for (index=0; index < 4; index++) {
          PyObject* item = __pyx_t_12(__pyx_t_11); if (unlikely(!item)) goto __pyx_L6_unpacking_failed;
          __Pyx_GOTREF(item);
          *(temps[index]) = item;
        }
        if (__Pyx_IternextUnpackEndCheck(__pyx_t_12(__pyx_t_11), 4) < 0) __PYX_ERR(0, 345, __pyx_L1_error)
        __pyx_t_12 = NULL;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        goto __pyx_L7_unpacking_done;
        __pyx_L6_unpacking_failed:;
        __Pyx_DECREF(__pyx_t_11); __pyx_t_11 = 0;
        __pyx_t_12 = NULL;
        if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
        __PYX_ERR(0, 345, __pyx_L1_error)
        __pyx_L7_unpacking_done:;
      }
      __Pyx_XDECREF_SET(__pyx_v_xmin, __pyx_t_7);
      __pyx_t_7 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ymin, __pyx_t_10);
      __pyx_t_10 = 0;
      __Pyx_XDECREF_SET(__pyx_v_xmax, __pyx_t_8);
      __pyx_t_8 = 0;
      __Pyx_XDECREF_SET(__pyx_v_ymax, __pyx_t_9);
      __pyx_t_9 = 0;

      /* "rasterio/_features.pyx":346
 *         for feature in geometry['features']:
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])
 *             xmins.append(xmin)             # <<<<<<<<<<<<<<
 *             ymins.append(ymin)
 *             xmaxs.append(xmax)
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_xmins, __pyx_v_xmin); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 346, __pyx_L1_error)

      /* "rasterio/_features.pyx":347
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])
 *             xmins.append(xmin)
 *             ymins.append(ymin)             # <<<<<<<<<<<<<<
 *             xmaxs.append(xmax)
 *             ymaxs.append(ymax)
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_ymins, __pyx_v_ymin); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 347, __pyx_L1_error)

      /* "rasterio/_features.pyx":348
 *             xmins.append(xmin)
 *             ymins.append(ymin)
 *             xmaxs.append(xmax)             # <<<<<<<<<<<<<<
 *             ymaxs.append(ymax)
 *         return min(xmins), min(ymins), max(xmaxs), max(ymaxs)
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_xmaxs, __pyx_v_xmax); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 348, __pyx_L1_error)

      /* "rasterio/_features.pyx":349
 *             ymins.append(ymin)
 *             xmaxs.append(xmax)
 *             ymaxs.append(ymax)             # <<<<<<<<<<<<<<
 *         return min(xmins), min(ymins), max(xmaxs), max(ymaxs)
 *     else:
 */
      __pyx_t_13 = __Pyx_PyList_Append(__pyx_v_ymaxs, __pyx_v_ymax); if (unlikely(__pyx_t_13 == -1)) __PYX_ERR(0, 349, __pyx_L1_error)

      /* "rasterio/_features.pyx":344
 *         xmaxs = []
 *         ymaxs = []
 *         for feature in geometry['features']:             # <<<<<<<<<<<<<<
 *             xmin, ymin, xmax, ymax = _bounds(feature['geometry'])
 *             xmins.append(xmin)
 */
    }
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

    /* "rasterio/_features.pyx":350
 *             xmaxs.append(xmax)
 *             ymaxs.append(ymax)
 *         return min(xmins), min(ymins), max(xmaxs), max(ymaxs)             # <<<<<<<<<<<<<<
 *     else:
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_xmins);
    __Pyx_GIVEREF(__pyx_v_xmins);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_xmins);
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_ymins);
    __Pyx_GIVEREF(__pyx_v_ymins);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_ymins);
    __pyx_t_9 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_4, NULL); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_xmaxs);
    __Pyx_GIVEREF(__pyx_v_xmaxs);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_xmaxs);
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_4, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_INCREF(__pyx_v_ymaxs);
    __Pyx_GIVEREF(__pyx_v_ymaxs);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_v_ymaxs);
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PyTuple_New(4); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 350, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_9);
    PyTuple_SET_ITEM(__pyx_t_4, 1, __pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_4, 2, __pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_4, 3, __pyx_t_10);
    __pyx_t_3 = 0;
    __pyx_t_9 = 0;
    __pyx_t_8 = 0;
    __pyx_t_10 = 0;
    __pyx_r = __pyx_t_4;
    __pyx_t_4 = 0;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":339
 *     TODO: add to Fiona.
 *     """
 *     if 'features' in geometry:             # <<<<<<<<<<<<<<
 *         xmins = []
 *         ymins = []
 */
  }

  /* "rasterio/_features.pyx":352
 *         return min(xmins), min(ymins), max(xmaxs), max(ymaxs)
 *     else:
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))             # <<<<<<<<<<<<<<
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])
 * 
 */
  /*else*/ {
    __pyx_t_10 = __Pyx_GetModuleGlobalName(__pyx_n_s_explode); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_8 = PyObject_GetItem(__pyx_v_geometry, __pyx_n_s_coordinates); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = NULL;
    if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_10))) {
      __pyx_t_9 = PyMethod_GET_SELF(__pyx_t_10);
      if (likely(__pyx_t_9)) {
        PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_10);
        __Pyx_INCREF(__pyx_t_9);
        __Pyx_INCREF(function);
        __Pyx_DECREF_SET(__pyx_t_10, function);
      }
    }
    if (!__pyx_t_9) {
      __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_10, __pyx_t_8); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_GOTREF(__pyx_t_4);
    } else {
      #if CYTHON_FAST_PYCALL
      if (PyFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_t_8};
        __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      #if CYTHON_FAST_PYCCALL
      if (__Pyx_PyFastCFunction_Check(__pyx_t_10)) {
        PyObject *__pyx_temp[2] = {__pyx_t_9, __pyx_t_8};
        __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_10, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
        __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      } else
      #endif
      {
        __pyx_t_3 = PyTuple_New(1+1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 352, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_GIVEREF(__pyx_t_9); PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_9); __pyx_t_9 = NULL;
        __Pyx_GIVEREF(__pyx_t_8);
        PyTuple_SET_ITEM(__pyx_t_3, 0+1, __pyx_t_8);
        __pyx_t_8 = 0;
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_10, __pyx_t_3, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      }
    }
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = PySequence_List(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_Tuple(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_zip, __pyx_t_4, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_4 = PySequence_Tuple(__pyx_t_10); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 352, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_v_xyz = ((PyObject*)__pyx_t_4);
    __pyx_t_4 = 0;

    /* "rasterio/_features.pyx":353
 *     else:
 *         xyz = tuple(zip(*list(_explode(geometry['coordinates']))))
 *         return min(xyz[0]), min(xyz[1]), max(xyz[0]), max(xyz[1])             # <<<<<<<<<<<<<<
 * 
 * 
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_4 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_10 = PyTuple_New(1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_10, 0, __pyx_t_4);
    __pyx_t_4 = 0;
    __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_10, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_10);
    __pyx_t_10 = 0;
    __pyx_t_10 = __Pyx_PyObject_Call(__pyx_builtin_min, __pyx_t_3, NULL); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_10);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_8 = PyTuple_New(1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_8, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_8, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_GetItemInt_Tuple(__pyx_v_xyz, 1, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __pyx_t_9 = PyTuple_New(1); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_8);
    __pyx_t_8 = 0;
    __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_max, __pyx_t_9, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_8);
    __Pyx_DECREF(__pyx_t_9); __pyx_t_9 = 0;
    __pyx_t_9 = PyTuple_New(4); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 353, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_9);
    __Pyx_GIVEREF(__pyx_t_4);
    PyTuple_SET_ITEM(__pyx_t_9, 0, __pyx_t_4);
    __Pyx_GIVEREF(__pyx_t_10);
    PyTuple_SET_ITEM(__pyx_t_9, 1, __pyx_t_10);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_9, 2, __pyx_t_3);
    __Pyx_GIVEREF(__pyx_t_8);
    PyTuple_SET_ITEM(__pyx_t_9, 3, __pyx_t_8);
    __pyx_t_4 = 0;
    __pyx_t_10 = 0;
    __pyx_t_3 = 0;
    __pyx_t_8 = 0;
    __pyx_r = __pyx_t_9;
    __pyx_t_9 = 0;
    goto __pyx_L0;
  }

  /* "rasterio/_features.pyx":334
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry.
 *     From Fiona 1.4.8 with updates here to handle feature collections.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_XDECREF(__pyx_t_11);
  __Pyx_AddTraceback("rasterio._features._bounds", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_xmins);
  __Pyx_XDECREF(__pyx_v_ymins);
  __Pyx_XDECREF(__pyx_v_xmaxs);
  __Pyx_XDECREF(__pyx_v_ymaxs);
  __Pyx_XDECREF(__pyx_v_feature);
  __Pyx_XDECREF(__pyx_v_xmin);
  __Pyx_XDECREF(__pyx_v_ymin);
  __Pyx_XDECREF(__pyx_v_xmax);
  __Pyx_XDECREF(__pyx_v_ymax);
  __Pyx_XDECREF(__pyx_v_xyz);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":386
 * 
 * 
 * cdef _deleteOgrGeom(OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *     """Delete an OGR geometry"""
 *     if geom != NULL:
 */

static PyObject *__pyx_f_8rasterio_9_features__deleteOgrGeom(OGRGeometryH __pyx_v_geom) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_deleteOgrGeom", 0);

  /* "rasterio/_features.pyx":388
 * cdef _deleteOgrGeom(OGRGeometryH geom):
 *     """Delete an OGR geometry"""
 *     if geom != NULL:             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(geom)
 *     geom = NULL
 */
  __pyx_t_1 = ((__pyx_v_geom != NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":389
 *     """Delete an OGR geometry"""
 *     if geom != NULL:
 *         OGR_G_DestroyGeometry(geom)             # <<<<<<<<<<<<<<
 *     geom = NULL
 * 
 */
    OGR_G_DestroyGeometry(__pyx_v_geom);

    /* "rasterio/_features.pyx":388
 * cdef _deleteOgrGeom(OGRGeometryH geom):
 *     """Delete an OGR geometry"""
 *     if geom != NULL:             # <<<<<<<<<<<<<<
 *         OGR_G_DestroyGeometry(geom)
 *     geom = NULL
 */
  }

  /* "rasterio/_features.pyx":390
 *     if geom != NULL:
 *         OGR_G_DestroyGeometry(geom)
 *     geom = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_geom = NULL;

  /* "rasterio/_features.pyx":386
 * 
 * 
 * cdef _deleteOgrGeom(OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *     """Delete an OGR geometry"""
 *     if geom != NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":396
 *     """Builds a GeoJSON (Fiona-style) geometry from an OGR geometry."""
 * 
 *     cdef _buildCoords(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         # Build a coordinate sequence
 *         cdef int i
 */

static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildCoords(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom) {
  int __pyx_v_i;
  int __pyx_v_npoints;
  PyObject *__pyx_v_coords = NULL;
  PyObject *__pyx_v_values = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  int __pyx_t_7;
  __Pyx_RefNannySetupContext("_buildCoords", 0);

  /* "rasterio/_features.pyx":399
 *         # Build a coordinate sequence
 *         cdef int i
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         npoints = OGR_G_GetPointCount(geom)
 */
  __pyx_t_1 = ((__pyx_v_geom == NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":400
 *         cdef int i
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         npoints = OGR_G_GetPointCount(geom)
 *         coords = []
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__25, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 400, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 400, __pyx_L1_error)

    /* "rasterio/_features.pyx":399
 *         # Build a coordinate sequence
 *         cdef int i
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         npoints = OGR_G_GetPointCount(geom)
 */
  }

  /* "rasterio/_features.pyx":401
 *         if geom == NULL:
 *             raise ValueError("Null geom")
 *         npoints = OGR_G_GetPointCount(geom)             # <<<<<<<<<<<<<<
 *         coords = []
 *         for i in range(npoints):
 */
  __pyx_v_npoints = OGR_G_GetPointCount(__pyx_v_geom);

  /* "rasterio/_features.pyx":402
 *             raise ValueError("Null geom")
 *         npoints = OGR_G_GetPointCount(geom)
 *         coords = []             # <<<<<<<<<<<<<<
 *         for i in range(npoints):
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 402, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_coords = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":403
 *         npoints = OGR_G_GetPointCount(geom)
 *         coords = []
 *         for i in range(npoints):             # <<<<<<<<<<<<<<
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]
 *             if self.ndims > 2:
 */
  __pyx_t_3 = __pyx_v_npoints;
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_i = __pyx_t_4;

    /* "rasterio/_features.pyx":404
 *         coords = []
 *         for i in range(npoints):
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]             # <<<<<<<<<<<<<<
 *             if self.ndims > 2:
 *                 values.append(OGR_G_GetZ(geom, i))
 */
    __pyx_t_2 = PyFloat_FromDouble(OGR_G_GetX(__pyx_v_geom, __pyx_v_i)); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = PyFloat_FromDouble(OGR_G_GetY(__pyx_v_geom, __pyx_v_i)); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __pyx_t_6 = PyList_New(2); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 404, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __Pyx_GIVEREF(__pyx_t_2);
    PyList_SET_ITEM(__pyx_t_6, 0, __pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_5);
    PyList_SET_ITEM(__pyx_t_6, 1, __pyx_t_5);
    __pyx_t_2 = 0;
    __pyx_t_5 = 0;
    __Pyx_XDECREF_SET(__pyx_v_values, ((PyObject*)__pyx_t_6));
    __pyx_t_6 = 0;

    /* "rasterio/_features.pyx":405
 *         for i in range(npoints):
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]
 *             if self.ndims > 2:             # <<<<<<<<<<<<<<
 *                 values.append(OGR_G_GetZ(geom, i))
 *             coords.append(tuple(values))
 */
    __pyx_t_6 = PyObject_RichCompare(__pyx_v_self->ndims, __pyx_int_2, Py_GT); __Pyx_XGOTREF(__pyx_t_6); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 405, __pyx_L1_error)
    __pyx_t_1 = __Pyx_PyObject_IsTrue(__pyx_t_6); if (unlikely(__pyx_t_1 < 0)) __PYX_ERR(0, 405, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
    if (__pyx_t_1) {

      /* "rasterio/_features.pyx":406
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]
 *             if self.ndims > 2:
 *                 values.append(OGR_G_GetZ(geom, i))             # <<<<<<<<<<<<<<
 *             coords.append(tuple(values))
 *         return coords
 */
      __pyx_t_6 = PyFloat_FromDouble(OGR_G_GetZ(__pyx_v_geom, __pyx_v_i)); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_values, __pyx_t_6); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 406, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;

      /* "rasterio/_features.pyx":405
 *         for i in range(npoints):
 *             values = [OGR_G_GetX(geom, i), OGR_G_GetY(geom, i)]
 *             if self.ndims > 2:             # <<<<<<<<<<<<<<
 *                 values.append(OGR_G_GetZ(geom, i))
 *             coords.append(tuple(values))
 */
    }

    /* "rasterio/_features.pyx":407
 *             if self.ndims > 2:
 *                 values.append(OGR_G_GetZ(geom, i))
 *             coords.append(tuple(values))             # <<<<<<<<<<<<<<
 *         return coords
 * 
 */
    __pyx_t_6 = PyList_AsTuple(__pyx_v_values); if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_6);
    __pyx_t_7 = __Pyx_PyList_Append(__pyx_v_coords, __pyx_t_6); if (unlikely(__pyx_t_7 == -1)) __PYX_ERR(0, 407, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
  }

  /* "rasterio/_features.pyx":408
 *                 values.append(OGR_G_GetZ(geom, i))
 *             coords.append(tuple(values))
 *         return coords             # <<<<<<<<<<<<<<
 * 
 *     cpdef _buildPoint(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_coords);
  __pyx_r = __pyx_v_coords;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":396
 *     """Builds a GeoJSON (Fiona-style) geometry from an OGR geometry."""
 * 
 *     cdef _buildCoords(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         # Build a coordinate sequence
 *         cdef int i
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildCoords", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coords);
  __Pyx_XDECREF(__pyx_v_values);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":410
 *         return coords
 * 
 *     cpdef _buildPoint(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'Point',
 */

static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_1_buildPoint(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_buildPoint", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buildPoint); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_1_buildPoint)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 410, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "rasterio/_features.pyx":411
 * 
 *     cpdef _buildPoint(self):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': 'Point',
 *             'coordinates': self._buildCoords(self.geom)[0]}
 */
  __Pyx_XDECREF(__pyx_r);

  /* "rasterio/_features.pyx":412
 *     cpdef _buildPoint(self):
 *         return {
 *             'type': 'Point',             # <<<<<<<<<<<<<<
 *             'coordinates': self._buildCoords(self.geom)[0]}
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_n_s_Point) < 0) __PYX_ERR(0, 412, __pyx_L1_error)

  /* "rasterio/_features.pyx":413
 *         return {
 *             'type': 'Point',
 *             'coordinates': self._buildCoords(self.geom)[0]}             # <<<<<<<<<<<<<<
 * 
 *     cpdef _buildLineString(self):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildCoords(__pyx_v_self, __pyx_v_self->geom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = __Pyx_GetItemInt(__pyx_t_2, 0, long, 1, __Pyx_PyInt_From_long, 0, 0, 1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 413, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_coordinates, __pyx_t_3) < 0) __PYX_ERR(0, 412, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":410
 *         return coords
 * 
 *     cpdef _buildPoint(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'Point',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildPoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_1_buildPoint(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_1_buildPoint(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_buildPoint (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_11GeomBuilder__buildPoint(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_buildPoint", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8rasterio_9_features_11GeomBuilder__buildPoint(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 410, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildPoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":415
 *             'coordinates': self._buildCoords(self.geom)[0]}
 * 
 *     cpdef _buildLineString(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'LineString',
 */

static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_3_buildLineString(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildLineString(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_buildLineString", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buildLineString); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_3_buildLineString)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 415, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "rasterio/_features.pyx":416
 * 
 *     cpdef _buildLineString(self):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': 'LineString',
 *             'coordinates': self._buildCoords(self.geom)}
 */
  __Pyx_XDECREF(__pyx_r);

  /* "rasterio/_features.pyx":417
 *     cpdef _buildLineString(self):
 *         return {
 *             'type': 'LineString',             # <<<<<<<<<<<<<<
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_n_s_LineString) < 0) __PYX_ERR(0, 417, __pyx_L1_error)

  /* "rasterio/_features.pyx":418
 *         return {
 *             'type': 'LineString',
 *             'coordinates': self._buildCoords(self.geom)}             # <<<<<<<<<<<<<<
 * 
 *     cpdef _buildLinearRing(self):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildCoords(__pyx_v_self, __pyx_v_self->geom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 418, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_coordinates, __pyx_t_2) < 0) __PYX_ERR(0, 417, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":415
 *             'coordinates': self._buildCoords(self.geom)[0]}
 * 
 *     cpdef _buildLineString(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'LineString',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildLineString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_3_buildLineString(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_3_buildLineString(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_buildLineString (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_11GeomBuilder_2_buildLineString(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_2_buildLineString(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_buildLineString", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8rasterio_9_features_11GeomBuilder__buildLineString(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 415, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildLineString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":420
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 *     cpdef _buildLinearRing(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'LinearRing',
 */

static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_5_buildLinearRing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildLinearRing(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("_buildLinearRing", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buildLinearRing); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_5_buildLinearRing)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 420, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "rasterio/_features.pyx":421
 * 
 *     cpdef _buildLinearRing(self):
 *         return {             # <<<<<<<<<<<<<<
 *             'type': 'LinearRing',
 *             'coordinates': self._buildCoords(self.geom)}
 */
  __Pyx_XDECREF(__pyx_r);

  /* "rasterio/_features.pyx":422
 *     cpdef _buildLinearRing(self):
 *         return {
 *             'type': 'LinearRing',             # <<<<<<<<<<<<<<
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 */
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_n_s_LinearRing) < 0) __PYX_ERR(0, 422, __pyx_L1_error)

  /* "rasterio/_features.pyx":423
 *         return {
 *             'type': 'LinearRing',
 *             'coordinates': self._buildCoords(self.geom)}             # <<<<<<<<<<<<<<
 * 
 *     cdef _buildParts(self, OGRGeometryH geom):
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildCoords(__pyx_v_self, __pyx_v_self->geom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 423, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_coordinates, __pyx_t_2) < 0) __PYX_ERR(0, 422, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":420
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 *     cpdef _buildLinearRing(self):             # <<<<<<<<<<<<<<
 *         return {
 *             'type': 'LinearRing',
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildLinearRing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_5_buildLinearRing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_5_buildLinearRing(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_buildLinearRing (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_11GeomBuilder_4_buildLinearRing(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_4_buildLinearRing(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_buildLinearRing", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8rasterio_9_features_11GeomBuilder__buildLinearRing(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 420, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildLinearRing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":425
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 *     cdef _buildParts(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         cdef int j
 *         cdef OGRGeometryH part
 */

static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildParts(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom) {
  int __pyx_v_j;
  OGRGeometryH __pyx_v_part;
  PyObject *__pyx_v_parts = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  int __pyx_t_6;
  __Pyx_RefNannySetupContext("_buildParts", 0);

  /* "rasterio/_features.pyx":428
 *         cdef int j
 *         cdef OGRGeometryH part
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         parts = []
 */
  __pyx_t_1 = ((__pyx_v_geom == NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":429
 *         cdef OGRGeometryH part
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         parts = []
 *         for j in range(OGR_G_GetGeometryCount(geom)):
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__26, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 429, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 429, __pyx_L1_error)

    /* "rasterio/_features.pyx":428
 *         cdef int j
 *         cdef OGRGeometryH part
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         parts = []
 */
  }

  /* "rasterio/_features.pyx":430
 *         if geom == NULL:
 *             raise ValueError("Null geom")
 *         parts = []             # <<<<<<<<<<<<<<
 *         for j in range(OGR_G_GetGeometryCount(geom)):
 *             part = OGR_G_GetGeometryRef(geom, j)
 */
  __pyx_t_2 = PyList_New(0); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 430, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_v_parts = ((PyObject*)__pyx_t_2);
  __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":431
 *             raise ValueError("Null geom")
 *         parts = []
 *         for j in range(OGR_G_GetGeometryCount(geom)):             # <<<<<<<<<<<<<<
 *             part = OGR_G_GetGeometryRef(geom, j)
 *             parts.append(GeomBuilder().build(part))
 */
  __pyx_t_3 = OGR_G_GetGeometryCount(__pyx_v_geom);
  for (__pyx_t_4 = 0; __pyx_t_4 < __pyx_t_3; __pyx_t_4+=1) {
    __pyx_v_j = __pyx_t_4;

    /* "rasterio/_features.pyx":432
 *         parts = []
 *         for j in range(OGR_G_GetGeometryCount(geom)):
 *             part = OGR_G_GetGeometryRef(geom, j)             # <<<<<<<<<<<<<<
 *             parts.append(GeomBuilder().build(part))
 *         return parts
 */
    __pyx_v_part = OGR_G_GetGeometryRef(__pyx_v_geom, __pyx_v_j);

    /* "rasterio/_features.pyx":433
 *         for j in range(OGR_G_GetGeometryCount(geom)):
 *             part = OGR_G_GetGeometryRef(geom, j)
 *             parts.append(GeomBuilder().build(part))             # <<<<<<<<<<<<<<
 *         return parts
 * 
 */
    __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_GeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_5 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_2)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_2), __pyx_v_part); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_5);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __pyx_t_6 = __Pyx_PyList_Append(__pyx_v_parts, __pyx_t_5); if (unlikely(__pyx_t_6 == -1)) __PYX_ERR(0, 433, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  }

  /* "rasterio/_features.pyx":434
 *             part = OGR_G_GetGeometryRef(geom, j)
 *             parts.append(GeomBuilder().build(part))
 *         return parts             # <<<<<<<<<<<<<<
 * 
 *     cpdef _buildPolygon(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(__pyx_v_parts);
  __pyx_r = __pyx_v_parts;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":425
 *             'coordinates': self._buildCoords(self.geom)}
 * 
 *     cdef _buildParts(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         cdef int j
 *         cdef OGRGeometryH part
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildParts", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_parts);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":436
 *         return parts
 * 
 *     cpdef _buildPolygon(self):             # <<<<<<<<<<<<<<
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'Polygon', 'coordinates': coordinates}
 */

static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_7_buildPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_coordinates = NULL;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildPolygon", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buildPolygon); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_7_buildPolygon)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 436, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "rasterio/_features.pyx":437
 * 
 *     cpdef _buildPolygon(self):
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]             # <<<<<<<<<<<<<<
 *         return {'type': 'Polygon', 'coordinates': coordinates}
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildParts(__pyx_v_self, __pyx_v_self->geom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 437, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 437, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 437, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 437, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_GetItem(__pyx_v_p, __pyx_n_s_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 437, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_coordinates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":438
 *     cpdef _buildPolygon(self):
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'Polygon', 'coordinates': coordinates}             # <<<<<<<<<<<<<<
 * 
 *     cpdef _buildMultiPolygon(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 438, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_n_s_Polygon) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_coordinates, __pyx_v_coordinates) < 0) __PYX_ERR(0, 438, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":436
 *         return parts
 * 
 *     cpdef _buildPolygon(self):             # <<<<<<<<<<<<<<
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'Polygon', 'coordinates': coordinates}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_7_buildPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_7_buildPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_buildPolygon (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_11GeomBuilder_6_buildPolygon(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_6_buildPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_buildPolygon", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8rasterio_9_features_11GeomBuilder__buildPolygon(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 436, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":440
 *         return {'type': 'Polygon', 'coordinates': coordinates}
 * 
 *     cpdef _buildMultiPolygon(self):             # <<<<<<<<<<<<<<
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}
 */

static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_9_buildMultiPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder__buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, int __pyx_skip_dispatch) {
  PyObject *__pyx_v_coordinates = NULL;
  PyObject *__pyx_v_p = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildMultiPolygon", 0);
  /* Check if called by wrapper */
  if (unlikely(__pyx_skip_dispatch)) ;
  /* Check if overridden in Python */
  else if (unlikely(Py_TYPE(((PyObject *)__pyx_v_self))->tp_dictoffset != 0)) {
    __pyx_t_1 = __Pyx_PyObject_GetAttrStr(((PyObject *)__pyx_v_self), __pyx_n_s_buildMultiPolygon); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    if (!PyCFunction_Check(__pyx_t_1) || (PyCFunction_GET_FUNCTION(__pyx_t_1) != (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_9_buildMultiPolygon)) {
      __Pyx_XDECREF(__pyx_r);
      __Pyx_INCREF(__pyx_t_1);
      __pyx_t_3 = __pyx_t_1; __pyx_t_4 = NULL;
      if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_3))) {
        __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_3);
        if (likely(__pyx_t_4)) {
          PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_3);
          __Pyx_INCREF(__pyx_t_4);
          __Pyx_INCREF(function);
          __Pyx_DECREF_SET(__pyx_t_3, function);
        }
      }
      if (__pyx_t_4) {
        __pyx_t_2 = __Pyx_PyObject_CallOneArg(__pyx_t_3, __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      } else {
        __pyx_t_2 = __Pyx_PyObject_CallNoArg(__pyx_t_3); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 440, __pyx_L1_error)
      }
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_r = __pyx_t_2;
      __pyx_t_2 = 0;
      __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
      goto __pyx_L0;
    }
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  }

  /* "rasterio/_features.pyx":441
 * 
 *     cpdef _buildMultiPolygon(self):
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]             # <<<<<<<<<<<<<<
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}
 * 
 */
  __pyx_t_1 = PyList_New(0); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildParts(__pyx_v_self, __pyx_v_self->geom); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  if (likely(PyList_CheckExact(__pyx_t_2)) || PyTuple_CheckExact(__pyx_t_2)) {
    __pyx_t_3 = __pyx_t_2; __Pyx_INCREF(__pyx_t_3); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_3 = PyObject_GetIter(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_6 = Py_TYPE(__pyx_t_3)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 441, __pyx_L1_error)
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_3))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyList_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_3)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_2 = PyTuple_GET_ITEM(__pyx_t_3, __pyx_t_5); __Pyx_INCREF(__pyx_t_2); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 441, __pyx_L1_error)
        #else
        __pyx_t_2 = PySequence_ITEM(__pyx_t_3, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_2);
        #endif
      }
    } else {
      __pyx_t_2 = __pyx_t_6(__pyx_t_3);
      if (unlikely(!__pyx_t_2)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 441, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_2);
    }
    __Pyx_XDECREF_SET(__pyx_v_p, __pyx_t_2);
    __pyx_t_2 = 0;
    __pyx_t_2 = PyObject_GetItem(__pyx_v_p, __pyx_n_s_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    if (unlikely(__Pyx_ListComp_Append(__pyx_t_1, (PyObject*)__pyx_t_2))) __PYX_ERR(0, 441, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  }
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_v_coordinates = ((PyObject*)__pyx_t_1);
  __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":442
 *     cpdef _buildMultiPolygon(self):
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}             # <<<<<<<<<<<<<<
 * 
 *     cdef build(self, OGRGeometryH geom):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyDict_New(); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 442, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_type, __pyx_n_s_MultiPolygon) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_1, __pyx_n_s_coordinates, __pyx_v_coordinates) < 0) __PYX_ERR(0, 442, __pyx_L1_error)
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":440
 *         return {'type': 'Polygon', 'coordinates': coordinates}
 * 
 *     cpdef _buildMultiPolygon(self):             # <<<<<<<<<<<<<<
 *         coordinates = [p['coordinates'] for p in self._buildParts(self.geom)]
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildMultiPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_XDECREF(__pyx_v_p);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_9_buildMultiPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_11GeomBuilder_9_buildMultiPolygon(PyObject *__pyx_v_self, CYTHON_UNUSED PyObject *unused) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("_buildMultiPolygon (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_11GeomBuilder_8_buildMultiPolygon(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_11GeomBuilder_8_buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("_buildMultiPolygon", 0);
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = __pyx_f_8rasterio_9_features_11GeomBuilder__buildMultiPolygon(__pyx_v_self, 1); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 440, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder._buildMultiPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":444
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}
 * 
 *     cdef build(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:
 */

static PyObject *__pyx_f_8rasterio_9_features_11GeomBuilder_build(struct __pyx_obj_8rasterio_9_features_GeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom) {
  unsigned int __pyx_v_etype;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  __Pyx_RefNannySetupContext("build", 0);

  /* "rasterio/_features.pyx":446
 *     cdef build(self, OGRGeometryH geom):
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 */
  __pyx_t_1 = ((__pyx_v_geom == NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":447
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 *         self.code = etype
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__27, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 447, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 447, __pyx_L1_error)

    /* "rasterio/_features.pyx":446
 *     cdef build(self, OGRGeometryH geom):
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:             # <<<<<<<<<<<<<<
 *             raise ValueError("Null geom")
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 */
  }

  /* "rasterio/_features.pyx":448
 *         if geom == NULL:
 *             raise ValueError("Null geom")
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)             # <<<<<<<<<<<<<<
 *         self.code = etype
 *         self.geomtypename = GEOMETRY_TYPES[self.code & (~0x80000000)]
 */
  __pyx_v_etype = OGR_G_GetGeometryType(__pyx_v_geom);

  /* "rasterio/_features.pyx":449
 *             raise ValueError("Null geom")
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 *         self.code = etype             # <<<<<<<<<<<<<<
 *         self.geomtypename = GEOMETRY_TYPES[self.code & (~0x80000000)]
 *         self.ndims = OGR_G_GetCoordinateDimension(geom)
 */
  __pyx_t_2 = __Pyx_PyInt_From_unsigned_int(__pyx_v_etype); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 449, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_GIVEREF(__pyx_t_2);
  __Pyx_GOTREF(__pyx_v_self->code);
  __Pyx_DECREF(__pyx_v_self->code);
  __pyx_v_self->code = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":450
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 *         self.code = etype
 *         self.geomtypename = GEOMETRY_TYPES[self.code & (~0x80000000)]             # <<<<<<<<<<<<<<
 *         self.ndims = OGR_G_GetCoordinateDimension(geom)
 *         self.geom = geom
 */
  __pyx_t_2 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOMETRY_TYPES); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __pyx_t_3 = PyNumber_Invert(__pyx_int_2147483648); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_4 = PyNumber_And(__pyx_v_self->code, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = PyObject_GetItem(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 450, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->geomtypename);
  __Pyx_DECREF(__pyx_v_self->geomtypename);
  __pyx_v_self->geomtypename = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "rasterio/_features.pyx":451
 *         self.code = etype
 *         self.geomtypename = GEOMETRY_TYPES[self.code & (~0x80000000)]
 *         self.ndims = OGR_G_GetCoordinateDimension(geom)             # <<<<<<<<<<<<<<
 *         self.geom = geom
 * 
 */
  __pyx_t_3 = __Pyx_PyInt_From_int(OGR_G_GetCoordinateDimension(__pyx_v_geom)); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 451, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_GIVEREF(__pyx_t_3);
  __Pyx_GOTREF(__pyx_v_self->ndims);
  __Pyx_DECREF(__pyx_v_self->ndims);
  __pyx_v_self->ndims = __pyx_t_3;
  __pyx_t_3 = 0;

  /* "rasterio/_features.pyx":452
 *         self.geomtypename = GEOMETRY_TYPES[self.code & (~0x80000000)]
 *         self.ndims = OGR_G_GetCoordinateDimension(geom)
 *         self.geom = geom             # <<<<<<<<<<<<<<
 * 
 *         return getattr(self, '_build' + self.geomtypename)()
 */
  __pyx_v_self->geom = __pyx_v_geom;

  /* "rasterio/_features.pyx":454
 *         self.geom = geom
 * 
 *         return getattr(self, '_build' + self.geomtypename)()             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_4 = PyNumber_Add(__pyx_n_s_build, __pyx_v_self->geomtypename); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __pyx_t_2 = __Pyx_GetAttr(((PyObject *)__pyx_v_self), __pyx_t_4); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 454, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  __pyx_t_4 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_2))) {
    __pyx_t_4 = PyMethod_GET_SELF(__pyx_t_2);
    if (likely(__pyx_t_4)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_2);
      __Pyx_INCREF(__pyx_t_4);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_2, function);
    }
  }
  if (__pyx_t_4) {
    __pyx_t_3 = __Pyx_PyObject_CallOneArg(__pyx_t_2, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  } else {
    __pyx_t_3 = __Pyx_PyObject_CallNoArg(__pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 454, __pyx_L1_error)
  }
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  __pyx_r = __pyx_t_3;
  __pyx_t_3 = 0;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":444
 *         return {'type': 'MultiPolygon', 'coordinates': coordinates}
 * 
 *     cdef build(self, OGRGeometryH geom):             # <<<<<<<<<<<<<<
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("rasterio._features.GeomBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":463
 *     """
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)
 *         if geom is NULL:
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__createOgrGeometry(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, int __pyx_v_geom_type) {
  OGRGeometryH __pyx_v_geom;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  __Pyx_RefNannySetupContext("_createOgrGeometry", 0);

  /* "rasterio/_features.pyx":464
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL:
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)             # <<<<<<<<<<<<<<
 *         if geom is NULL:
 *             raise Exception(
 */
  __pyx_v_geom = OGR_G_CreateGeometry(__pyx_v_geom_type);

  /* "rasterio/_features.pyx":465
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL:
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)
 *         if geom is NULL:             # <<<<<<<<<<<<<<
 *             raise Exception(
 *                 "Could not create OGR Geometry of type: %i" % geom_type)
 */
  __pyx_t_1 = ((__pyx_v_geom == NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":467
 *         if geom is NULL:
 *             raise Exception(
 *                 "Could not create OGR Geometry of type: %i" % geom_type)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    __pyx_t_2 = __Pyx_PyInt_From_int(__pyx_v_geom_type); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_3 = __Pyx_PyString_Format(__pyx_kp_s_Could_not_create_OGR_Geometry_of, __pyx_t_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 467, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

    /* "rasterio/_features.pyx":466
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)
 *         if geom is NULL:
 *             raise Exception(             # <<<<<<<<<<<<<<
 *                 "Could not create OGR Geometry of type: %i" % geom_type)
 *         return geom
 */
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_t_3 = __Pyx_PyObject_Call(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])), __pyx_t_2, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 466, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(0, 466, __pyx_L1_error)

    /* "rasterio/_features.pyx":465
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL:
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)
 *         if geom is NULL:             # <<<<<<<<<<<<<<
 *             raise Exception(
 *                 "Could not create OGR Geometry of type: %i" % geom_type)
 */
  }

  /* "rasterio/_features.pyx":468
 *             raise Exception(
 *                 "Could not create OGR Geometry of type: %i" % geom_type)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":463
 *     """
 * 
 *     cdef OGRGeometryH _createOgrGeometry(self, int geom_type) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = OGR_G_CreateGeometry(geom_type)
 *         if geom is NULL:
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._createOgrGeometry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":470
 *         return geom
 * 
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):             # <<<<<<<<<<<<<<
 *         if len(coordinate) == 2:
 *             x, y = coordinate
 */

static PyObject *__pyx_f_8rasterio_9_features_14OGRGeomBuilder__addPointToGeometry(CYTHON_UNUSED struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, OGRGeometryH __pyx_v_geom, PyObject *__pyx_v_coordinate) {
  PyObject *__pyx_v_x = NULL;
  PyObject *__pyx_v_y = NULL;
  PyObject *__pyx_v_z = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  Py_ssize_t __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *(*__pyx_t_6)(PyObject *);
  double __pyx_t_7;
  double __pyx_t_8;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  double __pyx_t_11;
  __Pyx_RefNannySetupContext("_addPointToGeometry", 0);

  /* "rasterio/_features.pyx":471
 * 
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):
 *         if len(coordinate) == 2:             # <<<<<<<<<<<<<<
 *             x, y = coordinate
 *             OGR_G_AddPoint_2D(geom, x, y)
 */
  __pyx_t_1 = PyObject_Length(__pyx_v_coordinate); if (unlikely(__pyx_t_1 == -1)) __PYX_ERR(0, 471, __pyx_L1_error)
  __pyx_t_2 = ((__pyx_t_1 == 2) != 0);
  if (__pyx_t_2) {

    /* "rasterio/_features.pyx":472
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):
 *         if len(coordinate) == 2:
 *             x, y = coordinate             # <<<<<<<<<<<<<<
 *             OGR_G_AddPoint_2D(geom, x, y)
 *         else:
 */
    if ((likely(PyTuple_CheckExact(__pyx_v_coordinate))) || (PyList_CheckExact(__pyx_v_coordinate))) {
      PyObject* sequence = __pyx_v_coordinate;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 472, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_4 = PyList_GET_ITEM(sequence, 1); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      Py_ssize_t index = -1;
      __pyx_t_5 = PyObject_GetIter(__pyx_v_coordinate); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 472, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_6 = Py_TYPE(__pyx_t_5)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_3)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_4 = __pyx_t_6(__pyx_t_5); if (unlikely(!__pyx_t_4)) goto __pyx_L4_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_4);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_5), 2) < 0) __PYX_ERR(0, 472, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      goto __pyx_L5_unpacking_done;
      __pyx_L4_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 472, __pyx_L1_error)
      __pyx_L5_unpacking_done:;
    }
    __pyx_v_x = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_y = __pyx_t_4;
    __pyx_t_4 = 0;

    /* "rasterio/_features.pyx":473
 *         if len(coordinate) == 2:
 *             x, y = coordinate
 *             OGR_G_AddPoint_2D(geom, x, y)             # <<<<<<<<<<<<<<
 *         else:
 *             x, y, z = coordinate[:3]
 */
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_v_x); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_y); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 473, __pyx_L1_error)
    OGR_G_AddPoint_2D(__pyx_v_geom, __pyx_t_7, __pyx_t_8);

    /* "rasterio/_features.pyx":471
 * 
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):
 *         if len(coordinate) == 2:             # <<<<<<<<<<<<<<
 *             x, y = coordinate
 *             OGR_G_AddPoint_2D(geom, x, y)
 */
    goto __pyx_L3;
  }

  /* "rasterio/_features.pyx":475
 *             OGR_G_AddPoint_2D(geom, x, y)
 *         else:
 *             x, y, z = coordinate[:3]             # <<<<<<<<<<<<<<
 *             OGR_G_AddPoint(geom, x, y, z)
 * 
 */
  /*else*/ {
    __pyx_t_4 = __Pyx_PyObject_GetSlice(__pyx_v_coordinate, 0, 3, NULL, NULL, &__pyx_slice__28, 0, 1, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 475, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    if ((likely(PyTuple_CheckExact(__pyx_t_4))) || (PyList_CheckExact(__pyx_t_4))) {
      PyObject* sequence = __pyx_t_4;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 3)) {
        if (size > 3) __Pyx_RaiseTooManyValuesError(3);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(0, 475, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      if (likely(PyTuple_CheckExact(sequence))) {
        __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyTuple_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyTuple_GET_ITEM(sequence, 2); 
      } else {
        __pyx_t_3 = PyList_GET_ITEM(sequence, 0); 
        __pyx_t_5 = PyList_GET_ITEM(sequence, 1); 
        __pyx_t_9 = PyList_GET_ITEM(sequence, 2); 
      }
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_9);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_5 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_5);
      __pyx_t_9 = PySequence_ITEM(sequence, 2); if (unlikely(!__pyx_t_9)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_9);
      #endif
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    } else {
      Py_ssize_t index = -1;
      __pyx_t_10 = PyObject_GetIter(__pyx_t_4); if (unlikely(!__pyx_t_10)) __PYX_ERR(0, 475, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_10);
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = Py_TYPE(__pyx_t_10)->tp_iternext;
      index = 0; __pyx_t_3 = __pyx_t_6(__pyx_t_10); if (unlikely(!__pyx_t_3)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_3);
      index = 1; __pyx_t_5 = __pyx_t_6(__pyx_t_10); if (unlikely(!__pyx_t_5)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_5);
      index = 2; __pyx_t_9 = __pyx_t_6(__pyx_t_10); if (unlikely(!__pyx_t_9)) goto __pyx_L6_unpacking_failed;
      __Pyx_GOTREF(__pyx_t_9);
      if (__Pyx_IternextUnpackEndCheck(__pyx_t_6(__pyx_t_10), 3) < 0) __PYX_ERR(0, 475, __pyx_L1_error)
      __pyx_t_6 = NULL;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      goto __pyx_L7_unpacking_done;
      __pyx_L6_unpacking_failed:;
      __Pyx_DECREF(__pyx_t_10); __pyx_t_10 = 0;
      __pyx_t_6 = NULL;
      if (__Pyx_IterFinish() == 0) __Pyx_RaiseNeedMoreValuesError(index);
      __PYX_ERR(0, 475, __pyx_L1_error)
      __pyx_L7_unpacking_done:;
    }
    __pyx_v_x = __pyx_t_3;
    __pyx_t_3 = 0;
    __pyx_v_y = __pyx_t_5;
    __pyx_t_5 = 0;
    __pyx_v_z = __pyx_t_9;
    __pyx_t_9 = 0;

    /* "rasterio/_features.pyx":476
 *         else:
 *             x, y, z = coordinate[:3]
 *             OGR_G_AddPoint(geom, x, y, z)             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildPoint(self, object coordinates) except NULL:
 */
    __pyx_t_8 = __pyx_PyFloat_AsDouble(__pyx_v_x); if (unlikely((__pyx_t_8 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
    __pyx_t_7 = __pyx_PyFloat_AsDouble(__pyx_v_y); if (unlikely((__pyx_t_7 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
    __pyx_t_11 = __pyx_PyFloat_AsDouble(__pyx_v_z); if (unlikely((__pyx_t_11 == (double)-1) && PyErr_Occurred())) __PYX_ERR(0, 476, __pyx_L1_error)
    OGR_G_AddPoint(__pyx_v_geom, __pyx_t_8, __pyx_t_7, __pyx_t_11);
  }
  __pyx_L3:;

  /* "rasterio/_features.pyx":470
 *         return geom
 * 
 *     cdef _addPointToGeometry(self, OGRGeometryH geom, object coordinate):             # <<<<<<<<<<<<<<
 *         if len(coordinate) == 2:
 *             x, y = coordinate
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_9);
  __Pyx_XDECREF(__pyx_t_10);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._addPointToGeometry", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_x);
  __Pyx_XDECREF(__pyx_v_y);
  __Pyx_XDECREF(__pyx_v_z);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":478
 *             OGR_G_AddPoint(geom, x, y, z)
 * 
 *     cdef OGRGeometryH _buildPoint(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPoint(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_geom;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  __Pyx_RefNannySetupContext("_buildPoint", 0);

  /* "rasterio/_features.pyx":480
 *     cdef OGRGeometryH _buildPoint(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])             # <<<<<<<<<<<<<<
 *         self._addPointToGeometry(geom, coordinates)
 *         return geom
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_Point); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 480, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":479
 * 
 *     cdef OGRGeometryH _buildPoint(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])
 *         self._addPointToGeometry(geom, coordinates)
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 479, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":481
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])
 *         self._addPointToGeometry(geom, coordinates)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
  __pyx_t_2 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_addPointToGeometry(__pyx_v_self, __pyx_v_geom, __pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 481, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":482
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])
 *         self._addPointToGeometry(geom, coordinates)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildLineString(self, object coordinates) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":478
 *             OGR_G_AddPoint(geom, x, y, z)
 * 
 *     cdef OGRGeometryH _buildPoint(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Point'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildPoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":484
 *         return geom
 * 
 *     cdef OGRGeometryH _buildLineString(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLineString(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_coordinate = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildLineString", 0);

  /* "rasterio/_features.pyx":486
 *     cdef OGRGeometryH _buildLineString(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])             # <<<<<<<<<<<<<<
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_LineString); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 486, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":485
 * 
 *     cdef OGRGeometryH _buildLineString(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 *         for coordinate in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 485, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":487
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             self._addPointToGeometry(geom, coordinate)
 *         return geom
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 487, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 487, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 487, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 487, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 487, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 487, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_coordinate, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":488
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_addPointToGeometry(__pyx_v_self, __pyx_v_geom, __pyx_v_coordinate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 488, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":487
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             self._addPointToGeometry(geom, coordinate)
 *         return geom
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":489
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildLinearRing(self, object coordinates) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":484
 *         return geom
 * 
 *     cdef OGRGeometryH _buildLineString(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LineString'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildLineString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coordinate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":491
 *         return geom
 * 
 *     cdef OGRGeometryH _buildLinearRing(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLinearRing(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_coordinate = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildLinearRing", 0);

  /* "rasterio/_features.pyx":493
 *     cdef OGRGeometryH _buildLinearRing(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])             # <<<<<<<<<<<<<<
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_LinearRing); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 493, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":492
 * 
 *     cdef OGRGeometryH _buildLinearRing(self, object coordinates) except NULL:
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 *         for coordinate in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 492, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":494
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             self._addPointToGeometry(geom, coordinate)
 *         OGR_G_CloseRings(geom)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 494, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 494, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 494, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 494, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 494, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 494, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_coordinate, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":495
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)             # <<<<<<<<<<<<<<
 *         OGR_G_CloseRings(geom)
 *         return geom
 */
    __pyx_t_1 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_addPointToGeometry(__pyx_v_self, __pyx_v_geom, __pyx_v_coordinate); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 495, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":494
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             self._addPointToGeometry(geom, coordinate)
 *         OGR_G_CloseRings(geom)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":496
 *         for coordinate in coordinates:
 *             self._addPointToGeometry(geom, coordinate)
 *         OGR_G_CloseRings(geom)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
  OGR_G_CloseRings(__pyx_v_geom);

  /* "rasterio/_features.pyx":497
 *             self._addPointToGeometry(geom, coordinate)
 *         OGR_G_CloseRings(geom)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildPolygon(self, object coordinates) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":491
 *         return geom
 * 
 *     cdef OGRGeometryH _buildLinearRing(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['LinearRing'])
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildLinearRing", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coordinate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":499
 *         return geom
 * 
 *     cdef OGRGeometryH _buildPolygon(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH ring = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPolygon(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_ring;
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_r = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildPolygon", 0);

  /* "rasterio/_features.pyx":500
 * 
 *     cdef OGRGeometryH _buildPolygon(self, object coordinates) except NULL:
 *         cdef OGRGeometryH ring = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])
 */
  __pyx_v_ring = NULL;

  /* "rasterio/_features.pyx":502
 *         cdef OGRGeometryH ring = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])             # <<<<<<<<<<<<<<
 *         for r in coordinates:
 *             ring = self._buildLinearRing(r)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_Polygon); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 502, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":501
 *     cdef OGRGeometryH _buildPolygon(self, object coordinates) except NULL:
 *         cdef OGRGeometryH ring = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])
 *         for r in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 501, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":503
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])
 *         for r in coordinates:             # <<<<<<<<<<<<<<
 *             ring = self._buildLinearRing(r)
 *             OGR_G_AddGeometryDirectly(geom, ring)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 503, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 503, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 503, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 503, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 503, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_r, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":504
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])
 *         for r in coordinates:
 *             ring = self._buildLinearRing(r)             # <<<<<<<<<<<<<<
 *             OGR_G_AddGeometryDirectly(geom, ring)
 *         return geom
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildLinearRing(__pyx_v_self, __pyx_v_r); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 504, __pyx_L1_error)
    __pyx_v_ring = __pyx_t_4;

    /* "rasterio/_features.pyx":505
 *         for r in coordinates:
 *             ring = self._buildLinearRing(r)
 *             OGR_G_AddGeometryDirectly(geom, ring)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    OGR_G_AddGeometryDirectly(__pyx_v_geom, __pyx_v_ring);

    /* "rasterio/_features.pyx":503
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['Polygon'])
 *         for r in coordinates:             # <<<<<<<<<<<<<<
 *             ring = self._buildLinearRing(r)
 *             OGR_G_AddGeometryDirectly(geom, ring)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":506
 *             ring = self._buildLinearRing(r)
 *             OGR_G_AddGeometryDirectly(geom, ring)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildMultiPoint(self, object coordinates) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":499
 *         return geom
 * 
 *     cdef OGRGeometryH _buildPolygon(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH ring = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":508
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiPoint(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPoint(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_part;
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_coordinate = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildMultiPoint", 0);

  /* "rasterio/_features.pyx":509
 * 
 *     cdef OGRGeometryH _buildMultiPoint(self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])
 */
  __pyx_v_part = NULL;

  /* "rasterio/_features.pyx":511
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])             # <<<<<<<<<<<<<<
 *         for coordinate in coordinates:
 *             part = self._buildPoint(coordinate)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_MultiPoint); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 511, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":510
 *     cdef OGRGeometryH _buildMultiPoint(self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])
 *         for coordinate in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 510, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":512
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildPoint(coordinate)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 512, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 512, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 512, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 512, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 512, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 512, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_coordinate, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":513
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])
 *         for coordinate in coordinates:
 *             part = self._buildPoint(coordinate)             # <<<<<<<<<<<<<<
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildPoint(__pyx_v_self, __pyx_v_coordinate); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 513, __pyx_L1_error)
    __pyx_v_part = __pyx_t_4;

    /* "rasterio/_features.pyx":514
 *         for coordinate in coordinates:
 *             part = self._buildPoint(coordinate)
 *             OGR_G_AddGeometryDirectly(geom, part)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    OGR_G_AddGeometryDirectly(__pyx_v_geom, __pyx_v_part);

    /* "rasterio/_features.pyx":512
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPoint'])
 *         for coordinate in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildPoint(coordinate)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":515
 *             part = self._buildPoint(coordinate)
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildMultiLineString(
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":508
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiPoint(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildMultiPoint", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_coordinate);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":517
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiLineString(             # <<<<<<<<<<<<<<
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiLineString(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_part;
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_line = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildMultiLineString", 0);

  /* "rasterio/_features.pyx":519
 *     cdef OGRGeometryH _buildMultiLineString(
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])
 */
  __pyx_v_part = NULL;

  /* "rasterio/_features.pyx":521
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])             # <<<<<<<<<<<<<<
 *         for line in coordinates:
 *             part = self._buildLineString(line)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_MultiLineString); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 521, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":520
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])
 *         for line in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 520, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":522
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])
 *         for line in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildLineString(line)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 522, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 522, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 522, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 522, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 522, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_line, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":523
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])
 *         for line in coordinates:
 *             part = self._buildLineString(line)             # <<<<<<<<<<<<<<
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildLineString(__pyx_v_self, __pyx_v_line); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 523, __pyx_L1_error)
    __pyx_v_part = __pyx_t_4;

    /* "rasterio/_features.pyx":524
 *         for line in coordinates:
 *             part = self._buildLineString(line)
 *             OGR_G_AddGeometryDirectly(geom, part)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    OGR_G_AddGeometryDirectly(__pyx_v_geom, __pyx_v_part);

    /* "rasterio/_features.pyx":522
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiLineString'])
 *         for line in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildLineString(line)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":525
 *             part = self._buildLineString(line)
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildMultiPolygon(self, object coordinates) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":517
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiLineString(             # <<<<<<<<<<<<<<
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildMultiLineString", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_line);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":527
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiPolygon(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPolygon(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_part;
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_poly = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildMultiPolygon", 0);

  /* "rasterio/_features.pyx":528
 * 
 *     cdef OGRGeometryH _buildMultiPolygon(self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])
 */
  __pyx_v_part = NULL;

  /* "rasterio/_features.pyx":530
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])             # <<<<<<<<<<<<<<
 *         for poly in coordinates:
 *             part = self._buildPolygon(poly)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_MultiPolygon); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 530, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":529
 *     cdef OGRGeometryH _buildMultiPolygon(self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])
 *         for poly in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 529, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":531
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])
 *         for poly in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildPolygon(poly)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 531, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 531, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 531, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 531, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 531, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_poly, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":532
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])
 *         for poly in coordinates:
 *             part = self._buildPolygon(poly)             # <<<<<<<<<<<<<<
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom
 */
    __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildPolygon(__pyx_v_self, __pyx_v_poly); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 532, __pyx_L1_error)
    __pyx_v_part = __pyx_t_4;

    /* "rasterio/_features.pyx":533
 *         for poly in coordinates:
 *             part = self._buildPolygon(poly)
 *             OGR_G_AddGeometryDirectly(geom, part)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    OGR_G_AddGeometryDirectly(__pyx_v_geom, __pyx_v_part);

    /* "rasterio/_features.pyx":531
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['MultiPolygon'])
 *         for poly in coordinates:             # <<<<<<<<<<<<<<
 *             part = self._buildPolygon(poly)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":534
 *             part = self._buildPolygon(poly)
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH _buildGeometryCollection(
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":527
 *         return geom
 * 
 *     cdef OGRGeometryH _buildMultiPolygon(self, object coordinates) except NULL:             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildMultiPolygon", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_poly);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":536
 *         return geom
 * 
 *     cdef OGRGeometryH _buildGeometryCollection(             # <<<<<<<<<<<<<<
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildGeometryCollection(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_coordinates) {
  OGRGeometryH __pyx_v_part;
  OGRGeometryH __pyx_v_geom;
  PyObject *__pyx_v_colxn = NULL;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  OGRGeometryH __pyx_t_4;
  Py_ssize_t __pyx_t_5;
  PyObject *(*__pyx_t_6)(PyObject *);
  __Pyx_RefNannySetupContext("_buildGeometryCollection", 0);

  /* "rasterio/_features.pyx":538
 *     cdef OGRGeometryH _buildGeometryCollection(
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])
 */
  __pyx_v_part = NULL;

  /* "rasterio/_features.pyx":540
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])             # <<<<<<<<<<<<<<
 *         for colxn in coordinates:
 *             part = OGRGeomBuilder().build(colxn)
 */
  __pyx_t_1 = __Pyx_GetModuleGlobalName(__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = PyObject_GetItem(__pyx_t_1, __pyx_n_s_GeometryCollection); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_3 = __Pyx_PyInt_As_int(__pyx_t_2); if (unlikely((__pyx_t_3 == (int)-1) && PyErr_Occurred())) __PYX_ERR(0, 540, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":539
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 *         cdef OGRGeometryH geom = self._createOgrGeometry(             # <<<<<<<<<<<<<<
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])
 *         for colxn in coordinates:
 */
  __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_createOgrGeometry(__pyx_v_self, __pyx_t_3); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 539, __pyx_L1_error)
  __pyx_v_geom = __pyx_t_4;

  /* "rasterio/_features.pyx":541
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])
 *         for colxn in coordinates:             # <<<<<<<<<<<<<<
 *             part = OGRGeomBuilder().build(colxn)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  if (likely(PyList_CheckExact(__pyx_v_coordinates)) || PyTuple_CheckExact(__pyx_v_coordinates)) {
    __pyx_t_2 = __pyx_v_coordinates; __Pyx_INCREF(__pyx_t_2); __pyx_t_5 = 0;
    __pyx_t_6 = NULL;
  } else {
    __pyx_t_5 = -1; __pyx_t_2 = PyObject_GetIter(__pyx_v_coordinates); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 541, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_6 = Py_TYPE(__pyx_t_2)->tp_iternext; if (unlikely(!__pyx_t_6)) __PYX_ERR(0, 541, __pyx_L1_error)
  }
  for (;;) {
    if (likely(!__pyx_t_6)) {
      if (likely(PyList_CheckExact(__pyx_t_2))) {
        if (__pyx_t_5 >= PyList_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyList_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 541, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      } else {
        if (__pyx_t_5 >= PyTuple_GET_SIZE(__pyx_t_2)) break;
        #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
        __pyx_t_1 = PyTuple_GET_ITEM(__pyx_t_2, __pyx_t_5); __Pyx_INCREF(__pyx_t_1); __pyx_t_5++; if (unlikely(0 < 0)) __PYX_ERR(0, 541, __pyx_L1_error)
        #else
        __pyx_t_1 = PySequence_ITEM(__pyx_t_2, __pyx_t_5); __pyx_t_5++; if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 541, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_1);
        #endif
      }
    } else {
      __pyx_t_1 = __pyx_t_6(__pyx_t_2);
      if (unlikely(!__pyx_t_1)) {
        PyObject* exc_type = PyErr_Occurred();
        if (exc_type) {
          if (likely(exc_type == PyExc_StopIteration || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration))) PyErr_Clear();
          else __PYX_ERR(0, 541, __pyx_L1_error)
        }
        break;
      }
      __Pyx_GOTREF(__pyx_t_1);
    }
    __Pyx_XDECREF_SET(__pyx_v_colxn, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":542
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])
 *         for colxn in coordinates:
 *             part = OGRGeomBuilder().build(colxn)             # <<<<<<<<<<<<<<
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom
 */
    __pyx_t_1 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_OGRGeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_4 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_1)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)__pyx_t_1), __pyx_v_colxn); if (unlikely(__pyx_t_4 == NULL)) __PYX_ERR(0, 542, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __pyx_v_part = __pyx_t_4;

    /* "rasterio/_features.pyx":543
 *         for colxn in coordinates:
 *             part = OGRGeomBuilder().build(colxn)
 *             OGR_G_AddGeometryDirectly(geom, part)             # <<<<<<<<<<<<<<
 *         return geom
 * 
 */
    OGR_G_AddGeometryDirectly(__pyx_v_geom, __pyx_v_part);

    /* "rasterio/_features.pyx":541
 *         cdef OGRGeometryH geom = self._createOgrGeometry(
 *             GEOJSON2OGR_GEOMETRY_TYPES['GeometryCollection'])
 *         for colxn in coordinates:             # <<<<<<<<<<<<<<
 *             part = OGRGeomBuilder().build(colxn)
 *             OGR_G_AddGeometryDirectly(geom, part)
 */
  }
  __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":544
 *             part = OGRGeomBuilder().build(colxn)
 *             OGR_G_AddGeometryDirectly(geom, part)
 *         return geom             # <<<<<<<<<<<<<<
 * 
 *     cdef OGRGeometryH build(self, object geometry) except NULL:
 */
  __pyx_r = __pyx_v_geom;
  goto __pyx_L0;

  /* "rasterio/_features.pyx":536
 *         return geom
 * 
 *     cdef OGRGeometryH _buildGeometryCollection(             # <<<<<<<<<<<<<<
 *             self, object coordinates) except NULL:
 *         cdef OGRGeometryH part = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder._buildGeometryCollection", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_colxn);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":546
 *         return geom
 * 
 *     cdef OGRGeometryH build(self, object geometry) except NULL:             # <<<<<<<<<<<<<<
 *         """Builds an OGR geometry from GeoJSON geometry."""
 *         cdef object typename = geometry['type']
 */

static OGRGeometryH __pyx_f_8rasterio_9_features_14OGRGeomBuilder_build(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *__pyx_v_self, PyObject *__pyx_v_geometry) {
  PyObject *__pyx_v_typename = 0;
  PyObject *__pyx_v_coordinates = 0;
  OGRGeometryH __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  int __pyx_t_3;
  int __pyx_t_4;
  int __pyx_t_5;
  OGRGeometryH __pyx_t_6;
  __Pyx_RefNannySetupContext("build", 0);

  /* "rasterio/_features.pyx":548
 *     cdef OGRGeometryH build(self, object geometry) except NULL:
 *         """Builds an OGR geometry from GeoJSON geometry."""
 *         cdef object typename = geometry['type']             # <<<<<<<<<<<<<<
 *         cdef object coordinates = geometry.get('coordinates')
 * 
 */
  __pyx_t_1 = PyObject_GetItem(__pyx_v_geometry, __pyx_n_s_type); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 548, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_v_typename = __pyx_t_1;
  __pyx_t_1 = 0;

  /* "rasterio/_features.pyx":549
 *         """Builds an OGR geometry from GeoJSON geometry."""
 *         cdef object typename = geometry['type']
 *         cdef object coordinates = geometry.get('coordinates')             # <<<<<<<<<<<<<<
 * 
 *         if not typename or not coordinates:
 */
  __pyx_t_1 = __Pyx_PyObject_GetAttrStr(__pyx_v_geometry, __pyx_n_s_get); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_t_2 = __Pyx_PyObject_Call(__pyx_t_1, __pyx_tuple__29, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_2);
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_v_coordinates = __pyx_t_2;
  __pyx_t_2 = 0;

  /* "rasterio/_features.pyx":551
 *         cdef object coordinates = geometry.get('coordinates')
 * 
 *         if not typename or not coordinates:             # <<<<<<<<<<<<<<
 *             raise ValueError("Input is not a valid geometry object")
 *         if typename == 'Point':
 */
  __pyx_t_4 = __Pyx_PyObject_IsTrue(__pyx_v_typename); if (unlikely(__pyx_t_4 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
  __pyx_t_5 = ((!__pyx_t_4) != 0);
  if (!__pyx_t_5) {
  } else {
    __pyx_t_3 = __pyx_t_5;
    goto __pyx_L4_bool_binop_done;
  }
  __pyx_t_5 = __Pyx_PyObject_IsTrue(__pyx_v_coordinates); if (unlikely(__pyx_t_5 < 0)) __PYX_ERR(0, 551, __pyx_L1_error)
  __pyx_t_4 = ((!__pyx_t_5) != 0);
  __pyx_t_3 = __pyx_t_4;
  __pyx_L4_bool_binop_done:;
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":552
 * 
 *         if not typename or not coordinates:
 *             raise ValueError("Input is not a valid geometry object")             # <<<<<<<<<<<<<<
 *         if typename == 'Point':
 *             return self._buildPoint(coordinates)
 */
    __pyx_t_2 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__30, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 552, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_Raise(__pyx_t_2, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __PYX_ERR(0, 552, __pyx_L1_error)

    /* "rasterio/_features.pyx":551
 *         cdef object coordinates = geometry.get('coordinates')
 * 
 *         if not typename or not coordinates:             # <<<<<<<<<<<<<<
 *             raise ValueError("Input is not a valid geometry object")
 *         if typename == 'Point':
 */
  }

  /* "rasterio/_features.pyx":553
 *         if not typename or not coordinates:
 *             raise ValueError("Input is not a valid geometry object")
 *         if typename == 'Point':             # <<<<<<<<<<<<<<
 *             return self._buildPoint(coordinates)
 *         elif typename == 'LineString':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_Point, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 553, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":554
 *             raise ValueError("Input is not a valid geometry object")
 *         if typename == 'Point':
 *             return self._buildPoint(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'LineString':
 *             return self._buildLineString(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildPoint(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 554, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":553
 *         if not typename or not coordinates:
 *             raise ValueError("Input is not a valid geometry object")
 *         if typename == 'Point':             # <<<<<<<<<<<<<<
 *             return self._buildPoint(coordinates)
 *         elif typename == 'LineString':
 */
  }

  /* "rasterio/_features.pyx":555
 *         if typename == 'Point':
 *             return self._buildPoint(coordinates)
 *         elif typename == 'LineString':             # <<<<<<<<<<<<<<
 *             return self._buildLineString(coordinates)
 *         elif typename == 'LinearRing':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_LineString, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 555, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":556
 *             return self._buildPoint(coordinates)
 *         elif typename == 'LineString':
 *             return self._buildLineString(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'LinearRing':
 *             return self._buildLinearRing(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildLineString(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 556, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":555
 *         if typename == 'Point':
 *             return self._buildPoint(coordinates)
 *         elif typename == 'LineString':             # <<<<<<<<<<<<<<
 *             return self._buildLineString(coordinates)
 *         elif typename == 'LinearRing':
 */
  }

  /* "rasterio/_features.pyx":557
 *         elif typename == 'LineString':
 *             return self._buildLineString(coordinates)
 *         elif typename == 'LinearRing':             # <<<<<<<<<<<<<<
 *             return self._buildLinearRing(coordinates)
 *         elif typename == 'Polygon':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_LinearRing, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 557, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":558
 *             return self._buildLineString(coordinates)
 *         elif typename == 'LinearRing':
 *             return self._buildLinearRing(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'Polygon':
 *             return self._buildPolygon(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildLinearRing(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 558, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":557
 *         elif typename == 'LineString':
 *             return self._buildLineString(coordinates)
 *         elif typename == 'LinearRing':             # <<<<<<<<<<<<<<
 *             return self._buildLinearRing(coordinates)
 *         elif typename == 'Polygon':
 */
  }

  /* "rasterio/_features.pyx":559
 *         elif typename == 'LinearRing':
 *             return self._buildLinearRing(coordinates)
 *         elif typename == 'Polygon':             # <<<<<<<<<<<<<<
 *             return self._buildPolygon(coordinates)
 *         elif typename == 'MultiPoint':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_Polygon, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 559, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":560
 *             return self._buildLinearRing(coordinates)
 *         elif typename == 'Polygon':
 *             return self._buildPolygon(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'MultiPoint':
 *             return self._buildMultiPoint(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildPolygon(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 560, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":559
 *         elif typename == 'LinearRing':
 *             return self._buildLinearRing(coordinates)
 *         elif typename == 'Polygon':             # <<<<<<<<<<<<<<
 *             return self._buildPolygon(coordinates)
 *         elif typename == 'MultiPoint':
 */
  }

  /* "rasterio/_features.pyx":561
 *         elif typename == 'Polygon':
 *             return self._buildPolygon(coordinates)
 *         elif typename == 'MultiPoint':             # <<<<<<<<<<<<<<
 *             return self._buildMultiPoint(coordinates)
 *         elif typename == 'MultiLineString':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_MultiPoint, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 561, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":562
 *             return self._buildPolygon(coordinates)
 *         elif typename == 'MultiPoint':
 *             return self._buildMultiPoint(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'MultiLineString':
 *             return self._buildMultiLineString(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildMultiPoint(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 562, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":561
 *         elif typename == 'Polygon':
 *             return self._buildPolygon(coordinates)
 *         elif typename == 'MultiPoint':             # <<<<<<<<<<<<<<
 *             return self._buildMultiPoint(coordinates)
 *         elif typename == 'MultiLineString':
 */
  }

  /* "rasterio/_features.pyx":563
 *         elif typename == 'MultiPoint':
 *             return self._buildMultiPoint(coordinates)
 *         elif typename == 'MultiLineString':             # <<<<<<<<<<<<<<
 *             return self._buildMultiLineString(coordinates)
 *         elif typename == 'MultiPolygon':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_MultiLineString, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 563, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":564
 *             return self._buildMultiPoint(coordinates)
 *         elif typename == 'MultiLineString':
 *             return self._buildMultiLineString(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'MultiPolygon':
 *             return self._buildMultiPolygon(coordinates)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildMultiLineString(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 564, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":563
 *         elif typename == 'MultiPoint':
 *             return self._buildMultiPoint(coordinates)
 *         elif typename == 'MultiLineString':             # <<<<<<<<<<<<<<
 *             return self._buildMultiLineString(coordinates)
 *         elif typename == 'MultiPolygon':
 */
  }

  /* "rasterio/_features.pyx":565
 *         elif typename == 'MultiLineString':
 *             return self._buildMultiLineString(coordinates)
 *         elif typename == 'MultiPolygon':             # <<<<<<<<<<<<<<
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_MultiPolygon, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 565, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":566
 *             return self._buildMultiLineString(coordinates)
 *         elif typename == 'MultiPolygon':
 *             return self._buildMultiPolygon(coordinates)             # <<<<<<<<<<<<<<
 *         elif typename == 'GeometryCollection':
 *             coordinates = geometry.get('geometries')
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildMultiPolygon(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 566, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":565
 *         elif typename == 'MultiLineString':
 *             return self._buildMultiLineString(coordinates)
 *         elif typename == 'MultiPolygon':             # <<<<<<<<<<<<<<
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':
 */
  }

  /* "rasterio/_features.pyx":567
 *         elif typename == 'MultiPolygon':
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':             # <<<<<<<<<<<<<<
 *             coordinates = geometry.get('geometries')
 *             return self._buildGeometryCollection(coordinates)
 */
  __pyx_t_3 = (__Pyx_PyString_Equals(__pyx_v_typename, __pyx_n_s_GeometryCollection, Py_EQ)); if (unlikely(__pyx_t_3 < 0)) __PYX_ERR(0, 567, __pyx_L1_error)
  if (__pyx_t_3) {

    /* "rasterio/_features.pyx":568
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':
 *             coordinates = geometry.get('geometries')             # <<<<<<<<<<<<<<
 *             return self._buildGeometryCollection(coordinates)
 *         else:
 */
    __pyx_t_2 = __Pyx_PyObject_GetAttrStr(__pyx_v_geometry, __pyx_n_s_get); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_t_2, __pyx_tuple__31, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 568, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_DECREF_SET(__pyx_v_coordinates, __pyx_t_1);
    __pyx_t_1 = 0;

    /* "rasterio/_features.pyx":569
 *         elif typename == 'GeometryCollection':
 *             coordinates = geometry.get('geometries')
 *             return self._buildGeometryCollection(coordinates)             # <<<<<<<<<<<<<<
 *         else:
 *             raise ValueError("Unsupported geometry type %s" % typename)
 */
    __pyx_t_6 = ((struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder *)__pyx_v_self->__pyx_vtab)->_buildGeometryCollection(__pyx_v_self, __pyx_v_coordinates); if (unlikely(__pyx_t_6 == NULL)) __PYX_ERR(0, 569, __pyx_L1_error)
    __pyx_r = __pyx_t_6;
    goto __pyx_L0;

    /* "rasterio/_features.pyx":567
 *         elif typename == 'MultiPolygon':
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':             # <<<<<<<<<<<<<<
 *             coordinates = geometry.get('geometries')
 *             return self._buildGeometryCollection(coordinates)
 */
  }

  /* "rasterio/_features.pyx":571
 *             return self._buildGeometryCollection(coordinates)
 *         else:
 *             raise ValueError("Unsupported geometry type %s" % typename)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __pyx_t_1 = __Pyx_PyString_Format(__pyx_kp_s_Unsupported_geometry_type_s, __pyx_v_typename); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __pyx_t_2 = PyTuple_New(1); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_GIVEREF(__pyx_t_1);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_t_1);
    __pyx_t_1 = 0;
    __pyx_t_1 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_2, NULL); if (unlikely(!__pyx_t_1)) __PYX_ERR(0, 571, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_1);
    __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_Raise(__pyx_t_1, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;
    __PYX_ERR(0, 571, __pyx_L1_error)
  }

  /* "rasterio/_features.pyx":546
 *         return geom
 * 
 *     cdef OGRGeometryH build(self, object geometry) except NULL:             # <<<<<<<<<<<<<<
 *         """Builds an OGR geometry from GeoJSON geometry."""
 *         cdef object typename = geometry['type']
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_AddTraceback("rasterio._features.OGRGeomBuilder.build", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_typename);
  __Pyx_XDECREF(__pyx_v_coordinates);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":576
 * # Feature extension classes and functions follow.
 * 
 * cdef _deleteOgrFeature(OGRFeatureH feat):             # <<<<<<<<<<<<<<
 *     """Delete an OGR feature"""
 *     if feat != NULL:
 */

static PyObject *__pyx_f_8rasterio_9_features__deleteOgrFeature(OGRFeatureH __pyx_v_feat) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("_deleteOgrFeature", 0);

  /* "rasterio/_features.pyx":578
 * cdef _deleteOgrFeature(OGRFeatureH feat):
 *     """Delete an OGR feature"""
 *     if feat != NULL:             # <<<<<<<<<<<<<<
 *         OGR_F_Destroy(feat)
 *     feat = NULL
 */
  __pyx_t_1 = ((__pyx_v_feat != NULL) != 0);
  if (__pyx_t_1) {

    /* "rasterio/_features.pyx":579
 *     """Delete an OGR feature"""
 *     if feat != NULL:
 *         OGR_F_Destroy(feat)             # <<<<<<<<<<<<<<
 *     feat = NULL
 * 
 */
    OGR_F_Destroy(__pyx_v_feat);

    /* "rasterio/_features.pyx":578
 * cdef _deleteOgrFeature(OGRFeatureH feat):
 *     """Delete an OGR feature"""
 *     if feat != NULL:             # <<<<<<<<<<<<<<
 *         OGR_F_Destroy(feat)
 *     feat = NULL
 */
  }

  /* "rasterio/_features.pyx":580
 *     if feat != NULL:
 *         OGR_F_Destroy(feat)
 *     feat = NULL             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_v_feat = NULL;

  /* "rasterio/_features.pyx":576
 * # Feature extension classes and functions follow.
 * 
 * cdef _deleteOgrFeature(OGRFeatureH feat):             # <<<<<<<<<<<<<<
 *     """Delete an OGR feature"""
 *     if feat != NULL:
 */

  /* function exit code */
  __pyx_r = Py_None; __Pyx_INCREF(Py_None);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":586
 *     """Provides an iterator over shapes in an OGR feature layer."""
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         OGR_L_ResetReading(self.layer)
 *         return self
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_13ShapeIterator_1__iter__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_13ShapeIterator_1__iter__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_13ShapeIterator___iter__(((struct __pyx_obj_8rasterio_9_features_ShapeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_13ShapeIterator___iter__(struct __pyx_obj_8rasterio_9_features_ShapeIterator *__pyx_v_self) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__iter__", 0);

  /* "rasterio/_features.pyx":587
 * 
 *     def __iter__(self):
 *         OGR_L_ResetReading(self.layer)             # <<<<<<<<<<<<<<
 *         return self
 * 
 */
  OGR_L_ResetReading(__pyx_v_self->layer);

  /* "rasterio/_features.pyx":588
 *     def __iter__(self):
 *         OGR_L_ResetReading(self.layer)
 *         return self             # <<<<<<<<<<<<<<
 * 
 *     def __next__(self):
 */
  __Pyx_XDECREF(__pyx_r);
  __Pyx_INCREF(((PyObject *)__pyx_v_self));
  __pyx_r = ((PyObject *)__pyx_v_self);
  goto __pyx_L0;

  /* "rasterio/_features.pyx":586
 *     """Provides an iterator over shapes in an OGR feature layer."""
 * 
 *     def __iter__(self):             # <<<<<<<<<<<<<<
 *         OGR_L_ResetReading(self.layer)
 *         return self
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "rasterio/_features.pyx":590
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef OGRFeatureH feat = NULL
 *         cdef OGRGeometryH geom = NULL
 */

/* Python wrapper */
static PyObject *__pyx_pw_8rasterio_9_features_13ShapeIterator_3__next__(PyObject *__pyx_v_self); /*proto*/
static PyObject *__pyx_pw_8rasterio_9_features_13ShapeIterator_3__next__(PyObject *__pyx_v_self) {
  PyObject *__pyx_r = 0;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__next__ (wrapper)", 0);
  __pyx_r = __pyx_pf_8rasterio_9_features_13ShapeIterator_2__next__(((struct __pyx_obj_8rasterio_9_features_ShapeIterator *)__pyx_v_self));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static PyObject *__pyx_pf_8rasterio_9_features_13ShapeIterator_2__next__(struct __pyx_obj_8rasterio_9_features_ShapeIterator *__pyx_v_self) {
  OGRFeatureH __pyx_v_feat;
  OGRGeometryH __pyx_v_geom;
  double __pyx_v_image_value;
  PyObject *__pyx_v_shape = NULL;
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  char const *__pyx_t_6;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  PyObject *__pyx_t_9 = NULL;
  PyObject *__pyx_t_10 = NULL;
  PyObject *__pyx_t_11 = NULL;
  PyObject *__pyx_t_12 = NULL;
  __Pyx_RefNannySetupContext("__next__", 0);

  /* "rasterio/_features.pyx":591
 * 
 *     def __next__(self):
 *         cdef OGRFeatureH feat = NULL             # <<<<<<<<<<<<<<
 *         cdef OGRGeometryH geom = NULL
 * 
 */
  __pyx_v_feat = NULL;

  /* "rasterio/_features.pyx":592
 *     def __next__(self):
 *         cdef OGRFeatureH feat = NULL
 *         cdef OGRGeometryH geom = NULL             # <<<<<<<<<<<<<<
 * 
 *         try:
 */
  __pyx_v_geom = NULL;

  /* "rasterio/_features.pyx":594
 *         cdef OGRGeometryH geom = NULL
 * 
 *         try:             # <<<<<<<<<<<<<<
 *             feat = OGR_L_GetNextFeature(self.layer)
 * 
 */
  /*try:*/ {

    /* "rasterio/_features.pyx":595
 * 
 *         try:
 *             feat = OGR_L_GetNextFeature(self.layer)             # <<<<<<<<<<<<<<
 * 
 *             if feat == NULL:
 */
    __pyx_v_feat = OGR_L_GetNextFeature(__pyx_v_self->layer);

    /* "rasterio/_features.pyx":597
 *             feat = OGR_L_GetNextFeature(self.layer)
 * 
 *             if feat == NULL:             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 * 
 */
    __pyx_t_1 = ((__pyx_v_feat == NULL) != 0);
    if (__pyx_t_1) {

      /* "rasterio/_features.pyx":598
 * 
 *             if feat == NULL:
 *                 raise StopIteration             # <<<<<<<<<<<<<<
 * 
 *             if self.fieldtype == 0:
 */
      __Pyx_Raise(__pyx_builtin_StopIteration, 0, 0, 0);
      __PYX_ERR(0, 598, __pyx_L4_error)

      /* "rasterio/_features.pyx":597
 *             feat = OGR_L_GetNextFeature(self.layer)
 * 
 *             if feat == NULL:             # <<<<<<<<<<<<<<
 *                 raise StopIteration
 * 
 */
    }

    /* "rasterio/_features.pyx":600
 *                 raise StopIteration
 * 
 *             if self.fieldtype == 0:             # <<<<<<<<<<<<<<
 *                 image_value = OGR_F_GetFieldAsInteger(feat, 0)
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_self->fieldtype == 0) != 0);
    if (__pyx_t_1) {

      /* "rasterio/_features.pyx":601
 * 
 *             if self.fieldtype == 0:
 *                 image_value = OGR_F_GetFieldAsInteger(feat, 0)             # <<<<<<<<<<<<<<
 *             else:
 *                 image_value = OGR_F_GetFieldAsDouble(feat, 0)
 */
      __pyx_v_image_value = OGR_F_GetFieldAsInteger(__pyx_v_feat, 0);

      /* "rasterio/_features.pyx":600
 *                 raise StopIteration
 * 
 *             if self.fieldtype == 0:             # <<<<<<<<<<<<<<
 *                 image_value = OGR_F_GetFieldAsInteger(feat, 0)
 *             else:
 */
      goto __pyx_L7;
    }

    /* "rasterio/_features.pyx":603
 *                 image_value = OGR_F_GetFieldAsInteger(feat, 0)
 *             else:
 *                 image_value = OGR_F_GetFieldAsDouble(feat, 0)             # <<<<<<<<<<<<<<
 *             geom = OGR_F_GetGeometryRef(feat)
 *             if geom != NULL:
 */
    /*else*/ {
      __pyx_v_image_value = OGR_F_GetFieldAsDouble(__pyx_v_feat, 0);
    }
    __pyx_L7:;

    /* "rasterio/_features.pyx":604
 *             else:
 *                 image_value = OGR_F_GetFieldAsDouble(feat, 0)
 *             geom = OGR_F_GetGeometryRef(feat)             # <<<<<<<<<<<<<<
 *             if geom != NULL:
 *                 shape = GeomBuilder().build(geom)
 */
    __pyx_v_geom = OGR_F_GetGeometryRef(__pyx_v_feat);

    /* "rasterio/_features.pyx":605
 *                 image_value = OGR_F_GetFieldAsDouble(feat, 0)
 *             geom = OGR_F_GetGeometryRef(feat)
 *             if geom != NULL:             # <<<<<<<<<<<<<<
 *                 shape = GeomBuilder().build(geom)
 *             else:
 */
    __pyx_t_1 = ((__pyx_v_geom != NULL) != 0);
    if (__pyx_t_1) {

      /* "rasterio/_features.pyx":606
 *             geom = OGR_F_GetGeometryRef(feat)
 *             if geom != NULL:
 *                 shape = GeomBuilder().build(geom)             # <<<<<<<<<<<<<<
 *             else:
 *                 shape = None
 */
      __pyx_t_2 = __Pyx_PyObject_Call(((PyObject *)__pyx_ptype_8rasterio_9_features_GeomBuilder), __pyx_empty_tuple, NULL); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 606, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_2);
      __pyx_t_3 = ((struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder *)((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_2)->__pyx_vtab)->build(((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)__pyx_t_2), __pyx_v_geom); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 606, __pyx_L4_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_v_shape = __pyx_t_3;
      __pyx_t_3 = 0;

      /* "rasterio/_features.pyx":605
 *                 image_value = OGR_F_GetFieldAsDouble(feat, 0)
 *             geom = OGR_F_GetGeometryRef(feat)
 *             if geom != NULL:             # <<<<<<<<<<<<<<
 *                 shape = GeomBuilder().build(geom)
 *             else:
 */
      goto __pyx_L8;
    }

    /* "rasterio/_features.pyx":608
 *                 shape = GeomBuilder().build(geom)
 *             else:
 *                 shape = None             # <<<<<<<<<<<<<<
 *             return shape, image_value
 * 
 */
    /*else*/ {
      __Pyx_INCREF(Py_None);
      __pyx_v_shape = Py_None;
    }
    __pyx_L8:;

    /* "rasterio/_features.pyx":609
 *             else:
 *                 shape = None
 *             return shape, image_value             # <<<<<<<<<<<<<<
 * 
 *         finally:
 */
    __Pyx_XDECREF(__pyx_r);
    __pyx_t_3 = PyFloat_FromDouble(__pyx_v_image_value); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 609, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_3);
    __pyx_t_2 = PyTuple_New(2); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 609, __pyx_L4_error)
    __Pyx_GOTREF(__pyx_t_2);
    __Pyx_INCREF(__pyx_v_shape);
    __Pyx_GIVEREF(__pyx_v_shape);
    PyTuple_SET_ITEM(__pyx_t_2, 0, __pyx_v_shape);
    __Pyx_GIVEREF(__pyx_t_3);
    PyTuple_SET_ITEM(__pyx_t_2, 1, __pyx_t_3);
    __pyx_t_3 = 0;
    __pyx_r = __pyx_t_2;
    __pyx_t_2 = 0;
    goto __pyx_L3_return;
  }

  /* "rasterio/_features.pyx":612
 * 
 *         finally:
 *             _deleteOgrFeature(feat)             # <<<<<<<<<<<<<<
 */
  /*finally:*/ {
    /*exception exit:*/{
      __Pyx_PyThreadState_declare
      __pyx_L4_error:;
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __Pyx_PyThreadState_assign
      __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
      if (PY_MAJOR_VERSION >= 3) __Pyx_ExceptionSwap(&__pyx_t_10, &__pyx_t_11, &__pyx_t_12);
      if ((PY_MAJOR_VERSION < 3) || unlikely(__Pyx_GetException(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9) < 0)) __Pyx_ErrFetch(&__pyx_t_7, &__pyx_t_8, &__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_7);
      __Pyx_XGOTREF(__pyx_t_8);
      __Pyx_XGOTREF(__pyx_t_9);
      __Pyx_XGOTREF(__pyx_t_10);
      __Pyx_XGOTREF(__pyx_t_11);
      __Pyx_XGOTREF(__pyx_t_12);
      __pyx_t_4 = __pyx_lineno; __pyx_t_5 = __pyx_clineno; __pyx_t_6 = __pyx_filename;
      {
        __pyx_t_2 = __pyx_f_8rasterio_9_features__deleteOgrFeature(__pyx_v_feat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L10_error)
        __Pyx_GOTREF(__pyx_t_2);
        __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      }
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XGIVEREF(__pyx_t_7);
      __Pyx_XGIVEREF(__pyx_t_8);
      __Pyx_XGIVEREF(__pyx_t_9);
      __Pyx_ErrRestore(__pyx_t_7, __pyx_t_8, __pyx_t_9);
      __pyx_t_7 = 0; __pyx_t_8 = 0; __pyx_t_9 = 0; __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      __pyx_lineno = __pyx_t_4; __pyx_clineno = __pyx_t_5; __pyx_filename = __pyx_t_6;
      goto __pyx_L1_error;
      __pyx_L10_error:;
      __Pyx_PyThreadState_assign
      if (PY_MAJOR_VERSION >= 3) {
        __Pyx_XGIVEREF(__pyx_t_10);
        __Pyx_XGIVEREF(__pyx_t_11);
        __Pyx_XGIVEREF(__pyx_t_12);
        __Pyx_ExceptionReset(__pyx_t_10, __pyx_t_11, __pyx_t_12);
      }
      __Pyx_XDECREF(__pyx_t_7); __pyx_t_7 = 0;
      __Pyx_XDECREF(__pyx_t_8); __pyx_t_8 = 0;
      __Pyx_XDECREF(__pyx_t_9); __pyx_t_9 = 0;
      __pyx_t_10 = 0; __pyx_t_11 = 0; __pyx_t_12 = 0;
      goto __pyx_L1_error;
    }
    __pyx_L3_return: {
      __pyx_t_12 = __pyx_r;
      __pyx_r = 0;
      __pyx_t_2 = __pyx_f_8rasterio_9_features__deleteOgrFeature(__pyx_v_feat); if (unlikely(!__pyx_t_2)) __PYX_ERR(0, 612, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_2);
      __Pyx_DECREF(__pyx_t_2); __pyx_t_2 = 0;
      __pyx_r = __pyx_t_12;
      __pyx_t_12 = 0;
      goto __pyx_L0;
    }
  }

  /* "rasterio/_features.pyx":590
 *         return self
 * 
 *     def __next__(self):             # <<<<<<<<<<<<<<
 *         cdef OGRFeatureH feat = NULL
 *         cdef OGRGeometryH geom = NULL
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_AddTraceback("rasterio._features.ShapeIterator.__next__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF(__pyx_v_shape);
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

/* Python wrapper */
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags); /*proto*/
static CYTHON_UNUSED int __pyx_pw_5numpy_7ndarray_1__getbuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__getbuffer__ (wrapper)", 0);
  __pyx_r = __pyx_pf_5numpy_7ndarray___getbuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info), ((int)__pyx_v_flags));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

static int __pyx_pf_5numpy_7ndarray___getbuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info, int __pyx_v_flags) {
  int __pyx_v_copy_shape;
  int __pyx_v_i;
  int __pyx_v_ndim;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  int __pyx_v_t;
  char *__pyx_v_f;
  PyArray_Descr *__pyx_v_descr = 0;
  int __pyx_v_offset;
  int __pyx_v_hasfields;
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  int __pyx_t_5;
  PyObject *__pyx_t_6 = NULL;
  char *__pyx_t_7;
  __Pyx_RefNannySetupContext("__getbuffer__", 0);
  if (__pyx_v_info != NULL) {
    __pyx_v_info->obj = Py_None; __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(__pyx_v_info->obj);
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":203
 *             # of flags
 * 
 *             if info == NULL: return             # <<<<<<<<<<<<<<
 * 
 *             cdef int copy_shape, i, ndim
 */
  __pyx_t_1 = ((__pyx_v_info == NULL) != 0);
  if (__pyx_t_1) {
    __pyx_r = 0;
    goto __pyx_L0;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":206
 * 
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":207
 *             cdef int copy_shape, i, ndim
 *             cdef int endian_detector = 1
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 * 
 *             ndim = PyArray_NDIM(self)
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":209
 *             cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 * 
 *             ndim = PyArray_NDIM(self)             # <<<<<<<<<<<<<<
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_v_ndim = PyArray_NDIM(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":212
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 copy_shape = 1             # <<<<<<<<<<<<<<
 *             else:
 *                 copy_shape = 0
 */
    __pyx_v_copy_shape = 1;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":211
 *             ndim = PyArray_NDIM(self)
 * 
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 copy_shape = 1
 *             else:
 */
    goto __pyx_L4;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":214
 *                 copy_shape = 1
 *             else:
 *                 copy_shape = 0             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 */
  /*else*/ {
    __pyx_v_copy_shape = 0;
  }
  __pyx_L4:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_C_CONTIGUOUS) == PyBUF_C_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L6_bool_binop_done;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":217
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_C_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L6_bool_binop_done:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__32, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 218, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 218, __pyx_L1_error)

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":216
 *                 copy_shape = 0
 * 
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  __pyx_t_2 = (((__pyx_v_flags & PyBUF_F_CONTIGUOUS) == PyBUF_F_CONTIGUOUS) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L9_bool_binop_done;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":221
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):             # <<<<<<<<<<<<<<
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 */
  __pyx_t_2 = ((!(PyArray_CHKFLAGS(__pyx_v_self, NPY_F_CONTIGUOUS) != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L9_bool_binop_done:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
    __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__33, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 222, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_Raise(__pyx_t_3, 0, 0, 0);
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __PYX_ERR(1, 222, __pyx_L1_error)

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":220
 *                 raise ValueError(u"ndarray is not C contiguous")
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)             # <<<<<<<<<<<<<<
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":224
 *                 raise ValueError(u"ndarray is not Fortran contiguous")
 * 
 *             info.buf = PyArray_DATA(self)             # <<<<<<<<<<<<<<
 *             info.ndim = ndim
 *             if copy_shape:
 */
  __pyx_v_info->buf = PyArray_DATA(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":225
 * 
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim             # <<<<<<<<<<<<<<
 *             if copy_shape:
 *                 # Allocate new buffer for strides and shape info.
 */
  __pyx_v_info->ndim = __pyx_v_ndim;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
  __pyx_t_1 = (__pyx_v_copy_shape != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":229
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)             # <<<<<<<<<<<<<<
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 */
    __pyx_v_info->strides = ((Py_ssize_t *)malloc((((sizeof(Py_ssize_t)) * ((size_t)__pyx_v_ndim)) * 2)));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":230
 *                 # This is allocated as one block, strides first.
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim             # <<<<<<<<<<<<<<
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 */
    __pyx_v_info->shape = (__pyx_v_info->strides + __pyx_v_ndim);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":231
 *                 info.strides = <Py_ssize_t*>stdlib.malloc(sizeof(Py_ssize_t) * <size_t>ndim * 2)
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):             # <<<<<<<<<<<<<<
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 */
    __pyx_t_4 = __pyx_v_ndim;
    for (__pyx_t_5 = 0; __pyx_t_5 < __pyx_t_4; __pyx_t_5+=1) {
      __pyx_v_i = __pyx_t_5;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":232
 *                 info.shape = info.strides + ndim
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]             # <<<<<<<<<<<<<<
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 */
      (__pyx_v_info->strides[__pyx_v_i]) = (PyArray_STRIDES(__pyx_v_self)[__pyx_v_i]);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":233
 *                 for i in range(ndim):
 *                     info.strides[i] = PyArray_STRIDES(self)[i]
 *                     info.shape[i] = PyArray_DIMS(self)[i]             # <<<<<<<<<<<<<<
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 */
      (__pyx_v_info->shape[__pyx_v_i]) = (PyArray_DIMS(__pyx_v_self)[__pyx_v_i]);
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":226
 *             info.buf = PyArray_DATA(self)
 *             info.ndim = ndim
 *             if copy_shape:             # <<<<<<<<<<<<<<
 *                 # Allocate new buffer for strides and shape info.
 *                 # This is allocated as one block, strides first.
 */
    goto __pyx_L11;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":235
 *                     info.shape[i] = PyArray_DIMS(self)[i]
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)             # <<<<<<<<<<<<<<
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 */
  /*else*/ {
    __pyx_v_info->strides = ((Py_ssize_t *)PyArray_STRIDES(__pyx_v_self));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":236
 *             else:
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)             # <<<<<<<<<<<<<<
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 */
    __pyx_v_info->shape = ((Py_ssize_t *)PyArray_DIMS(__pyx_v_self));
  }
  __pyx_L11:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":237
 *                 info.strides = <Py_ssize_t*>PyArray_STRIDES(self)
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL             # <<<<<<<<<<<<<<
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 */
  __pyx_v_info->suboffsets = NULL;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":238
 *                 info.shape = <Py_ssize_t*>PyArray_DIMS(self)
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)             # <<<<<<<<<<<<<<
 *             info.readonly = not PyArray_ISWRITEABLE(self)
 * 
 */
  __pyx_v_info->itemsize = PyArray_ITEMSIZE(__pyx_v_self);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":239
 *             info.suboffsets = NULL
 *             info.itemsize = PyArray_ITEMSIZE(self)
 *             info.readonly = not PyArray_ISWRITEABLE(self)             # <<<<<<<<<<<<<<
 * 
 *             cdef int t
 */
  __pyx_v_info->readonly = (!(PyArray_ISWRITEABLE(__pyx_v_self) != 0));

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":242
 * 
 *             cdef int t
 *             cdef char* f = NULL             # <<<<<<<<<<<<<<
 *             cdef dtype descr = self.descr
 *             cdef int offset
 */
  __pyx_v_f = NULL;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":243
 *             cdef int t
 *             cdef char* f = NULL
 *             cdef dtype descr = self.descr             # <<<<<<<<<<<<<<
 *             cdef int offset
 * 
 */
  __pyx_t_3 = ((PyObject *)__pyx_v_self->descr);
  __Pyx_INCREF(__pyx_t_3);
  __pyx_v_descr = ((PyArray_Descr *)__pyx_t_3);
  __pyx_t_3 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":246
 *             cdef int offset
 * 
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields and not copy_shape:
 */
  __pyx_v_hasfields = PyDataType_HASFIELDS(__pyx_v_descr);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
  __pyx_t_2 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_2) {
  } else {
    __pyx_t_1 = __pyx_t_2;
    goto __pyx_L15_bool_binop_done;
  }
  __pyx_t_2 = ((!(__pyx_v_copy_shape != 0)) != 0);
  __pyx_t_1 = __pyx_t_2;
  __pyx_L15_bool_binop_done:;
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":250
 *             if not hasfields and not copy_shape:
 *                 # do not call releasebuffer
 *                 info.obj = None             # <<<<<<<<<<<<<<
 *             else:
 *                 # need to call releasebuffer
 */
    __Pyx_INCREF(Py_None);
    __Pyx_GIVEREF(Py_None);
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = Py_None;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":248
 *             cdef bint hasfields = PyDataType_HASFIELDS(descr)
 * 
 *             if not hasfields and not copy_shape:             # <<<<<<<<<<<<<<
 *                 # do not call releasebuffer
 *                 info.obj = None
 */
    goto __pyx_L14;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":253
 *             else:
 *                 # need to call releasebuffer
 *                 info.obj = self             # <<<<<<<<<<<<<<
 * 
 *             if not hasfields:
 */
  /*else*/ {
    __Pyx_INCREF(((PyObject *)__pyx_v_self));
    __Pyx_GIVEREF(((PyObject *)__pyx_v_self));
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj);
    __pyx_v_info->obj = ((PyObject *)__pyx_v_self);
  }
  __pyx_L14:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  __pyx_t_1 = ((!(__pyx_v_hasfields != 0)) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":256
 * 
 *             if not hasfields:
 *                 t = descr.type_num             # <<<<<<<<<<<<<<
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 */
    __pyx_t_4 = __pyx_v_descr->type_num;
    __pyx_v_t = __pyx_t_4;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '>') != 0);
    if (!__pyx_t_2) {
      goto __pyx_L20_next_or;
    } else {
    }
    __pyx_t_2 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_L20_next_or:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":258
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 */
    __pyx_t_2 = ((__pyx_v_descr->byteorder == '<') != 0);
    if (__pyx_t_2) {
    } else {
      __pyx_t_1 = __pyx_t_2;
      goto __pyx_L19_bool_binop_done;
    }
    __pyx_t_2 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_1 = __pyx_t_2;
    __pyx_L19_bool_binop_done:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_1) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__34, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 259, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 259, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":257
 *             if not hasfields:
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":260
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 */
    switch (__pyx_v_t) {
      case NPY_BYTE:
      __pyx_v_f = ((char *)"b");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":261
 *                     raise ValueError(u"Non-native byte order not supported")
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 */
      case NPY_UBYTE:
      __pyx_v_f = ((char *)"B");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":262
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 */
      case NPY_SHORT:
      __pyx_v_f = ((char *)"h");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":263
 *                 elif t == NPY_UBYTE:       f = "B"
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 */
      case NPY_USHORT:
      __pyx_v_f = ((char *)"H");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":264
 *                 elif t == NPY_SHORT:       f = "h"
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 */
      case NPY_INT:
      __pyx_v_f = ((char *)"i");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":265
 *                 elif t == NPY_USHORT:      f = "H"
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 */
      case NPY_UINT:
      __pyx_v_f = ((char *)"I");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":266
 *                 elif t == NPY_INT:         f = "i"
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 */
      case NPY_LONG:
      __pyx_v_f = ((char *)"l");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":267
 *                 elif t == NPY_UINT:        f = "I"
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 */
      case NPY_ULONG:
      __pyx_v_f = ((char *)"L");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":268
 *                 elif t == NPY_LONG:        f = "l"
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 */
      case NPY_LONGLONG:
      __pyx_v_f = ((char *)"q");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":269
 *                 elif t == NPY_ULONG:       f = "L"
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 */
      case NPY_ULONGLONG:
      __pyx_v_f = ((char *)"Q");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":270
 *                 elif t == NPY_LONGLONG:    f = "q"
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 */
      case NPY_FLOAT:
      __pyx_v_f = ((char *)"f");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":271
 *                 elif t == NPY_ULONGLONG:   f = "Q"
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 */
      case NPY_DOUBLE:
      __pyx_v_f = ((char *)"d");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":272
 *                 elif t == NPY_FLOAT:       f = "f"
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 */
      case NPY_LONGDOUBLE:
      __pyx_v_f = ((char *)"g");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":273
 *                 elif t == NPY_DOUBLE:      f = "d"
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 */
      case NPY_CFLOAT:
      __pyx_v_f = ((char *)"Zf");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":274
 *                 elif t == NPY_LONGDOUBLE:  f = "g"
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"
 */
      case NPY_CDOUBLE:
      __pyx_v_f = ((char *)"Zd");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":275
 *                 elif t == NPY_CFLOAT:      f = "Zf"
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"             # <<<<<<<<<<<<<<
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 */
      case NPY_CLONGDOUBLE:
      __pyx_v_f = ((char *)"Zg");
      break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":276
 *                 elif t == NPY_CDOUBLE:     f = "Zd"
 *                 elif t == NPY_CLONGDOUBLE: f = "Zg"
 *                 elif t == NPY_OBJECT:      f = "O"             # <<<<<<<<<<<<<<
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      case NPY_OBJECT:
      __pyx_v_f = ((char *)"O");
      break;
      default:

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":278
 *                 elif t == NPY_OBJECT:      f = "O"
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *                 info.format = f
 *                 return
 */
      __pyx_t_3 = __Pyx_PyInt_From_int(__pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_6 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_t_3); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_3 = PyTuple_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_GIVEREF(__pyx_t_6);
      PyTuple_SET_ITEM(__pyx_t_3, 0, __pyx_t_6);
      __pyx_t_6 = 0;
      __pyx_t_6 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_3, NULL); if (unlikely(!__pyx_t_6)) __PYX_ERR(1, 278, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __Pyx_Raise(__pyx_t_6, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_6); __pyx_t_6 = 0;
      __PYX_ERR(1, 278, __pyx_L1_error)
      break;
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":279
 *                 else:
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f             # <<<<<<<<<<<<<<
 *                 return
 *             else:
 */
    __pyx_v_info->format = __pyx_v_f;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":280
 *                     raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *                 info.format = f
 *                 return             # <<<<<<<<<<<<<<
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 */
    __pyx_r = 0;
    goto __pyx_L0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":255
 *                 info.obj = self
 * 
 *             if not hasfields:             # <<<<<<<<<<<<<<
 *                 t = descr.type_num
 *                 if ((descr.byteorder == c'>' and little_endian) or
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":282
 *                 return
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)             # <<<<<<<<<<<<<<
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 */
  /*else*/ {
    __pyx_v_info->format = ((char *)malloc(0xFF));

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":283
 *             else:
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment             # <<<<<<<<<<<<<<
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,
 */
    (__pyx_v_info->format[0]) = '^';

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":284
 *                 info.format = <char*>stdlib.malloc(_buffer_format_string_len)
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0             # <<<<<<<<<<<<<<
 *                 f = _util_dtypestring(descr, info.format + 1,
 *                                       info.format + _buffer_format_string_len,
 */
    __pyx_v_offset = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":285
 *                 info.format[0] = c'^' # Native data types, manual alignment
 *                 offset = 0
 *                 f = _util_dtypestring(descr, info.format + 1,             # <<<<<<<<<<<<<<
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 */
    __pyx_t_7 = __pyx_f_5numpy__util_dtypestring(__pyx_v_descr, (__pyx_v_info->format + 1), (__pyx_v_info->format + 0xFF), (&__pyx_v_offset)); if (unlikely(__pyx_t_7 == NULL)) __PYX_ERR(1, 285, __pyx_L1_error)
    __pyx_v_f = __pyx_t_7;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":288
 *                                       info.format + _buffer_format_string_len,
 *                                       &offset)
 *                 f[0] = c'\0' # Terminate format string             # <<<<<<<<<<<<<<
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 */
    (__pyx_v_f[0]) = '\x00';
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":197
 *         # experimental exception made for __getbuffer__ and __releasebuffer__
 *         # -- the details of this may change.
 *         def __getbuffer__(ndarray self, Py_buffer* info, int flags):             # <<<<<<<<<<<<<<
 *             # This implementation of getbuffer is geared towards Cython
 *             # requirements, and does not yet fullfill the PEP.
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_AddTraceback("numpy.ndarray.__getbuffer__", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  if (__pyx_v_info != NULL && __pyx_v_info->obj != NULL) {
    __Pyx_GOTREF(__pyx_v_info->obj);
    __Pyx_DECREF(__pyx_v_info->obj); __pyx_v_info->obj = NULL;
  }
  goto __pyx_L2;
  __pyx_L0:;
  if (__pyx_v_info != NULL && __pyx_v_info->obj == Py_None) {
    __Pyx_GOTREF(Py_None);
    __Pyx_DECREF(Py_None); __pyx_v_info->obj = NULL;
  }
  __pyx_L2:;
  __Pyx_XDECREF((PyObject *)__pyx_v_descr);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

/* Python wrapper */
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info); /*proto*/
static CYTHON_UNUSED void __pyx_pw_5numpy_7ndarray_3__releasebuffer__(PyObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__releasebuffer__ (wrapper)", 0);
  __pyx_pf_5numpy_7ndarray_2__releasebuffer__(((PyArrayObject *)__pyx_v_self), ((Py_buffer *)__pyx_v_info));

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

static void __pyx_pf_5numpy_7ndarray_2__releasebuffer__(PyArrayObject *__pyx_v_self, Py_buffer *__pyx_v_info) {
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("__releasebuffer__", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  __pyx_t_1 = (PyArray_HASFIELDS(__pyx_v_self) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":292
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)             # <<<<<<<<<<<<<<
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)
 */
    free(__pyx_v_info->format);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":291
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):
 *             if PyArray_HASFIELDS(self):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  __pyx_t_1 = (((sizeof(npy_intp)) != (sizeof(Py_ssize_t))) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":294
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):
 *                 stdlib.free(info.strides)             # <<<<<<<<<<<<<<
 *                 # info.shape was stored after info.strides in the same block
 * 
 */
    free(__pyx_v_info->strides);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":293
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 *             if sizeof(npy_intp) != sizeof(Py_ssize_t):             # <<<<<<<<<<<<<<
 *                 stdlib.free(info.strides)
 *                 # info.shape was stored after info.strides in the same block
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":290
 *                 f[0] = c'\0' # Terminate format string
 * 
 *         def __releasebuffer__(ndarray self, Py_buffer* info):             # <<<<<<<<<<<<<<
 *             if PyArray_HASFIELDS(self):
 *                 stdlib.free(info.format)
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew1(PyObject *__pyx_v_a) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew1", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":771
 * 
 * cdef inline object PyArray_MultiIterNew1(a):
 *     return PyArray_MultiIterNew(1, <void*>a)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(1, ((void *)__pyx_v_a)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 771, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":770
 * ctypedef npy_cdouble     complex_t
 * 
 * cdef inline object PyArray_MultiIterNew1(a):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew1", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew2(PyObject *__pyx_v_a, PyObject *__pyx_v_b) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew2", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":774
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(2, ((void *)__pyx_v_a), ((void *)__pyx_v_b)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 774, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":773
 *     return PyArray_MultiIterNew(1, <void*>a)
 * 
 * cdef inline object PyArray_MultiIterNew2(a, b):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew2", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew3(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew3", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":777
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(3, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 777, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":776
 *     return PyArray_MultiIterNew(2, <void*>a, <void*>b)
 * 
 * cdef inline object PyArray_MultiIterNew3(a, b, c):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew3", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew4(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew4", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":780
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)             # <<<<<<<<<<<<<<
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(4, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 780, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":779
 *     return PyArray_MultiIterNew(3, <void*>a, <void*>b, <void*> c)
 * 
 * cdef inline object PyArray_MultiIterNew4(a, b, c, d):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew4", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_PyArray_MultiIterNew5(PyObject *__pyx_v_a, PyObject *__pyx_v_b, PyObject *__pyx_v_c, PyObject *__pyx_v_d, PyObject *__pyx_v_e) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  __Pyx_RefNannySetupContext("PyArray_MultiIterNew5", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":783
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)             # <<<<<<<<<<<<<<
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:
 */
  __Pyx_XDECREF(__pyx_r);
  __pyx_t_1 = PyArray_MultiIterNew(5, ((void *)__pyx_v_a), ((void *)__pyx_v_b), ((void *)__pyx_v_c), ((void *)__pyx_v_d), ((void *)__pyx_v_e)); if (unlikely(!__pyx_t_1)) __PYX_ERR(1, 783, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_1);
  __pyx_r = __pyx_t_1;
  __pyx_t_1 = 0;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":782
 *     return PyArray_MultiIterNew(4, <void*>a, <void*>b, <void*>c, <void*> d)
 * 
 * cdef inline object PyArray_MultiIterNew5(a, b, c, d, e):             # <<<<<<<<<<<<<<
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_AddTraceback("numpy.PyArray_MultiIterNew5", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = 0;
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

static CYTHON_INLINE char *__pyx_f_5numpy__util_dtypestring(PyArray_Descr *__pyx_v_descr, char *__pyx_v_f, char *__pyx_v_end, int *__pyx_v_offset) {
  PyArray_Descr *__pyx_v_child = 0;
  int __pyx_v_endian_detector;
  int __pyx_v_little_endian;
  PyObject *__pyx_v_fields = 0;
  PyObject *__pyx_v_childname = NULL;
  PyObject *__pyx_v_new_offset = NULL;
  PyObject *__pyx_v_t = NULL;
  char *__pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  Py_ssize_t __pyx_t_2;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  int __pyx_t_5;
  int __pyx_t_6;
  int __pyx_t_7;
  long __pyx_t_8;
  char *__pyx_t_9;
  __Pyx_RefNannySetupContext("_util_dtypestring", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":790
 * 
 *     cdef dtype child
 *     cdef int endian_detector = 1             # <<<<<<<<<<<<<<
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)
 *     cdef tuple fields
 */
  __pyx_v_endian_detector = 1;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":791
 *     cdef dtype child
 *     cdef int endian_detector = 1
 *     cdef bint little_endian = ((<char*>&endian_detector)[0] != 0)             # <<<<<<<<<<<<<<
 *     cdef tuple fields
 * 
 */
  __pyx_v_little_endian = ((((char *)(&__pyx_v_endian_detector))[0]) != 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  if (unlikely(__pyx_v_descr->names == Py_None)) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
    __PYX_ERR(1, 794, __pyx_L1_error)
  }
  __pyx_t_1 = __pyx_v_descr->names; __Pyx_INCREF(__pyx_t_1); __pyx_t_2 = 0;
  for (;;) {
    if (__pyx_t_2 >= PyTuple_GET_SIZE(__pyx_t_1)) break;
    #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    __pyx_t_3 = PyTuple_GET_ITEM(__pyx_t_1, __pyx_t_2); __Pyx_INCREF(__pyx_t_3); __pyx_t_2++; if (unlikely(0 < 0)) __PYX_ERR(1, 794, __pyx_L1_error)
    #else
    __pyx_t_3 = PySequence_ITEM(__pyx_t_1, __pyx_t_2); __pyx_t_2++; if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 794, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    #endif
    __Pyx_XDECREF_SET(__pyx_v_childname, __pyx_t_3);
    __pyx_t_3 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":795
 * 
 *     for childname in descr.names:
 *         fields = descr.fields[childname]             # <<<<<<<<<<<<<<
 *         child, new_offset = fields
 * 
 */
    if (unlikely(__pyx_v_descr->fields == Py_None)) {
      PyErr_SetString(PyExc_TypeError, "'NoneType' object is not subscriptable");
      __PYX_ERR(1, 795, __pyx_L1_error)
    }
    __pyx_t_3 = __Pyx_PyDict_GetItem(__pyx_v_descr->fields, __pyx_v_childname); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    if (!(likely(PyTuple_CheckExact(__pyx_t_3))||((__pyx_t_3) == Py_None)||(PyErr_Format(PyExc_TypeError, "Expected %.16s, got %.200s", "tuple", Py_TYPE(__pyx_t_3)->tp_name), 0))) __PYX_ERR(1, 795, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_fields, ((PyObject*)__pyx_t_3));
    __pyx_t_3 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":796
 *     for childname in descr.names:
 *         fields = descr.fields[childname]
 *         child, new_offset = fields             # <<<<<<<<<<<<<<
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 */
    if (likely(__pyx_v_fields != Py_None)) {
      PyObject* sequence = __pyx_v_fields;
      #if !CYTHON_COMPILING_IN_PYPY
      Py_ssize_t size = Py_SIZE(sequence);
      #else
      Py_ssize_t size = PySequence_Size(sequence);
      #endif
      if (unlikely(size != 2)) {
        if (size > 2) __Pyx_RaiseTooManyValuesError(2);
        else if (size >= 0) __Pyx_RaiseNeedMoreValuesError(size);
        __PYX_ERR(1, 796, __pyx_L1_error)
      }
      #if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
      __pyx_t_3 = PyTuple_GET_ITEM(sequence, 0); 
      __pyx_t_4 = PyTuple_GET_ITEM(sequence, 1); 
      __Pyx_INCREF(__pyx_t_3);
      __Pyx_INCREF(__pyx_t_4);
      #else
      __pyx_t_3 = PySequence_ITEM(sequence, 0); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PySequence_ITEM(sequence, 1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 796, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      #endif
    } else {
      __Pyx_RaiseNoneNotIterableError(); __PYX_ERR(1, 796, __pyx_L1_error)
    }
    if (!(likely(((__pyx_t_3) == Py_None) || likely(__Pyx_TypeTest(__pyx_t_3, __pyx_ptype_5numpy_dtype))))) __PYX_ERR(1, 796, __pyx_L1_error)
    __Pyx_XDECREF_SET(__pyx_v_child, ((PyArray_Descr *)__pyx_t_3));
    __pyx_t_3 = 0;
    __Pyx_XDECREF_SET(__pyx_v_new_offset, __pyx_t_4);
    __pyx_t_4 = 0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    __pyx_t_4 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_4);
    __pyx_t_3 = PyNumber_Subtract(__pyx_v_new_offset, __pyx_t_4); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_GOTREF(__pyx_t_3);
    __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
    __pyx_t_5 = __Pyx_PyInt_As_int(__pyx_t_3); if (unlikely((__pyx_t_5 == (int)-1) && PyErr_Occurred())) __PYX_ERR(1, 798, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __pyx_t_6 = ((((__pyx_v_end - __pyx_v_f) - ((int)__pyx_t_5)) < 15) != 0);
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__35, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 799, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 799, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":798
 *         child, new_offset = fields
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:             # <<<<<<<<<<<<<<
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '>') != 0);
    if (!__pyx_t_7) {
      goto __pyx_L8_next_or;
    } else {
    }
    __pyx_t_7 = (__pyx_v_little_endian != 0);
    if (!__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_L8_next_or:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":802
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):             # <<<<<<<<<<<<<<
 *             raise ValueError(u"Non-native byte order not supported")
 *             # One could encode it in the format string and have Cython
 */
    __pyx_t_7 = ((__pyx_v_child->byteorder == '<') != 0);
    if (__pyx_t_7) {
    } else {
      __pyx_t_6 = __pyx_t_7;
      goto __pyx_L7_bool_binop_done;
    }
    __pyx_t_7 = ((!(__pyx_v_little_endian != 0)) != 0);
    __pyx_t_6 = __pyx_t_7;
    __pyx_L7_bool_binop_done:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
      __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_tuple__36, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 803, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __Pyx_Raise(__pyx_t_3, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __PYX_ERR(1, 803, __pyx_L1_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":801
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")
 * 
 *         if ((child.byteorder == c'>' and little_endian) or             # <<<<<<<<<<<<<<
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")
 */
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":813
 * 
 *         # Output padding bytes
 *         while offset[0] < new_offset:             # <<<<<<<<<<<<<<
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 */
    while (1) {
      __pyx_t_3 = __Pyx_PyInt_From_int((__pyx_v_offset[0])); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_t_3, __pyx_v_new_offset, Py_LT); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 813, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (!__pyx_t_6) break;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":814
 *         # Output padding bytes
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte             # <<<<<<<<<<<<<<
 *             f += 1
 *             offset[0] += 1
 */
      (__pyx_v_f[0]) = 0x78;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":815
 *         while offset[0] < new_offset:
 *             f[0] = 120 # "x"; pad byte
 *             f += 1             # <<<<<<<<<<<<<<
 *             offset[0] += 1
 * 
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":816
 *             f[0] = 120 # "x"; pad byte
 *             f += 1
 *             offset[0] += 1             # <<<<<<<<<<<<<<
 * 
 *         offset[0] += child.itemsize
 */
      __pyx_t_8 = 0;
      (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + 1);
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":818
 *             offset[0] += 1
 * 
 *         offset[0] += child.itemsize             # <<<<<<<<<<<<<<
 * 
 *         if not PyDataType_HASFIELDS(child):
 */
    __pyx_t_8 = 0;
    (__pyx_v_offset[__pyx_t_8]) = ((__pyx_v_offset[__pyx_t_8]) + __pyx_v_child->elsize);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
    __pyx_t_6 = ((!(PyDataType_HASFIELDS(__pyx_v_child) != 0)) != 0);
    if (__pyx_t_6) {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":821
 * 
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num             # <<<<<<<<<<<<<<
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")
 */
      __pyx_t_4 = __Pyx_PyInt_From_int(__pyx_v_child->type_num); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 821, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_XDECREF_SET(__pyx_v_t, __pyx_t_4);
      __pyx_t_4 = 0;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      __pyx_t_6 = (((__pyx_v_end - __pyx_v_f) < 5) != 0);
      if (__pyx_t_6) {

        /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
        __pyx_t_4 = __Pyx_PyObject_Call(__pyx_builtin_RuntimeError, __pyx_tuple__37, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 823, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_Raise(__pyx_t_4, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __PYX_ERR(1, 823, __pyx_L1_error)

        /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":822
 *         if not PyDataType_HASFIELDS(child):
 *             t = child.type_num
 *             if end - f < 5:             # <<<<<<<<<<<<<<
 *                 raise RuntimeError(u"Format string allocated too short.")
 * 
 */
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":826
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_BYTE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 826, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 98;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":827
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"             # <<<<<<<<<<<<<<
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UBYTE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 827, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 66;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":828
 *             if   t == NPY_BYTE:        f[0] =  98 #"b"
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"             # <<<<<<<<<<<<<<
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_SHORT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 828, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x68;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":829
 *             elif t == NPY_UBYTE:       f[0] =  66 #"B"
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"             # <<<<<<<<<<<<<<
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_USHORT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 829, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 72;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":830
 *             elif t == NPY_SHORT:       f[0] = 104 #"h"
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"             # <<<<<<<<<<<<<<
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_INT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 830, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x69;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":831
 *             elif t == NPY_USHORT:      f[0] =  72 #"H"
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_UINT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 831, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 73;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":832
 *             elif t == NPY_INT:         f[0] = 105 #"i"
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 832, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x6C;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":833
 *             elif t == NPY_UINT:        f[0] =  73 #"I"
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 833, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 76;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":834
 *             elif t == NPY_LONG:        f[0] = 108 #"l"
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGLONG); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 834, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x71;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":835
 *             elif t == NPY_ULONG:       f[0] = 76  #"L"
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"             # <<<<<<<<<<<<<<
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_ULONGLONG); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 835, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 81;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":836
 *             elif t == NPY_LONGLONG:    f[0] = 113 #"q"
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"             # <<<<<<<<<<<<<<
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_FLOAT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 836, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x66;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":837
 *             elif t == NPY_ULONGLONG:   f[0] = 81  #"Q"
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"             # <<<<<<<<<<<<<<
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_DOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 837, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x64;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":838
 *             elif t == NPY_FLOAT:       f[0] = 102 #"f"
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"             # <<<<<<<<<<<<<<
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_LONGDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 838, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 0x67;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":839
 *             elif t == NPY_DOUBLE:      f[0] = 100 #"d"
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf             # <<<<<<<<<<<<<<
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CFLOAT); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 839, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x66;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":840
 *             elif t == NPY_LONGDOUBLE:  f[0] = 103 #"g"
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd             # <<<<<<<<<<<<<<
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CDOUBLE); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 840, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x64;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":841
 *             elif t == NPY_CFLOAT:      f[0] = 90; f[1] = 102; f += 1 # Zf
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg             # <<<<<<<<<<<<<<
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 */
      __pyx_t_3 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_CLONGDOUBLE); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_3);
      __pyx_t_4 = PyObject_RichCompare(__pyx_v_t, __pyx_t_3, Py_EQ); __Pyx_XGOTREF(__pyx_t_4); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_4); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 841, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 90;
        (__pyx_v_f[1]) = 0x67;
        __pyx_v_f = (__pyx_v_f + 1);
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":842
 *             elif t == NPY_CDOUBLE:     f[0] = 90; f[1] = 100; f += 1 # Zd
 *             elif t == NPY_CLONGDOUBLE: f[0] = 90; f[1] = 103; f += 1 # Zg
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"             # <<<<<<<<<<<<<<
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 */
      __pyx_t_4 = __Pyx_PyInt_From_enum__NPY_TYPES(NPY_OBJECT); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __pyx_t_3 = PyObject_RichCompare(__pyx_v_t, __pyx_t_4, Py_EQ); __Pyx_XGOTREF(__pyx_t_3); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
      __pyx_t_6 = __Pyx_PyObject_IsTrue(__pyx_t_3); if (unlikely(__pyx_t_6 < 0)) __PYX_ERR(1, 842, __pyx_L1_error)
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
      if (__pyx_t_6) {
        (__pyx_v_f[0]) = 79;
        goto __pyx_L15;
      }

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":844
 *             elif t == NPY_OBJECT:      f[0] = 79 #"O"
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)             # <<<<<<<<<<<<<<
 *             f += 1
 *         else:
 */
      /*else*/ {
        __pyx_t_3 = PyUnicode_Format(__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_v_t); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __pyx_t_4 = PyTuple_New(1); if (unlikely(!__pyx_t_4)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_4);
        __Pyx_GIVEREF(__pyx_t_3);
        PyTuple_SET_ITEM(__pyx_t_4, 0, __pyx_t_3);
        __pyx_t_3 = 0;
        __pyx_t_3 = __Pyx_PyObject_Call(__pyx_builtin_ValueError, __pyx_t_4, NULL); if (unlikely(!__pyx_t_3)) __PYX_ERR(1, 844, __pyx_L1_error)
        __Pyx_GOTREF(__pyx_t_3);
        __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
        __Pyx_Raise(__pyx_t_3, 0, 0, 0);
        __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
        __PYX_ERR(1, 844, __pyx_L1_error)
      }
      __pyx_L15:;

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":845
 *             else:
 *                 raise ValueError(u"unknown dtype code in numpy.pxd (%d)" % t)
 *             f += 1             # <<<<<<<<<<<<<<
 *         else:
 *             # Cython ignores struct boundary information ("T{...}"),
 */
      __pyx_v_f = (__pyx_v_f + 1);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":820
 *         offset[0] += child.itemsize
 * 
 *         if not PyDataType_HASFIELDS(child):             # <<<<<<<<<<<<<<
 *             t = child.type_num
 *             if end - f < 5:
 */
      goto __pyx_L13;
    }

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":849
 *             # Cython ignores struct boundary information ("T{...}"),
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)             # <<<<<<<<<<<<<<
 *     return f
 * 
 */
    /*else*/ {
      __pyx_t_9 = __pyx_f_5numpy__util_dtypestring(__pyx_v_child, __pyx_v_f, __pyx_v_end, __pyx_v_offset); if (unlikely(__pyx_t_9 == NULL)) __PYX_ERR(1, 849, __pyx_L1_error)
      __pyx_v_f = __pyx_t_9;
    }
    __pyx_L13:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":794
 *     cdef tuple fields
 * 
 *     for childname in descr.names:             # <<<<<<<<<<<<<<
 *         fields = descr.fields[childname]
 *         child, new_offset = fields
 */
  }
  __Pyx_DECREF(__pyx_t_1); __pyx_t_1 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":850
 *             # so don't output it
 *             f = _util_dtypestring(child, f, end, offset)
 *     return f             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_r = __pyx_v_f;
  goto __pyx_L0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":785
 *     return PyArray_MultiIterNew(5, <void*>a, <void*>b, <void*>c, <void*> d, <void*> e)
 * 
 * cdef inline char* _util_dtypestring(dtype descr, char* f, char* end, int* offset) except NULL:             # <<<<<<<<<<<<<<
 *     # Recursive utility function used in __getbuffer__ to get format
 *     # string. The new location in the format string is returned.
 */

  /* function exit code */
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_AddTraceback("numpy._util_dtypestring", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = NULL;
  __pyx_L0:;
  __Pyx_XDECREF((PyObject *)__pyx_v_child);
  __Pyx_XDECREF(__pyx_v_fields);
  __Pyx_XDECREF(__pyx_v_childname);
  __Pyx_XDECREF(__pyx_v_new_offset);
  __Pyx_XDECREF(__pyx_v_t);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

static CYTHON_INLINE void __pyx_f_5numpy_set_array_base(PyArrayObject *__pyx_v_arr, PyObject *__pyx_v_base) {
  PyObject *__pyx_v_baseptr;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  int __pyx_t_2;
  __Pyx_RefNannySetupContext("set_array_base", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
  __pyx_t_1 = (__pyx_v_base == Py_None);
  __pyx_t_2 = (__pyx_t_1 != 0);
  if (__pyx_t_2) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":969
 *      cdef PyObject* baseptr
 *      if base is None:
 *          baseptr = NULL             # <<<<<<<<<<<<<<
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 */
    __pyx_v_baseptr = NULL;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":968
 * cdef inline void set_array_base(ndarray arr, object base):
 *      cdef PyObject* baseptr
 *      if base is None:             # <<<<<<<<<<<<<<
 *          baseptr = NULL
 *      else:
 */
    goto __pyx_L3;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":971
 *          baseptr = NULL
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!             # <<<<<<<<<<<<<<
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 */
  /*else*/ {
    Py_INCREF(__pyx_v_base);

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":972
 *      else:
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base             # <<<<<<<<<<<<<<
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr
 */
    __pyx_v_baseptr = ((PyObject *)__pyx_v_base);
  }
  __pyx_L3:;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":973
 *          Py_INCREF(base) # important to do this before decref below!
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)             # <<<<<<<<<<<<<<
 *      arr.base = baseptr
 * 
 */
  Py_XDECREF(__pyx_v_arr->base);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":974
 *          baseptr = <PyObject*>base
 *      Py_XDECREF(arr.base)
 *      arr.base = baseptr             # <<<<<<<<<<<<<<
 * 
 * cdef inline object get_array_base(ndarray arr):
 */
  __pyx_v_arr->base = __pyx_v_baseptr;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":966
 * 
 * 
 * cdef inline void set_array_base(ndarray arr, object base):             # <<<<<<<<<<<<<<
 *      cdef PyObject* baseptr
 *      if base is None:
 */

  /* function exit code */
  __Pyx_RefNannyFinishContext();
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

static CYTHON_INLINE PyObject *__pyx_f_5numpy_get_array_base(PyArrayObject *__pyx_v_arr) {
  PyObject *__pyx_r = NULL;
  __Pyx_RefNannyDeclarations
  int __pyx_t_1;
  __Pyx_RefNannySetupContext("get_array_base", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  __pyx_t_1 = ((__pyx_v_arr->base == NULL) != 0);
  if (__pyx_t_1) {

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":978
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:
 *         return None             # <<<<<<<<<<<<<<
 *     else:
 *         return <object>arr.base
 */
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(Py_None);
    __pyx_r = Py_None;
    goto __pyx_L0;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":977
 * 
 * cdef inline object get_array_base(ndarray arr):
 *     if arr.base is NULL:             # <<<<<<<<<<<<<<
 *         return None
 *     else:
 */
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":980
 *         return None
 *     else:
 *         return <object>arr.base             # <<<<<<<<<<<<<<
 * 
 * 
 */
  /*else*/ {
    __Pyx_XDECREF(__pyx_r);
    __Pyx_INCREF(((PyObject *)__pyx_v_arr->base));
    __pyx_r = ((PyObject *)__pyx_v_arr->base);
    goto __pyx_L0;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":976
 *      arr.base = baseptr
 * 
 * cdef inline object get_array_base(ndarray arr):             # <<<<<<<<<<<<<<
 *     if arr.base is NULL:
 *         return None
 */

  /* function exit code */
  __pyx_L0:;
  __Pyx_XGIVEREF(__pyx_r);
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_array(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_array", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":987
 * cdef inline int import_array() except -1:
 *     try:
 *         _import_array()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")
 */
      __pyx_t_4 = _import_array(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 987, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":988
 *     try:
 *         _import_array()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 988, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__38, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 989, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 989, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":986
 * # Cython code.
 * cdef inline int import_array() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_array()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":985
 * # Versions of the import_* functions which are more suitable for
 * # Cython code.
 * cdef inline int import_array() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_array()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_array", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_umath(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_umath", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":993
 * cdef inline int import_umath() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 993, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":994
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 994, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__39, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 995, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 995, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":992
 * 
 * cdef inline int import_umath() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":991
 *         raise ImportError("numpy.core.multiarray failed to import")
 * 
 * cdef inline int import_umath() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_umath", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}

/* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

static CYTHON_INLINE int __pyx_f_5numpy_import_ufunc(void) {
  int __pyx_r;
  __Pyx_RefNannyDeclarations
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  int __pyx_t_4;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  PyObject *__pyx_t_8 = NULL;
  __Pyx_RefNannySetupContext("import_ufunc", 0);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
  {
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ExceptionSave(&__pyx_t_1, &__pyx_t_2, &__pyx_t_3);
    __Pyx_XGOTREF(__pyx_t_1);
    __Pyx_XGOTREF(__pyx_t_2);
    __Pyx_XGOTREF(__pyx_t_3);
    /*try:*/ {

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":999
 * cdef inline int import_ufunc() except -1:
 *     try:
 *         _import_umath()             # <<<<<<<<<<<<<<
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")
 */
      __pyx_t_4 = _import_umath(); if (unlikely(__pyx_t_4 == -1)) __PYX_ERR(1, 999, __pyx_L3_error)

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    }
    __Pyx_XDECREF(__pyx_t_1); __pyx_t_1 = 0;
    __Pyx_XDECREF(__pyx_t_2); __pyx_t_2 = 0;
    __Pyx_XDECREF(__pyx_t_3); __pyx_t_3 = 0;
    goto __pyx_L10_try_end;
    __pyx_L3_error:;
    __Pyx_PyThreadState_assign

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1000
 *     try:
 *         _import_umath()
 *     except Exception:             # <<<<<<<<<<<<<<
 *         raise ImportError("numpy.core.umath failed to import")
 */
    __pyx_t_4 = __Pyx_PyErr_ExceptionMatches(((PyObject *)(&((PyTypeObject*)PyExc_Exception)[0])));
    if (__pyx_t_4) {
      __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
      if (__Pyx_GetException(&__pyx_t_5, &__pyx_t_6, &__pyx_t_7) < 0) __PYX_ERR(1, 1000, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_5);
      __Pyx_GOTREF(__pyx_t_6);
      __Pyx_GOTREF(__pyx_t_7);

      /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
      __pyx_t_8 = __Pyx_PyObject_Call(__pyx_builtin_ImportError, __pyx_tuple__40, NULL); if (unlikely(!__pyx_t_8)) __PYX_ERR(1, 1001, __pyx_L5_except_error)
      __Pyx_GOTREF(__pyx_t_8);
      __Pyx_Raise(__pyx_t_8, 0, 0, 0);
      __Pyx_DECREF(__pyx_t_8); __pyx_t_8 = 0;
      __PYX_ERR(1, 1001, __pyx_L5_except_error)
    }
    goto __pyx_L5_except_error;
    __pyx_L5_except_error:;

    /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":998
 * 
 * cdef inline int import_ufunc() except -1:
 *     try:             # <<<<<<<<<<<<<<
 *         _import_umath()
 *     except Exception:
 */
    __Pyx_PyThreadState_assign
    __Pyx_XGIVEREF(__pyx_t_1);
    __Pyx_XGIVEREF(__pyx_t_2);
    __Pyx_XGIVEREF(__pyx_t_3);
    __Pyx_ExceptionReset(__pyx_t_1, __pyx_t_2, __pyx_t_3);
    goto __pyx_L1_error;
    __pyx_L10_try_end:;
  }

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /* function exit code */
  __pyx_r = 0;
  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  __Pyx_XDECREF(__pyx_t_8);
  __Pyx_AddTraceback("numpy.import_ufunc", __pyx_clineno, __pyx_lineno, __pyx_filename);
  __pyx_r = -1;
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  return __pyx_r;
}
static struct __pyx_vtabstruct_8rasterio_9_features_GeomBuilder __pyx_vtable_8rasterio_9_features_GeomBuilder;

static PyObject *__pyx_tp_new_8rasterio_9_features_GeomBuilder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8rasterio_9_features_GeomBuilder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8rasterio_9_features_GeomBuilder *)o);
  p->__pyx_vtab = __pyx_vtabptr_8rasterio_9_features_GeomBuilder;
  p->code = Py_None; Py_INCREF(Py_None);
  p->geomtypename = Py_None; Py_INCREF(Py_None);
  p->ndims = Py_None; Py_INCREF(Py_None);
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features_GeomBuilder(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features_GeomBuilder *p = (struct __pyx_obj_8rasterio_9_features_GeomBuilder *)o;
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && !_PyGC_FINALIZED(o)) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->code);
  Py_CLEAR(p->geomtypename);
  Py_CLEAR(p->ndims);
  (*Py_TYPE(o)->tp_free)(o);
}

static int __pyx_tp_traverse_8rasterio_9_features_GeomBuilder(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features_GeomBuilder *p = (struct __pyx_obj_8rasterio_9_features_GeomBuilder *)o;
  if (p->code) {
    e = (*v)(p->code, a); if (e) return e;
  }
  if (p->geomtypename) {
    e = (*v)(p->geomtypename, a); if (e) return e;
  }
  if (p->ndims) {
    e = (*v)(p->ndims, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features_GeomBuilder(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features_GeomBuilder *p = (struct __pyx_obj_8rasterio_9_features_GeomBuilder *)o;
  tmp = ((PyObject*)p->code);
  p->code = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->geomtypename);
  p->geomtypename = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->ndims);
  p->ndims = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyMethodDef __pyx_methods_8rasterio_9_features_GeomBuilder[] = {
  {"_buildPoint", (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_1_buildPoint, METH_NOARGS, 0},
  {"_buildLineString", (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_3_buildLineString, METH_NOARGS, 0},
  {"_buildLinearRing", (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_5_buildLinearRing, METH_NOARGS, 0},
  {"_buildPolygon", (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_7_buildPolygon, METH_NOARGS, 0},
  {"_buildMultiPolygon", (PyCFunction)__pyx_pw_8rasterio_9_features_11GeomBuilder_9_buildMultiPolygon, METH_NOARGS, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8rasterio_9_features_GeomBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.GeomBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features_GeomBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features_GeomBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  "Builds a GeoJSON (Fiona-style) geometry from an OGR geometry.", /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features_GeomBuilder, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features_GeomBuilder, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  __pyx_methods_8rasterio_9_features_GeomBuilder, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features_GeomBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};
static struct __pyx_vtabstruct_8rasterio_9_features_OGRGeomBuilder __pyx_vtable_8rasterio_9_features_OGRGeomBuilder;

static PyObject *__pyx_tp_new_8rasterio_9_features_OGRGeomBuilder(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *p;
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  p = ((struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *)o);
  p->__pyx_vtab = __pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder;
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features_OGRGeomBuilder(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyTypeObject __pyx_type_8rasterio_9_features_OGRGeomBuilder = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.OGRGeomBuilder", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features_OGRGeomBuilder, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "\n    Builds an OGR geometry from GeoJSON geometry.\n    From Fiona: https://github.com/Toblerity/Fiona/blob/master/src/fiona/ogrext.pyx\n    ", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features_OGRGeomBuilder, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyObject *__pyx_tp_new_8rasterio_9_features_ShapeIterator(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (likely((t->tp_flags & Py_TPFLAGS_IS_ABSTRACT) == 0)) {
    o = (*t->tp_alloc)(t, 0);
  } else {
    o = (PyObject *) PyBaseObject_Type.tp_new(t, __pyx_empty_tuple, 0);
  }
  if (unlikely(!o)) return 0;
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features_ShapeIterator(PyObject *o) {
  #if PY_VERSION_HEX >= 0x030400a1
  if (unlikely(Py_TYPE(o)->tp_finalize) && (!PyType_IS_GC(Py_TYPE(o)) || !_PyGC_FINALIZED(o))) {
    if (PyObject_CallFinalizerFromDealloc(o)) return;
  }
  #endif
  (*Py_TYPE(o)->tp_free)(o);
}

static PyMethodDef __pyx_methods_8rasterio_9_features_ShapeIterator[] = {
  {"__next__", (PyCFunction)__pyx_pw_8rasterio_9_features_13ShapeIterator_3__next__, METH_NOARGS|METH_COEXIST, 0},
  {0, 0, 0, 0}
};

static PyTypeObject __pyx_type_8rasterio_9_features_ShapeIterator = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.ShapeIterator", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features_ShapeIterator), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features_ShapeIterator, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_BASETYPE, /*tp_flags*/
  "Provides an iterator over shapes in an OGR feature layer.", /*tp_doc*/
  0, /*tp_traverse*/
  0, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  __pyx_pw_8rasterio_9_features_13ShapeIterator_1__iter__, /*tp_iter*/
  __pyx_pw_8rasterio_9_features_13ShapeIterator_3__next__, /*tp_iternext*/
  __pyx_methods_8rasterio_9_features_ShapeIterator, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features_ShapeIterator, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *__pyx_freelist_8rasterio_9_features___pyx_scope_struct___shapes[8];
static int __pyx_freecount_8rasterio_9_features___pyx_scope_struct___shapes = 0;

static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct___shapes(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_9_features___pyx_scope_struct___shapes > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_9_features___pyx_scope_struct___shapes[--__pyx_freecount_8rasterio_9_features___pyx_scope_struct___shapes];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct___shapes(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_connectivity);
  Py_CLEAR(p->__pyx_v_image);
  Py_CLEAR(p->__pyx_v_mask);
  Py_CLEAR(p->__pyx_v_mask_ds);
  Py_CLEAR(p->__pyx_v_mem_ds);
  Py_CLEAR(p->__pyx_v_mrdr);
  Py_CLEAR(p->__pyx_v_rdr);
  Py_CLEAR(p->__pyx_v_s);
  Py_CLEAR(p->__pyx_v_shape_iter);
  Py_CLEAR(p->__pyx_v_transform);
  Py_CLEAR(p->__pyx_v_v);
  Py_CLEAR(p->__pyx_v_valid_dtypes);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_9_features___pyx_scope_struct___shapes < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes)))) {
    __pyx_freelist_8rasterio_9_features___pyx_scope_struct___shapes[__pyx_freecount_8rasterio_9_features___pyx_scope_struct___shapes++] = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct___shapes(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)o;
  if (p->__pyx_v_connectivity) {
    e = (*v)(p->__pyx_v_connectivity, a); if (e) return e;
  }
  if (p->__pyx_v_image) {
    e = (*v)(p->__pyx_v_image, a); if (e) return e;
  }
  if (p->__pyx_v_mask) {
    e = (*v)(p->__pyx_v_mask, a); if (e) return e;
  }
  if (p->__pyx_v_mask_ds) {
    e = (*v)(((PyObject*)p->__pyx_v_mask_ds), a); if (e) return e;
  }
  if (p->__pyx_v_mem_ds) {
    e = (*v)(((PyObject*)p->__pyx_v_mem_ds), a); if (e) return e;
  }
  if (p->__pyx_v_mrdr) {
    e = (*v)(p->__pyx_v_mrdr, a); if (e) return e;
  }
  if (p->__pyx_v_rdr) {
    e = (*v)(p->__pyx_v_rdr, a); if (e) return e;
  }
  if (p->__pyx_v_s) {
    e = (*v)(p->__pyx_v_s, a); if (e) return e;
  }
  if (p->__pyx_v_shape_iter) {
    e = (*v)(((PyObject*)p->__pyx_v_shape_iter), a); if (e) return e;
  }
  if (p->__pyx_v_transform) {
    e = (*v)(p->__pyx_v_transform, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  if (p->__pyx_v_valid_dtypes) {
    e = (*v)(p->__pyx_v_valid_dtypes, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct___shapes(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes *)o;
  tmp = ((PyObject*)p->__pyx_v_connectivity);
  p->__pyx_v_connectivity = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_image);
  p->__pyx_v_image = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_mask);
  p->__pyx_v_mask = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_mask_ds);
  p->__pyx_v_mask_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_mem_ds);
  p->__pyx_v_mem_ds = ((struct __pyx_obj_8rasterio_3_io_InMemoryRaster *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_mrdr);
  p->__pyx_v_mrdr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_rdr);
  p->__pyx_v_rdr = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_s);
  p->__pyx_v_s = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_shape_iter);
  p->__pyx_v_shape_iter = ((struct __pyx_obj_8rasterio_9_features_ShapeIterator *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_transform);
  p->__pyx_v_transform = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_valid_dtypes);
  p->__pyx_v_valid_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_9_features___pyx_scope_struct___shapes = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.__pyx_scope_struct___shapes", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct___shapes), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct___shapes, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct___shapes, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct___shapes, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features___pyx_scope_struct___shapes, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *__pyx_freelist_8rasterio_9_features___pyx_scope_struct_1__sieve[8];
static int __pyx_freecount_8rasterio_9_features___pyx_scope_struct_1__sieve = 0;

static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_1__sieve(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_1__sieve > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_9_features___pyx_scope_struct_1__sieve[--__pyx_freecount_8rasterio_9_features___pyx_scope_struct_1__sieve];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_1__sieve(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_valid_dtypes);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_1__sieve < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve)))) {
    __pyx_freelist_8rasterio_9_features___pyx_scope_struct_1__sieve[__pyx_freecount_8rasterio_9_features___pyx_scope_struct_1__sieve++] = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_1__sieve(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)o;
  if (p->__pyx_v_valid_dtypes) {
    e = (*v)(p->__pyx_v_valid_dtypes, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_1__sieve(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)o;
  tmp = ((PyObject*)p->__pyx_v_valid_dtypes);
  p->__pyx_v_valid_dtypes = ((PyObject*)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_9_features___pyx_scope_struct_1__sieve = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.__pyx_scope_struct_1__sieve", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_1__sieve, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_1__sieve, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_1__sieve, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features___pyx_scope_struct_1__sieve, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *__pyx_freelist_8rasterio_9_features___pyx_scope_struct_2_genexpr[8];
static int __pyx_freecount_8rasterio_9_features___pyx_scope_struct_2_genexpr = 0;

static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_2_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_2_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_9_features___pyx_scope_struct_2_genexpr[--__pyx_freecount_8rasterio_9_features___pyx_scope_struct_2_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_2_genexpr(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_outer_scope);
  Py_CLEAR(p->__pyx_v_t);
  Py_CLEAR(p->__pyx_t_0);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_2_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr)))) {
    __pyx_freelist_8rasterio_9_features___pyx_scope_struct_2_genexpr[__pyx_freecount_8rasterio_9_features___pyx_scope_struct_2_genexpr++] = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_2_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)o;
  if (p->__pyx_outer_scope) {
    e = (*v)(((PyObject*)p->__pyx_outer_scope), a); if (e) return e;
  }
  if (p->__pyx_v_t) {
    e = (*v)(p->__pyx_v_t, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_2_genexpr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr *)o;
  tmp = ((PyObject*)p->__pyx_outer_scope);
  p->__pyx_outer_scope = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_1__sieve *)Py_None); Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_t);
  p->__pyx_v_t = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_9_features___pyx_scope_struct_2_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.__pyx_scope_struct_2_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_2_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_2_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_2_genexpr, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_2_genexpr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features___pyx_scope_struct_2_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *__pyx_freelist_8rasterio_9_features___pyx_scope_struct_3__explode[8];
static int __pyx_freecount_8rasterio_9_features___pyx_scope_struct_3__explode = 0;

static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_3__explode(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_3__explode > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_9_features___pyx_scope_struct_3__explode[--__pyx_freecount_8rasterio_9_features___pyx_scope_struct_3__explode];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_3__explode(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_coords);
  Py_CLEAR(p->__pyx_v_e);
  Py_CLEAR(p->__pyx_v_f);
  Py_CLEAR(p->__pyx_t_0);
  Py_CLEAR(p->__pyx_t_3);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_3__explode < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode)))) {
    __pyx_freelist_8rasterio_9_features___pyx_scope_struct_3__explode[__pyx_freecount_8rasterio_9_features___pyx_scope_struct_3__explode++] = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_3__explode(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)o;
  if (p->__pyx_v_coords) {
    e = (*v)(p->__pyx_v_coords, a); if (e) return e;
  }
  if (p->__pyx_v_e) {
    e = (*v)(p->__pyx_v_e, a); if (e) return e;
  }
  if (p->__pyx_v_f) {
    e = (*v)(p->__pyx_v_f, a); if (e) return e;
  }
  if (p->__pyx_t_0) {
    e = (*v)(p->__pyx_t_0, a); if (e) return e;
  }
  if (p->__pyx_t_3) {
    e = (*v)(p->__pyx_t_3, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_3__explode(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode *)o;
  tmp = ((PyObject*)p->__pyx_v_coords);
  p->__pyx_v_coords = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_e);
  p->__pyx_v_e = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_f);
  p->__pyx_v_f = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_0);
  p->__pyx_t_0 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_t_3);
  p->__pyx_t_3 = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_9_features___pyx_scope_struct_3__explode = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.__pyx_scope_struct_3__explode", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_3__explode), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_3__explode, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_3__explode, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_3__explode, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features___pyx_scope_struct_3__explode, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *__pyx_freelist_8rasterio_9_features___pyx_scope_struct_4_genexpr[8];
static int __pyx_freecount_8rasterio_9_features___pyx_scope_struct_4_genexpr = 0;

static PyObject *__pyx_tp_new_8rasterio_9_features___pyx_scope_struct_4_genexpr(PyTypeObject *t, CYTHON_UNUSED PyObject *a, CYTHON_UNUSED PyObject *k) {
  PyObject *o;
  if (CYTHON_COMPILING_IN_CPYTHON && likely((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_4_genexpr > 0) & (t->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr)))) {
    o = (PyObject*)__pyx_freelist_8rasterio_9_features___pyx_scope_struct_4_genexpr[--__pyx_freecount_8rasterio_9_features___pyx_scope_struct_4_genexpr];
    memset(o, 0, sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr));
    (void) PyObject_INIT(o, t);
    PyObject_GC_Track(o);
  } else {
    o = (*t->tp_alloc)(t, 0);
    if (unlikely(!o)) return 0;
  }
  return o;
}

static void __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_4_genexpr(PyObject *o) {
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)o;
  PyObject_GC_UnTrack(o);
  Py_CLEAR(p->__pyx_v_k);
  Py_CLEAR(p->__pyx_v_v);
  if (CYTHON_COMPILING_IN_CPYTHON && ((__pyx_freecount_8rasterio_9_features___pyx_scope_struct_4_genexpr < 8) & (Py_TYPE(o)->tp_basicsize == sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr)))) {
    __pyx_freelist_8rasterio_9_features___pyx_scope_struct_4_genexpr[__pyx_freecount_8rasterio_9_features___pyx_scope_struct_4_genexpr++] = ((struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)o);
  } else {
    (*Py_TYPE(o)->tp_free)(o);
  }
}

static int __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_4_genexpr(PyObject *o, visitproc v, void *a) {
  int e;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)o;
  if (p->__pyx_v_k) {
    e = (*v)(p->__pyx_v_k, a); if (e) return e;
  }
  if (p->__pyx_v_v) {
    e = (*v)(p->__pyx_v_v, a); if (e) return e;
  }
  return 0;
}

static int __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_4_genexpr(PyObject *o) {
  PyObject* tmp;
  struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *p = (struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr *)o;
  tmp = ((PyObject*)p->__pyx_v_k);
  p->__pyx_v_k = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  tmp = ((PyObject*)p->__pyx_v_v);
  p->__pyx_v_v = Py_None; Py_INCREF(Py_None);
  Py_XDECREF(tmp);
  return 0;
}

static PyTypeObject __pyx_type_8rasterio_9_features___pyx_scope_struct_4_genexpr = {
  PyVarObject_HEAD_INIT(0, 0)
  "rasterio._features.__pyx_scope_struct_4_genexpr", /*tp_name*/
  sizeof(struct __pyx_obj_8rasterio_9_features___pyx_scope_struct_4_genexpr), /*tp_basicsize*/
  0, /*tp_itemsize*/
  __pyx_tp_dealloc_8rasterio_9_features___pyx_scope_struct_4_genexpr, /*tp_dealloc*/
  0, /*tp_print*/
  0, /*tp_getattr*/
  0, /*tp_setattr*/
  #if PY_MAJOR_VERSION < 3
  0, /*tp_compare*/
  #endif
  #if PY_MAJOR_VERSION >= 3
  0, /*tp_as_async*/
  #endif
  0, /*tp_repr*/
  0, /*tp_as_number*/
  0, /*tp_as_sequence*/
  0, /*tp_as_mapping*/
  0, /*tp_hash*/
  0, /*tp_call*/
  0, /*tp_str*/
  0, /*tp_getattro*/
  0, /*tp_setattro*/
  0, /*tp_as_buffer*/
  Py_TPFLAGS_DEFAULT|Py_TPFLAGS_HAVE_VERSION_TAG|Py_TPFLAGS_CHECKTYPES|Py_TPFLAGS_HAVE_NEWBUFFER|Py_TPFLAGS_HAVE_GC, /*tp_flags*/
  0, /*tp_doc*/
  __pyx_tp_traverse_8rasterio_9_features___pyx_scope_struct_4_genexpr, /*tp_traverse*/
  __pyx_tp_clear_8rasterio_9_features___pyx_scope_struct_4_genexpr, /*tp_clear*/
  0, /*tp_richcompare*/
  0, /*tp_weaklistoffset*/
  0, /*tp_iter*/
  0, /*tp_iternext*/
  0, /*tp_methods*/
  0, /*tp_members*/
  0, /*tp_getset*/
  0, /*tp_base*/
  0, /*tp_dict*/
  0, /*tp_descr_get*/
  0, /*tp_descr_set*/
  0, /*tp_dictoffset*/
  0, /*tp_init*/
  0, /*tp_alloc*/
  __pyx_tp_new_8rasterio_9_features___pyx_scope_struct_4_genexpr, /*tp_new*/
  0, /*tp_free*/
  0, /*tp_is_gc*/
  0, /*tp_bases*/
  0, /*tp_mro*/
  0, /*tp_cache*/
  0, /*tp_subclasses*/
  0, /*tp_weaklist*/
  0, /*tp_del*/
  0, /*tp_version_tag*/
  #if PY_VERSION_HEX >= 0x030400a1
  0, /*tp_finalize*/
  #endif
};

static PyMethodDef __pyx_methods[] = {
  {0, 0, 0, 0}
};

#if PY_MAJOR_VERSION >= 3
static struct PyModuleDef __pyx_moduledef = {
  #if PY_VERSION_HEX < 0x03020000
    { PyObject_HEAD_INIT(NULL) NULL, 0, NULL },
  #else
    PyModuleDef_HEAD_INIT,
  #endif
    "_features",
    __pyx_k_Feature_extraction, /* m_doc */
    -1, /* m_size */
    __pyx_methods /* m_methods */,
    NULL, /* m_reload */
    NULL, /* m_traverse */
    NULL, /* m_clear */
    NULL /* m_free */
};
#endif

static __Pyx_StringTabEntry __pyx_string_tab[] = {
  {&__pyx_kp_s_3D_GeometryCollection, __pyx_k_3D_GeometryCollection, sizeof(__pyx_k_3D_GeometryCollection), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_LineString, __pyx_k_3D_LineString, sizeof(__pyx_k_3D_LineString), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_MultiLineString, __pyx_k_3D_MultiLineString, sizeof(__pyx_k_3D_MultiLineString), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_MultiPoint, __pyx_k_3D_MultiPoint, sizeof(__pyx_k_3D_MultiPoint), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_MultiPolygon, __pyx_k_3D_MultiPolygon, sizeof(__pyx_k_3D_MultiPolygon), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_Point, __pyx_k_3D_Point, sizeof(__pyx_k_3D_Point), 0, 0, 1, 0},
  {&__pyx_kp_s_3D_Polygon, __pyx_k_3D_Polygon, sizeof(__pyx_k_3D_Polygon), 0, 0, 1, 0},
  {&__pyx_kp_s_Connectivity_Option_must_be_4_or, __pyx_k_Connectivity_Option_must_be_4_or, sizeof(__pyx_k_Connectivity_Option_must_be_4_or), 0, 0, 1, 0},
  {&__pyx_kp_s_Could_not_create_OGR_Geometry_of, __pyx_k_Could_not_create_OGR_Geometry_of, sizeof(__pyx_k_Could_not_create_OGR_Geometry_of), 0, 0, 1, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor, __pyx_k_Format_string_allocated_too_shor, sizeof(__pyx_k_Format_string_allocated_too_shor), 0, 1, 0, 0},
  {&__pyx_kp_u_Format_string_allocated_too_shor_2, __pyx_k_Format_string_allocated_too_shor_2, sizeof(__pyx_k_Format_string_allocated_too_shor_2), 0, 1, 0, 0},
  {&__pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES, __pyx_k_GEOJSON2OGR_GEOMETRY_TYPES, sizeof(__pyx_k_GEOJSON2OGR_GEOMETRY_TYPES), 0, 0, 1, 1},
  {&__pyx_n_s_GEOMETRY_TYPES, __pyx_k_GEOMETRY_TYPES, sizeof(__pyx_k_GEOMETRY_TYPES), 0, 0, 1, 1},
  {&__pyx_n_s_GeometryCollection, __pyx_k_GeometryCollection, sizeof(__pyx_k_GeometryCollection), 0, 0, 1, 1},
  {&__pyx_kp_s_Geometry_r_at_index_d_with_value, __pyx_k_Geometry_r_at_index_d_with_value, sizeof(__pyx_k_Geometry_r_at_index_d_with_value), 0, 0, 1, 0},
  {&__pyx_n_s_ImportError, __pyx_k_ImportError, sizeof(__pyx_k_ImportError), 0, 0, 1, 1},
  {&__pyx_kp_s_Input_is_not_a_valid_geometry_ob, __pyx_k_Input_is_not_a_valid_geometry_ob, sizeof(__pyx_k_Input_is_not_a_valid_geometry_ob), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_out_image, __pyx_k_Invalid_out_image, sizeof(__pyx_k_Invalid_out_image), 0, 0, 1, 0},
  {&__pyx_kp_s_Invalid_source_image, __pyx_k_Invalid_source_image, sizeof(__pyx_k_Invalid_source_image), 0, 0, 1, 0},
  {&__pyx_n_s_LineString, __pyx_k_LineString, sizeof(__pyx_k_LineString), 0, 0, 1, 1},
  {&__pyx_n_s_LinearRing, __pyx_k_LinearRing, sizeof(__pyx_k_LinearRing), 0, 0, 1, 1},
  {&__pyx_kp_s_Mask_must_be_dtype_rasterio_bool, __pyx_k_Mask_must_be_dtype_rasterio_bool, sizeof(__pyx_k_Mask_must_be_dtype_rasterio_bool), 0, 0, 1, 0},
  {&__pyx_kp_s_Mask_must_have_same_shape_as_ima, __pyx_k_Mask_must_have_same_shape_as_ima, sizeof(__pyx_k_Mask_must_have_same_shape_as_ima), 0, 0, 1, 0},
  {&__pyx_n_s_MultiLineString, __pyx_k_MultiLineString, sizeof(__pyx_k_MultiLineString), 0, 0, 1, 1},
  {&__pyx_n_s_MultiPoint, __pyx_k_MultiPoint, sizeof(__pyx_k_MultiPoint), 0, 0, 1, 1},
  {&__pyx_n_s_MultiPolygon, __pyx_k_MultiPolygon, sizeof(__pyx_k_MultiPolygon), 0, 0, 1, 1},
  {&__pyx_kp_s_NULL_driver, __pyx_k_NULL_driver, sizeof(__pyx_k_NULL_driver), 0, 0, 1, 0},
  {&__pyx_kp_s_NULL_feature_dataset, __pyx_k_NULL_feature_dataset, sizeof(__pyx_k_NULL_feature_dataset), 0, 0, 1, 0},
  {&__pyx_kp_s_NULL_field_definition, __pyx_k_NULL_field_definition, sizeof(__pyx_k_NULL_field_definition), 0, 0, 1, 0},
  {&__pyx_kp_s_NULL_layer, __pyx_k_NULL_layer, sizeof(__pyx_k_NULL_layer), 0, 0, 1, 0},
  {&__pyx_kp_u_Non_native_byte_order_not_suppor, __pyx_k_Non_native_byte_order_not_suppor, sizeof(__pyx_k_Non_native_byte_order_not_suppor), 0, 1, 0, 0},
  {&__pyx_n_s_None, __pyx_k_None, sizeof(__pyx_k_None), 0, 0, 1, 1},
  {&__pyx_kp_s_Null_geom, __pyx_k_Null_geom, sizeof(__pyx_k_Null_geom), 0, 0, 1, 0},
  {&__pyx_n_s_Point, __pyx_k_Point, sizeof(__pyx_k_Point), 0, 0, 1, 1},
  {&__pyx_n_s_Polygon, __pyx_k_Polygon, sizeof(__pyx_k_Polygon), 0, 0, 1, 1},
  {&__pyx_n_s_RuntimeError, __pyx_k_RuntimeError, sizeof(__pyx_k_RuntimeError), 0, 0, 1, 1},
  {&__pyx_n_s_StopIteration, __pyx_k_StopIteration, sizeof(__pyx_k_StopIteration), 0, 0, 1, 1},
  {&__pyx_n_s_Unknown, __pyx_k_Unknown, sizeof(__pyx_k_Unknown), 0, 0, 1, 1},
  {&__pyx_kp_s_Unsupported_geometry_type_s, __pyx_k_Unsupported_geometry_type_s, sizeof(__pyx_k_Unsupported_geometry_type_s), 0, 0, 1, 0},
  {&__pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_k_Users_sean_code_frs_wheel_build, sizeof(__pyx_k_Users_sean_code_frs_wheel_build), 0, 0, 1, 0},
  {&__pyx_n_s_ValueError, __pyx_k_ValueError, sizeof(__pyx_k_ValueError), 0, 0, 1, 1},
  {&__pyx_kp_s__2, __pyx_k__2, sizeof(__pyx_k__2), 0, 0, 1, 0},
  {&__pyx_n_s_all_shapes, __pyx_k_all_shapes, sizeof(__pyx_k_all_shapes), 0, 0, 1, 1},
  {&__pyx_n_s_all_touched, __pyx_k_all_touched, sizeof(__pyx_k_all_touched), 0, 0, 1, 1},
  {&__pyx_n_s_args, __pyx_k_args, sizeof(__pyx_k_args), 0, 0, 1, 1},
  {&__pyx_n_s_astype, __pyx_k_astype, sizeof(__pyx_k_astype), 0, 0, 1, 1},
  {&__pyx_n_s_band, __pyx_k_band, sizeof(__pyx_k_band), 0, 0, 1, 1},
  {&__pyx_n_s_bidx, __pyx_k_bidx, sizeof(__pyx_k_bidx), 0, 0, 1, 1},
  {&__pyx_n_s_bool, __pyx_k_bool, sizeof(__pyx_k_bool), 0, 0, 1, 1},
  {&__pyx_n_s_bounds, __pyx_k_bounds, sizeof(__pyx_k_bounds), 0, 0, 1, 1},
  {&__pyx_n_s_build, __pyx_k_build, sizeof(__pyx_k_build), 0, 0, 1, 1},
  {&__pyx_n_s_buildLineString, __pyx_k_buildLineString, sizeof(__pyx_k_buildLineString), 0, 0, 1, 1},
  {&__pyx_n_s_buildLinearRing, __pyx_k_buildLinearRing, sizeof(__pyx_k_buildLinearRing), 0, 0, 1, 1},
  {&__pyx_n_s_buildMultiPolygon, __pyx_k_buildMultiPolygon, sizeof(__pyx_k_buildMultiPolygon), 0, 0, 1, 1},
  {&__pyx_n_s_buildPoint, __pyx_k_buildPoint, sizeof(__pyx_k_buildPoint), 0, 0, 1, 1},
  {&__pyx_n_s_buildPolygon, __pyx_k_buildPolygon, sizeof(__pyx_k_buildPolygon), 0, 0, 1, 1},
  {&__pyx_n_s_close, __pyx_k_close, sizeof(__pyx_k_close), 0, 0, 1, 1},
  {&__pyx_n_s_cols, __pyx_k_cols, sizeof(__pyx_k_cols), 0, 0, 1, 1},
  {&__pyx_n_s_connectivity, __pyx_k_connectivity, sizeof(__pyx_k_connectivity), 0, 0, 1, 1},
  {&__pyx_kp_s_connectivity_must_be_4_or_8, __pyx_k_connectivity_must_be_4_or_8, sizeof(__pyx_k_connectivity_must_be_4_or_8), 0, 0, 1, 0},
  {&__pyx_n_s_coordinates, __pyx_k_coordinates, sizeof(__pyx_k_coordinates), 0, 0, 1, 1},
  {&__pyx_n_s_coords, __pyx_k_coords, sizeof(__pyx_k_coords), 0, 0, 1, 1},
  {&__pyx_n_s_debug, __pyx_k_debug, sizeof(__pyx_k_debug), 0, 0, 1, 1},
  {&__pyx_n_s_driver, __pyx_k_driver, sizeof(__pyx_k_driver), 0, 0, 1, 1},
  {&__pyx_n_s_ds, __pyx_k_ds, sizeof(__pyx_k_ds), 0, 0, 1, 1},
  {&__pyx_n_s_dtype, __pyx_k_dtype, sizeof(__pyx_k_dtype), 0, 0, 1, 1},
  {&__pyx_n_s_dtypes, __pyx_k_dtypes, sizeof(__pyx_k_dtypes), 0, 0, 1, 1},
  {&__pyx_n_s_e, __pyx_k_e, sizeof(__pyx_k_e), 0, 0, 1, 1},
  {&__pyx_n_s_enter, __pyx_k_enter, sizeof(__pyx_k_enter), 0, 0, 1, 1},
  {&__pyx_n_s_enumerate, __pyx_k_enumerate, sizeof(__pyx_k_enumerate), 0, 0, 1, 1},
  {&__pyx_n_s_error, __pyx_k_error, sizeof(__pyx_k_error), 0, 0, 1, 1},
  {&__pyx_n_s_exit, __pyx_k_exit, sizeof(__pyx_k_exit), 0, 0, 1, 1},
  {&__pyx_n_s_explode, __pyx_k_explode, sizeof(__pyx_k_explode), 0, 0, 1, 1},
  {&__pyx_n_s_f, __pyx_k_f, sizeof(__pyx_k_f), 0, 0, 1, 1},
  {&__pyx_n_s_feature, __pyx_k_feature, sizeof(__pyx_k_feature), 0, 0, 1, 1},
  {&__pyx_n_s_features, __pyx_k_features, sizeof(__pyx_k_features), 0, 0, 1, 1},
  {&__pyx_n_s_fielddefn, __pyx_k_fielddefn, sizeof(__pyx_k_fielddefn), 0, 0, 1, 1},
  {&__pyx_n_s_fieldtp, __pyx_k_fieldtp, sizeof(__pyx_k_fieldtp), 0, 0, 1, 1},
  {&__pyx_n_s_float32, __pyx_k_float32, sizeof(__pyx_k_float32), 0, 0, 1, 1},
  {&__pyx_n_s_format, __pyx_k_format, sizeof(__pyx_k_format), 0, 0, 1, 1},
  {&__pyx_n_s_fs, __pyx_k_fs, sizeof(__pyx_k_fs), 0, 0, 1, 1},
  {&__pyx_n_s_genexpr, __pyx_k_genexpr, sizeof(__pyx_k_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_geometries, __pyx_k_geometries, sizeof(__pyx_k_geometries), 0, 0, 1, 1},
  {&__pyx_n_s_geometry, __pyx_k_geometry, sizeof(__pyx_k_geometry), 0, 0, 1, 1},
  {&__pyx_n_s_geoms, __pyx_k_geoms, sizeof(__pyx_k_geoms), 0, 0, 1, 1},
  {&__pyx_n_s_get, __pyx_k_get, sizeof(__pyx_k_get), 0, 0, 1, 1},
  {&__pyx_n_s_getLogger, __pyx_k_getLogger, sizeof(__pyx_k_getLogger), 0, 0, 1, 1},
  {&__pyx_n_s_i, __pyx_k_i, sizeof(__pyx_k_i), 0, 0, 1, 1},
  {&__pyx_n_s_image, __pyx_k_image, sizeof(__pyx_k_image), 0, 0, 1, 1},
  {&__pyx_kp_s_image_dtype_must_be_one_of_0, __pyx_k_image_dtype_must_be_one_of_0, sizeof(__pyx_k_image_dtype_must_be_one_of_0), 0, 0, 1, 0},
  {&__pyx_n_s_import, __pyx_k_import, sizeof(__pyx_k_import), 0, 0, 1, 1},
  {&__pyx_n_s_in_band, __pyx_k_in_band, sizeof(__pyx_k_in_band), 0, 0, 1, 1},
  {&__pyx_n_s_in_mem_ds, __pyx_k_in_mem_ds, sizeof(__pyx_k_in_mem_ds), 0, 0, 1, 1},
  {&__pyx_n_s_int16, __pyx_k_int16, sizeof(__pyx_k_int16), 0, 0, 1, 1},
  {&__pyx_n_s_int32, __pyx_k_int32, sizeof(__pyx_k_int32), 0, 0, 1, 1},
  {&__pyx_n_s_is_float, __pyx_k_is_float, sizeof(__pyx_k_is_float), 0, 0, 1, 1},
  {&__pyx_n_s_is_ndarray, __pyx_k_is_ndarray, sizeof(__pyx_k_is_ndarray), 0, 0, 1, 1},
  {&__pyx_n_s_iteritems, __pyx_k_iteritems, sizeof(__pyx_k_iteritems), 0, 0, 1, 1},
  {&__pyx_n_s_join, __pyx_k_join, sizeof(__pyx_k_join), 0, 0, 1, 1},
  {&__pyx_n_s_layer, __pyx_k_layer, sizeof(__pyx_k_layer), 0, 0, 1, 1},
  {&__pyx_n_s_log, __pyx_k_log, sizeof(__pyx_k_log), 0, 0, 1, 1},
  {&__pyx_n_s_logging, __pyx_k_logging, sizeof(__pyx_k_logging), 0, 0, 1, 1},
  {&__pyx_n_s_main, __pyx_k_main, sizeof(__pyx_k_main), 0, 0, 1, 1},
  {&__pyx_n_s_mask, __pyx_k_mask, sizeof(__pyx_k_mask), 0, 0, 1, 1},
  {&__pyx_n_s_mask_band, __pyx_k_mask_band, sizeof(__pyx_k_mask_band), 0, 0, 1, 1},
  {&__pyx_n_s_mask_ds, __pyx_k_mask_ds, sizeof(__pyx_k_mask_ds), 0, 0, 1, 1},
  {&__pyx_n_s_mask_mem_ds, __pyx_k_mask_mem_ds, sizeof(__pyx_k_mask_mem_ds), 0, 0, 1, 1},
  {&__pyx_n_s_mask_reader, __pyx_k_mask_reader, sizeof(__pyx_k_mask_reader), 0, 0, 1, 1},
  {&__pyx_n_s_maskband, __pyx_k_maskband, sizeof(__pyx_k_maskband), 0, 0, 1, 1},
  {&__pyx_n_s_max, __pyx_k_max, sizeof(__pyx_k_max), 0, 0, 1, 1},
  {&__pyx_n_s_mem, __pyx_k_mem, sizeof(__pyx_k_mem), 0, 0, 1, 1},
  {&__pyx_n_s_mem_ds, __pyx_k_mem_ds, sizeof(__pyx_k_mem_ds), 0, 0, 1, 1},
  {&__pyx_n_s_min, __pyx_k_min, sizeof(__pyx_k_min), 0, 0, 1, 1},
  {&__pyx_n_s_mrdr, __pyx_k_mrdr, sizeof(__pyx_k_mrdr), 0, 0, 1, 1},
  {&__pyx_n_s_name, __pyx_k_name, sizeof(__pyx_k_name), 0, 0, 1, 1},
  {&__pyx_n_s_name_2, __pyx_k_name_2, sizeof(__pyx_k_name_2), 0, 0, 1, 1},
  {&__pyx_kp_u_ndarray_is_not_C_contiguous, __pyx_k_ndarray_is_not_C_contiguous, sizeof(__pyx_k_ndarray_is_not_C_contiguous), 0, 1, 0, 0},
  {&__pyx_kp_u_ndarray_is_not_Fortran_contiguou, __pyx_k_ndarray_is_not_Fortran_contiguou, sizeof(__pyx_k_ndarray_is_not_Fortran_contiguou), 0, 1, 0, 0},
  {&__pyx_n_s_np, __pyx_k_np, sizeof(__pyx_k_np), 0, 0, 1, 1},
  {&__pyx_n_s_num_geoms, __pyx_k_num_geoms, sizeof(__pyx_k_num_geoms), 0, 0, 1, 1},
  {&__pyx_n_s_numpy, __pyx_k_numpy, sizeof(__pyx_k_numpy), 0, 0, 1, 1},
  {&__pyx_kp_s_numpy_core_multiarray_failed_to, __pyx_k_numpy_core_multiarray_failed_to, sizeof(__pyx_k_numpy_core_multiarray_failed_to), 0, 0, 1, 0},
  {&__pyx_kp_s_numpy_core_umath_failed_to_impor, __pyx_k_numpy_core_umath_failed_to_impor, sizeof(__pyx_k_numpy_core_umath_failed_to_impor), 0, 0, 1, 0},
  {&__pyx_n_s_options, __pyx_k_options, sizeof(__pyx_k_options), 0, 0, 1, 1},
  {&__pyx_n_s_out, __pyx_k_out, sizeof(__pyx_k_out), 0, 0, 1, 1},
  {&__pyx_kp_s_out_array_r, __pyx_k_out_array_r, sizeof(__pyx_k_out_array_r), 0, 0, 1, 0},
  {&__pyx_n_s_out_band, __pyx_k_out_band, sizeof(__pyx_k_out_band), 0, 0, 1, 1},
  {&__pyx_n_s_out_mem_ds, __pyx_k_out_mem_ds, sizeof(__pyx_k_out_mem_ds), 0, 0, 1, 1},
  {&__pyx_kp_s_out_raster_must_match_dtype_of_i, __pyx_k_out_raster_must_match_dtype_of_i, sizeof(__pyx_k_out_raster_must_match_dtype_of_i), 0, 0, 1, 0},
  {&__pyx_kp_s_out_raster_shape_must_be_same_as, __pyx_k_out_raster_shape_must_be_same_as, sizeof(__pyx_k_out_raster_shape_must_be_same_as), 0, 0, 1, 0},
  {&__pyx_n_s_pixel_values, __pyx_k_pixel_values, sizeof(__pyx_k_pixel_values), 0, 0, 1, 1},
  {&__pyx_n_s_pyx_vtable, __pyx_k_pyx_vtable, sizeof(__pyx_k_pyx_vtable), 0, 0, 1, 1},
  {&__pyx_n_s_range, __pyx_k_range, sizeof(__pyx_k_range), 0, 0, 1, 1},
  {&__pyx_n_s_rasterio, __pyx_k_rasterio, sizeof(__pyx_k_rasterio), 0, 0, 1, 1},
  {&__pyx_kp_s_rasterio_0, __pyx_k_rasterio_0, sizeof(__pyx_k_rasterio_0), 0, 0, 1, 0},
  {&__pyx_n_s_rasterio__features, __pyx_k_rasterio__features, sizeof(__pyx_k_rasterio__features), 0, 0, 1, 1},
  {&__pyx_n_s_rasterize, __pyx_k_rasterize, sizeof(__pyx_k_rasterize), 0, 0, 1, 1},
  {&__pyx_n_s_rdr, __pyx_k_rdr, sizeof(__pyx_k_rdr), 0, 0, 1, 1},
  {&__pyx_n_s_read, __pyx_k_read, sizeof(__pyx_k_read), 0, 0, 1, 1},
  {&__pyx_n_s_retval, __pyx_k_retval, sizeof(__pyx_k_retval), 0, 0, 1, 1},
  {&__pyx_n_s_rows, __pyx_k_rows, sizeof(__pyx_k_rows), 0, 0, 1, 1},
  {&__pyx_n_s_s, __pyx_k_s, sizeof(__pyx_k_s), 0, 0, 1, 1},
  {&__pyx_n_s_send, __pyx_k_send, sizeof(__pyx_k_send), 0, 0, 1, 1},
  {&__pyx_n_s_shape, __pyx_k_shape, sizeof(__pyx_k_shape), 0, 0, 1, 1},
  {&__pyx_n_s_shape_iter, __pyx_k_shape_iter, sizeof(__pyx_k_shape_iter), 0, 0, 1, 1},
  {&__pyx_n_s_shapes, __pyx_k_shapes, sizeof(__pyx_k_shapes), 0, 0, 1, 1},
  {&__pyx_n_s_shapes_2, __pyx_k_shapes_2, sizeof(__pyx_k_shapes_2), 0, 0, 1, 1},
  {&__pyx_n_s_sieve, __pyx_k_sieve, sizeof(__pyx_k_sieve), 0, 0, 1, 1},
  {&__pyx_n_s_sieve_locals_genexpr, __pyx_k_sieve_locals_genexpr, sizeof(__pyx_k_sieve_locals_genexpr), 0, 0, 1, 1},
  {&__pyx_n_s_size, __pyx_k_size, sizeof(__pyx_k_size), 0, 0, 1, 1},
  {&__pyx_kp_s_size_must_be_an_integer_number_o, __pyx_k_size_must_be_an_integer_number_o, sizeof(__pyx_k_size_must_be_an_integer_number_o), 0, 0, 1, 0},
  {&__pyx_kp_s_size_must_be_greater_than_0, __pyx_k_size_must_be_greater_than_0, sizeof(__pyx_k_size_must_be_greater_than_0), 0, 0, 1, 0},
  {&__pyx_kp_s_size_must_be_smaller_than_size_o, __pyx_k_size_must_be_smaller_than_size_o, sizeof(__pyx_k_size_must_be_smaller_than_size_o), 0, 0, 1, 0},
  {&__pyx_n_s_test, __pyx_k_test, sizeof(__pyx_k_test), 0, 0, 1, 1},
  {&__pyx_n_s_throw, __pyx_k_throw, sizeof(__pyx_k_throw), 0, 0, 1, 1},
  {&__pyx_n_s_transform, __pyx_k_transform, sizeof(__pyx_k_transform), 0, 0, 1, 1},
  {&__pyx_n_s_type, __pyx_k_type, sizeof(__pyx_k_type), 0, 0, 1, 1},
  {&__pyx_n_s_udr, __pyx_k_udr, sizeof(__pyx_k_udr), 0, 0, 1, 1},
  {&__pyx_n_s_uint16, __pyx_k_uint16, sizeof(__pyx_k_uint16), 0, 0, 1, 1},
  {&__pyx_n_s_uint8, __pyx_k_uint8, sizeof(__pyx_k_uint8), 0, 0, 1, 1},
  {&__pyx_kp_u_unknown_dtype_code_in_numpy_pxd, __pyx_k_unknown_dtype_code_in_numpy_pxd, sizeof(__pyx_k_unknown_dtype_code_in_numpy_pxd), 0, 1, 0, 0},
  {&__pyx_n_s_v, __pyx_k_v, sizeof(__pyx_k_v), 0, 0, 1, 1},
  {&__pyx_n_s_valid_dtypes, __pyx_k_valid_dtypes, sizeof(__pyx_k_valid_dtypes), 0, 0, 1, 1},
  {&__pyx_n_s_valid_types_str, __pyx_k_valid_types_str, sizeof(__pyx_k_valid_types_str), 0, 0, 1, 1},
  {&__pyx_n_s_value, __pyx_k_value, sizeof(__pyx_k_value), 0, 0, 1, 1},
  {&__pyx_n_s_xmax, __pyx_k_xmax, sizeof(__pyx_k_xmax), 0, 0, 1, 1},
  {&__pyx_n_s_xmaxs, __pyx_k_xmaxs, sizeof(__pyx_k_xmaxs), 0, 0, 1, 1},
  {&__pyx_n_s_xmin, __pyx_k_xmin, sizeof(__pyx_k_xmin), 0, 0, 1, 1},
  {&__pyx_n_s_xmins, __pyx_k_xmins, sizeof(__pyx_k_xmins), 0, 0, 1, 1},
  {&__pyx_n_s_xyz, __pyx_k_xyz, sizeof(__pyx_k_xyz), 0, 0, 1, 1},
  {&__pyx_n_s_ymax, __pyx_k_ymax, sizeof(__pyx_k_ymax), 0, 0, 1, 1},
  {&__pyx_n_s_ymaxs, __pyx_k_ymaxs, sizeof(__pyx_k_ymaxs), 0, 0, 1, 1},
  {&__pyx_n_s_ymin, __pyx_k_ymin, sizeof(__pyx_k_ymin), 0, 0, 1, 1},
  {&__pyx_n_s_ymins, __pyx_k_ymins, sizeof(__pyx_k_ymins), 0, 0, 1, 1},
  {&__pyx_n_s_zip, __pyx_k_zip, sizeof(__pyx_k_zip), 0, 0, 1, 1},
  {0, 0, 0, 0, 0, 0, 0}
};
static int __Pyx_InitCachedBuiltins(void) {
  __pyx_builtin_ValueError = __Pyx_GetBuiltinName(__pyx_n_s_ValueError); if (!__pyx_builtin_ValueError) __PYX_ERR(0, 70, __pyx_L1_error)
  __pyx_builtin_enumerate = __Pyx_GetBuiltinName(__pyx_n_s_enumerate); if (!__pyx_builtin_enumerate) __PYX_ERR(0, 295, __pyx_L1_error)
  __pyx_builtin_range = __Pyx_GetBuiltinName(__pyx_n_s_range); if (!__pyx_builtin_range) __PYX_ERR(0, 313, __pyx_L1_error)
  __pyx_builtin_min = __Pyx_GetBuiltinName(__pyx_n_s_min); if (!__pyx_builtin_min) __PYX_ERR(0, 350, __pyx_L1_error)
  __pyx_builtin_max = __Pyx_GetBuiltinName(__pyx_n_s_max); if (!__pyx_builtin_max) __PYX_ERR(0, 350, __pyx_L1_error)
  __pyx_builtin_zip = __Pyx_GetBuiltinName(__pyx_n_s_zip); if (!__pyx_builtin_zip) __PYX_ERR(0, 352, __pyx_L1_error)
  __pyx_builtin_StopIteration = __Pyx_GetBuiltinName(__pyx_n_s_StopIteration); if (!__pyx_builtin_StopIteration) __PYX_ERR(0, 598, __pyx_L1_error)
  __pyx_builtin_RuntimeError = __Pyx_GetBuiltinName(__pyx_n_s_RuntimeError); if (!__pyx_builtin_RuntimeError) __PYX_ERR(1, 799, __pyx_L1_error)
  __pyx_builtin_ImportError = __Pyx_GetBuiltinName(__pyx_n_s_ImportError); if (!__pyx_builtin_ImportError) __PYX_ERR(1, 989, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

static int __Pyx_InitCachedConstants(void) {
  __Pyx_RefNannyDeclarations
  __Pyx_RefNannySetupContext("__Pyx_InitCachedConstants", 0);

  /* "rasterio/_features.pyx":67
 *     cdef int fieldtp = 2 if is_float else 0
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16', 'float32')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 */
  __pyx_tuple_ = PyTuple_Pack(5, __pyx_n_s_int16, __pyx_n_s_int32, __pyx_n_s_uint8, __pyx_n_s_uint16, __pyx_n_s_float32); if (unlikely(!__pyx_tuple_)) __PYX_ERR(0, 67, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple_);
  __Pyx_GIVEREF(__pyx_tuple_);

  /* "rasterio/_features.pyx":74
 * 
 *     if connectivity not in (4, 8):
 *         raise ValueError("Connectivity Option must be 4 or 8")             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(image):
 */
  __pyx_tuple__3 = PyTuple_Pack(1, __pyx_kp_s_Connectivity_Option_must_be_4_or); if (unlikely(!__pyx_tuple__3)) __PYX_ERR(0, 74, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__3);
  __Pyx_GIVEREF(__pyx_tuple__3);

  /* "rasterio/_features.pyx":83
 *         band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 *         raise ValueError("Invalid source image")             # <<<<<<<<<<<<<<
 * 
 *     if mask is not None:
 */
  __pyx_tuple__4 = PyTuple_Pack(1, __pyx_kp_s_Invalid_source_image); if (unlikely(!__pyx_tuple__4)) __PYX_ERR(0, 83, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__4);
  __Pyx_GIVEREF(__pyx_tuple__4);

  /* "rasterio/_features.pyx":87
 *     if mask is not None:
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")             # <<<<<<<<<<<<<<
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):
 */
  __pyx_tuple__5 = PyTuple_Pack(1, __pyx_kp_s_Mask_must_have_same_shape_as_ima); if (unlikely(!__pyx_tuple__5)) __PYX_ERR(0, 87, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__5);
  __Pyx_GIVEREF(__pyx_tuple__5);

  /* "rasterio/_features.pyx":90
 * 
 *         if np.dtype(mask.dtype).name not in ('bool', 'uint8'):
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "             # <<<<<<<<<<<<<<
 *                              "rasterio.uint8")
 * 
 */
  __pyx_tuple__6 = PyTuple_Pack(1, __pyx_kp_s_Mask_must_be_dtype_rasterio_bool); if (unlikely(!__pyx_tuple__6)) __PYX_ERR(0, 90, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__6);
  __Pyx_GIVEREF(__pyx_tuple__6);

  /* "rasterio/_features.pyx":95
 *         if dtypes.is_ndarray(mask):
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_ds = InMemoryRaster(image=mask.astype('uint8'),             # <<<<<<<<<<<<<<
 *                                      transform=transform)
 *             maskband = mask_ds.band(1)
 */
  __pyx_tuple__7 = PyTuple_Pack(1, __pyx_n_s_uint8); if (unlikely(!__pyx_tuple__7)) __PYX_ERR(0, 95, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__7);
  __Pyx_GIVEREF(__pyx_tuple__7);

  /* "rasterio/_features.pyx":105
 *     driver = OGRGetDriverByName("Memory")
 *     if driver == NULL:
 *         raise ValueError("NULL driver")             # <<<<<<<<<<<<<<
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:
 */
  __pyx_tuple__8 = PyTuple_Pack(1, __pyx_kp_s_NULL_driver); if (unlikely(!__pyx_tuple__8)) __PYX_ERR(0, 105, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__8);
  __Pyx_GIVEREF(__pyx_tuple__8);

  /* "rasterio/_features.pyx":108
 *     fs = OGR_Dr_CreateDataSource(driver, "temp", NULL)
 *     if fs == NULL:
 *         raise ValueError("NULL feature dataset")             # <<<<<<<<<<<<<<
 * 
 *     # And a layer.
 */
  __pyx_tuple__9 = PyTuple_Pack(1, __pyx_kp_s_NULL_feature_dataset); if (unlikely(!__pyx_tuple__9)) __PYX_ERR(0, 108, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__9);
  __Pyx_GIVEREF(__pyx_tuple__9);

  /* "rasterio/_features.pyx":113
 *     layer = OGR_DS_CreateLayer(fs, "polygons", NULL, 3, NULL)
 *     if layer == NULL:
 *         raise ValueError("NULL layer")             # <<<<<<<<<<<<<<
 * 
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 */
  __pyx_tuple__10 = PyTuple_Pack(1, __pyx_kp_s_NULL_layer); if (unlikely(!__pyx_tuple__10)) __PYX_ERR(0, 113, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__10);
  __Pyx_GIVEREF(__pyx_tuple__10);

  /* "rasterio/_features.pyx":117
 *     fielddefn = OGR_Fld_Create("image_value", fieldtp)
 *     if fielddefn == NULL:
 *         raise ValueError("NULL field definition")             # <<<<<<<<<<<<<<
 *     OGR_L_CreateField(layer, fielddefn, 1)
 *     OGR_Fld_Destroy(fielddefn)
 */
  __pyx_tuple__11 = PyTuple_Pack(1, __pyx_kp_s_NULL_field_definition); if (unlikely(!__pyx_tuple__11)) __PYX_ERR(0, 117, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__11);
  __Pyx_GIVEREF(__pyx_tuple__11);

  /* "rasterio/_features.pyx":179
 *     cdef GDALRasterBandH mask_band = NULL
 * 
 *     valid_dtypes = ('int16', 'int32', 'uint8', 'uint16')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name not in valid_dtypes:
 */
  __pyx_tuple__12 = PyTuple_Pack(4, __pyx_n_s_int16, __pyx_n_s_int32, __pyx_n_s_uint8, __pyx_n_s_uint16); if (unlikely(!__pyx_tuple__12)) __PYX_ERR(0, 179, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__12);
  __Pyx_GIVEREF(__pyx_tuple__12);

  /* "rasterio/_features.pyx":188
 * 
 *     if size <= 0:
 *         raise ValueError('size must be greater than 0')             # <<<<<<<<<<<<<<
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')
 */
  __pyx_tuple__13 = PyTuple_Pack(1, __pyx_kp_s_size_must_be_greater_than_0); if (unlikely(!__pyx_tuple__13)) __PYX_ERR(0, 188, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__13);
  __Pyx_GIVEREF(__pyx_tuple__13);

  /* "rasterio/_features.pyx":190
 *         raise ValueError('size must be greater than 0')
 *     elif type(size) == float:
 *         raise ValueError('size must be an integer number of pixels')             # <<<<<<<<<<<<<<
 *     elif size > (image.shape[0] * image.shape[1]):
 *         raise ValueError('size must be smaller than size of image')
 */
  __pyx_tuple__14 = PyTuple_Pack(1, __pyx_kp_s_size_must_be_an_integer_number_o); if (unlikely(!__pyx_tuple__14)) __PYX_ERR(0, 190, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__14);
  __Pyx_GIVEREF(__pyx_tuple__14);

  /* "rasterio/_features.pyx":192
 *         raise ValueError('size must be an integer number of pixels')
 *     elif size > (image.shape[0] * image.shape[1]):
 *         raise ValueError('size must be smaller than size of image')             # <<<<<<<<<<<<<<
 * 
 *     if connectivity not in (4, 8):
 */
  __pyx_tuple__15 = PyTuple_Pack(1, __pyx_kp_s_size_must_be_smaller_than_size_o); if (unlikely(!__pyx_tuple__15)) __PYX_ERR(0, 192, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__15);
  __Pyx_GIVEREF(__pyx_tuple__15);

  /* "rasterio/_features.pyx":195
 * 
 *     if connectivity not in (4, 8):
 *         raise ValueError('connectivity must be 4 or 8')             # <<<<<<<<<<<<<<
 * 
 *     if out.shape != image.shape:
 */
  __pyx_tuple__16 = PyTuple_Pack(1, __pyx_kp_s_connectivity_must_be_4_or_8); if (unlikely(!__pyx_tuple__16)) __PYX_ERR(0, 195, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__16);
  __Pyx_GIVEREF(__pyx_tuple__16);

  /* "rasterio/_features.pyx":198
 * 
 *     if out.shape != image.shape:
 *         raise ValueError('out raster shape must be same as image shape')             # <<<<<<<<<<<<<<
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:
 */
  __pyx_tuple__17 = PyTuple_Pack(1, __pyx_kp_s_out_raster_shape_must_be_same_as); if (unlikely(!__pyx_tuple__17)) __PYX_ERR(0, 198, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__17);
  __Pyx_GIVEREF(__pyx_tuple__17);

  /* "rasterio/_features.pyx":201
 * 
 *     if np.dtype(image.dtype).name != np.dtype(out.dtype).name:
 *         raise ValueError('out raster must match dtype of image')             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(image):
 */
  __pyx_tuple__18 = PyTuple_Pack(1, __pyx_kp_s_out_raster_must_match_dtype_of_i); if (unlikely(!__pyx_tuple__18)) __PYX_ERR(0, 201, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__18);
  __Pyx_GIVEREF(__pyx_tuple__18);

  /* "rasterio/_features.pyx":210
 *         in_band = (<DatasetReaderBase?>rdr).band(image.bidx)
 *     else:
 *         raise ValueError("Invalid source image")             # <<<<<<<<<<<<<<
 * 
 *     if dtypes.is_ndarray(out):
 */
  __pyx_tuple__19 = PyTuple_Pack(1, __pyx_kp_s_Invalid_source_image); if (unlikely(!__pyx_tuple__19)) __PYX_ERR(0, 210, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__19);
  __Pyx_GIVEREF(__pyx_tuple__19);

  /* "rasterio/_features.pyx":220
 *         out_band = (<DatasetReaderBase?>udr).band(out.bidx)
 *     else:
 *         raise ValueError("Invalid out image")             # <<<<<<<<<<<<<<
 * 
 *     if mask is not None:
 */
  __pyx_tuple__20 = PyTuple_Pack(1, __pyx_kp_s_Invalid_out_image); if (unlikely(!__pyx_tuple__20)) __PYX_ERR(0, 220, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__20);
  __Pyx_GIVEREF(__pyx_tuple__20);

  /* "rasterio/_features.pyx":224
 *     if mask is not None:
 *         if mask.shape != image.shape:
 *             raise ValueError("Mask must have same shape as image")             # <<<<<<<<<<<<<<
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):
 */
  __pyx_tuple__21 = PyTuple_Pack(1, __pyx_kp_s_Mask_must_have_same_shape_as_ima); if (unlikely(!__pyx_tuple__21)) __PYX_ERR(0, 224, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__21);
  __Pyx_GIVEREF(__pyx_tuple__21);

  /* "rasterio/_features.pyx":227
 * 
 *         if np.dtype(mask.dtype) not in ('bool', 'uint8'):
 *             raise ValueError("Mask must be dtype rasterio.bool_ or "             # <<<<<<<<<<<<<<
 *                              "rasterio.uint8")
 * 
 */
  __pyx_tuple__22 = PyTuple_Pack(1, __pyx_kp_s_Mask_must_be_dtype_rasterio_bool); if (unlikely(!__pyx_tuple__22)) __PYX_ERR(0, 227, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__22);
  __Pyx_GIVEREF(__pyx_tuple__22);

  /* "rasterio/_features.pyx":232
 *         if dtypes.is_ndarray(mask):
 *             # A boolean mask must be converted to uint8 for GDAL
 *             mask_mem_ds = InMemoryRaster(image=mask.astype('uint8'))             # <<<<<<<<<<<<<<
 *             mask_band = mask_mem_ds.band(1)
 * 
 */
  __pyx_tuple__23 = PyTuple_Pack(1, __pyx_n_s_uint8); if (unlikely(!__pyx_tuple__23)) __PYX_ERR(0, 232, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__23);
  __Pyx_GIVEREF(__pyx_tuple__23);

  /* "rasterio/_features.pyx":303
 *                     geometry, i, value)
 * 
 *         with InMemoryRaster(image=image, transform=transform) as mem:             # <<<<<<<<<<<<<<
 *             exc_wrap_int(
 *                 GDALRasterizeGeometries(
 */
  __pyx_tuple__24 = PyTuple_Pack(3, Py_None, Py_None, Py_None); if (unlikely(!__pyx_tuple__24)) __PYX_ERR(0, 303, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__24);
  __Pyx_GIVEREF(__pyx_tuple__24);

  /* "rasterio/_features.pyx":400
 *         cdef int i
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         npoints = OGR_G_GetPointCount(geom)
 *         coords = []
 */
  __pyx_tuple__25 = PyTuple_Pack(1, __pyx_kp_s_Null_geom); if (unlikely(!__pyx_tuple__25)) __PYX_ERR(0, 400, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__25);
  __Pyx_GIVEREF(__pyx_tuple__25);

  /* "rasterio/_features.pyx":429
 *         cdef OGRGeometryH part
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         parts = []
 *         for j in range(OGR_G_GetGeometryCount(geom)):
 */
  __pyx_tuple__26 = PyTuple_Pack(1, __pyx_kp_s_Null_geom); if (unlikely(!__pyx_tuple__26)) __PYX_ERR(0, 429, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__26);
  __Pyx_GIVEREF(__pyx_tuple__26);

  /* "rasterio/_features.pyx":447
 *         """Builds a GeoJSON object from an OGR geometry object."""
 *         if geom == NULL:
 *             raise ValueError("Null geom")             # <<<<<<<<<<<<<<
 *         cdef unsigned int etype = OGR_G_GetGeometryType(geom)
 *         self.code = etype
 */
  __pyx_tuple__27 = PyTuple_Pack(1, __pyx_kp_s_Null_geom); if (unlikely(!__pyx_tuple__27)) __PYX_ERR(0, 447, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__27);
  __Pyx_GIVEREF(__pyx_tuple__27);

  /* "rasterio/_features.pyx":475
 *             OGR_G_AddPoint_2D(geom, x, y)
 *         else:
 *             x, y, z = coordinate[:3]             # <<<<<<<<<<<<<<
 *             OGR_G_AddPoint(geom, x, y, z)
 * 
 */
  __pyx_slice__28 = PySlice_New(Py_None, __pyx_int_3, Py_None); if (unlikely(!__pyx_slice__28)) __PYX_ERR(0, 475, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_slice__28);
  __Pyx_GIVEREF(__pyx_slice__28);

  /* "rasterio/_features.pyx":549
 *         """Builds an OGR geometry from GeoJSON geometry."""
 *         cdef object typename = geometry['type']
 *         cdef object coordinates = geometry.get('coordinates')             # <<<<<<<<<<<<<<
 * 
 *         if not typename or not coordinates:
 */
  __pyx_tuple__29 = PyTuple_Pack(1, __pyx_n_s_coordinates); if (unlikely(!__pyx_tuple__29)) __PYX_ERR(0, 549, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__29);
  __Pyx_GIVEREF(__pyx_tuple__29);

  /* "rasterio/_features.pyx":552
 * 
 *         if not typename or not coordinates:
 *             raise ValueError("Input is not a valid geometry object")             # <<<<<<<<<<<<<<
 *         if typename == 'Point':
 *             return self._buildPoint(coordinates)
 */
  __pyx_tuple__30 = PyTuple_Pack(1, __pyx_kp_s_Input_is_not_a_valid_geometry_ob); if (unlikely(!__pyx_tuple__30)) __PYX_ERR(0, 552, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__30);
  __Pyx_GIVEREF(__pyx_tuple__30);

  /* "rasterio/_features.pyx":568
 *             return self._buildMultiPolygon(coordinates)
 *         elif typename == 'GeometryCollection':
 *             coordinates = geometry.get('geometries')             # <<<<<<<<<<<<<<
 *             return self._buildGeometryCollection(coordinates)
 *         else:
 */
  __pyx_tuple__31 = PyTuple_Pack(1, __pyx_n_s_geometries); if (unlikely(!__pyx_tuple__31)) __PYX_ERR(0, 568, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__31);
  __Pyx_GIVEREF(__pyx_tuple__31);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":218
 *             if ((flags & pybuf.PyBUF_C_CONTIGUOUS == pybuf.PyBUF_C_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_C_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not C contiguous")             # <<<<<<<<<<<<<<
 * 
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 */
  __pyx_tuple__32 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_C_contiguous); if (unlikely(!__pyx_tuple__32)) __PYX_ERR(1, 218, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__32);
  __Pyx_GIVEREF(__pyx_tuple__32);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":222
 *             if ((flags & pybuf.PyBUF_F_CONTIGUOUS == pybuf.PyBUF_F_CONTIGUOUS)
 *                 and not PyArray_CHKFLAGS(self, NPY_F_CONTIGUOUS)):
 *                 raise ValueError(u"ndarray is not Fortran contiguous")             # <<<<<<<<<<<<<<
 * 
 *             info.buf = PyArray_DATA(self)
 */
  __pyx_tuple__33 = PyTuple_Pack(1, __pyx_kp_u_ndarray_is_not_Fortran_contiguou); if (unlikely(!__pyx_tuple__33)) __PYX_ERR(1, 222, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__33);
  __Pyx_GIVEREF(__pyx_tuple__33);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":259
 *                 if ((descr.byteorder == c'>' and little_endian) or
 *                     (descr.byteorder == c'<' and not little_endian)):
 *                     raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *                 if   t == NPY_BYTE:        f = "b"
 *                 elif t == NPY_UBYTE:       f = "B"
 */
  __pyx_tuple__34 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__34)) __PYX_ERR(1, 259, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__34);
  __Pyx_GIVEREF(__pyx_tuple__34);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":799
 * 
 *         if (end - f) - <int>(new_offset - offset[0]) < 15:
 *             raise RuntimeError(u"Format string allocated too short, see comment in numpy.pxd")             # <<<<<<<<<<<<<<
 * 
 *         if ((child.byteorder == c'>' and little_endian) or
 */
  __pyx_tuple__35 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor); if (unlikely(!__pyx_tuple__35)) __PYX_ERR(1, 799, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__35);
  __Pyx_GIVEREF(__pyx_tuple__35);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":803
 *         if ((child.byteorder == c'>' and little_endian) or
 *             (child.byteorder == c'<' and not little_endian)):
 *             raise ValueError(u"Non-native byte order not supported")             # <<<<<<<<<<<<<<
 *             # One could encode it in the format string and have Cython
 *             # complain instead, BUT: < and > in format strings also imply
 */
  __pyx_tuple__36 = PyTuple_Pack(1, __pyx_kp_u_Non_native_byte_order_not_suppor); if (unlikely(!__pyx_tuple__36)) __PYX_ERR(1, 803, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__36);
  __Pyx_GIVEREF(__pyx_tuple__36);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":823
 *             t = child.type_num
 *             if end - f < 5:
 *                 raise RuntimeError(u"Format string allocated too short.")             # <<<<<<<<<<<<<<
 * 
 *             # Until ticket #99 is fixed, use integers to avoid warnings
 */
  __pyx_tuple__37 = PyTuple_Pack(1, __pyx_kp_u_Format_string_allocated_too_shor_2); if (unlikely(!__pyx_tuple__37)) __PYX_ERR(1, 823, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__37);
  __Pyx_GIVEREF(__pyx_tuple__37);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":989
 *         _import_array()
 *     except Exception:
 *         raise ImportError("numpy.core.multiarray failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_umath() except -1:
 */
  __pyx_tuple__38 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_multiarray_failed_to); if (unlikely(!__pyx_tuple__38)) __PYX_ERR(1, 989, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__38);
  __Pyx_GIVEREF(__pyx_tuple__38);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":995
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 * 
 * cdef inline int import_ufunc() except -1:
 */
  __pyx_tuple__39 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__39)) __PYX_ERR(1, 995, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__39);
  __Pyx_GIVEREF(__pyx_tuple__39);

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":1001
 *         _import_umath()
 *     except Exception:
 *         raise ImportError("numpy.core.umath failed to import")             # <<<<<<<<<<<<<<
 */
  __pyx_tuple__40 = PyTuple_Pack(1, __pyx_kp_s_numpy_core_umath_failed_to_impor); if (unlikely(!__pyx_tuple__40)) __PYX_ERR(1, 1001, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__40);
  __Pyx_GIVEREF(__pyx_tuple__40);

  /* "rasterio/_features.pyx":20
 * 
 * 
 * def _shapes(image, mask, connectivity, transform):             # <<<<<<<<<<<<<<
 *     """
 *     Return a generator of (polygon, value) for each each set of adjacent pixels
 */
  __pyx_tuple__41 = PyTuple_Pack(24, __pyx_n_s_image, __pyx_n_s_mask, __pyx_n_s_connectivity, __pyx_n_s_transform, __pyx_n_s_retval, __pyx_n_s_rows, __pyx_n_s_cols, __pyx_n_s_band, __pyx_n_s_maskband, __pyx_n_s_driver, __pyx_n_s_fs, __pyx_n_s_layer, __pyx_n_s_fielddefn, __pyx_n_s_options, __pyx_n_s_mem_ds, __pyx_n_s_mask_ds, __pyx_n_s_shape_iter, __pyx_n_s_is_float, __pyx_n_s_fieldtp, __pyx_n_s_valid_dtypes, __pyx_n_s_rdr, __pyx_n_s_mrdr, __pyx_n_s_s, __pyx_n_s_v); if (unlikely(!__pyx_tuple__41)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__41);
  __Pyx_GIVEREF(__pyx_tuple__41);
  __pyx_codeobj__42 = (PyObject*)__Pyx_PyCode_New(4, 0, 24, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__41, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_shapes, 20, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__42)) __PYX_ERR(0, 20, __pyx_L1_error)

  /* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */
  __pyx_tuple__43 = PyTuple_Pack(21, __pyx_n_s_image, __pyx_n_s_size, __pyx_n_s_out, __pyx_n_s_mask, __pyx_n_s_connectivity, __pyx_n_s_retval, __pyx_n_s_rows, __pyx_n_s_cols, __pyx_n_s_in_mem_ds, __pyx_n_s_out_mem_ds, __pyx_n_s_mask_mem_ds, __pyx_n_s_in_band, __pyx_n_s_out_band, __pyx_n_s_mask_band, __pyx_n_s_valid_dtypes, __pyx_n_s_valid_types_str, __pyx_n_s_rdr, __pyx_n_s_udr, __pyx_n_s_mask_reader, __pyx_n_s_genexpr, __pyx_n_s_genexpr); if (unlikely(!__pyx_tuple__43)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__43);
  __Pyx_GIVEREF(__pyx_tuple__43);
  __pyx_codeobj__44 = (PyObject*)__Pyx_PyCode_New(5, 0, 21, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__43, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_sieve, 146, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__44)) __PYX_ERR(0, 146, __pyx_L1_error)

  /* "rasterio/_features.pyx":253
 * 
 * 
 * def _rasterize(shapes, image, transform, all_touched):             # <<<<<<<<<<<<<<
 *     """
 *     Burns input geometries into `image`.
 */
  __pyx_tuple__45 = PyTuple_Pack(14, __pyx_n_s_shapes_2, __pyx_n_s_image, __pyx_n_s_transform, __pyx_n_s_all_touched, __pyx_n_s_retval, __pyx_n_s_i, __pyx_n_s_num_geoms, __pyx_n_s_geoms, __pyx_n_s_options, __pyx_n_s_pixel_values, __pyx_n_s_mem, __pyx_n_s_all_shapes, __pyx_n_s_geometry, __pyx_n_s_value); if (unlikely(!__pyx_tuple__45)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__45);
  __Pyx_GIVEREF(__pyx_tuple__45);
  __pyx_codeobj__46 = (PyObject*)__Pyx_PyCode_New(4, 0, 14, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__45, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_rasterize, 253, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__46)) __PYX_ERR(0, 253, __pyx_L1_error)

  /* "rasterio/_features.pyx":321
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
  __pyx_tuple__47 = PyTuple_Pack(3, __pyx_n_s_coords, __pyx_n_s_e, __pyx_n_s_f); if (unlikely(!__pyx_tuple__47)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__47);
  __Pyx_GIVEREF(__pyx_tuple__47);
  __pyx_codeobj__48 = (PyObject*)__Pyx_PyCode_New(1, 0, 3, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__47, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_explode, 321, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__48)) __PYX_ERR(0, 321, __pyx_L1_error)

  /* "rasterio/_features.pyx":334
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry.
 *     From Fiona 1.4.8 with updates here to handle feature collections.
 */
  __pyx_tuple__49 = PyTuple_Pack(11, __pyx_n_s_geometry, __pyx_n_s_xmins, __pyx_n_s_ymins, __pyx_n_s_xmaxs, __pyx_n_s_ymaxs, __pyx_n_s_feature, __pyx_n_s_xmin, __pyx_n_s_ymin, __pyx_n_s_xmax, __pyx_n_s_ymax, __pyx_n_s_xyz); if (unlikely(!__pyx_tuple__49)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_tuple__49);
  __Pyx_GIVEREF(__pyx_tuple__49);
  __pyx_codeobj__50 = (PyObject*)__Pyx_PyCode_New(1, 0, 11, 0, 0, __pyx_empty_bytes, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_tuple__49, __pyx_empty_tuple, __pyx_empty_tuple, __pyx_kp_s_Users_sean_code_frs_wheel_build, __pyx_n_s_bounds, 334, __pyx_empty_bytes); if (unlikely(!__pyx_codeobj__50)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_RefNannyFinishContext();
  return 0;
  __pyx_L1_error:;
  __Pyx_RefNannyFinishContext();
  return -1;
}

static int __Pyx_InitGlobals(void) {
  if (__Pyx_InitStrings(__pyx_string_tab) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  __pyx_int_0 = PyInt_FromLong(0); if (unlikely(!__pyx_int_0)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_1 = PyInt_FromLong(1); if (unlikely(!__pyx_int_1)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2 = PyInt_FromLong(2); if (unlikely(!__pyx_int_2)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_3 = PyInt_FromLong(3); if (unlikely(!__pyx_int_3)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_4 = PyInt_FromLong(4); if (unlikely(!__pyx_int_4)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_5 = PyInt_FromLong(5); if (unlikely(!__pyx_int_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_6 = PyInt_FromLong(6); if (unlikely(!__pyx_int_6)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_7 = PyInt_FromLong(7); if (unlikely(!__pyx_int_7)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_8 = PyInt_FromLong(8); if (unlikely(!__pyx_int_8)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_100 = PyInt_FromLong(100); if (unlikely(!__pyx_int_100)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_101 = PyInt_FromLong(101); if (unlikely(!__pyx_int_101)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483648 = PyInt_FromString((char *)"2147483648", 0, 0); if (unlikely(!__pyx_int_2147483648)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483649 = PyInt_FromString((char *)"2147483649", 0, 0); if (unlikely(!__pyx_int_2147483649)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483650 = PyInt_FromString((char *)"2147483650", 0, 0); if (unlikely(!__pyx_int_2147483650)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483651 = PyInt_FromString((char *)"2147483651", 0, 0); if (unlikely(!__pyx_int_2147483651)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483652 = PyInt_FromString((char *)"2147483652", 0, 0); if (unlikely(!__pyx_int_2147483652)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483653 = PyInt_FromString((char *)"2147483653", 0, 0); if (unlikely(!__pyx_int_2147483653)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483654 = PyInt_FromString((char *)"2147483654", 0, 0); if (unlikely(!__pyx_int_2147483654)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_int_2147483655 = PyInt_FromString((char *)"2147483655", 0, 0); if (unlikely(!__pyx_int_2147483655)) __PYX_ERR(0, 1, __pyx_L1_error)
  return 0;
  __pyx_L1_error:;
  return -1;
}

#if PY_MAJOR_VERSION < 3
PyMODINIT_FUNC init_features(void); /*proto*/
PyMODINIT_FUNC init_features(void)
#else
PyMODINIT_FUNC PyInit__features(void); /*proto*/
PyMODINIT_FUNC PyInit__features(void)
#endif
{
  PyObject *__pyx_t_1 = NULL;
  PyObject *__pyx_t_2 = NULL;
  PyObject *__pyx_t_3 = NULL;
  PyObject *__pyx_t_4 = NULL;
  PyObject *__pyx_t_5 = NULL;
  PyObject *__pyx_t_6 = NULL;
  PyObject *__pyx_t_7 = NULL;
  __Pyx_RefNannyDeclarations
  #if CYTHON_REFNANNY
  __Pyx_RefNanny = __Pyx_RefNannyImportAPI("refnanny");
  if (!__Pyx_RefNanny) {
      PyErr_Clear();
      __Pyx_RefNanny = __Pyx_RefNannyImportAPI("Cython.Runtime.refnanny");
      if (!__Pyx_RefNanny)
          Py_FatalError("failed to import 'refnanny' module");
  }
  #endif
  __Pyx_RefNannySetupContext("PyMODINIT_FUNC PyInit__features(void)", 0);
  if (__Pyx_check_binary_version() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_tuple = PyTuple_New(0); if (unlikely(!__pyx_empty_tuple)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_bytes = PyBytes_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_bytes)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_empty_unicode = PyUnicode_FromStringAndSize("", 0); if (unlikely(!__pyx_empty_unicode)) __PYX_ERR(0, 1, __pyx_L1_error)
  #ifdef __Pyx_CyFunction_USED
  if (__pyx_CyFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_FusedFunction_USED
  if (__pyx_FusedFunction_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Coroutine_USED
  if (__pyx_Coroutine_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_Generator_USED
  if (__pyx_Generator_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  #ifdef __Pyx_StopAsyncIteration_USED
  if (__pyx_StopAsyncIteration_init() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  /*--- Library function declarations ---*/
  /*--- Threads initialization code ---*/
  #if defined(__PYX_FORCE_INIT_THREADS) && __PYX_FORCE_INIT_THREADS
  #ifdef WITH_THREAD /* Python build with threading support? */
  PyEval_InitThreads();
  #endif
  #endif
  /*--- Module creation code ---*/
  #if PY_MAJOR_VERSION < 3
  __pyx_m = Py_InitModule4("_features", __pyx_methods, __pyx_k_Feature_extraction, 0, PYTHON_API_VERSION); Py_XINCREF(__pyx_m);
  #else
  __pyx_m = PyModule_Create(&__pyx_moduledef);
  #endif
  if (unlikely(!__pyx_m)) __PYX_ERR(0, 1, __pyx_L1_error)
  __pyx_d = PyModule_GetDict(__pyx_m); if (unlikely(!__pyx_d)) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_INCREF(__pyx_d);
  __pyx_b = PyImport_AddModule(__Pyx_BUILTIN_MODULE_NAME); if (unlikely(!__pyx_b)) __PYX_ERR(0, 1, __pyx_L1_error)
  #if CYTHON_COMPILING_IN_PYPY
  Py_INCREF(__pyx_b);
  #endif
  if (PyObject_SetAttrString(__pyx_m, "__builtins__", __pyx_b) < 0) __PYX_ERR(0, 1, __pyx_L1_error);
  /*--- Initialize various global constants etc. ---*/
  if (__Pyx_InitGlobals() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #if PY_MAJOR_VERSION < 3 && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
  if (__Pyx_init_sys_getdefaultencoding_params() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif
  if (__pyx_module_is_main_rasterio___features) {
    if (PyObject_SetAttrString(__pyx_m, "__name__", __pyx_n_s_main) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  }
  #if PY_MAJOR_VERSION >= 3
  {
    PyObject *modules = PyImport_GetModuleDict(); if (unlikely(!modules)) __PYX_ERR(0, 1, __pyx_L1_error)
    if (!PyDict_GetItemString(modules, "rasterio._features")) {
      if (unlikely(PyDict_SetItemString(modules, "rasterio._features", __pyx_m) < 0)) __PYX_ERR(0, 1, __pyx_L1_error)
    }
  }
  #endif
  /*--- Builtin init code ---*/
  if (__Pyx_InitCachedBuiltins() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Constants init code ---*/
  if (__Pyx_InitCachedConstants() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  /*--- Global init code ---*/
  /*--- Variable export code ---*/
  /*--- Function export code ---*/
  /*--- Type init code ---*/
  __pyx_vtabptr_8rasterio_9_features_GeomBuilder = &__pyx_vtable_8rasterio_9_features_GeomBuilder;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildCoords = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH))__pyx_f_8rasterio_9_features_11GeomBuilder__buildCoords;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildPoint = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch))__pyx_f_8rasterio_9_features_11GeomBuilder__buildPoint;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildLineString = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch))__pyx_f_8rasterio_9_features_11GeomBuilder__buildLineString;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildLinearRing = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch))__pyx_f_8rasterio_9_features_11GeomBuilder__buildLinearRing;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildParts = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH))__pyx_f_8rasterio_9_features_11GeomBuilder__buildParts;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildPolygon = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch))__pyx_f_8rasterio_9_features_11GeomBuilder__buildPolygon;
  __pyx_vtable_8rasterio_9_features_GeomBuilder._buildMultiPolygon = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, int __pyx_skip_dispatch))__pyx_f_8rasterio_9_features_11GeomBuilder__buildMultiPolygon;
  __pyx_vtable_8rasterio_9_features_GeomBuilder.build = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_GeomBuilder *, OGRGeometryH))__pyx_f_8rasterio_9_features_11GeomBuilder_build;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features_GeomBuilder) < 0) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_type_8rasterio_9_features_GeomBuilder.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_8rasterio_9_features_GeomBuilder.tp_dict, __pyx_vtabptr_8rasterio_9_features_GeomBuilder) < 0) __PYX_ERR(0, 393, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "GeomBuilder", (PyObject *)&__pyx_type_8rasterio_9_features_GeomBuilder) < 0) __PYX_ERR(0, 393, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_GeomBuilder = &__pyx_type_8rasterio_9_features_GeomBuilder;
  __pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder = &__pyx_vtable_8rasterio_9_features_OGRGeomBuilder;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._createOgrGeometry = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, int))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__createOgrGeometry;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._addPointToGeometry = (PyObject *(*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, OGRGeometryH, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__addPointToGeometry;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildPoint = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPoint;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildLineString = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLineString;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildLinearRing = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildLinearRing;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildPolygon = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildPolygon;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildMultiPoint = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPoint;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildMultiLineString = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiLineString;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildMultiPolygon = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildMultiPolygon;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder._buildGeometryCollection = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder__buildGeometryCollection;
  __pyx_vtable_8rasterio_9_features_OGRGeomBuilder.build = (OGRGeometryH (*)(struct __pyx_obj_8rasterio_9_features_OGRGeomBuilder *, PyObject *))__pyx_f_8rasterio_9_features_14OGRGeomBuilder_build;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features_OGRGeomBuilder) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_type_8rasterio_9_features_OGRGeomBuilder.tp_print = 0;
  if (__Pyx_SetVtable(__pyx_type_8rasterio_9_features_OGRGeomBuilder.tp_dict, __pyx_vtabptr_8rasterio_9_features_OGRGeomBuilder) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  if (PyObject_SetAttrString(__pyx_m, "OGRGeomBuilder", (PyObject *)&__pyx_type_8rasterio_9_features_OGRGeomBuilder) < 0) __PYX_ERR(0, 457, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_OGRGeomBuilder = &__pyx_type_8rasterio_9_features_OGRGeomBuilder;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features_ShapeIterator) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
  __pyx_type_8rasterio_9_features_ShapeIterator.tp_print = 0;
  if (PyObject_SetAttrString(__pyx_m, "ShapeIterator", (PyObject *)&__pyx_type_8rasterio_9_features_ShapeIterator) < 0) __PYX_ERR(0, 583, __pyx_L1_error)
  __pyx_ptype_8rasterio_9_features_ShapeIterator = &__pyx_type_8rasterio_9_features_ShapeIterator;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features___pyx_scope_struct___shapes) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __pyx_type_8rasterio_9_features___pyx_scope_struct___shapes.tp_print = 0;
  __pyx_ptype_8rasterio_9_features___pyx_scope_struct___shapes = &__pyx_type_8rasterio_9_features___pyx_scope_struct___shapes;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features___pyx_scope_struct_1__sieve) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __pyx_type_8rasterio_9_features___pyx_scope_struct_1__sieve.tp_print = 0;
  __pyx_ptype_8rasterio_9_features___pyx_scope_struct_1__sieve = &__pyx_type_8rasterio_9_features___pyx_scope_struct_1__sieve;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features___pyx_scope_struct_2_genexpr) < 0) __PYX_ERR(0, 182, __pyx_L1_error)
  __pyx_type_8rasterio_9_features___pyx_scope_struct_2_genexpr.tp_print = 0;
  __pyx_ptype_8rasterio_9_features___pyx_scope_struct_2_genexpr = &__pyx_type_8rasterio_9_features___pyx_scope_struct_2_genexpr;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features___pyx_scope_struct_3__explode) < 0) __PYX_ERR(0, 321, __pyx_L1_error)
  __pyx_type_8rasterio_9_features___pyx_scope_struct_3__explode.tp_print = 0;
  __pyx_ptype_8rasterio_9_features___pyx_scope_struct_3__explode = &__pyx_type_8rasterio_9_features___pyx_scope_struct_3__explode;
  if (PyType_Ready(&__pyx_type_8rasterio_9_features___pyx_scope_struct_4_genexpr) < 0) __PYX_ERR(0, 379, __pyx_L1_error)
  __pyx_type_8rasterio_9_features___pyx_scope_struct_4_genexpr.tp_print = 0;
  __pyx_ptype_8rasterio_9_features___pyx_scope_struct_4_genexpr = &__pyx_type_8rasterio_9_features___pyx_scope_struct_4_genexpr;
  /*--- Type import code ---*/
  __pyx_ptype_7cpython_4type_type = __Pyx_ImportType(__Pyx_BUILTIN_MODULE_NAME, "type", 
  #if CYTHON_COMPILING_IN_PYPY
  sizeof(PyTypeObject),
  #else
  sizeof(PyHeapTypeObject),
  #endif
  0); if (unlikely(!__pyx_ptype_7cpython_4type_type)) __PYX_ERR(2, 9, __pyx_L1_error)
  __pyx_ptype_5numpy_dtype = __Pyx_ImportType("numpy", "dtype", sizeof(PyArray_Descr), 0); if (unlikely(!__pyx_ptype_5numpy_dtype)) __PYX_ERR(1, 155, __pyx_L1_error)
  __pyx_ptype_5numpy_flatiter = __Pyx_ImportType("numpy", "flatiter", sizeof(PyArrayIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_flatiter)) __PYX_ERR(1, 168, __pyx_L1_error)
  __pyx_ptype_5numpy_broadcast = __Pyx_ImportType("numpy", "broadcast", sizeof(PyArrayMultiIterObject), 0); if (unlikely(!__pyx_ptype_5numpy_broadcast)) __PYX_ERR(1, 172, __pyx_L1_error)
  __pyx_ptype_5numpy_ndarray = __Pyx_ImportType("numpy", "ndarray", sizeof(PyArrayObject), 0); if (unlikely(!__pyx_ptype_5numpy_ndarray)) __PYX_ERR(1, 181, __pyx_L1_error)
  __pyx_ptype_5numpy_ufunc = __Pyx_ImportType("numpy", "ufunc", sizeof(PyUFuncObject), 0); if (unlikely(!__pyx_ptype_5numpy_ufunc)) __PYX_ERR(1, 861, __pyx_L1_error)
  __pyx_ptype_8rasterio_5_base_DatasetBase = __Pyx_ImportType("rasterio._base", "DatasetBase", sizeof(struct __pyx_obj_8rasterio_5_base_DatasetBase), 1); if (unlikely(!__pyx_ptype_8rasterio_5_base_DatasetBase)) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_5_base_DatasetBase = (struct __pyx_vtabstruct_8rasterio_5_base_DatasetBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_5_base_DatasetBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_5_base_DatasetBase)) __PYX_ERR(3, 6, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_DatasetReaderBase = __Pyx_ImportType("rasterio._io", "DatasetReaderBase", sizeof(struct __pyx_obj_8rasterio_3_io_DatasetReaderBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_DatasetReaderBase)) __PYX_ERR(4, 8, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_DatasetReaderBase = (struct __pyx_vtabstruct_8rasterio_3_io_DatasetReaderBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_DatasetReaderBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_DatasetReaderBase)) __PYX_ERR(4, 8, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_DatasetWriterBase = __Pyx_ImportType("rasterio._io", "DatasetWriterBase", sizeof(struct __pyx_obj_8rasterio_3_io_DatasetWriterBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_DatasetWriterBase)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_DatasetWriterBase = (struct __pyx_vtabstruct_8rasterio_3_io_DatasetWriterBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_DatasetWriterBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_DatasetWriterBase)) __PYX_ERR(4, 12, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase = __Pyx_ImportType("rasterio._io", "BufferedDatasetWriterBase", sizeof(struct __pyx_obj_8rasterio_3_io_BufferedDatasetWriterBase), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase)) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase = (struct __pyx_vtabstruct_8rasterio_3_io_BufferedDatasetWriterBase*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_BufferedDatasetWriterBase->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_BufferedDatasetWriterBase)) __PYX_ERR(4, 18, __pyx_L1_error)
  __pyx_ptype_8rasterio_3_io_InMemoryRaster = __Pyx_ImportType("rasterio._io", "InMemoryRaster", sizeof(struct __pyx_obj_8rasterio_3_io_InMemoryRaster), 1); if (unlikely(!__pyx_ptype_8rasterio_3_io_InMemoryRaster)) __PYX_ERR(4, 22, __pyx_L1_error)
  __pyx_vtabptr_8rasterio_3_io_InMemoryRaster = (struct __pyx_vtabstruct_8rasterio_3_io_InMemoryRaster*)__Pyx_GetVtable(__pyx_ptype_8rasterio_3_io_InMemoryRaster->tp_dict); if (unlikely(!__pyx_vtabptr_8rasterio_3_io_InMemoryRaster)) __PYX_ERR(4, 22, __pyx_L1_error)
  /*--- Variable import code ---*/
  /*--- Function import code ---*/
  __pyx_t_1 = __Pyx_ImportModule("rasterio._err"); if (!__pyx_t_1) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "exc_wrap_int", (void (**)(void))&__pyx_f_8rasterio_4_err_exc_wrap_int, "int (int)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_1, "exc_wrap_pointer", (void (**)(void))&__pyx_f_8rasterio_4_err_exc_wrap_pointer, "void *(void *)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_1); __pyx_t_1 = 0;
  __pyx_t_2 = __Pyx_ImportModule("rasterio._io"); if (!__pyx_t_2) __PYX_ERR(0, 1, __pyx_L1_error)
  if (__Pyx_ImportFunction(__pyx_t_2, "io_auto", (void (**)(void))&__pyx_f_8rasterio_3_io_io_auto, "int (PyObject *, GDALRasterBandH, int, struct __pyx_opt_args_8rasterio_3_io_io_auto *__pyx_optional_args)") < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  Py_DECREF(__pyx_t_2); __pyx_t_2 = 0;
  /*--- Execution code ---*/
  #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
  if (__Pyx_patch_abc() < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  #endif

  /* "rasterio/_features.pyx":5
 * include "gdal.pxi"
 * 
 * import logging             # <<<<<<<<<<<<<<
 * 
 * import numpy as np
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_logging, 0, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_logging, __pyx_t_3) < 0) __PYX_ERR(0, 5, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "rasterio/_features.pyx":7
 * import logging
 * 
 * import numpy as np             # <<<<<<<<<<<<<<
 * 
 * from rasterio import dtypes
 */
  __pyx_t_3 = __Pyx_Import(__pyx_n_s_numpy, 0, -1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_np, __pyx_t_3) < 0) __PYX_ERR(0, 7, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;

  /* "rasterio/_features.pyx":9
 * import numpy as np
 * 
 * from rasterio import dtypes             # <<<<<<<<<<<<<<
 * 
 * cimport numpy as np
 */
  __pyx_t_3 = PyList_New(1); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __Pyx_INCREF(__pyx_n_s_dtypes);
  __Pyx_GIVEREF(__pyx_n_s_dtypes);
  PyList_SET_ITEM(__pyx_t_3, 0, __pyx_n_s_dtypes);
  __pyx_t_4 = __Pyx_Import(__pyx_n_s_rasterio, __pyx_t_3, -1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_ImportFrom(__pyx_t_4, __pyx_n_s_dtypes); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_dtypes, __pyx_t_3) < 0) __PYX_ERR(0, 9, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":17
 * 
 * 
 * log = logging.getLogger(__name__)             # <<<<<<<<<<<<<<
 * 
 * 
 */
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_logging); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_5 = __Pyx_PyObject_GetAttrStr(__pyx_t_3, __pyx_n_s_getLogger); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
  __pyx_t_3 = __Pyx_GetModuleGlobalName(__pyx_n_s_name_2); if (unlikely(!__pyx_t_3)) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_3);
  __pyx_t_6 = NULL;
  if (CYTHON_UNPACK_METHODS && unlikely(PyMethod_Check(__pyx_t_5))) {
    __pyx_t_6 = PyMethod_GET_SELF(__pyx_t_5);
    if (likely(__pyx_t_6)) {
      PyObject* function = PyMethod_GET_FUNCTION(__pyx_t_5);
      __Pyx_INCREF(__pyx_t_6);
      __Pyx_INCREF(function);
      __Pyx_DECREF_SET(__pyx_t_5, function);
    }
  }
  if (!__pyx_t_6) {
    __pyx_t_4 = __Pyx_PyObject_CallOneArg(__pyx_t_5, __pyx_t_3); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
    __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    __Pyx_GOTREF(__pyx_t_4);
  } else {
    #if CYTHON_FAST_PYCALL
    if (PyFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_3};
      __pyx_t_4 = __Pyx_PyFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    #if CYTHON_FAST_PYCCALL
    if (__Pyx_PyFastCFunction_Check(__pyx_t_5)) {
      PyObject *__pyx_temp[2] = {__pyx_t_6, __pyx_t_3};
      __pyx_t_4 = __Pyx_PyCFunction_FastCall(__pyx_t_5, __pyx_temp+1-1, 1+1); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
      __Pyx_XDECREF(__pyx_t_6); __pyx_t_6 = 0;
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_3); __pyx_t_3 = 0;
    } else
    #endif
    {
      __pyx_t_7 = PyTuple_New(1+1); if (unlikely(!__pyx_t_7)) __PYX_ERR(0, 17, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_7);
      __Pyx_GIVEREF(__pyx_t_6); PyTuple_SET_ITEM(__pyx_t_7, 0, __pyx_t_6); __pyx_t_6 = NULL;
      __Pyx_GIVEREF(__pyx_t_3);
      PyTuple_SET_ITEM(__pyx_t_7, 0+1, __pyx_t_3);
      __pyx_t_3 = 0;
      __pyx_t_4 = __Pyx_PyObject_Call(__pyx_t_5, __pyx_t_7, NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 17, __pyx_L1_error)
      __Pyx_GOTREF(__pyx_t_4);
      __Pyx_DECREF(__pyx_t_7); __pyx_t_7 = 0;
    }
  }
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_log, __pyx_t_4) < 0) __PYX_ERR(0, 17, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":20
 * 
 * 
 * def _shapes(image, mask, connectivity, transform):             # <<<<<<<<<<<<<<
 *     """
 *     Return a generator of (polygon, value) for each each set of adjacent pixels
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_9_features_1_shapes, NULL, __pyx_n_s_rasterio__features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_shapes, __pyx_t_4) < 0) __PYX_ERR(0, 20, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":146
 * 
 * 
 * def _sieve(image, size, out, mask, connectivity):             # <<<<<<<<<<<<<<
 *     """
 *     Replaces small polygons in `image` with the value of their largest
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_9_features_4_sieve, NULL, __pyx_n_s_rasterio__features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_sieve, __pyx_t_4) < 0) __PYX_ERR(0, 146, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":253
 * 
 * 
 * def _rasterize(shapes, image, transform, all_touched):             # <<<<<<<<<<<<<<
 *     """
 *     Burns input geometries into `image`.
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_9_features_6_rasterize, NULL, __pyx_n_s_rasterio__features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_rasterize, __pyx_t_4) < 0) __PYX_ERR(0, 253, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":321
 * 
 * 
 * def _explode(coords):             # <<<<<<<<<<<<<<
 *     """Explode a GeoJSON geometry's coordinates object and yield
 *     coordinate tuples. As long as the input is conforming, the type of
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_9_features_8_explode, NULL, __pyx_n_s_rasterio__features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_explode, __pyx_t_4) < 0) __PYX_ERR(0, 321, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":334
 * 
 * 
 * def _bounds(geometry):             # <<<<<<<<<<<<<<
 *     """Bounding box of a GeoJSON geometry.
 *     From Fiona 1.4.8 with updates here to handle feature collections.
 */
  __pyx_t_4 = PyCFunction_NewEx(&__pyx_mdef_8rasterio_9_features_11_bounds, NULL, __pyx_n_s_rasterio__features); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_bounds, __pyx_t_4) < 0) __PYX_ERR(0, 334, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":358
 * # Mapping of OGR integer geometry types to GeoJSON type names.
 * GEOMETRY_TYPES = {
 *     0: 'Unknown',             # <<<<<<<<<<<<<<
 *     1: 'Point',
 *     2: 'LineString',
 */
  __pyx_t_4 = PyDict_New(); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 358, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_0, __pyx_n_s_Unknown) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_1, __pyx_n_s_Point) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2, __pyx_n_s_LineString) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_3, __pyx_n_s_Polygon) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_4, __pyx_n_s_MultiPoint) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_5, __pyx_n_s_MultiLineString) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_6, __pyx_n_s_MultiPolygon) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_7, __pyx_n_s_GeometryCollection) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_100, __pyx_n_s_None) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_101, __pyx_n_s_LinearRing) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483649, __pyx_kp_s_3D_Point) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483650, __pyx_kp_s_3D_LineString) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483651, __pyx_kp_s_3D_Polygon) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483652, __pyx_kp_s_3D_MultiPoint) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483653, __pyx_kp_s_3D_MultiLineString) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483654, __pyx_kp_s_3D_MultiPolygon) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_t_4, __pyx_int_2147483655, __pyx_kp_s_3D_GeometryCollection) < 0) __PYX_ERR(0, 358, __pyx_L1_error)
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEOMETRY_TYPES, __pyx_t_4) < 0) __PYX_ERR(0, 357, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;

  /* "rasterio/_features.pyx":379
 * # Mapping of GeoJSON type names to OGR integer geometry types
 * GEOJSON2OGR_GEOMETRY_TYPES = dict(
 *     (v, k) for k, v in GEOMETRY_TYPES.iteritems()             # <<<<<<<<<<<<<<
 * )
 * 
 */
  __pyx_t_4 = __pyx_pf_8rasterio_9_features_12genexpr(NULL); if (unlikely(!__pyx_t_4)) __PYX_ERR(0, 379, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_4);

  /* "rasterio/_features.pyx":378
 * 
 * # Mapping of GeoJSON type names to OGR integer geometry types
 * GEOJSON2OGR_GEOMETRY_TYPES = dict(             # <<<<<<<<<<<<<<
 *     (v, k) for k, v in GEOMETRY_TYPES.iteritems()
 * )
 */
  __pyx_t_5 = __Pyx_Generator_Next(__pyx_t_4); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  __Pyx_DECREF(__pyx_t_4); __pyx_t_4 = 0;
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_GEOJSON2OGR_GEOMETRY_TYPES, __pyx_t_5) < 0) __PYX_ERR(0, 378, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "rasterio/_features.pyx":1
 * """Feature extraction"""             # <<<<<<<<<<<<<<
 * 
 * include "gdal.pxi"
 */
  __pyx_t_5 = PyDict_New(); if (unlikely(!__pyx_t_5)) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_GOTREF(__pyx_t_5);
  if (PyDict_SetItem(__pyx_d, __pyx_n_s_test, __pyx_t_5) < 0) __PYX_ERR(0, 1, __pyx_L1_error)
  __Pyx_DECREF(__pyx_t_5); __pyx_t_5 = 0;

  /* "../../../../../../var/folders/zq/fv6jds3j4wb_0z0zh35xky_c0000gn/T/frswb.LKeVrF/sdist/lib/python3.5/site-packages/Cython/Includes/numpy/__init__.pxd":997
 *         raise ImportError("numpy.core.umath failed to import")
 * 
 * cdef inline int import_ufunc() except -1:             # <<<<<<<<<<<<<<
 *     try:
 *         _import_umath()
 */

  /*--- Wrapped vars code ---*/

  goto __pyx_L0;
  __pyx_L1_error:;
  __Pyx_XDECREF(__pyx_t_1);
  __Pyx_XDECREF(__pyx_t_2);
  __Pyx_XDECREF(__pyx_t_3);
  __Pyx_XDECREF(__pyx_t_4);
  __Pyx_XDECREF(__pyx_t_5);
  __Pyx_XDECREF(__pyx_t_6);
  __Pyx_XDECREF(__pyx_t_7);
  if (__pyx_m) {
    if (__pyx_d) {
      __Pyx_AddTraceback("init rasterio._features", __pyx_clineno, __pyx_lineno, __pyx_filename);
    }
    Py_DECREF(__pyx_m); __pyx_m = 0;
  } else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_ImportError, "init rasterio._features");
  }
  __pyx_L0:;
  __Pyx_RefNannyFinishContext();
  #if PY_MAJOR_VERSION < 3
  return;
  #else
  return __pyx_m;
  #endif
}

/* --- Runtime support code --- */
/* Refnanny */
#if CYTHON_REFNANNY
static __Pyx_RefNannyAPIStruct *__Pyx_RefNannyImportAPI(const char *modname) {
    PyObject *m = NULL, *p = NULL;
    void *r = NULL;
    m = PyImport_ImportModule((char *)modname);
    if (!m) goto end;
    p = PyObject_GetAttrString(m, (char *)"RefNannyAPI");
    if (!p) goto end;
    r = PyLong_AsVoidPtr(p);
end:
    Py_XDECREF(p);
    Py_XDECREF(m);
    return (__Pyx_RefNannyAPIStruct *)r;
}
#endif

/* GetBuiltinName */
static PyObject *__Pyx_GetBuiltinName(PyObject *name) {
    PyObject* result = __Pyx_PyObject_GetAttrStr(__pyx_b, name);
    if (unlikely(!result)) {
        PyErr_Format(PyExc_NameError,
#if PY_MAJOR_VERSION >= 3
            "name '%U' is not defined", name);
#else
            "name '%.200s' is not defined", PyString_AS_STRING(name));
#endif
    }
    return result;
}

/* IterFinish */
static CYTHON_INLINE int __Pyx_IterFinish(void) {
#if CYTHON_FAST_THREAD_STATE
    PyThreadState *tstate = PyThreadState_GET();
    PyObject* exc_type = tstate->curexc_type;
    if (unlikely(exc_type)) {
        if (likely(exc_type == PyExc_StopIteration) || PyErr_GivenExceptionMatches(exc_type, PyExc_StopIteration)) {
            PyObject *exc_value, *exc_tb;
            exc_value = tstate->curexc_value;
            exc_tb = tstate->curexc_traceback;
            tstate->curexc_type = 0;
            tstate->curexc_value = 0;
            tstate->curexc_traceback = 0;
            Py_DECREF(exc_type);
            Py_XDECREF(exc_value);
            Py_XDECREF(exc_tb);
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#else
    if (unlikely(PyErr_Occurred())) {
        if (likely(PyErr_ExceptionMatches(PyExc_StopIteration))) {
            PyErr_Clear();
            return 0;
        } else {
            return -1;
        }
    }
    return 0;
#endif
}

/* PyFunctionFastCall */
#if CYTHON_FAST_PYCALL
#include "frameobject.h"
static PyObject* __Pyx_PyFunction_FastCallNoKw(PyCodeObject *co, PyObject **args, Py_ssize_t na,
                                               PyObject *globals) {
    PyFrameObject *f;
    PyThreadState *tstate = PyThreadState_GET();
    PyObject **fastlocals;
    Py_ssize_t i;
    PyObject *result;
    assert(globals != NULL);
    /* XXX Perhaps we should create a specialized
       PyFrame_New() that doesn't take locals, but does
       take builtins without sanity checking them.
       */
    assert(tstate != NULL);
    f = PyFrame_New(tstate, co, globals, NULL);
    if (f == NULL) {
        return NULL;
    }
    fastlocals = f->f_localsplus;
    for (i = 0; i < na; i++) {
        Py_INCREF(*args);
        fastlocals[i] = *args++;
    }
    result = PyEval_EvalFrameEx(f,0);
    ++tstate->recursion_depth;
    Py_DECREF(f);
    --tstate->recursion_depth;
    return result;
}
#if 1 || PY_VERSION_HEX < 0x030600B1
static PyObject *__Pyx_PyFunction_FastCallDict(PyObject *func, PyObject **args, int nargs, PyObject *kwargs) {
    PyCodeObject *co = (PyCodeObject *)PyFunction_GET_CODE(func);
    PyObject *globals = PyFunction_GET_GLOBALS(func);
    PyObject *argdefs = PyFunction_GET_DEFAULTS(func);
    PyObject *closure;
#if PY_MAJOR_VERSION >= 3
    PyObject *kwdefs;
#endif
    PyObject *kwtuple, **k;
    PyObject **d;
    Py_ssize_t nd;
    Py_ssize_t nk;
    PyObject *result;
    assert(kwargs == NULL || PyDict_Check(kwargs));
    nk = kwargs ? PyDict_Size(kwargs) : 0;
    if (Py_EnterRecursiveCall((char*)" while calling a Python object")) {
        return NULL;
    }
    if (
#if PY_MAJOR_VERSION >= 3
            co->co_kwonlyargcount == 0 &&
#endif
            likely(kwargs == NULL || nk == 0) &&
            co->co_flags == (CO_OPTIMIZED | CO_NEWLOCALS | CO_NOFREE)) {
        if (argdefs == NULL && co->co_argcount == nargs) {
            result = __Pyx_PyFunction_FastCallNoKw(co, args, nargs, globals);
            goto done;
        }
        else if (nargs == 0 && argdefs != NULL
                 && co->co_argcount == Py_SIZE(argdefs)) {
            /* function called with no arguments, but all parameters have
               a default value: use default values as arguments .*/
            args = &PyTuple_GET_ITEM(argdefs, 0);
            result =__Pyx_PyFunction_FastCallNoKw(co, args, Py_SIZE(argdefs), globals);
            goto done;
        }
    }
    if (kwargs != NULL) {
        Py_ssize_t pos, i;
        kwtuple = PyTuple_New(2 * nk);
        if (kwtuple == NULL) {
            result = NULL;
            goto done;
        }
        k = &PyTuple_GET_ITEM(kwtuple, 0);
        pos = i = 0;
        while (PyDict_Next(kwargs, &pos, &k[i], &k[i+1])) {
            Py_INCREF(k[i]);
            Py_INCREF(k[i+1]);
            i += 2;
        }
        nk = i / 2;
    }
    else {
        kwtuple = NULL;
        k = NULL;
    }
    closure = PyFunction_GET_CLOSURE(func);
#if PY_MAJOR_VERSION >= 3
    kwdefs = PyFunction_GET_KW_DEFAULTS(func);
#endif
    if (argdefs != NULL) {
        d = &PyTuple_GET_ITEM(argdefs, 0);
        nd = Py_SIZE(argdefs);
    }
    else {
        d = NULL;
        nd = 0;
    }
#if PY_MAJOR_VERSION >= 3
    result = PyEval_EvalCodeEx((PyObject*)co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, kwdefs, closure);
#else
    result = PyEval_EvalCodeEx(co, globals, (PyObject *)NULL,
                               args, nargs,
                               k, (int)nk,
                               d, (int)nd, closure);
#endif
    Py_XDECREF(kwtuple);
done:
    Py_LeaveRecursiveCall();
    return result;
}
#endif  // CPython < 3.6
#endif  // CYTHON_FAST_PYCALL

/* PyObjectCall */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_Call(PyObject *func, PyObject *arg, PyObject *kw) {
    PyObject *result;
    ternaryfunc call = func->ob_type->tp_call;
    if (unlikely(!call))
        return PyObject_Call(func, arg, kw);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = (*call)(func, arg, kw);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallMethO */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallMethO(PyObject *func, PyObject *arg) {
    PyObject *self, *result;
    PyCFunction cfunc;
    cfunc = PyCFunction_GET_FUNCTION(func);
    self = PyCFunction_GET_SELF(func);
    if (unlikely(Py_EnterRecursiveCall((char*)" while calling a Python object")))
        return NULL;
    result = cfunc(self, arg);
    Py_LeaveRecursiveCall();
    if (unlikely(!result) && unlikely(!PyErr_Occurred())) {
        PyErr_SetString(
            PyExc_SystemError,
            "NULL result without error in PyObject_Call");
    }
    return result;
}
#endif

/* PyObjectCallNoArg */
#if CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallNoArg(PyObject *func) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, NULL, 0);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_NOARGS)) {
            return __Pyx_PyObject_CallMethO(func, NULL);
        }
    }
    return __Pyx_PyObject_Call(func, __pyx_empty_tuple, NULL);
}
#endif

/* PyCFunctionFastCall */
  #if CYTHON_FAST_PYCCALL
static CYTHON_INLINE PyObject * __Pyx_PyCFunction_FastCall(PyObject *func_obj, PyObject **args, Py_ssize_t nargs) {
    PyCFunctionObject *func = (PyCFunctionObject*)func_obj;
    PyCFunction meth = PyCFunction_GET_FUNCTION(func);
    PyObject *self = PyCFunction_GET_SELF(func);
    assert(PyCFunction_Check(func));
    assert(METH_FASTCALL == (PyCFunction_GET_FLAGS(func) & ~(METH_CLASS | METH_STATIC | METH_COEXIST)));
    assert(nargs >= 0);
    assert(nargs == 0 || args != NULL);
    /* _PyCFunction_FastCallDict() must not be called with an exception set,
       because it may clear it (directly or indirectly) and so the
       caller loses its exception */
    assert(!PyErr_Occurred());
    return (*((__Pyx_PyCFunctionFast)meth)) (self, args, nargs, NULL);
}
#endif  // CYTHON_FAST_PYCCALL

/* PyObjectCallOneArg */
  #if CYTHON_COMPILING_IN_CPYTHON
static PyObject* __Pyx__PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_New(1);
    if (unlikely(!args)) return NULL;
    Py_INCREF(arg);
    PyTuple_SET_ITEM(args, 0, arg);
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
#if CYTHON_FAST_PYCALL
    if (PyFunction_Check(func)) {
        return __Pyx_PyFunction_FastCall(func, &arg, 1);
    }
#endif
#ifdef __Pyx_CyFunction_USED
    if (likely(PyCFunction_Check(func) || PyObject_TypeCheck(func, __pyx_CyFunctionType))) {
#else
    if (likely(PyCFunction_Check(func))) {
#endif
        if (likely(PyCFunction_GET_FLAGS(func) & METH_O)) {
            return __Pyx_PyObject_CallMethO(func, arg);
#if CYTHON_FAST_PYCCALL
        } else if (PyCFunction_GET_FLAGS(func) & METH_FASTCALL) {
            return __Pyx_PyCFunction_FastCall(func, &arg, 1);
#endif
        }
    }
    return __Pyx__PyObject_CallOneArg(func, arg);
}
#else
static CYTHON_INLINE PyObject* __Pyx_PyObject_CallOneArg(PyObject *func, PyObject *arg) {
    PyObject *result;
    PyObject *args = PyTuple_Pack(1, arg);
    if (unlikely(!args)) return NULL;
    result = __Pyx_PyObject_Call(func, args, NULL);
    Py_DECREF(args);
    return result;
}
#endif

/* PyObjectCallMethod0 */
    static PyObject* __Pyx_PyObject_CallMethod0(PyObject* obj, PyObject* method_name) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto bad;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *function = PyMethod_GET_FUNCTION(method);
            result = __Pyx_PyObject_CallOneArg(function, self);
            Py_DECREF(method);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallNoArg(method);
    Py_DECREF(method);
bad:
    return result;
}

/* RaiseNeedMoreValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseNeedMoreValuesError(Py_ssize_t index) {
    PyErr_Format(PyExc_ValueError,
                 "need more than %" CYTHON_FORMAT_SSIZE_T "d value%.1s to unpack",
                 index, (index == 1) ? "" : "s");
}

/* RaiseTooManyValuesToUnpack */
    static CYTHON_INLINE void __Pyx_RaiseTooManyValuesError(Py_ssize_t expected) {
    PyErr_Format(PyExc_ValueError,
                 "too many values to unpack (expected %" CYTHON_FORMAT_SSIZE_T "d)", expected);
}

/* UnpackItemEndCheck */
    static int __Pyx_IternextUnpackEndCheck(PyObject *retval, Py_ssize_t expected) {
    if (unlikely(retval)) {
        Py_DECREF(retval);
        __Pyx_RaiseTooManyValuesError(expected);
        return -1;
    } else {
        return __Pyx_IterFinish();
    }
    return 0;
}

/* RaiseNoneIterError */
    static CYTHON_INLINE void __Pyx_RaiseNoneNotIterableError(void) {
    PyErr_SetString(PyExc_TypeError, "'NoneType' object is not iterable");
}

/* UnpackTupleError */
    static void __Pyx_UnpackTupleError(PyObject *t, Py_ssize_t index) {
    if (t == Py_None) {
      __Pyx_RaiseNoneNotIterableError();
    } else if (PyTuple_GET_SIZE(t) < index) {
      __Pyx_RaiseNeedMoreValuesError(PyTuple_GET_SIZE(t));
    } else {
      __Pyx_RaiseTooManyValuesError(index);
    }
}

/* UnpackTuple2 */
    static CYTHON_INLINE int __Pyx_unpack_tuple2(PyObject* tuple, PyObject** pvalue1, PyObject** pvalue2,
                                             int is_tuple, int has_known_size, int decref_tuple) {
    Py_ssize_t index;
    PyObject *value1 = NULL, *value2 = NULL, *iter = NULL;
    if (!is_tuple && unlikely(!PyTuple_Check(tuple))) {
        iternextfunc iternext;
        iter = PyObject_GetIter(tuple);
        if (unlikely(!iter)) goto bad;
        if (decref_tuple) { Py_DECREF(tuple); tuple = NULL; }
        iternext = Py_TYPE(iter)->tp_iternext;
        value1 = iternext(iter); if (unlikely(!value1)) { index = 0; goto unpacking_failed; }
        value2 = iternext(iter); if (unlikely(!value2)) { index = 1; goto unpacking_failed; }
        if (!has_known_size && unlikely(__Pyx_IternextUnpackEndCheck(iternext(iter), 2))) goto bad;
        Py_DECREF(iter);
    } else {
        if (!has_known_size && unlikely(PyTuple_GET_SIZE(tuple) != 2)) {
            __Pyx_UnpackTupleError(tuple, 2);
            goto bad;
        }
#if CYTHON_COMPILING_IN_PYPY
        value1 = PySequence_ITEM(tuple, 0);
        if (unlikely(!value1)) goto bad;
        value2 = PySequence_ITEM(tuple, 1);
        if (unlikely(!value2)) goto bad;
#else
        value1 = PyTuple_GET_ITEM(tuple, 0);
        value2 = PyTuple_GET_ITEM(tuple, 1);
        Py_INCREF(value1);
        Py_INCREF(value2);
#endif
        if (decref_tuple) { Py_DECREF(tuple); }
    }
    *pvalue1 = value1;
    *pvalue2 = value2;
    return 0;
unpacking_failed:
    if (!has_known_size && __Pyx_IterFinish() == 0)
        __Pyx_RaiseNeedMoreValuesError(index);
bad:
    Py_XDECREF(iter);
    Py_XDECREF(value1);
    Py_XDECREF(value2);
    if (decref_tuple) { Py_XDECREF(tuple); }
    return -1;
}

/* dict_iter */
    static CYTHON_INLINE PyObject* __Pyx_dict_iterator(PyObject* iterable, int is_dict, PyObject* method_name,
                                                   Py_ssize_t* p_orig_length, int* p_source_is_dict) {
    is_dict = is_dict || likely(PyDict_CheckExact(iterable));
    *p_source_is_dict = is_dict;
#if !CYTHON_COMPILING_IN_PYPY
    if (is_dict) {
        *p_orig_length = PyDict_Size(iterable);
        Py_INCREF(iterable);
        return iterable;
    }
#endif
    *p_orig_length = 0;
    if (method_name) {
        PyObject* iter;
        iterable = __Pyx_PyObject_CallMethod0(iterable, method_name);
        if (!iterable)
            return NULL;
#if !CYTHON_COMPILING_IN_PYPY
        if (PyTuple_CheckExact(iterable) || PyList_CheckExact(iterable))
            return iterable;
#endif
        iter = PyObject_GetIter(iterable);
        Py_DECREF(iterable);
        return iter;
    }
    return PyObject_GetIter(iterable);
}
static CYTHON_INLINE int __Pyx_dict_iter_next(
        PyObject* iter_obj, CYTHON_NCP_UNUSED Py_ssize_t orig_length, CYTHON_NCP_UNUSED Py_ssize_t* ppos,
        PyObject** pkey, PyObject** pvalue, PyObject** pitem, int source_is_dict) {
    PyObject* next_item;
#if !CYTHON_COMPILING_IN_PYPY
    if (source_is_dict) {
        PyObject *key, *value;
        if (unlikely(orig_length != PyDict_Size(iter_obj))) {
            PyErr_SetString(PyExc_RuntimeError, "dictionary changed size during iteration");
            return -1;
        }
        if (unlikely(!PyDict_Next(iter_obj, ppos, &key, &value))) {
            return 0;
        }
        if (pitem) {
            PyObject* tuple = PyTuple_New(2);
            if (unlikely(!tuple)) {
                return -1;
            }
            Py_INCREF(key);
            Py_INCREF(value);
            PyTuple_SET_ITEM(tuple, 0, key);
            PyTuple_SET_ITEM(tuple, 1, value);
            *pitem = tuple;
        } else {
            if (pkey) {
                Py_INCREF(key);
                *pkey = key;
            }
            if (pvalue) {
                Py_INCREF(value);
                *pvalue = value;
            }
        }
        return 1;
    } else if (PyTuple_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyTuple_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyTuple_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else if (PyList_CheckExact(iter_obj)) {
        Py_ssize_t pos = *ppos;
        if (unlikely(pos >= PyList_GET_SIZE(iter_obj))) return 0;
        *ppos = pos + 1;
        next_item = PyList_GET_ITEM(iter_obj, pos);
        Py_INCREF(next_item);
    } else
#endif
    {
        next_item = PyIter_Next(iter_obj);
        if (unlikely(!next_item)) {
            return __Pyx_IterFinish();
        }
    }
    if (pitem) {
        *pitem = next_item;
    } else if (pkey && pvalue) {
        if (__Pyx_unpack_tuple2(next_item, pkey, pvalue, source_is_dict, source_is_dict, 1))
            return -1;
    } else if (pkey) {
        *pkey = next_item;
    } else {
        *pvalue = next_item;
    }
    return 1;
}

/* GetModuleGlobalName */
    static CYTHON_INLINE PyObject *__Pyx_GetModuleGlobalName(PyObject *name) {
    PyObject *result;
#if !CYTHON_AVOID_BORROWED_REFS
    result = PyDict_GetItem(__pyx_d, name);
    if (likely(result)) {
        Py_INCREF(result);
    } else {
#else
    result = PyObject_GetItem(__pyx_d, name);
    if (!result) {
        PyErr_Clear();
#endif
        result = __Pyx_GetBuiltinName(name);
    }
    return result;
}

/* RaiseArgTupleInvalid */
      static void __Pyx_RaiseArgtupleInvalid(
    const char* func_name,
    int exact,
    Py_ssize_t num_min,
    Py_ssize_t num_max,
    Py_ssize_t num_found)
{
    Py_ssize_t num_expected;
    const char *more_or_less;
    if (num_found < num_min) {
        num_expected = num_min;
        more_or_less = "at least";
    } else {
        num_expected = num_max;
        more_or_less = "at most";
    }
    if (exact) {
        more_or_less = "exactly";
    }
    PyErr_Format(PyExc_TypeError,
                 "%.200s() takes %.8s %" CYTHON_FORMAT_SSIZE_T "d positional argument%.1s (%" CYTHON_FORMAT_SSIZE_T "d given)",
                 func_name, more_or_less, num_expected,
                 (num_expected == 1) ? "" : "s", num_found);
}

/* RaiseDoubleKeywords */
      static void __Pyx_RaiseDoubleKeywordsError(
    const char* func_name,
    PyObject* kw_name)
{
    PyErr_Format(PyExc_TypeError,
        #if PY_MAJOR_VERSION >= 3
        "%s() got multiple values for keyword argument '%U'", func_name, kw_name);
        #else
        "%s() got multiple values for keyword argument '%s'", func_name,
        PyString_AsString(kw_name));
        #endif
}

/* ParseKeywords */
      static int __Pyx_ParseOptionalKeywords(
    PyObject *kwds,
    PyObject **argnames[],
    PyObject *kwds2,
    PyObject *values[],
    Py_ssize_t num_pos_args,
    const char* function_name)
{
    PyObject *key = 0, *value = 0;
    Py_ssize_t pos = 0;
    PyObject*** name;
    PyObject*** first_kw_arg = argnames + num_pos_args;
    while (PyDict_Next(kwds, &pos, &key, &value)) {
        name = first_kw_arg;
        while (*name && (**name != key)) name++;
        if (*name) {
            values[name-argnames] = value;
            continue;
        }
        name = first_kw_arg;
        #if PY_MAJOR_VERSION < 3
        if (likely(PyString_CheckExact(key)) || likely(PyString_Check(key))) {
            while (*name) {
                if ((CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**name) == PyString_GET_SIZE(key))
                        && _PyString_Eq(**name, key)) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    if ((**argname == key) || (
                            (CYTHON_COMPILING_IN_PYPY || PyString_GET_SIZE(**argname) == PyString_GET_SIZE(key))
                             && _PyString_Eq(**argname, key))) {
                        goto arg_passed_twice;
                    }
                    argname++;
                }
            }
        } else
        #endif
        if (likely(PyUnicode_Check(key))) {
            while (*name) {
                int cmp = (**name == key) ? 0 :
                #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                    (PyUnicode_GET_SIZE(**name) != PyUnicode_GET_SIZE(key)) ? 1 :
                #endif
                    PyUnicode_Compare(**name, key);
                if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                if (cmp == 0) {
                    values[name-argnames] = value;
                    break;
                }
                name++;
            }
            if (*name) continue;
            else {
                PyObject*** argname = argnames;
                while (argname != first_kw_arg) {
                    int cmp = (**argname == key) ? 0 :
                    #if !CYTHON_COMPILING_IN_PYPY && PY_MAJOR_VERSION >= 3
                        (PyUnicode_GET_SIZE(**argname) != PyUnicode_GET_SIZE(key)) ? 1 :
                    #endif
                        PyUnicode_Compare(**argname, key);
                    if (cmp < 0 && unlikely(PyErr_Occurred())) goto bad;
                    if (cmp == 0) goto arg_passed_twice;
                    argname++;
                }
            }
        } else
            goto invalid_keyword_type;
        if (kwds2) {
            if (unlikely(PyDict_SetItem(kwds2, key, value))) goto bad;
        } else {
            goto invalid_keyword;
        }
    }
    return 0;
arg_passed_twice:
    __Pyx_RaiseDoubleKeywordsError(function_name, key);
    goto bad;
invalid_keyword_type:
    PyErr_Format(PyExc_TypeError,
        "%.200s() keywords must be strings", function_name);
    goto bad;
invalid_keyword:
    PyErr_Format(PyExc_TypeError,
    #if PY_MAJOR_VERSION < 3
        "%.200s() got an unexpected keyword argument '%.200s'",
        function_name, PyString_AsString(key));
    #else
        "%s() got an unexpected keyword argument '%U'",
        function_name, key);
    #endif
bad:
    return -1;
}

/* StringJoin */
      #if !CYTHON_COMPILING_IN_CPYTHON
static CYTHON_INLINE PyObject* __Pyx_PyBytes_Join(PyObject* sep, PyObject* values) {
    return PyObject_CallMethodObjArgs(sep, __pyx_n_s_join, values, NULL);
}
#endif

/* PyErrFetchRestore */
      #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx_ErrRestoreInState(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->curexc_type;
    tmp_value = tstate->curexc_value;
    tmp_tb = tstate->curexc_traceback;
    tstate->curexc_type = type;
    tstate->curexc_value = value;
    tstate->curexc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
static CYTHON_INLINE void __Pyx_ErrFetchInState(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->curexc_type;
    *value = tstate->curexc_value;
    *tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
}
#endif

/* RaiseException */
      #if PY_MAJOR_VERSION < 3
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb,
                        CYTHON_UNUSED PyObject *cause) {
    __Pyx_PyThreadState_declare
    Py_XINCREF(type);
    if (!value || value == Py_None)
        value = NULL;
    else
        Py_INCREF(value);
    if (!tb || tb == Py_None)
        tb = NULL;
    else {
        Py_INCREF(tb);
        if (!PyTraceBack_Check(tb)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: arg 3 must be a traceback or None");
            goto raise_error;
        }
    }
    if (PyType_Check(type)) {
#if CYTHON_COMPILING_IN_PYPY
        if (!value) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#endif
        PyErr_NormalizeException(&type, &value, &tb);
    } else {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto raise_error;
        }
        value = type;
        type = (PyObject*) Py_TYPE(type);
        Py_INCREF(type);
        if (!PyType_IsSubtype((PyTypeObject *)type, (PyTypeObject *)PyExc_BaseException)) {
            PyErr_SetString(PyExc_TypeError,
                "raise: exception class must be a subclass of BaseException");
            goto raise_error;
        }
    }
    __Pyx_PyThreadState_assign
    __Pyx_ErrRestore(type, value, tb);
    return;
raise_error:
    Py_XDECREF(value);
    Py_XDECREF(type);
    Py_XDECREF(tb);
    return;
}
#else
static void __Pyx_Raise(PyObject *type, PyObject *value, PyObject *tb, PyObject *cause) {
    PyObject* owned_instance = NULL;
    if (tb == Py_None) {
        tb = 0;
    } else if (tb && !PyTraceBack_Check(tb)) {
        PyErr_SetString(PyExc_TypeError,
            "raise: arg 3 must be a traceback or None");
        goto bad;
    }
    if (value == Py_None)
        value = 0;
    if (PyExceptionInstance_Check(type)) {
        if (value) {
            PyErr_SetString(PyExc_TypeError,
                "instance exception may not have a separate value");
            goto bad;
        }
        value = type;
        type = (PyObject*) Py_TYPE(value);
    } else if (PyExceptionClass_Check(type)) {
        PyObject *instance_class = NULL;
        if (value && PyExceptionInstance_Check(value)) {
            instance_class = (PyObject*) Py_TYPE(value);
            if (instance_class != type) {
                int is_subclass = PyObject_IsSubclass(instance_class, type);
                if (!is_subclass) {
                    instance_class = NULL;
                } else if (unlikely(is_subclass == -1)) {
                    goto bad;
                } else {
                    type = instance_class;
                }
            }
        }
        if (!instance_class) {
            PyObject *args;
            if (!value)
                args = PyTuple_New(0);
            else if (PyTuple_Check(value)) {
                Py_INCREF(value);
                args = value;
            } else
                args = PyTuple_Pack(1, value);
            if (!args)
                goto bad;
            owned_instance = PyObject_Call(type, args, NULL);
            Py_DECREF(args);
            if (!owned_instance)
                goto bad;
            value = owned_instance;
            if (!PyExceptionInstance_Check(value)) {
                PyErr_Format(PyExc_TypeError,
                             "calling %R should have returned an instance of "
                             "BaseException, not %R",
                             type, Py_TYPE(value));
                goto bad;
            }
        }
    } else {
        PyErr_SetString(PyExc_TypeError,
            "raise: exception class must be a subclass of BaseException");
        goto bad;
    }
#if PY_VERSION_HEX >= 0x03030000
    if (cause) {
#else
    if (cause && cause != Py_None) {
#endif
        PyObject *fixed_cause;
        if (cause == Py_None) {
            fixed_cause = NULL;
        } else if (PyExceptionClass_Check(cause)) {
            fixed_cause = PyObject_CallObject(cause, NULL);
            if (fixed_cause == NULL)
                goto bad;
        } else if (PyExceptionInstance_Check(cause)) {
            fixed_cause = cause;
            Py_INCREF(fixed_cause);
        } else {
            PyErr_SetString(PyExc_TypeError,
                            "exception causes must derive from "
                            "BaseException");
            goto bad;
        }
        PyException_SetCause(value, fixed_cause);
    }
    PyErr_SetObject(type, value);
    if (tb) {
#if CYTHON_COMPILING_IN_PYPY
        PyObject *tmp_type, *tmp_value, *tmp_tb;
        PyErr_Fetch(&tmp_type, &tmp_value, &tmp_tb);
        Py_INCREF(tb);
        PyErr_Restore(tmp_type, tmp_value, tb);
        Py_XDECREF(tmp_tb);
#else
        PyThreadState *tstate = PyThreadState_GET();
        PyObject* tmp_tb = tstate->curexc_traceback;
        if (tb != tmp_tb) {
            Py_INCREF(tb);
            tstate->curexc_traceback = tb;
            Py_XDECREF(tmp_tb);
        }
#endif
    }
bad:
    Py_XDECREF(owned_instance);
    return;
}
#endif

/* ExtTypeTest */
        static CYTHON_INLINE int __Pyx_TypeTest(PyObject *obj, PyTypeObject *type) {
    if (unlikely(!type)) {
        PyErr_SetString(PyExc_SystemError, "Missing type object");
        return 0;
    }
    if (likely(PyObject_TypeCheck(obj, type)))
        return 1;
    PyErr_Format(PyExc_TypeError, "Cannot convert %.200s to %.200s",
                 Py_TYPE(obj)->tp_name, type->tp_name);
    return 0;
}

/* BytesEquals */
        static CYTHON_INLINE int __Pyx_PyBytes_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
    if (s1 == s2) {
        return (equals == Py_EQ);
    } else if (PyBytes_CheckExact(s1) & PyBytes_CheckExact(s2)) {
        const char *ps1, *ps2;
        Py_ssize_t length = PyBytes_GET_SIZE(s1);
        if (length != PyBytes_GET_SIZE(s2))
            return (equals == Py_NE);
        ps1 = PyBytes_AS_STRING(s1);
        ps2 = PyBytes_AS_STRING(s2);
        if (ps1[0] != ps2[0]) {
            return (equals == Py_NE);
        } else if (length == 1) {
            return (equals == Py_EQ);
        } else {
            int result = memcmp(ps1, ps2, (size_t)length);
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & PyBytes_CheckExact(s2)) {
        return (equals == Py_NE);
    } else if ((s2 == Py_None) & PyBytes_CheckExact(s1)) {
        return (equals == Py_NE);
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
#endif
}

/* UnicodeEquals */
        static CYTHON_INLINE int __Pyx_PyUnicode_Equals(PyObject* s1, PyObject* s2, int equals) {
#if CYTHON_COMPILING_IN_PYPY
    return PyObject_RichCompareBool(s1, s2, equals);
#else
#if PY_MAJOR_VERSION < 3
    PyObject* owned_ref = NULL;
#endif
    int s1_is_unicode, s2_is_unicode;
    if (s1 == s2) {
        goto return_eq;
    }
    s1_is_unicode = PyUnicode_CheckExact(s1);
    s2_is_unicode = PyUnicode_CheckExact(s2);
#if PY_MAJOR_VERSION < 3
    if ((s1_is_unicode & (!s2_is_unicode)) && PyString_CheckExact(s2)) {
        owned_ref = PyUnicode_FromObject(s2);
        if (unlikely(!owned_ref))
            return -1;
        s2 = owned_ref;
        s2_is_unicode = 1;
    } else if ((s2_is_unicode & (!s1_is_unicode)) && PyString_CheckExact(s1)) {
        owned_ref = PyUnicode_FromObject(s1);
        if (unlikely(!owned_ref))
            return -1;
        s1 = owned_ref;
        s1_is_unicode = 1;
    } else if (((!s2_is_unicode) & (!s1_is_unicode))) {
        return __Pyx_PyBytes_Equals(s1, s2, equals);
    }
#endif
    if (s1_is_unicode & s2_is_unicode) {
        Py_ssize_t length;
        int kind;
        void *data1, *data2;
        if (unlikely(__Pyx_PyUnicode_READY(s1) < 0) || unlikely(__Pyx_PyUnicode_READY(s2) < 0))
            return -1;
        length = __Pyx_PyUnicode_GET_LENGTH(s1);
        if (length != __Pyx_PyUnicode_GET_LENGTH(s2)) {
            goto return_ne;
        }
        kind = __Pyx_PyUnicode_KIND(s1);
        if (kind != __Pyx_PyUnicode_KIND(s2)) {
            goto return_ne;
        }
        data1 = __Pyx_PyUnicode_DATA(s1);
        data2 = __Pyx_PyUnicode_DATA(s2);
        if (__Pyx_PyUnicode_READ(kind, data1, 0) != __Pyx_PyUnicode_READ(kind, data2, 0)) {
            goto return_ne;
        } else if (length == 1) {
            goto return_eq;
        } else {
            int result = memcmp(data1, data2, (size_t)(length * kind));
            #if PY_MAJOR_VERSION < 3
            Py_XDECREF(owned_ref);
            #endif
            return (equals == Py_EQ) ? (result == 0) : (result != 0);
        }
    } else if ((s1 == Py_None) & s2_is_unicode) {
        goto return_ne;
    } else if ((s2 == Py_None) & s1_is_unicode) {
        goto return_ne;
    } else {
        int result;
        PyObject* py_result = PyObject_RichCompare(s1, s2, equals);
        if (!py_result)
            return -1;
        result = __Pyx_PyObject_IsTrue(py_result);
        Py_DECREF(py_result);
        return result;
    }
return_eq:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_EQ);
return_ne:
    #if PY_MAJOR_VERSION < 3
    Py_XDECREF(owned_ref);
    #endif
    return (equals == Py_NE);
#endif
}

/* PyIntBinop */
        #if !CYTHON_COMPILING_IN_PYPY
static PyObject* __Pyx_PyInt_EqObjC(PyObject *op1, PyObject *op2, CYTHON_UNUSED long intval, CYTHON_UNUSED int inplace) {
    if (op1 == op2) {
        Py_RETURN_TRUE;
    }
    #if PY_MAJOR_VERSION < 3
    if (likely(PyInt_CheckExact(op1))) {
        const long b = intval;
        long a = PyInt_AS_LONG(op1);
        if (a == b) {
            Py_RETURN_TRUE;
        } else {
            Py_RETURN_FALSE;
        }
    }
    #endif
    #if CYTHON_USE_PYLONG_INTERNALS
    if (likely(PyLong_CheckExact(op1))) {
        const long b = intval;
        long a;
        const digit* digits = ((PyLongObject*)op1)->ob_digit;
        const Py_ssize_t size = Py_SIZE(op1);
        if (likely(__Pyx_sst_abs(size) <= 1)) {
            a = likely(size) ? digits[0] : 0;
            if (size == -1) a = -a;
        } else {
            switch (size) {
                case -2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 2:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        a = (long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 3:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        a = (long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case -4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                case 4:
                    if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                        a = (long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0]));
                        break;
                    }
                #if PyLong_SHIFT < 30 && PyLong_SHIFT != 15
                default: return PyLong_Type.tp_richcompare(op1, op2, Py_EQ);
                #else
                default: Py_RETURN_FALSE;
                #endif
            }
        }
            if (a == b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    #endif
    if (PyFloat_CheckExact(op1)) {
        const long b = intval;
        double a = PyFloat_AS_DOUBLE(op1);
            if ((double)a == (double)b) {
                Py_RETURN_TRUE;
            } else {
                Py_RETURN_FALSE;
            }
    }
    return PyObject_RichCompare(op1, op2, Py_EQ);
}
#endif

/* None */
        static CYTHON_INLINE void __Pyx_RaiseClosureNameError(const char *varname) {
    PyErr_Format(PyExc_NameError, "free variable '%s' referenced before assignment in enclosing scope", varname);
}

/* GetItemInt */
        static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Generic(PyObject *o, PyObject* j) {
    PyObject *r;
    if (!j) return NULL;
    r = PyObject_GetItem(o, j);
    Py_DECREF(j);
    return r;
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_List_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyList_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyList_GET_SIZE(o)))) {
        PyObject *r = PyList_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Tuple_Fast(PyObject *o, Py_ssize_t i,
                                                              CYTHON_NCP_UNUSED int wraparound,
                                                              CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
    if (wraparound & unlikely(i < 0)) i += PyTuple_GET_SIZE(o);
    if ((!boundscheck) || likely((0 <= i) & (i < PyTuple_GET_SIZE(o)))) {
        PyObject *r = PyTuple_GET_ITEM(o, i);
        Py_INCREF(r);
        return r;
    }
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
#else
    return PySequence_GetItem(o, i);
#endif
}
static CYTHON_INLINE PyObject *__Pyx_GetItemInt_Fast(PyObject *o, Py_ssize_t i, int is_list,
                                                     CYTHON_NCP_UNUSED int wraparound,
                                                     CYTHON_NCP_UNUSED int boundscheck) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS && CYTHON_USE_TYPE_SLOTS
    if (is_list || PyList_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyList_GET_SIZE(o);
        if ((!boundscheck) || (likely((n >= 0) & (n < PyList_GET_SIZE(o))))) {
            PyObject *r = PyList_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    }
    else if (PyTuple_CheckExact(o)) {
        Py_ssize_t n = ((!wraparound) | likely(i >= 0)) ? i : i + PyTuple_GET_SIZE(o);
        if ((!boundscheck) || likely((n >= 0) & (n < PyTuple_GET_SIZE(o)))) {
            PyObject *r = PyTuple_GET_ITEM(o, n);
            Py_INCREF(r);
            return r;
        }
    } else {
        PySequenceMethods *m = Py_TYPE(o)->tp_as_sequence;
        if (likely(m && m->sq_item)) {
            if (wraparound && unlikely(i < 0) && likely(m->sq_length)) {
                Py_ssize_t l = m->sq_length(o);
                if (likely(l >= 0)) {
                    i += l;
                } else {
                    if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                        return NULL;
                    PyErr_Clear();
                }
            }
            return m->sq_item(o, i);
        }
    }
#else
    if (is_list || PySequence_Check(o)) {
        return PySequence_GetItem(o, i);
    }
#endif
    return __Pyx_GetItemInt_Generic(o, PyInt_FromSsize_t(i));
}

/* SaveResetException */
        #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSave(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    *type = tstate->exc_type;
    *value = tstate->exc_value;
    *tb = tstate->exc_traceback;
    Py_XINCREF(*type);
    Py_XINCREF(*value);
    Py_XINCREF(*tb);
}
static CYTHON_INLINE void __Pyx__ExceptionReset(PyThreadState *tstate, PyObject *type, PyObject *value, PyObject *tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = type;
    tstate->exc_value = value;
    tstate->exc_traceback = tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
}
#endif

/* GetException */
        #if CYTHON_FAST_THREAD_STATE
static int __Pyx__GetException(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
#else
static int __Pyx_GetException(PyObject **type, PyObject **value, PyObject **tb) {
#endif
    PyObject *local_type, *local_value, *local_tb;
#if CYTHON_FAST_THREAD_STATE
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    local_type = tstate->curexc_type;
    local_value = tstate->curexc_value;
    local_tb = tstate->curexc_traceback;
    tstate->curexc_type = 0;
    tstate->curexc_value = 0;
    tstate->curexc_traceback = 0;
#else
    PyErr_Fetch(&local_type, &local_value, &local_tb);
#endif
    PyErr_NormalizeException(&local_type, &local_value, &local_tb);
#if CYTHON_FAST_THREAD_STATE
    if (unlikely(tstate->curexc_type))
#else
    if (unlikely(PyErr_Occurred()))
#endif
        goto bad;
    #if PY_MAJOR_VERSION >= 3
    if (local_tb) {
        if (unlikely(PyException_SetTraceback(local_value, local_tb) < 0))
            goto bad;
    }
    #endif
    Py_XINCREF(local_tb);
    Py_XINCREF(local_type);
    Py_XINCREF(local_value);
    *type = local_type;
    *value = local_value;
    *tb = local_tb;
#if CYTHON_FAST_THREAD_STATE
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = local_type;
    tstate->exc_value = local_value;
    tstate->exc_traceback = local_tb;
    Py_XDECREF(tmp_type);
    Py_XDECREF(tmp_value);
    Py_XDECREF(tmp_tb);
#else
    PyErr_SetExcInfo(local_type, local_value, local_tb);
#endif
    return 0;
bad:
    *type = 0;
    *value = 0;
    *tb = 0;
    Py_XDECREF(local_type);
    Py_XDECREF(local_value);
    Py_XDECREF(local_tb);
    return -1;
}

/* SwapException */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE void __Pyx__ExceptionSwap(PyThreadState *tstate, PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    tmp_type = tstate->exc_type;
    tmp_value = tstate->exc_value;
    tmp_tb = tstate->exc_traceback;
    tstate->exc_type = *type;
    tstate->exc_value = *value;
    tstate->exc_traceback = *tb;
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#else
static CYTHON_INLINE void __Pyx_ExceptionSwap(PyObject **type, PyObject **value, PyObject **tb) {
    PyObject *tmp_type, *tmp_value, *tmp_tb;
    PyErr_GetExcInfo(&tmp_type, &tmp_value, &tmp_tb);
    PyErr_SetExcInfo(*type, *value, *tb);
    *type = tmp_type;
    *value = tmp_value;
    *tb = tmp_tb;
}
#endif

/* GetAttr */
          static CYTHON_INLINE PyObject *__Pyx_GetAttr(PyObject *o, PyObject *n) {
#if CYTHON_COMPILING_IN_CPYTHON
#if PY_MAJOR_VERSION >= 3
    if (likely(PyUnicode_Check(n)))
#else
    if (likely(PyString_Check(n)))
#endif
        return __Pyx_PyObject_GetAttrStr(o, n);
#endif
    return PyObject_GetAttr(o, n);
}

/* SliceObject */
          static CYTHON_INLINE PyObject* __Pyx_PyObject_GetSlice(PyObject* obj,
        Py_ssize_t cstart, Py_ssize_t cstop,
        PyObject** _py_start, PyObject** _py_stop, PyObject** _py_slice,
        int has_cstart, int has_cstop, CYTHON_UNUSED int wraparound) {
#if CYTHON_USE_TYPE_SLOTS
    PyMappingMethods* mp;
#if PY_MAJOR_VERSION < 3
    PySequenceMethods* ms = Py_TYPE(obj)->tp_as_sequence;
    if (likely(ms && ms->sq_slice)) {
        if (!has_cstart) {
            if (_py_start && (*_py_start != Py_None)) {
                cstart = __Pyx_PyIndex_AsSsize_t(*_py_start);
                if ((cstart == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstart = 0;
        }
        if (!has_cstop) {
            if (_py_stop && (*_py_stop != Py_None)) {
                cstop = __Pyx_PyIndex_AsSsize_t(*_py_stop);
                if ((cstop == (Py_ssize_t)-1) && PyErr_Occurred()) goto bad;
            } else
                cstop = PY_SSIZE_T_MAX;
        }
        if (wraparound && unlikely((cstart < 0) | (cstop < 0)) && likely(ms->sq_length)) {
            Py_ssize_t l = ms->sq_length(obj);
            if (likely(l >= 0)) {
                if (cstop < 0) {
                    cstop += l;
                    if (cstop < 0) cstop = 0;
                }
                if (cstart < 0) {
                    cstart += l;
                    if (cstart < 0) cstart = 0;
                }
            } else {
                if (!PyErr_ExceptionMatches(PyExc_OverflowError))
                    goto bad;
                PyErr_Clear();
            }
        }
        return ms->sq_slice(obj, cstart, cstop);
    }
#endif
    mp = Py_TYPE(obj)->tp_as_mapping;
    if (likely(mp && mp->mp_subscript))
#endif
    {
        PyObject* result;
        PyObject *py_slice, *py_start, *py_stop;
        if (_py_slice) {
            py_slice = *_py_slice;
        } else {
            PyObject* owned_start = NULL;
            PyObject* owned_stop = NULL;
            if (_py_start) {
                py_start = *_py_start;
            } else {
                if (has_cstart) {
                    owned_start = py_start = PyInt_FromSsize_t(cstart);
                    if (unlikely(!py_start)) goto bad;
                } else
                    py_start = Py_None;
            }
            if (_py_stop) {
                py_stop = *_py_stop;
            } else {
                if (has_cstop) {
                    owned_stop = py_stop = PyInt_FromSsize_t(cstop);
                    if (unlikely(!py_stop)) {
                        Py_XDECREF(owned_start);
                        goto bad;
                    }
                } else
                    py_stop = Py_None;
            }
            py_slice = PySlice_New(py_start, py_stop, Py_None);
            Py_XDECREF(owned_start);
            Py_XDECREF(owned_stop);
            if (unlikely(!py_slice)) goto bad;
        }
#if CYTHON_USE_TYPE_SLOTS
        result = mp->mp_subscript(obj, py_slice);
#else
        result = PyObject_GetItem(obj, py_slice);
#endif
        if (!_py_slice) {
            Py_DECREF(py_slice);
        }
        return result;
    }
    PyErr_Format(PyExc_TypeError,
        "'%.200s' object is unsliceable", Py_TYPE(obj)->tp_name);
bad:
    return NULL;
}

/* PyErrExceptionMatches */
          #if CYTHON_FAST_THREAD_STATE
static CYTHON_INLINE int __Pyx_PyErr_ExceptionMatchesInState(PyThreadState* tstate, PyObject* err) {
    PyObject *exc_type = tstate->curexc_type;
    if (exc_type == err) return 1;
    if (unlikely(!exc_type)) return 0;
    return PyErr_GivenExceptionMatches(exc_type, err);
}
#endif

/* SetVTable */
          static int __Pyx_SetVtable(PyObject *dict, void *vtable) {
#if PY_VERSION_HEX >= 0x02070000
    PyObject *ob = PyCapsule_New(vtable, 0, 0);
#else
    PyObject *ob = PyCObject_FromVoidPtr(vtable, 0);
#endif
    if (!ob)
        goto bad;
    if (PyDict_SetItem(dict, __pyx_n_s_pyx_vtable, ob) < 0)
        goto bad;
    Py_DECREF(ob);
    return 0;
bad:
    Py_XDECREF(ob);
    return -1;
}

/* GetVTable */
          static void* __Pyx_GetVtable(PyObject *dict) {
    void* ptr;
    PyObject *ob = PyObject_GetItem(dict, __pyx_n_s_pyx_vtable);
    if (!ob)
        goto bad;
#if PY_VERSION_HEX >= 0x02070000
    ptr = PyCapsule_GetPointer(ob, 0);
#else
    ptr = PyCObject_AsVoidPtr(ob);
#endif
    if (!ptr && !PyErr_Occurred())
        PyErr_SetString(PyExc_RuntimeError, "invalid vtable found for imported type");
    Py_DECREF(ob);
    return ptr;
bad:
    Py_XDECREF(ob);
    return NULL;
}

/* Import */
          static PyObject *__Pyx_Import(PyObject *name, PyObject *from_list, int level) {
    PyObject *empty_list = 0;
    PyObject *module = 0;
    PyObject *global_dict = 0;
    PyObject *empty_dict = 0;
    PyObject *list;
    #if PY_VERSION_HEX < 0x03030000
    PyObject *py_import;
    py_import = __Pyx_PyObject_GetAttrStr(__pyx_b, __pyx_n_s_import);
    if (!py_import)
        goto bad;
    #endif
    if (from_list)
        list = from_list;
    else {
        empty_list = PyList_New(0);
        if (!empty_list)
            goto bad;
        list = empty_list;
    }
    global_dict = PyModule_GetDict(__pyx_m);
    if (!global_dict)
        goto bad;
    empty_dict = PyDict_New();
    if (!empty_dict)
        goto bad;
    {
        #if PY_MAJOR_VERSION >= 3
        if (level == -1) {
            if (strchr(__Pyx_MODULE_NAME, '.')) {
                #if PY_VERSION_HEX < 0x03030000
                PyObject *py_level = PyInt_FromLong(1);
                if (!py_level)
                    goto bad;
                module = PyObject_CallFunctionObjArgs(py_import,
                    name, global_dict, empty_dict, list, py_level, NULL);
                Py_DECREF(py_level);
                #else
                module = PyImport_ImportModuleLevelObject(
                    name, global_dict, empty_dict, list, 1);
                #endif
                if (!module) {
                    if (!PyErr_ExceptionMatches(PyExc_ImportError))
                        goto bad;
                    PyErr_Clear();
                }
            }
            level = 0;
        }
        #endif
        if (!module) {
            #if PY_VERSION_HEX < 0x03030000
            PyObject *py_level = PyInt_FromLong(level);
            if (!py_level)
                goto bad;
            module = PyObject_CallFunctionObjArgs(py_import,
                name, global_dict, empty_dict, list, py_level, NULL);
            Py_DECREF(py_level);
            #else
            module = PyImport_ImportModuleLevelObject(
                name, global_dict, empty_dict, list, level);
            #endif
        }
    }
bad:
    #if PY_VERSION_HEX < 0x03030000
    Py_XDECREF(py_import);
    #endif
    Py_XDECREF(empty_list);
    Py_XDECREF(empty_dict);
    return module;
}

/* ImportFrom */
          static PyObject* __Pyx_ImportFrom(PyObject* module, PyObject* name) {
    PyObject* value = __Pyx_PyObject_GetAttrStr(module, name);
    if (unlikely(!value) && PyErr_ExceptionMatches(PyExc_AttributeError)) {
        PyErr_Format(PyExc_ImportError,
        #if PY_MAJOR_VERSION < 3
            "cannot import name %.230s", PyString_AS_STRING(name));
        #else
            "cannot import name %S", name);
        #endif
    }
    return value;
}

/* CodeObjectCache */
          static int __pyx_bisect_code_objects(__Pyx_CodeObjectCacheEntry* entries, int count, int code_line) {
    int start = 0, mid = 0, end = count - 1;
    if (end >= 0 && code_line > entries[end].code_line) {
        return count;
    }
    while (start < end) {
        mid = start + (end - start) / 2;
        if (code_line < entries[mid].code_line) {
            end = mid;
        } else if (code_line > entries[mid].code_line) {
             start = mid + 1;
        } else {
            return mid;
        }
    }
    if (code_line <= entries[mid].code_line) {
        return mid;
    } else {
        return mid + 1;
    }
}
static PyCodeObject *__pyx_find_code_object(int code_line) {
    PyCodeObject* code_object;
    int pos;
    if (unlikely(!code_line) || unlikely(!__pyx_code_cache.entries)) {
        return NULL;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if (unlikely(pos >= __pyx_code_cache.count) || unlikely(__pyx_code_cache.entries[pos].code_line != code_line)) {
        return NULL;
    }
    code_object = __pyx_code_cache.entries[pos].code_object;
    Py_INCREF(code_object);
    return code_object;
}
static void __pyx_insert_code_object(int code_line, PyCodeObject* code_object) {
    int pos, i;
    __Pyx_CodeObjectCacheEntry* entries = __pyx_code_cache.entries;
    if (unlikely(!code_line)) {
        return;
    }
    if (unlikely(!entries)) {
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Malloc(64*sizeof(__Pyx_CodeObjectCacheEntry));
        if (likely(entries)) {
            __pyx_code_cache.entries = entries;
            __pyx_code_cache.max_count = 64;
            __pyx_code_cache.count = 1;
            entries[0].code_line = code_line;
            entries[0].code_object = code_object;
            Py_INCREF(code_object);
        }
        return;
    }
    pos = __pyx_bisect_code_objects(__pyx_code_cache.entries, __pyx_code_cache.count, code_line);
    if ((pos < __pyx_code_cache.count) && unlikely(__pyx_code_cache.entries[pos].code_line == code_line)) {
        PyCodeObject* tmp = entries[pos].code_object;
        entries[pos].code_object = code_object;
        Py_DECREF(tmp);
        return;
    }
    if (__pyx_code_cache.count == __pyx_code_cache.max_count) {
        int new_max = __pyx_code_cache.max_count + 64;
        entries = (__Pyx_CodeObjectCacheEntry*)PyMem_Realloc(
            __pyx_code_cache.entries, (size_t)new_max*sizeof(__Pyx_CodeObjectCacheEntry));
        if (unlikely(!entries)) {
            return;
        }
        __pyx_code_cache.entries = entries;
        __pyx_code_cache.max_count = new_max;
    }
    for (i=__pyx_code_cache.count; i>pos; i--) {
        entries[i] = entries[i-1];
    }
    entries[pos].code_line = code_line;
    entries[pos].code_object = code_object;
    __pyx_code_cache.count++;
    Py_INCREF(code_object);
}

/* AddTraceback */
          #include "compile.h"
#include "frameobject.h"
#include "traceback.h"
static PyCodeObject* __Pyx_CreateCodeObjectForTraceback(
            const char *funcname, int c_line,
            int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyObject *py_srcfile = 0;
    PyObject *py_funcname = 0;
    #if PY_MAJOR_VERSION < 3
    py_srcfile = PyString_FromString(filename);
    #else
    py_srcfile = PyUnicode_FromString(filename);
    #endif
    if (!py_srcfile) goto bad;
    if (c_line) {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #else
        py_funcname = PyUnicode_FromFormat( "%s (%s:%d)", funcname, __pyx_cfilenm, c_line);
        #endif
    }
    else {
        #if PY_MAJOR_VERSION < 3
        py_funcname = PyString_FromString(funcname);
        #else
        py_funcname = PyUnicode_FromString(funcname);
        #endif
    }
    if (!py_funcname) goto bad;
    py_code = __Pyx_PyCode_New(
        0,
        0,
        0,
        0,
        0,
        __pyx_empty_bytes, /*PyObject *code,*/
        __pyx_empty_tuple, /*PyObject *consts,*/
        __pyx_empty_tuple, /*PyObject *names,*/
        __pyx_empty_tuple, /*PyObject *varnames,*/
        __pyx_empty_tuple, /*PyObject *freevars,*/
        __pyx_empty_tuple, /*PyObject *cellvars,*/
        py_srcfile,   /*PyObject *filename,*/
        py_funcname,  /*PyObject *name,*/
        py_line,
        __pyx_empty_bytes  /*PyObject *lnotab*/
    );
    Py_DECREF(py_srcfile);
    Py_DECREF(py_funcname);
    return py_code;
bad:
    Py_XDECREF(py_srcfile);
    Py_XDECREF(py_funcname);
    return NULL;
}
static void __Pyx_AddTraceback(const char *funcname, int c_line,
                               int py_line, const char *filename) {
    PyCodeObject *py_code = 0;
    PyFrameObject *py_frame = 0;
    py_code = __pyx_find_code_object(c_line ? c_line : py_line);
    if (!py_code) {
        py_code = __Pyx_CreateCodeObjectForTraceback(
            funcname, c_line, py_line, filename);
        if (!py_code) goto bad;
        __pyx_insert_code_object(c_line ? c_line : py_line, py_code);
    }
    py_frame = PyFrame_New(
        PyThreadState_GET(), /*PyThreadState *tstate,*/
        py_code,             /*PyCodeObject *code,*/
        __pyx_d,      /*PyObject *globals,*/
        0                    /*PyObject *locals*/
    );
    if (!py_frame) goto bad;
    __Pyx_PyFrame_SetLineNumber(py_frame, py_line);
    PyTraceBack_Here(py_frame);
bad:
    Py_XDECREF(py_code);
    Py_XDECREF(py_frame);
}

/* CIntFromPyVerify */
          #define __PYX_VERIFY_RETURN_INT(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 0)
#define __PYX_VERIFY_RETURN_INT_EXC(target_type, func_type, func_value)\
    __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, 1)
#define __PYX__VERIFY_RETURN_INT(target_type, func_type, func_value, exc)\
    {\
        func_type value = func_value;\
        if (sizeof(target_type) < sizeof(func_type)) {\
            if (unlikely(value != (func_type) (target_type) value)) {\
                func_type zero = 0;\
                if (exc && unlikely(value == (func_type)-1 && PyErr_Occurred()))\
                    return (target_type) -1;\
                if (is_unsigned && unlikely(value < zero))\
                    goto raise_neg_overflow;\
                else\
                    goto raise_overflow;\
            }\
        }\
        return (target_type) value;\
    }

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_long(long value) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(long) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(long) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(long) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(long),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_int(int value) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(int),
                                     little, !is_unsigned);
    }
}

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_unsigned_int(unsigned int value) {
    const unsigned int neg_one = (unsigned int) -1, const_zero = (unsigned int) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(unsigned int) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(unsigned int) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(unsigned int) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(unsigned int) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(unsigned int),
                                     little, !is_unsigned);
    }
}

/* Declarations */
          #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return ::std::complex< float >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      return x + y*(__pyx_t_float_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_float_complex __pyx_t_float_complex_from_parts(float x, float y) {
      __pyx_t_float_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
          #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_sum_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_diff_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_prod_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        __pyx_t_float_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabsf(b.real) >= fabsf(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                float r = b.imag / b.real;
                float s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_float_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            float r = b.real / b.imag;
            float s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_float_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_quot_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
        if (b.imag == 0) {
            return __pyx_t_float_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            float denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_float_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_neg_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_float(__pyx_t_float_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_conj_float(__pyx_t_float_complex a) {
        __pyx_t_float_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE float __Pyx_c_abs_float(__pyx_t_float_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrtf(z.real*z.real + z.imag*z.imag);
          #else
            return hypotf(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_float_complex __Pyx_c_pow_float(__pyx_t_float_complex a, __pyx_t_float_complex b) {
            __pyx_t_float_complex z;
            float r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    float denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(a, a);
                    case 3:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, a);
                    case 4:
                        z = __Pyx_c_prod_float(a, a);
                        return __Pyx_c_prod_float(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = powf(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2f(0, -1);
                }
            } else {
                r = __Pyx_c_abs_float(a);
                theta = atan2f(a.imag, a.real);
            }
            lnr = logf(r);
            z_r = expf(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cosf(z_theta);
            z.imag = z_r * sinf(z_theta);
            return z;
        }
    #endif
#endif

/* Declarations */
          #if CYTHON_CCOMPLEX
  #ifdef __cplusplus
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return ::std::complex< double >(x, y);
    }
  #else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      return x + y*(__pyx_t_double_complex)_Complex_I;
    }
  #endif
#else
    static CYTHON_INLINE __pyx_t_double_complex __pyx_t_double_complex_from_parts(double x, double y) {
      __pyx_t_double_complex z;
      z.real = x;
      z.imag = y;
      return z;
    }
#endif

/* Arithmetic */
          #if CYTHON_CCOMPLEX
#else
    static CYTHON_INLINE int __Pyx_c_eq_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
       return (a.real == b.real) && (a.imag == b.imag);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_sum_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real + b.real;
        z.imag = a.imag + b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_diff_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real - b.real;
        z.imag = a.imag - b.imag;
        return z;
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_prod_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        __pyx_t_double_complex z;
        z.real = a.real * b.real - a.imag * b.imag;
        z.imag = a.real * b.imag + a.imag * b.real;
        return z;
    }
    #if 1
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else if (fabs(b.real) >= fabs(b.imag)) {
            if (b.real == 0 && b.imag == 0) {
                return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.imag);
            } else {
                double r = b.imag / b.real;
                double s = 1.0 / (b.real + b.imag * r);
                return __pyx_t_double_complex_from_parts(
                    (a.real + a.imag * r) * s, (a.imag - a.real * r) * s);
            }
        } else {
            double r = b.real / b.imag;
            double s = 1.0 / (b.imag + b.real * r);
            return __pyx_t_double_complex_from_parts(
                (a.real * r + a.imag) * s, (a.imag * r - a.real) * s);
        }
    }
    #else
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_quot_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
        if (b.imag == 0) {
            return __pyx_t_double_complex_from_parts(a.real / b.real, a.imag / b.real);
        } else {
            double denom = b.real * b.real + b.imag * b.imag;
            return __pyx_t_double_complex_from_parts(
                (a.real * b.real + a.imag * b.imag) / denom,
                (a.imag * b.real - a.real * b.imag) / denom);
        }
    }
    #endif
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_neg_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real = -a.real;
        z.imag = -a.imag;
        return z;
    }
    static CYTHON_INLINE int __Pyx_c_is_zero_double(__pyx_t_double_complex a) {
       return (a.real == 0) && (a.imag == 0);
    }
    static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_conj_double(__pyx_t_double_complex a) {
        __pyx_t_double_complex z;
        z.real =  a.real;
        z.imag = -a.imag;
        return z;
    }
    #if 1
        static CYTHON_INLINE double __Pyx_c_abs_double(__pyx_t_double_complex z) {
          #if !defined(HAVE_HYPOT) || defined(_MSC_VER)
            return sqrt(z.real*z.real + z.imag*z.imag);
          #else
            return hypot(z.real, z.imag);
          #endif
        }
        static CYTHON_INLINE __pyx_t_double_complex __Pyx_c_pow_double(__pyx_t_double_complex a, __pyx_t_double_complex b) {
            __pyx_t_double_complex z;
            double r, lnr, theta, z_r, z_theta;
            if (b.imag == 0 && b.real == (int)b.real) {
                if (b.real < 0) {
                    double denom = a.real * a.real + a.imag * a.imag;
                    a.real = a.real / denom;
                    a.imag = -a.imag / denom;
                    b.real = -b.real;
                }
                switch ((int)b.real) {
                    case 0:
                        z.real = 1;
                        z.imag = 0;
                        return z;
                    case 1:
                        return a;
                    case 2:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(a, a);
                    case 3:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, a);
                    case 4:
                        z = __Pyx_c_prod_double(a, a);
                        return __Pyx_c_prod_double(z, z);
                }
            }
            if (a.imag == 0) {
                if (a.real == 0) {
                    return a;
                } else if (b.imag == 0) {
                    z.real = pow(a.real, b.real);
                    z.imag = 0;
                    return z;
                } else if (a.real > 0) {
                    r = a.real;
                    theta = 0;
                } else {
                    r = -a.real;
                    theta = atan2(0, -1);
                }
            } else {
                r = __Pyx_c_abs_double(a);
                theta = atan2(a.imag, a.real);
            }
            lnr = log(r);
            z_r = exp(lnr * b.real - theta * b.imag);
            z_theta = theta * b.real + lnr * b.imag;
            z.real = z_r * cos(z_theta);
            z.imag = z_r * sin(z_theta);
            return z;
        }
    #endif
#endif

/* CIntToPy */
          static CYTHON_INLINE PyObject* __Pyx_PyInt_From_enum__NPY_TYPES(enum NPY_TYPES value) {
    const enum NPY_TYPES neg_one = (enum NPY_TYPES) -1, const_zero = (enum NPY_TYPES) 0;
    const int is_unsigned = neg_one > const_zero;
    if (is_unsigned) {
        if (sizeof(enum NPY_TYPES) < sizeof(long)) {
            return PyInt_FromLong((long) value);
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned long)) {
            return PyLong_FromUnsignedLong((unsigned long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(unsigned PY_LONG_LONG)) {
            return PyLong_FromUnsignedLongLong((unsigned PY_LONG_LONG) value);
#endif
        }
    } else {
        if (sizeof(enum NPY_TYPES) <= sizeof(long)) {
            return PyInt_FromLong((long) value);
#ifdef HAVE_LONG_LONG
        } else if (sizeof(enum NPY_TYPES) <= sizeof(PY_LONG_LONG)) {
            return PyLong_FromLongLong((PY_LONG_LONG) value);
#endif
        }
    }
    {
        int one = 1; int little = (int)*(unsigned char *)&one;
        unsigned char *bytes = (unsigned char *)&value;
        return _PyLong_FromByteArray(bytes, sizeof(enum NPY_TYPES),
                                     little, !is_unsigned);
    }
}

/* CIntFromPy */
          static CYTHON_INLINE int __Pyx_PyInt_As_int(PyObject *x) {
    const int neg_one = (int) -1, const_zero = (int) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(int) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(int, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (int) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case  1: __PYX_VERIFY_RETURN_INT(int, digit, digits[0])
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 2 * PyLong_SHIFT) {
                            return (int) (((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 3 * PyLong_SHIFT) {
                            return (int) (((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) >= 4 * PyLong_SHIFT) {
                            return (int) (((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (int) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(int) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (int) 0;
                case -1: __PYX_VERIFY_RETURN_INT(int, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(int,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(int) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(int) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                            return (int) ((((((int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(int) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(int) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                            return (int) ((((((((int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(int) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) (((int)-1)*(((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(int) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(int, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(int) - 1 > 4 * PyLong_SHIFT) {
                            return (int) ((((((((((int)digits[3]) << PyLong_SHIFT) | (int)digits[2]) << PyLong_SHIFT) | (int)digits[1]) << PyLong_SHIFT) | (int)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(int) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(int) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(int, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            int val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (int) -1;
        }
    } else {
        int val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (int) -1;
        val = __Pyx_PyInt_As_int(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to int");
    return (int) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to int");
    return (int) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE size_t __Pyx_PyInt_As_size_t(PyObject *x) {
    const size_t neg_one = (size_t) -1, const_zero = (size_t) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(size_t) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(size_t, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (size_t) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case  1: __PYX_VERIFY_RETURN_INT(size_t, digit, digits[0])
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 2 * PyLong_SHIFT) {
                            return (size_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 3 * PyLong_SHIFT) {
                            return (size_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) >= 4 * PyLong_SHIFT) {
                            return (size_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (size_t) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(size_t) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (size_t) 0;
                case -1: __PYX_VERIFY_RETURN_INT(size_t, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(size_t,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(size_t) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(size_t) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                            return (size_t) ((((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(size_t) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(size_t) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                            return (size_t) ((((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(size_t) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) (((size_t)-1)*(((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(size_t) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(size_t, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(size_t) - 1 > 4 * PyLong_SHIFT) {
                            return (size_t) ((((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(size_t) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(size_t) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(size_t, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            size_t val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (size_t) -1;
        }
    } else {
        size_t val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (size_t) -1;
        val = __Pyx_PyInt_As_size_t(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to size_t");
    return (size_t) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to size_t");
    return (size_t) -1;
}

/* CIntFromPy */
          static CYTHON_INLINE long __Pyx_PyInt_As_long(PyObject *x) {
    const long neg_one = (long) -1, const_zero = (long) 0;
    const int is_unsigned = neg_one > const_zero;
#if PY_MAJOR_VERSION < 3
    if (likely(PyInt_Check(x))) {
        if (sizeof(long) < sizeof(long)) {
            __PYX_VERIFY_RETURN_INT(long, long, PyInt_AS_LONG(x))
        } else {
            long val = PyInt_AS_LONG(x);
            if (is_unsigned && unlikely(val < 0)) {
                goto raise_neg_overflow;
            }
            return (long) val;
        }
    } else
#endif
    if (likely(PyLong_Check(x))) {
        if (is_unsigned) {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case  1: __PYX_VERIFY_RETURN_INT(long, digit, digits[0])
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 2 * PyLong_SHIFT) {
                            return (long) (((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 3 * PyLong_SHIFT) {
                            return (long) (((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) >= 4 * PyLong_SHIFT) {
                            return (long) (((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0]));
                        }
                    }
                    break;
            }
#endif
#if CYTHON_COMPILING_IN_CPYTHON
            if (unlikely(Py_SIZE(x) < 0)) {
                goto raise_neg_overflow;
            }
#else
            {
                int result = PyObject_RichCompareBool(x, Py_False, Py_LT);
                if (unlikely(result < 0))
                    return (long) -1;
                if (unlikely(result == 1))
                    goto raise_neg_overflow;
            }
#endif
            if (sizeof(long) <= sizeof(unsigned long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned long, PyLong_AsUnsignedLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(unsigned PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, unsigned PY_LONG_LONG, PyLong_AsUnsignedLongLong(x))
#endif
            }
        } else {
#if CYTHON_USE_PYLONG_INTERNALS
            const digit* digits = ((PyLongObject*)x)->ob_digit;
            switch (Py_SIZE(x)) {
                case  0: return (long) 0;
                case -1: __PYX_VERIFY_RETURN_INT(long, sdigit, (sdigit) (-(sdigit)digits[0]))
                case  1: __PYX_VERIFY_RETURN_INT(long,  digit, +digits[0])
                case -2:
                    if (8 * sizeof(long) - 1 > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 2:
                    if (8 * sizeof(long) > 1 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 2 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                            return (long) ((((((long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -3:
                    if (8 * sizeof(long) - 1 > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 3:
                    if (8 * sizeof(long) > 2 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 3 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                            return (long) ((((((((long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case -4:
                    if (8 * sizeof(long) - 1 > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, long, -(long) (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) (((long)-1)*(((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
                case 4:
                    if (8 * sizeof(long) > 3 * PyLong_SHIFT) {
                        if (8 * sizeof(unsigned long) > 4 * PyLong_SHIFT) {
                            __PYX_VERIFY_RETURN_INT(long, unsigned long, (((((((((unsigned long)digits[3]) << PyLong_SHIFT) | (unsigned long)digits[2]) << PyLong_SHIFT) | (unsigned long)digits[1]) << PyLong_SHIFT) | (unsigned long)digits[0])))
                        } else if (8 * sizeof(long) - 1 > 4 * PyLong_SHIFT) {
                            return (long) ((((((((((long)digits[3]) << PyLong_SHIFT) | (long)digits[2]) << PyLong_SHIFT) | (long)digits[1]) << PyLong_SHIFT) | (long)digits[0])));
                        }
                    }
                    break;
            }
#endif
            if (sizeof(long) <= sizeof(long)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, long, PyLong_AsLong(x))
#ifdef HAVE_LONG_LONG
            } else if (sizeof(long) <= sizeof(PY_LONG_LONG)) {
                __PYX_VERIFY_RETURN_INT_EXC(long, PY_LONG_LONG, PyLong_AsLongLong(x))
#endif
            }
        }
        {
#if CYTHON_COMPILING_IN_PYPY && !defined(_PyLong_AsByteArray)
            PyErr_SetString(PyExc_RuntimeError,
                            "_PyLong_AsByteArray() not available in PyPy, cannot convert large numbers");
#else
            long val;
            PyObject *v = __Pyx_PyNumber_IntOrLong(x);
 #if PY_MAJOR_VERSION < 3
            if (likely(v) && !PyLong_Check(v)) {
                PyObject *tmp = v;
                v = PyNumber_Long(tmp);
                Py_DECREF(tmp);
            }
 #endif
            if (likely(v)) {
                int one = 1; int is_little = (int)*(unsigned char *)&one;
                unsigned char *bytes = (unsigned char *)&val;
                int ret = _PyLong_AsByteArray((PyLongObject *)v,
                                              bytes, sizeof(val),
                                              is_little, !is_unsigned);
                Py_DECREF(v);
                if (likely(!ret))
                    return val;
            }
#endif
            return (long) -1;
        }
    } else {
        long val;
        PyObject *tmp = __Pyx_PyNumber_IntOrLong(x);
        if (!tmp) return (long) -1;
        val = __Pyx_PyInt_As_long(tmp);
        Py_DECREF(tmp);
        return val;
    }
raise_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "value too large to convert to long");
    return (long) -1;
raise_neg_overflow:
    PyErr_SetString(PyExc_OverflowError,
        "can't convert negative value to long");
    return (long) -1;
}

/* FetchCommonType */
          static PyTypeObject* __Pyx_FetchCommonType(PyTypeObject* type) {
    PyObject* fake_module;
    PyTypeObject* cached_type = NULL;
    fake_module = PyImport_AddModule((char*) "_cython_" CYTHON_ABI);
    if (!fake_module) return NULL;
    Py_INCREF(fake_module);
    cached_type = (PyTypeObject*) PyObject_GetAttrString(fake_module, type->tp_name);
    if (cached_type) {
        if (!PyType_Check((PyObject*)cached_type)) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s is not a type object",
                type->tp_name);
            goto bad;
        }
        if (cached_type->tp_basicsize != type->tp_basicsize) {
            PyErr_Format(PyExc_TypeError,
                "Shared Cython type %.200s has the wrong size, try recompiling",
                type->tp_name);
            goto bad;
        }
    } else {
        if (!PyErr_ExceptionMatches(PyExc_AttributeError)) goto bad;
        PyErr_Clear();
        if (PyType_Ready(type) < 0) goto bad;
        if (PyObject_SetAttrString(fake_module, type->tp_name, (PyObject*) type) < 0)
            goto bad;
        Py_INCREF(type);
        cached_type = type;
    }
done:
    Py_DECREF(fake_module);
    return cached_type;
bad:
    Py_XDECREF(cached_type);
    cached_type = NULL;
    goto done;
}

/* PyObjectCallMethod1 */
          static PyObject* __Pyx_PyObject_CallMethod1(PyObject* obj, PyObject* method_name, PyObject* arg) {
    PyObject *method, *result = NULL;
    method = __Pyx_PyObject_GetAttrStr(obj, method_name);
    if (unlikely(!method)) goto done;
#if CYTHON_UNPACK_METHODS
    if (likely(PyMethod_Check(method))) {
        PyObject *self = PyMethod_GET_SELF(method);
        if (likely(self)) {
            PyObject *args;
            PyObject *function = PyMethod_GET_FUNCTION(method);
            #if CYTHON_FAST_PYCALL
            if (PyFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            #if CYTHON_FAST_PYCCALL
            if (__Pyx_PyFastCFunction_Check(function)) {
                PyObject *args[2] = {self, arg};
                result = __Pyx_PyCFunction_FastCall(function, args, 2);
                goto done;
            }
            #endif
            args = PyTuple_New(2);
            if (unlikely(!args)) goto done;
            Py_INCREF(self);
            PyTuple_SET_ITEM(args, 0, self);
            Py_INCREF(arg);
            PyTuple_SET_ITEM(args, 1, arg);
            Py_INCREF(function);
            Py_DECREF(method); method = NULL;
            result = __Pyx_PyObject_Call(function, args, NULL);
            Py_DECREF(args);
            Py_DECREF(function);
            return result;
        }
    }
#endif
    result = __Pyx_PyObject_CallOneArg(method, arg);
done:
    Py_XDECREF(method);
    return result;
}

/* CoroutineBase */
          #include <structmember.h>
#include <frameobject.h>
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value);
static PyObject *__Pyx_Coroutine_Close(PyObject *self);
static PyObject *__Pyx_Coroutine_Throw(PyObject *gen, PyObject *args);
#define __Pyx_Coroutine_Undelegate(gen) Py_CLEAR((gen)->yieldfrom)
#if 1 || PY_VERSION_HEX < 0x030300B0
static int __Pyx_PyGen_FetchStopIterationValue(PyObject **pvalue) {
    PyObject *et, *ev, *tb;
    PyObject *value = NULL;
    __Pyx_PyThreadState_declare
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&et, &ev, &tb);
    if (!et) {
        Py_XDECREF(tb);
        Py_XDECREF(ev);
        Py_INCREF(Py_None);
        *pvalue = Py_None;
        return 0;
    }
    if (likely(et == PyExc_StopIteration)) {
        if (!ev) {
            Py_INCREF(Py_None);
            value = Py_None;
        }
#if PY_VERSION_HEX >= 0x030300A0
        else if (Py_TYPE(ev) == (PyTypeObject*)PyExc_StopIteration) {
            value = ((PyStopIterationObject *)ev)->value;
            Py_INCREF(value);
            Py_DECREF(ev);
        }
#endif
        else if (unlikely(PyTuple_Check(ev))) {
            if (PyTuple_GET_SIZE(ev) >= 1) {
#if CYTHON_ASSUME_SAFE_MACROS && !CYTHON_AVOID_BORROWED_REFS
                value = PyTuple_GET_ITEM(ev, 0);
                Py_INCREF(value);
#else
                value = PySequence_ITEM(ev, 0);
#endif
            } else {
                Py_INCREF(Py_None);
                value = Py_None;
            }
            Py_DECREF(ev);
        }
        else if (!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration)) {
            value = ev;
        }
        if (likely(value)) {
            Py_XDECREF(tb);
            Py_DECREF(et);
            *pvalue = value;
            return 0;
        }
    } else if (!PyErr_GivenExceptionMatches(et, PyExc_StopIteration)) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    PyErr_NormalizeException(&et, &ev, &tb);
    if (unlikely(!PyObject_TypeCheck(ev, (PyTypeObject*)PyExc_StopIteration))) {
        __Pyx_ErrRestore(et, ev, tb);
        return -1;
    }
    Py_XDECREF(tb);
    Py_DECREF(et);
#if PY_VERSION_HEX >= 0x030300A0
    value = ((PyStopIterationObject *)ev)->value;
    Py_INCREF(value);
    Py_DECREF(ev);
#else
    {
        PyObject* args = __Pyx_PyObject_GetAttrStr(ev, __pyx_n_s_args);
        Py_DECREF(ev);
        if (likely(args)) {
            value = PySequence_GetItem(args, 0);
            Py_DECREF(args);
        }
        if (unlikely(!value)) {
            __Pyx_ErrRestore(NULL, NULL, NULL);
            Py_INCREF(Py_None);
            value = Py_None;
        }
    }
#endif
    *pvalue = value;
    return 0;
}
#endif
static CYTHON_INLINE
void __Pyx_Coroutine_ExceptionClear(__pyx_CoroutineObject *self) {
    PyObject *exc_type = self->exc_type;
    PyObject *exc_value = self->exc_value;
    PyObject *exc_traceback = self->exc_traceback;
    self->exc_type = NULL;
    self->exc_value = NULL;
    self->exc_traceback = NULL;
    Py_XDECREF(exc_type);
    Py_XDECREF(exc_value);
    Py_XDECREF(exc_traceback);
}
static CYTHON_INLINE
int __Pyx_Coroutine_CheckRunning(__pyx_CoroutineObject *gen) {
    if (unlikely(gen->is_running)) {
        PyErr_SetString(PyExc_ValueError,
                        "generator already executing");
        return 1;
    }
    return 0;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_SendEx(__pyx_CoroutineObject *self, PyObject *value) {
    PyObject *retval;
    __Pyx_PyThreadState_declare
    assert(!self->is_running);
    if (unlikely(self->resume_label == 0)) {
        if (unlikely(value && value != Py_None)) {
            PyErr_SetString(PyExc_TypeError,
                            "can't send non-None value to a "
                            "just-started generator");
            return NULL;
        }
    }
    if (unlikely(self->resume_label == -1)) {
        PyErr_SetNone(PyExc_StopIteration);
        return NULL;
    }
    __Pyx_PyThreadState_assign
    if (value) {
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_XINCREF(__pyx_tstate->frame);
            assert(f->f_back == NULL);
            f->f_back = __pyx_tstate->frame;
        }
#endif
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    self->is_running = 1;
    retval = self->body((PyObject *) self, value);
    self->is_running = 0;
    if (retval) {
        __Pyx_ExceptionSwap(&self->exc_type, &self->exc_value,
                            &self->exc_traceback);
#if CYTHON_COMPILING_IN_PYPY || CYTHON_COMPILING_IN_PYSTON
#else
        if (self->exc_traceback) {
            PyTracebackObject *tb = (PyTracebackObject *) self->exc_traceback;
            PyFrameObject *f = tb->tb_frame;
            Py_CLEAR(f->f_back);
        }
#endif
    } else {
        __Pyx_Coroutine_ExceptionClear(self);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_MethodReturn(PyObject *retval) {
    if (unlikely(!retval && !PyErr_Occurred())) {
        PyErr_SetNone(PyExc_StopIteration);
    }
    return retval;
}
static CYTHON_INLINE
PyObject *__Pyx_Coroutine_FinishDelegation(__pyx_CoroutineObject *gen) {
    PyObject *ret;
    PyObject *val = NULL;
    __Pyx_Coroutine_Undelegate(gen);
    __Pyx_PyGen_FetchStopIterationValue(&val);
    ret = __Pyx_Coroutine_SendEx(gen, val);
    Py_XDECREF(val);
    return ret;
}
static PyObject *__Pyx_Coroutine_Send(PyObject *self, PyObject *value) {
    PyObject *retval;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Send(yf, value);
        } else
        #endif
        {
            if (value == Py_None)
                ret = Py_TYPE(yf)->tp_iternext(yf);
            else
                ret = __Pyx_PyObject_CallMethod1(yf, __pyx_n_s_send, value);
        }
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        retval = __Pyx_Coroutine_FinishDelegation(gen);
    } else {
        retval = __Pyx_Coroutine_SendEx(gen, value);
    }
    return __Pyx_Coroutine_MethodReturn(retval);
}
static int __Pyx_Coroutine_CloseIter(__pyx_CoroutineObject *gen, PyObject *yf) {
    PyObject *retval = NULL;
    int err = 0;
    #ifdef __Pyx_Generator_USED
    if (__Pyx_Generator_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    #ifdef __Pyx_Coroutine_USED
    if (__Pyx_Coroutine_CheckExact(yf)) {
        retval = __Pyx_Coroutine_Close(yf);
        if (!retval)
            return -1;
    } else
    #endif
    {
        PyObject *meth;
        gen->is_running = 1;
        meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_close);
        if (unlikely(!meth)) {
            if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                PyErr_WriteUnraisable(yf);
            }
            PyErr_Clear();
        } else {
            retval = PyObject_CallFunction(meth, NULL);
            Py_DECREF(meth);
            if (!retval)
                err = -1;
        }
        gen->is_running = 0;
    }
    Py_XDECREF(retval);
    return err;
}
static PyObject *__Pyx_Generator_Next(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject*) self;
    PyObject *yf = gen->yieldfrom;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Generator_Next(yf);
        } else
        #endif
            ret = Py_TYPE(yf)->tp_iternext(yf);
        gen->is_running = 0;
        if (likely(ret)) {
            return ret;
        }
        return __Pyx_Coroutine_FinishDelegation(gen);
    }
    return __Pyx_Coroutine_SendEx(gen, Py_None);
}
static PyObject *__Pyx_Coroutine_Close(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *retval, *raised_exception;
    PyObject *yf = gen->yieldfrom;
    int err = 0;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        Py_INCREF(yf);
        err = __Pyx_Coroutine_CloseIter(gen, yf);
        __Pyx_Coroutine_Undelegate(gen);
        Py_DECREF(yf);
    }
    if (err == 0)
        PyErr_SetNone(PyExc_GeneratorExit);
    retval = __Pyx_Coroutine_SendEx(gen, NULL);
    if (retval) {
        Py_DECREF(retval);
        PyErr_SetString(PyExc_RuntimeError,
                        "generator ignored GeneratorExit");
        return NULL;
    }
    raised_exception = PyErr_Occurred();
    if (!raised_exception
        || raised_exception == PyExc_StopIteration
        || raised_exception == PyExc_GeneratorExit
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_GeneratorExit)
        || PyErr_GivenExceptionMatches(raised_exception, PyExc_StopIteration))
    {
        if (raised_exception) PyErr_Clear();
        Py_INCREF(Py_None);
        return Py_None;
    }
    return NULL;
}
static PyObject *__Pyx_Coroutine_Throw(PyObject *self, PyObject *args) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject *typ;
    PyObject *tb = NULL;
    PyObject *val = NULL;
    PyObject *yf = gen->yieldfrom;
    if (!PyArg_UnpackTuple(args, (char *)"throw", 1, 3, &typ, &val, &tb))
        return NULL;
    if (unlikely(__Pyx_Coroutine_CheckRunning(gen)))
        return NULL;
    if (yf) {
        PyObject *ret;
        Py_INCREF(yf);
        if (PyErr_GivenExceptionMatches(typ, PyExc_GeneratorExit)) {
            int err = __Pyx_Coroutine_CloseIter(gen, yf);
            Py_DECREF(yf);
            __Pyx_Coroutine_Undelegate(gen);
            if (err < 0)
                return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
            goto throw_here;
        }
        gen->is_running = 1;
        #ifdef __Pyx_Generator_USED
        if (__Pyx_Generator_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        #ifdef __Pyx_Coroutine_USED
        if (__Pyx_Coroutine_CheckExact(yf)) {
            ret = __Pyx_Coroutine_Throw(yf, args);
        } else
        #endif
        {
            PyObject *meth = __Pyx_PyObject_GetAttrStr(yf, __pyx_n_s_throw);
            if (unlikely(!meth)) {
                Py_DECREF(yf);
                if (!PyErr_ExceptionMatches(PyExc_AttributeError)) {
                    gen->is_running = 0;
                    return NULL;
                }
                PyErr_Clear();
                __Pyx_Coroutine_Undelegate(gen);
                gen->is_running = 0;
                goto throw_here;
            }
            ret = PyObject_CallObject(meth, args);
            Py_DECREF(meth);
        }
        gen->is_running = 0;
        Py_DECREF(yf);
        if (!ret) {
            ret = __Pyx_Coroutine_FinishDelegation(gen);
        }
        return __Pyx_Coroutine_MethodReturn(ret);
    }
throw_here:
    __Pyx_Raise(typ, val, tb, NULL);
    return __Pyx_Coroutine_MethodReturn(__Pyx_Coroutine_SendEx(gen, NULL));
}
static int __Pyx_Coroutine_traverse(PyObject *self, visitproc visit, void *arg) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_VISIT(gen->closure);
    Py_VISIT(gen->classobj);
    Py_VISIT(gen->yieldfrom);
    Py_VISIT(gen->exc_type);
    Py_VISIT(gen->exc_value);
    Py_VISIT(gen->exc_traceback);
    return 0;
}
static int __Pyx_Coroutine_clear(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    Py_CLEAR(gen->closure);
    Py_CLEAR(gen->classobj);
    Py_CLEAR(gen->yieldfrom);
    Py_CLEAR(gen->exc_type);
    Py_CLEAR(gen->exc_value);
    Py_CLEAR(gen->exc_traceback);
    Py_CLEAR(gen->gi_name);
    Py_CLEAR(gen->gi_qualname);
    return 0;
}
static void __Pyx_Coroutine_dealloc(PyObject *self) {
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    PyObject_GC_UnTrack(gen);
    if (gen->gi_weakreflist != NULL)
        PyObject_ClearWeakRefs(self);
    if (gen->resume_label > 0) {
        PyObject_GC_Track(self);
#if PY_VERSION_HEX >= 0x030400a1
        if (PyObject_CallFinalizerFromDealloc(self))
#else
        Py_TYPE(gen)->tp_del(self);
        if (self->ob_refcnt > 0)
#endif
        {
            return;
        }
        PyObject_GC_UnTrack(self);
    }
    __Pyx_Coroutine_clear(self);
    PyObject_GC_Del(gen);
}
static void __Pyx_Coroutine_del(PyObject *self) {
    PyObject *res;
    PyObject *error_type, *error_value, *error_traceback;
    __pyx_CoroutineObject *gen = (__pyx_CoroutineObject *) self;
    __Pyx_PyThreadState_declare
    if (gen->resume_label <= 0)
        return ;
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt == 0);
    self->ob_refcnt = 1;
#endif
    __Pyx_PyThreadState_assign
    __Pyx_ErrFetch(&error_type, &error_value, &error_traceback);
    res = __Pyx_Coroutine_Close(self);
    if (res == NULL)
        PyErr_WriteUnraisable(self);
    else
        Py_DECREF(res);
    __Pyx_ErrRestore(error_type, error_value, error_traceback);
#if PY_VERSION_HEX < 0x030400a1
    assert(self->ob_refcnt > 0);
    if (--self->ob_refcnt == 0) {
        return;
    }
    {
        Py_ssize_t refcnt = self->ob_refcnt;
        _Py_NewReference(self);
        self->ob_refcnt = refcnt;
    }
#if CYTHON_COMPILING_IN_CPYTHON
    assert(PyType_IS_GC(self->ob_type) &&
           _Py_AS_GC(self)->gc.gc_refs != _PyGC_REFS_UNTRACKED);
    _Py_DEC_REFTOTAL;
#endif
#ifdef COUNT_ALLOCS
    --Py_TYPE(self)->tp_frees;
    --Py_TYPE(self)->tp_allocs;
#endif
#endif
}
static PyObject *
__Pyx_Coroutine_get_name(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_name;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_name(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__name__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_name;
    Py_INCREF(value);
    self->gi_name = value;
    Py_XDECREF(tmp);
    return 0;
}
static PyObject *
__Pyx_Coroutine_get_qualname(__pyx_CoroutineObject *self)
{
    PyObject *name = self->gi_qualname;
    if (unlikely(!name)) name = Py_None;
    Py_INCREF(name);
    return name;
}
static int
__Pyx_Coroutine_set_qualname(__pyx_CoroutineObject *self, PyObject *value)
{
    PyObject *tmp;
#if PY_MAJOR_VERSION >= 3
    if (unlikely(value == NULL || !PyUnicode_Check(value))) {
#else
    if (unlikely(value == NULL || !PyString_Check(value))) {
#endif
        PyErr_SetString(PyExc_TypeError,
                        "__qualname__ must be set to a string object");
        return -1;
    }
    tmp = self->gi_qualname;
    Py_INCREF(value);
    self->gi_qualname = value;
    Py_XDECREF(tmp);
    return 0;
}
static __pyx_CoroutineObject *__Pyx__Coroutine_New(
            PyTypeObject* type, __pyx_coroutine_body_t body, PyObject *closure,
            PyObject *name, PyObject *qualname, PyObject *module_name) {
    __pyx_CoroutineObject *gen = PyObject_GC_New(__pyx_CoroutineObject, type);
    if (gen == NULL)
        return NULL;
    gen->body = body;
    gen->closure = closure;
    Py_XINCREF(closure);
    gen->is_running = 0;
    gen->resume_label = 0;
    gen->classobj = NULL;
    gen->yieldfrom = NULL;
    gen->exc_type = NULL;
    gen->exc_value = NULL;
    gen->exc_traceback = NULL;
    gen->gi_weakreflist = NULL;
    Py_XINCREF(qualname);
    gen->gi_qualname = qualname;
    Py_XINCREF(name);
    gen->gi_name = name;
    Py_XINCREF(module_name);
    gen->gi_modulename = module_name;
    PyObject_GC_Track(gen);
    return gen;
}

/* PatchModuleWithCoroutine */
              static PyObject* __Pyx_Coroutine_patch_module(PyObject* module, const char* py_code) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    int result;
    PyObject *globals, *result_obj;
    globals = PyDict_New();  if (unlikely(!globals)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_coroutine_type",
    #ifdef __Pyx_Coroutine_USED
        (PyObject*)__pyx_CoroutineType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    result = PyDict_SetItemString(globals, "_cython_generator_type",
    #ifdef __Pyx_Generator_USED
        (PyObject*)__pyx_GeneratorType);
    #else
        Py_None);
    #endif
    if (unlikely(result < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "_module", module) < 0)) goto ignore;
    if (unlikely(PyDict_SetItemString(globals, "__builtins__", __pyx_b) < 0)) goto ignore;
    result_obj = PyRun_String(py_code, Py_file_input, globals, globals);
    if (unlikely(!result_obj)) goto ignore;
    Py_DECREF(result_obj);
    Py_DECREF(globals);
    return module;
ignore:
    Py_XDECREF(globals);
    PyErr_WriteUnraisable(module);
    if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning, "Cython module failed to patch module with custom type", 1) < 0)) {
        Py_DECREF(module);
        module = NULL;
    }
#else
    py_code++;
#endif
    return module;
}

/* PatchGeneratorABC */
              #if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
static PyObject* __Pyx_patch_abc_module(PyObject *module);
static PyObject* __Pyx_patch_abc_module(PyObject *module) {
    module = __Pyx_Coroutine_patch_module(
        module, ""
"if _cython_generator_type is not None:\n"
"    try: Generator = _module.Generator\n"
"    except AttributeError: pass\n"
"    else: Generator.register(_cython_generator_type)\n"
"if _cython_coroutine_type is not None:\n"
"    try: Coroutine = _module.Coroutine\n"
"    except AttributeError: pass\n"
"    else: Coroutine.register(_cython_coroutine_type)\n"
    );
    return module;
}
#endif
static int __Pyx_patch_abc(void) {
#if defined(__Pyx_Generator_USED) || defined(__Pyx_Coroutine_USED)
    static int abc_patched = 0;
    if (!abc_patched) {
        PyObject *module;
        module = PyImport_ImportModule((PY_VERSION_HEX >= 0x03030000) ? "collections.abc" : "collections");
        if (!module) {
            PyErr_WriteUnraisable(NULL);
            if (unlikely(PyErr_WarnEx(PyExc_RuntimeWarning,
                    ((PY_VERSION_HEX >= 0x03030000) ?
                        "Cython module failed to register with collections.abc module" :
                        "Cython module failed to register with collections module"), 1) < 0)) {
                return -1;
            }
        } else {
            module = __Pyx_patch_abc_module(module);
            abc_patched = 1;
            if (unlikely(!module))
                return -1;
            Py_DECREF(module);
        }
        module = PyImport_ImportModule("backports_abc");
        if (module) {
            module = __Pyx_patch_abc_module(module);
            Py_XDECREF(module);
        }
        if (!module) {
            PyErr_Clear();
        }
    }
#else
    if (0) __Pyx_Coroutine_patch_module(NULL, NULL);
#endif
    return 0;
}

/* Generator */
              static PyMethodDef __pyx_Generator_methods[] = {
    {"send", (PyCFunction) __Pyx_Coroutine_Send, METH_O,
     (char*) PyDoc_STR("send(arg) -> send 'arg' into generator,\nreturn next yielded value or raise StopIteration.")},
    {"throw", (PyCFunction) __Pyx_Coroutine_Throw, METH_VARARGS,
     (char*) PyDoc_STR("throw(typ[,val[,tb]]) -> raise exception in generator,\nreturn next yielded value or raise StopIteration.")},
    {"close", (PyCFunction) __Pyx_Coroutine_Close, METH_NOARGS,
     (char*) PyDoc_STR("close() -> raise GeneratorExit inside generator.")},
    {0, 0, 0, 0}
};
static PyMemberDef __pyx_Generator_memberlist[] = {
    {(char *) "gi_running", T_BOOL, offsetof(__pyx_CoroutineObject, is_running), READONLY, NULL},
    {(char*) "gi_yieldfrom", T_OBJECT, offsetof(__pyx_CoroutineObject, yieldfrom), READONLY,
     (char*) PyDoc_STR("object being iterated by 'yield from', or None")},
    {0, 0, 0, 0, 0}
};
static PyGetSetDef __pyx_Generator_getsets[] = {
    {(char *) "__name__", (getter)__Pyx_Coroutine_get_name, (setter)__Pyx_Coroutine_set_name,
     (char*) PyDoc_STR("name of the generator"), 0},
    {(char *) "__qualname__", (getter)__Pyx_Coroutine_get_qualname, (setter)__Pyx_Coroutine_set_qualname,
     (char*) PyDoc_STR("qualified name of the generator"), 0},
    {0, 0, 0, 0, 0}
};
static PyTypeObject __pyx_GeneratorType_type = {
    PyVarObject_HEAD_INIT(0, 0)
    "generator",
    sizeof(__pyx_CoroutineObject),
    0,
    (destructor) __Pyx_Coroutine_dealloc,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    Py_TPFLAGS_DEFAULT | Py_TPFLAGS_HAVE_GC | Py_TPFLAGS_HAVE_FINALIZE,
    0,
    (traverseproc) __Pyx_Coroutine_traverse,
    0,
    0,
    offsetof(__pyx_CoroutineObject, gi_weakreflist),
    0,
    (iternextfunc) __Pyx_Generator_Next,
    __pyx_Generator_methods,
    __pyx_Generator_memberlist,
    __pyx_Generator_getsets,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
    0,
#if PY_VERSION_HEX >= 0x030400a1
    0,
#else
    __Pyx_Coroutine_del,
#endif
    0,
#if PY_VERSION_HEX >= 0x030400a1
    __Pyx_Coroutine_del,
#endif
};
static int __pyx_Generator_init(void) {
    __pyx_GeneratorType_type.tp_getattro = PyObject_GenericGetAttr;
    __pyx_GeneratorType_type.tp_iter = PyObject_SelfIter;
    __pyx_GeneratorType = __Pyx_FetchCommonType(&__pyx_GeneratorType_type);
    if (unlikely(!__pyx_GeneratorType)) {
        return -1;
    }
    return 0;
}

/* CheckBinaryVersion */
              static int __Pyx_check_binary_version(void) {
    char ctversion[4], rtversion[4];
    PyOS_snprintf(ctversion, 4, "%d.%d", PY_MAJOR_VERSION, PY_MINOR_VERSION);
    PyOS_snprintf(rtversion, 4, "%s", Py_GetVersion());
    if (ctversion[0] != rtversion[0] || ctversion[2] != rtversion[2]) {
        char message[200];
        PyOS_snprintf(message, sizeof(message),
                      "compiletime version %s of module '%.100s' "
                      "does not match runtime version %s",
                      ctversion, __Pyx_MODULE_NAME, rtversion);
        return PyErr_WarnEx(NULL, message, 1);
    }
    return 0;
}

/* ModuleImport */
              #ifndef __PYX_HAVE_RT_ImportModule
#define __PYX_HAVE_RT_ImportModule
static PyObject *__Pyx_ImportModule(const char *name) {
    PyObject *py_name = 0;
    PyObject *py_module = 0;
    py_name = __Pyx_PyIdentifier_FromString(name);
    if (!py_name)
        goto bad;
    py_module = PyImport_Import(py_name);
    Py_DECREF(py_name);
    return py_module;
bad:
    Py_XDECREF(py_name);
    return 0;
}
#endif

/* TypeImport */
              #ifndef __PYX_HAVE_RT_ImportType
#define __PYX_HAVE_RT_ImportType
static PyTypeObject *__Pyx_ImportType(const char *module_name, const char *class_name,
    size_t size, int strict)
{
    PyObject *py_module = 0;
    PyObject *result = 0;
    PyObject *py_name = 0;
    char warning[200];
    Py_ssize_t basicsize;
#ifdef Py_LIMITED_API
    PyObject *py_basicsize;
#endif
    py_module = __Pyx_ImportModule(module_name);
    if (!py_module)
        goto bad;
    py_name = __Pyx_PyIdentifier_FromString(class_name);
    if (!py_name)
        goto bad;
    result = PyObject_GetAttr(py_module, py_name);
    Py_DECREF(py_name);
    py_name = 0;
    Py_DECREF(py_module);
    py_module = 0;
    if (!result)
        goto bad;
    if (!PyType_Check(result)) {
        PyErr_Format(PyExc_TypeError,
            "%.200s.%.200s is not a type object",
            module_name, class_name);
        goto bad;
    }
#ifndef Py_LIMITED_API
    basicsize = ((PyTypeObject *)result)->tp_basicsize;
#else
    py_basicsize = PyObject_GetAttrString(result, "__basicsize__");
    if (!py_basicsize)
        goto bad;
    basicsize = PyLong_AsSsize_t(py_basicsize);
    Py_DECREF(py_basicsize);
    py_basicsize = 0;
    if (basicsize == (Py_ssize_t)-1 && PyErr_Occurred())
        goto bad;
#endif
    if (!strict && (size_t)basicsize > size) {
        PyOS_snprintf(warning, sizeof(warning),
            "%s.%s size changed, may indicate binary incompatibility. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        if (PyErr_WarnEx(NULL, warning, 0) < 0) goto bad;
    }
    else if ((size_t)basicsize != size) {
        PyErr_Format(PyExc_ValueError,
            "%.200s.%.200s has the wrong size, try recompiling. Expected %zd, got %zd",
            module_name, class_name, basicsize, size);
        goto bad;
    }
    return (PyTypeObject *)result;
bad:
    Py_XDECREF(py_module);
    Py_XDECREF(result);
    return NULL;
}
#endif

/* FunctionImport */
              #ifndef __PYX_HAVE_RT_ImportFunction
#define __PYX_HAVE_RT_ImportFunction
static int __Pyx_ImportFunction(PyObject *module, const char *funcname, void (**f)(void), const char *sig) {
    PyObject *d = 0;
    PyObject *cobj = 0;
    union {
        void (*fp)(void);
        void *p;
    } tmp;
    d = PyObject_GetAttrString(module, (char *)"__pyx_capi__");
    if (!d)
        goto bad;
    cobj = PyDict_GetItemString(d, funcname);
    if (!cobj) {
        PyErr_Format(PyExc_ImportError,
            "%.200s does not export expected C function %.200s",
                PyModule_GetName(module), funcname);
        goto bad;
    }
#if PY_VERSION_HEX >= 0x02070000
    if (!PyCapsule_IsValid(cobj, sig)) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, PyCapsule_GetName(cobj));
        goto bad;
    }
    tmp.p = PyCapsule_GetPointer(cobj, sig);
#else
    {const char *desc, *s1, *s2;
    desc = (const char *)PyCObject_GetDesc(cobj);
    if (!desc)
        goto bad;
    s1 = desc; s2 = sig;
    while (*s1 != '\0' && *s1 == *s2) { s1++; s2++; }
    if (*s1 != *s2) {
        PyErr_Format(PyExc_TypeError,
            "C function %.200s.%.200s has wrong signature (expected %.500s, got %.500s)",
             PyModule_GetName(module), funcname, sig, desc);
        goto bad;
    }
    tmp.p = PyCObject_AsVoidPtr(cobj);}
#endif
    *f = tmp.fp;
    if (!(*f))
        goto bad;
    Py_DECREF(d);
    return 0;
bad:
    Py_XDECREF(d);
    return -1;
}
#endif

/* InitStrings */
              static int __Pyx_InitStrings(__Pyx_StringTabEntry *t) {
    while (t->p) {
        #if PY_MAJOR_VERSION < 3
        if (t->is_unicode) {
            *t->p = PyUnicode_DecodeUTF8(t->s, t->n - 1, NULL);
        } else if (t->intern) {
            *t->p = PyString_InternFromString(t->s);
        } else {
            *t->p = PyString_FromStringAndSize(t->s, t->n - 1);
        }
        #else
        if (t->is_unicode | t->is_str) {
            if (t->intern) {
                *t->p = PyUnicode_InternFromString(t->s);
            } else if (t->encoding) {
                *t->p = PyUnicode_Decode(t->s, t->n - 1, t->encoding, NULL);
            } else {
                *t->p = PyUnicode_FromStringAndSize(t->s, t->n - 1);
            }
        } else {
            *t->p = PyBytes_FromStringAndSize(t->s, t->n - 1);
        }
        #endif
        if (!*t->p)
            return -1;
        ++t;
    }
    return 0;
}

static CYTHON_INLINE PyObject* __Pyx_PyUnicode_FromString(const char* c_str) {
    return __Pyx_PyUnicode_FromStringAndSize(c_str, (Py_ssize_t)strlen(c_str));
}
static CYTHON_INLINE char* __Pyx_PyObject_AsString(PyObject* o) {
    Py_ssize_t ignore;
    return __Pyx_PyObject_AsStringAndSize(o, &ignore);
}
static CYTHON_INLINE char* __Pyx_PyObject_AsStringAndSize(PyObject* o, Py_ssize_t *length) {
#if CYTHON_COMPILING_IN_CPYTHON && (__PYX_DEFAULT_STRING_ENCODING_IS_ASCII || __PYX_DEFAULT_STRING_ENCODING_IS_DEFAULT)
    if (
#if PY_MAJOR_VERSION < 3 && __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
            __Pyx_sys_getdefaultencoding_not_ascii &&
#endif
            PyUnicode_Check(o)) {
#if PY_VERSION_HEX < 0x03030000
        char* defenc_c;
        PyObject* defenc = _PyUnicode_AsDefaultEncodedString(o, NULL);
        if (!defenc) return NULL;
        defenc_c = PyBytes_AS_STRING(defenc);
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        {
            char* end = defenc_c + PyBytes_GET_SIZE(defenc);
            char* c;
            for (c = defenc_c; c < end; c++) {
                if ((unsigned char) (*c) >= 128) {
                    PyUnicode_AsASCIIString(o);
                    return NULL;
                }
            }
        }
#endif
        *length = PyBytes_GET_SIZE(defenc);
        return defenc_c;
#else
        if (__Pyx_PyUnicode_READY(o) == -1) return NULL;
#if __PYX_DEFAULT_STRING_ENCODING_IS_ASCII
        if (PyUnicode_IS_ASCII(o)) {
            *length = PyUnicode_GET_LENGTH(o);
            return PyUnicode_AsUTF8(o);
        } else {
            PyUnicode_AsASCIIString(o);
            return NULL;
        }
#else
        return PyUnicode_AsUTF8AndSize(o, length);
#endif
#endif
    } else
#endif
#if (!CYTHON_COMPILING_IN_PYPY) || (defined(PyByteArray_AS_STRING) && defined(PyByteArray_GET_SIZE))
    if (PyByteArray_Check(o)) {
        *length = PyByteArray_GET_SIZE(o);
        return PyByteArray_AS_STRING(o);
    } else
#endif
    {
        char* result;
        int r = PyBytes_AsStringAndSize(o, &result, length);
        if (unlikely(r < 0)) {
            return NULL;
        } else {
            return result;
        }
    }
}
static CYTHON_INLINE int __Pyx_PyObject_IsTrue(PyObject* x) {
   int is_true = x == Py_True;
   if (is_true | (x == Py_False) | (x == Py_None)) return is_true;
   else return PyObject_IsTrue(x);
}
static CYTHON_INLINE PyObject* __Pyx_PyNumber_IntOrLong(PyObject* x) {
#if CYTHON_USE_TYPE_SLOTS
  PyNumberMethods *m;
#endif
  const char *name = NULL;
  PyObject *res = NULL;
#if PY_MAJOR_VERSION < 3
  if (PyInt_Check(x) || PyLong_Check(x))
#else
  if (PyLong_Check(x))
#endif
    return __Pyx_NewRef(x);
#if CYTHON_USE_TYPE_SLOTS
  m = Py_TYPE(x)->tp_as_number;
  #if PY_MAJOR_VERSION < 3
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Int(x);
  }
  else if (m && m->nb_long) {
    name = "long";
    res = PyNumber_Long(x);
  }
  #else
  if (m && m->nb_int) {
    name = "int";
    res = PyNumber_Long(x);
  }
  #endif
#else
  res = PyNumber_Int(x);
#endif
  if (res) {
#if PY_MAJOR_VERSION < 3
    if (!PyInt_Check(res) && !PyLong_Check(res)) {
#else
    if (!PyLong_Check(res)) {
#endif
      PyErr_Format(PyExc_TypeError,
                   "__%.4s__ returned non-%.4s (type %.200s)",
                   name, name, Py_TYPE(res)->tp_name);
      Py_DECREF(res);
      return NULL;
    }
  }
  else if (!PyErr_Occurred()) {
    PyErr_SetString(PyExc_TypeError,
                    "an integer is required");
  }
  return res;
}
static CYTHON_INLINE Py_ssize_t __Pyx_PyIndex_AsSsize_t(PyObject* b) {
  Py_ssize_t ival;
  PyObject *x;
#if PY_MAJOR_VERSION < 3
  if (likely(PyInt_CheckExact(b))) {
    if (sizeof(Py_ssize_t) >= sizeof(long))
        return PyInt_AS_LONG(b);
    else
        return PyInt_AsSsize_t(x);
  }
#endif
  if (likely(PyLong_CheckExact(b))) {
    #if CYTHON_USE_PYLONG_INTERNALS
    const digit* digits = ((PyLongObject*)b)->ob_digit;
    const Py_ssize_t size = Py_SIZE(b);
    if (likely(__Pyx_sst_abs(size) <= 1)) {
        ival = likely(size) ? digits[0] : 0;
        if (size == -1) ival = -ival;
        return ival;
    } else {
      switch (size) {
         case 2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -2:
           if (8 * sizeof(Py_ssize_t) > 2 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -3:
           if (8 * sizeof(Py_ssize_t) > 3 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case 4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return (Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
         case -4:
           if (8 * sizeof(Py_ssize_t) > 4 * PyLong_SHIFT) {
             return -(Py_ssize_t) (((((((((size_t)digits[3]) << PyLong_SHIFT) | (size_t)digits[2]) << PyLong_SHIFT) | (size_t)digits[1]) << PyLong_SHIFT) | (size_t)digits[0]));
           }
           break;
      }
    }
    #endif
    return PyLong_AsSsize_t(b);
  }
  x = PyNumber_Index(b);
  if (!x) return -1;
  ival = PyInt_AsSsize_t(x);
  Py_DECREF(x);
  return ival;
}
static CYTHON_INLINE PyObject * __Pyx_PyInt_FromSize_t(size_t ival) {
    return PyInt_FromSize_t(ival);
}


#endif /* Py_PYTHON_H */
